! **************************************************
!         Hamiltonian for Cartesian System
!
!                  Stewart K. Reed
!                   14th Oct 2008
!
! **************************************************
module cartesianham
  use hamiltoniandata
  
  private ! Keep most functions private
  
  ! public interface 
  public :: hamcart,dhamcartdz,dhamcartdzc,readhamcart
  ! Classical hamiltonian
  public :: chamcart,dchamcartdz,dchamcartdzc,readchamcart

  ! For constraints
  public :: ncart_constraint,cart_constraint
  public :: readhampcart
  
  real(kind=8),dimension(:),allocatable :: modekinzpe ! kinetic contribution to zero point energy
  real(kind=8) :: kinzpe ! Total
  real(kind=8),dimension(:),allocatable :: cart_gammamodes
  real(kind=8) :: totalmass ! total mass of the system.

  integer :: ncart_constraint=0
  integer,dimension(:),allocatable :: constraint_p1, constraint_p2
  real(kind=8),dimension(:),allocatable :: constraint_sep
  logical :: constraincom=.true. ! Constrain centre of mass momentum
contains
! ==================================================
!          Cartesian Hamiltomian 
! ==================================================
function hamcart(zic,zj)
  use coherentstates, ONLY: ndf,gamma
  use options, ONLY: ncartdim,eu
  use arrays, ONLY: t
  implicit none 
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  complex(kind=8)::hamcart
!debug complex(kind=8):: ke

  ! ---- local variables ----
  integer :: i,g,cdof    ! loop indicies
  integer :: alpha,beta  ! indicies for the degrees of freedom.
  complex :: Vmorse

  ! kinetic energy p^2 =hbar^2*gamma/4m -(1/4m)*hbar^2*gamma*(zic-zj)^2
  !  hamcart = 0.25*sum( (1-(zic(:)-zj(:))**2)*gamma(:)/m(:)) 
  hamcart = kinzpe-0.25*sum(((zic(:)-zj(:))**2)*gamma(:)/m(:)) 
!debug  ke = hamcart
  ! ---- Loop of the Morse potentials ----
  do i = 1, cart_nints
     ! ---- Do the sum of gaussian potentials ----
     do g = 1, cart_ntint(i)
        ! ---- loop of cartesian dimensions ----
        Vmorse = cart_gsamp(g,i)
        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim

           Vmorse =  Vmorse &
                * hamcartgauss(zic(alpha),zic(beta), zj(alpha)  ,zj(beta),&
                            alpha,beta,cart_gstwosigmasq(g,i))
        end do
        hamcart=hamcart + Vmorse !-cart_gsamp(g,i)
     end do
  end do
!debug  write(301,*) t,real(ke)*eu,real(hamcart-ke)*eu,real(hamcart)*eu
end function hamcart

! ==================================================
!          Cartesian Hamiltomian Derivative
! ==================================================
function dhamcartdz(zic,zj)   
  use coherentstates, ONLY: ndf,gamma,sgamma
  use options, ONLY: ncartdim
  implicit none 
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  complex(kind=8)::dhamcartdz(ndf)
  integer :: i,g,cdof
  integer ::alpha,beta
  complex(kind=8) :: Vmorse
  
!  dhamcartdz(:)= (0.0d0,0.0d0)
!  return
  ! ---- derivative of the kinetic energy ----
  dhamcartdz(:) =0.5d0*gamma(:)*(zic(:)-zj(:))/m(:)

  ! ---- Loop of the Morse potentials ----
  do i = 1, cart_nints
     ! ---- Do the sum of gaussian potentials ----
     do g = 1, cart_ntint(i)
        ! ---- First work out gaussian potential ----
        Vmorse = cart_gsamp(g,i)

        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
                      
           
           Vmorse =  Vmorse  &
                * hamcartgauss(zic(alpha),zic(beta), zj(alpha)  ,zj(beta),&
                alpha,beta,cart_gstwosigmasq(g,i))
        end do
        
        ! ---- Now work out contributions to derivative ----
        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
        
           dhamcartdz(alpha) = dhamcartdz(alpha)&
                - ((gamma(beta)*(zic(alpha)+zj(alpha))) &
                -sgamma(alpha)*sgamma(beta)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)+gamma(alpha)+gamma(beta)) * Vmorse

           dhamcartdz(beta) = dhamcartdz(beta)&
                +  ((sgamma(alpha)*sgamma(beta)*(zic(alpha)+zj(alpha))) &
                -gamma(alpha)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)+gamma(alpha)+gamma(beta)) * Vmorse
        end do
     end do
  end do

end function dhamcartdz

! ==================================================
!          Cartesian Hamiltomian Derivative
! ==================================================
function dhamcartdzc(zic,zj)   
  use coherentstates, ONLY: ndf,gamma,sgamma
  use options, ONLY: ncartdim
  implicit none 
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  complex(kind=8)::dhamcartdzc(ndf)
  integer :: i,g,cdof
  integer ::alpha,beta
  complex(kind=8) :: Vmorse
  
!  dhamcartdzc(:)= (0.0d0,0.0d0)
!  return
  ! ---- derivative of the kinetic energy ----
  dhamcartdzc(:) =-0.5d0*gamma(:)*(zic(:)-zj(:))/m(:)

  ! ---- Loop of the Morse potentials ----
  do i = 1, cart_nints
     ! ---- Do the sum of gaussian potentials ----

     do g = 1, cart_ntint(i)
        ! ---- First work out gaussian potential ----
        Vmorse = cart_gsamp(g,i)

        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
                      
           
           Vmorse =  Vmorse  &
                * hamcartgauss(zic(alpha),zic(beta), zj(alpha)  ,zj(beta),&
                alpha,beta,cart_gstwosigmasq(g,i))

        end do
        
        ! ---- Now work out contributions to derivative ----
        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
        
           dhamcartdzc(alpha) = dhamcartdzc(alpha)&
                - ((gamma(beta)*(zic(alpha)+zj(alpha))) &
                -sgamma(alpha)*sgamma(beta)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)+gamma(alpha)+gamma(beta)) * Vmorse

           dhamcartdzc(beta) = dhamcartdzc(beta)&
                +  ((sgamma(alpha)*sgamma(beta)*(zic(alpha)+zj(alpha))) &
                -gamma(alpha)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)+gamma(alpha)+gamma(beta)) * Vmorse
        end do
     end do
  end do

end function dhamcartdzc

! ==================================================
!          hamcartgauss
! Function to calculate the gauassian potential
 
! ==================================================
function hamcartgauss(z1ac,z1bc,z2a,z2b,p1,p2,twosigmasq)
  use coherentstates,ONLY:gamma,sgamma
  implicit none
  complex(kind=8),intent(in) :: z1ac,z1bc,z2a,z2b
  integer,intent(in) :: p1,p2
  real(kind=8),intent(in) :: twosigmasq
  complex(kind=8) :: hamcartgauss ! the return type

  !prefactor
  hamcartgauss = dsqrt(twosigmasq /(twosigmasq+1/gamma(p1)+1/gamma(p2))) &
       *exp(-0.5d0*(sgamma(p2)*(z1ac+z2a)-sgamma(p1)*(z1bc+z2b))**2 &
             /(gamma(p1)*gamma(p2)*twosigmasq+gamma(p1)+gamma(p2)))

end function hamcartgauss


! ==================================================
!     Classical cartesian Hamiltomian 
! ==================================================
function chamcart(zic,zj)
  use coherentstates, ONLY: ndf,gamma
  use options, ONLY: ncartdim,eu
  use arrays, ONLY: t
  implicit none 
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  complex(kind=8)::chamcart
!debug complex(kind=8):: ke

  ! ---- local variables ----
  integer :: i,g,cdof    ! loop indicies
  integer :: alpha,beta  ! indicies for the degrees of freedom.
  complex :: Vmorse

  ! kinetic energy p^2 =hbar^2*gamma/4m -(1/4m)*hbar^2*gamma*(zic-zj)^2
  !  hamcart = 0.25*sum( (1-(zic(:)-zj(:))**2)*gamma(:)/m(:)) 
  ! For classical calculations, don't include the zero point energy, kinzpe
  chamcart = -0.25*sum(((zic(:)-zj(:))**2)*gamma(:)/m(:)) 

!debug  ke = hamcart
  ! ---- Loop of the Morse potentials ----
  do i = 1, cart_nints
     ! ---- Do the sum of gaussian potentials ----
     do g = 1, cart_ntint(i)
        ! ---- loop of cartesian dimensions ----
        Vmorse = cart_gsamp(g,i)
        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim

           Vmorse =  Vmorse &
                * classhamcartgauss(zic(alpha),zic(beta), zj(alpha)  ,zj(beta),&
                            alpha,beta,cart_gstwosigmasq(g,i))
        end do
        chamcart=chamcart + Vmorse !-cart_gsamp(g,i)
     end do
  end do
!debug  write(301,*) t,real(ke)*eu,real(hamcart-ke)*eu,real(hamcart)*eu
end function chamcart

! ==================================================
!      Classical cartesian Hamiltomian Derivative
! ==================================================
function dchamcartdz(zic,zj)   
  use coherentstates, ONLY: ndf,gamma,sgamma
  use options, ONLY: ncartdim
  implicit none 
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  complex(kind=8)::dchamcartdz(ndf)
  integer :: i,g,cdof
  integer ::alpha,beta
  complex(kind=8) :: Vmorse
  
!  dhamcartdz(:)= (0.0d0,0.0d0)
!  return
  ! ---- derivative of the kinetic energy ----
  dchamcartdz(:) =0.5d0*gamma(:)*(zic(:)-zj(:))/m(:)

  ! ---- Loop of the Morse potentials ----
  do i = 1, cart_nints
     ! ---- Do the sum of gaussian potentials ----
     do g = 1, cart_ntint(i)
        ! ---- First work out gaussian potential ----
        Vmorse = cart_gsamp(g,i)

        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
                      
           
           Vmorse =  Vmorse  &
                * classhamcartgauss(zic(alpha),zic(beta), zj(alpha)  ,zj(beta),&
                alpha,beta,cart_gstwosigmasq(g,i))
        end do
        
        ! ---- Now work out contributions to derivative ----
        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
        
           dchamcartdz(alpha) = dchamcartdz(alpha)&
                - ((gamma(beta)*(zic(alpha)+zj(alpha))) &
                -sgamma(alpha)*sgamma(beta)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)) * Vmorse

           dchamcartdz(beta) = dchamcartdz(beta)&
                +  ((sgamma(alpha)*sgamma(beta)*(zic(alpha)+zj(alpha))) &
                -gamma(alpha)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)) * Vmorse
        end do
     end do
  end do

end function dchamcartdz

! ==================================================
!      Classical cartesian Hamiltomian Derivative
! ==================================================
function dchamcartdzc(zic,zj)   
  use coherentstates, ONLY: ndf,gamma,sgamma
  use options, ONLY: ncartdim
  implicit none 
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  complex(kind=8)::dchamcartdzc(ndf)
  integer :: i,g,cdof
  integer ::alpha,beta
  complex(kind=8) :: Vmorse
  
!  dhamcartdzc(:)= (0.0d0,0.0d0)
!  return
  ! ---- derivative of the kinetic energy ----
  dchamcartdzc(:) =-0.5d0*gamma(:)*(zic(:)-zj(:))/m(:)

  ! ---- Loop of the Morse potentials ----
  do i = 1, cart_nints
     ! ---- Do the sum of gaussian potentials ----

     do g = 1, cart_ntint(i)
        ! ---- First work out gaussian potential ----
        Vmorse = cart_gsamp(g,i)

        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
                      
           
           Vmorse =  Vmorse  &
                * classhamcartgauss(zic(alpha),zic(beta), zj(alpha)  ,zj(beta),&
                alpha,beta,cart_gstwosigmasq(g,i))

        end do
        
        ! ---- Now work out contributions to derivative ----
        do cdof = 1,ncartdim
           ! determine global indicies
           alpha = ncartdim*cart_part1(i)+cdof-ncartdim
           beta  = ncartdim*cart_part2(i)+cdof-ncartdim
        
           dchamcartdzc(alpha) = dchamcartdzc(alpha)&
                - ((gamma(beta)*(zic(alpha)+zj(alpha))) &
                -sgamma(alpha)*sgamma(beta)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)) * Vmorse

           dchamcartdzc(beta) = dchamcartdzc(beta)&
                +  ((sgamma(alpha)*sgamma(beta)*(zic(alpha)+zj(alpha))) &
                -gamma(alpha)*( zic(beta)+zj(beta))) &
             /(gamma(alpha)*gamma(beta)*cart_gstwosigmasq(g,i)) * Vmorse
        end do
     end do
  end do

end function dchamcartdzc

! ==================================================
!          classhamcartgauss
! Function to calculate the unaveraged 
! gauassian potential
! ==================================================
function classhamcartgauss(z1ac,z1bc,z2a,z2b,p1,p2,twosigmasq)
  use coherentstates,ONLY:gamma,sgamma
  implicit none
  complex(kind=8),intent(in) :: z1ac,z1bc,z2a,z2b
  integer,intent(in) :: p1,p2
  real(kind=8),intent(in) :: twosigmasq
  complex(kind=8) :: classhamcartgauss ! the return type

  !prefactor
  classhamcartgauss = &
       exp(-0.5d0*(sgamma(p2)*(z1ac+z2a)-sgamma(p1)*(z1bc+z2b))**2 &
             /(gamma(p1)*gamma(p2)*twosigmasq))

end function classhamcartgauss

! ==================================================
!          Apply constraints 
! ==================================================
subroutine cart_constraint(bs,bsnew,dbsdt,dt)
  use coherentstatestype
  use coherentstates, ONLY: s2gammarec,shgamma
  use options, ONLY: ncartdim,nparticles
!  use coherentstates, ONLY: pos,mom
  implicit none
  type(coherentstatebasisset),intent(in) :: bs ! Old basis set
  type(coherentstatebasisset),intent(inout) :: bsnew ! New basis set
  type(coherentstatebasisset),intent(inout) :: dbsdt ! Gradient of basis set
  real(kind=8),intent(in) :: dt ! time step 
  ! ---- local variables ----
  real(kind=8) :: vqsepf(ncartdim),vqsep(ncartdim) ! Separation of basis functions at t+dt before corrections

  real(kind=8) :: sep,sep2,sepf2

  !complex(kind=8),save :: dzsep(ncartdim)
  integer :: cdof,i
  real(kind=8) :: zre, mom
  
  real(kind=8) :: lambda1, lambda2 
  integer :: bf ! Loop over basis functions
  integer :: p1,p2
  real(kind=8) :: qa,qb,qc
  real(kind=8) :: po1,po2,pf1,pf2
  real(kind=8),dimension(ncartdim) :: reds2dz1,reds2dz2
  real(kind=8) :: denom,pdotqterm
  real(kind=8) :: comp ! Centre of mass momentum

  complex(kind=8) :: dcomr,dcomp ! Centre of mass force/acceleration

!!  write(400,*) "Applying constraint"

  ! Offsets for particles. Need to add 1 to get the first position.
  p1 = ncartdim*(constraint_p1(1)-1)
  p2 = ncartdim*(constraint_p2(1)-1)
  do bf = 1,bs%nbasis 
     
     ! Calculate uncorrected new and old separations
     do cdof = 1,ncartdim
        ! Old

        ! ******** But really want cartesian separation rather than phase space separation ***********
        vqsep(cdof) = real(s2gammarec(p1+1)*(conjg(bs%zb(p1+cdof,bf))+bs%zb(p1+cdof,bf)) &
             -s2gammarec(p2+1)*(conjg(bs%zb(p2+cdof,bf))+bs%zb(p2+cdof,bf)))
        ! New
        vqsepf(cdof) = real(s2gammarec(p1+1)*(conjg(bsnew%zb(p1+cdof,bf))+bsnew%zb(p1+cdof,bf)) &
             -s2gammarec(p2+1)*(conjg(bsnew%zb(p2+cdof,bf))+bsnew%zb(p2+cdof,bf)))
     end do

     ! Now calculate the lengths of theses vectors
     sep2 = sum(vqsep(:)*vqsep(:))
     sep = sqrt(sep2)
     sepf2 = sum(vqsepf(:)*vqsepf(:))
!!     write(400,*) "Separations old/new",sep,sqrt(sepf2),constraint_sep(1)
     
     ! ---- First calculate lambda2 ----
     qb =  -2*dt*(shgamma(p1+1)/m(p1+1)+shgamma(p2+1)/m(p2+1))*sum(vqsepf(:)*vqsep(:))/sep
     qc = sepf2 - constraint_sep(1)**2
     qa = 4*dt**2*(shgamma(p1+1)/m(p1+1)+shgamma(p2+1)/m(p2+1))**2/sep2

     !     d=sepf2 - constraint_sep(1)**2
!!      write (400,*) bf,"lambda2", -(qb+ sqrt(qb**2-4*qa*qc))*0.5/qa, -(qb- sqrt(qb**2-4*qa*qc))*0.5/qa
     if (abs(sepf2 - constraint_sep(1)**2) .lt. 1e-10) then
        lambda2 = 0.0d0
     else
        if (abs( -(qb+ sqrt(qb**2-4*qa*qc))*0.5/qa) .le. (abs( -(qb- sqrt(qb**2-4*qa*qc))*0.5/qa))) then
           lambda2 =  -(qb+ sqrt(qb**2-4*qa*qc))*0.5/qa
        else
           lambda2 = -(qb- sqrt(qb**2-4*qa*qc))*0.5/qa
        end if
!!$        if (sepf2 .lt. constraint_sep(1)**2) then
!!$           lambda2 = -(qb+ sqrt(qb**2-4*qa*qc))*0.5/qa
!!$           write (400,*) bf,"-lambda2",lambda2
!!$        else
!!$           lambda2 = -(qb- sqrt(qb**2-4*qa*qc))*0.5/qa
!!$           write (400,*) bf,"+lambda2",lambda2           
!!$        end if
     end if

     ! Now we can correct the positions
     do cdof = 1,ncartdim
        bsnew%zb(p1 + cdof,bf ) =  bsnew%zb(p1 + cdof,bf )  &
             - lambda2 * dt*shgamma(p1+1)/m(p1+1)*vqsep(cdof)/sep
!!        write(400,*) "Rez1 correction", -lambda2 * dt*shgamma(p1+1)/m(p1+1)*vqsep(cdof)/sep
        bsnew%zb(p2 + cdof ,bf) =  bsnew%zb(p2 + cdof,bf)  &
             + lambda2 * dt*shgamma(p2+1)/m(p2+1)*vqsep(cdof)/sep
!!        write(400,*) "Rez2 correction", +lambda2 *dt* shgamma(p1+1)/m(p1+1)*vqsep(cdof)/sep
     end do


     ! ---- Now we want to calculate lambda1 ----
     ! Real part of dsigma/dzstar_
     pdotqterm =  0.0d0
     do cdof = 1,ncartdim
        ! momenta of atoms 1 and 2
        ! at time t
        po1 = 2*shgamma(p1+1)*imag(bs%zb(p1+cdof,bf))
        po2 = 2*shgamma(p2+1)*imag(bs%zb(p2+cdof,bf))
        ! and at time t+dt but before correction
        pf1 = 2*shgamma(p1+1)*imag(bsnew%zb(p1+cdof,bf))
        pf2 = 2*shgamma(p2+1)*imag(bsnew%zb(p2+cdof,bf))
        
  !!      write(400,*) "P dof Old 1,2 / New 1,2",cdof,po1,po2,pf1,pf2
        ! Numerator for third term.
        pdotqterm = pdotqterm + (po1/m(p1+1) -po2/m(p2+1))*vqsep(cdof)
        
        
        reds2dz1(cdof) = (po1/m(p1+1)-po2/m(p2+1))/sep*s2gammarec(p1+1)/sep
        reds2dz2(cdof) = -(po1/m(p1+1)-po2/m(p2+1))/sep*s2gammarec(p2+1)/sep
        
     end do

     reds2dz1(:) =  reds2dz1(:) - pdotqterm/(sep*sep2)*s2gammarec(p1+1)
     reds2dz1(:) =  reds2dz1(:) + pdotqterm/(sep*sep2)*s2gammarec(p2+1)
     
     ! Work out the numerator for lambda1 but call it lambda1 for simplicity and the denominator
     lambda1 = 0.0d0  
     denom=0.0d0
     do cdof = 1,ncartdim
        pf1 = 2*shgamma(p1+1)*imag(bsnew%zb(p1+cdof,bf))
        pf2 = 2*shgamma(p2+1)*imag(bsnew%zb(p2+cdof,bf))
        lambda1 = lambda1 + &
             (pf1/m(p1+1) - pf2/m(p2+1) &
             - dt*2*lambda2*(shgamma(p1+1)/m(p1+1)*reds2dz1(cdof) - (shgamma(p2+1)/m(p2+1)*reds2dz2(cdof))))*vqsepf(cdof)
        
        denom = denom + vqsep(cdof)*vqsepf(cdof)
     end do
     
     ! multiply the denominator by the appropriate constants
     denom = denom*2*dt*(1/m(p1+1)+1/m(p2+1))
     
     ! Now work out lambda1
     lambda1= lambda1/denom
     
  !!   write (400,*) bf,"lambda1",lambda1
     ! Finally we want to correct the velocity and the derivative
     do cdof = 1,ncartdim
        bsnew%zb(p1 + cdof,bf ) =  bsnew%zb(p1 + cdof,bf )  &
             - (0.0d0,1.0d0) * dt*(lambda1 * vqsep(cdof)/shgamma(p1+1)+ lambda2*reds2dz1(cdof))
  !!      write(400,*) "Imdz", - (0.0d0,1.0d0) * dt*(lambda1 * vqsep(cdof)/shgamma(p1+1) + lambda2*reds2dz1(cdof))
        bsnew%zb(p2 + cdof,bf ) =  bsnew%zb(p2 + cdof,bf )  &
             + (0.0d0,1.0d0) * dt*(lambda1 * vqsep(cdof)/shgamma(p2+1) - lambda2*reds2dz2(cdof))
  !!      write(400,*) "Imdz", (0.0d0,1.0d0) * dt*(lambda1 * vqsep(cdof)/shgamma(p2+1) - lambda2*reds2dz2(cdof))
!         bsnew%zb(p1 + cdof,bf ) = real( bsnew%zb(p1 + cdof,bf ))
!         bsnew%zb(p2 + cdof,bf ) = real( bsnew%zb(p2 + cdof,bf ))

   !!     write(400,*) "dbsdt1 ",cdof, "old",  dbsdt%zb(p1+cdof,bf)," new ",&
   !!          (bsnew%zb(p1+cdof,bf) - bs%zb(p1+cdof,bf))/dt
   !!     write(400,*) "dbsdt2 ",cdof, "old",  dbsdt%zb(p2+cdof,bf)," new ",&
   !!          (bsnew%zb(p2+cdof,bf) - bs%zb(p2+cdof,bf))/dt
        dbsdt%zb(p1+cdof,bf) = (bsnew%zb(p1+cdof,bf) - bs%zb(p1+cdof,bf))!/dt
        dbsdt%zb(p2+cdof,bf) = (bsnew%zb(p2+cdof,bf) - bs%zb(p2+cdof,bf))!/dt
     end do
  
   !!  write(400,*) "p1", 2*shgamma(p1+1)*sqrt(sum(imag(bsnew%zb(p1+1:p1+3,bf))**2)),&
   !!       "p2", 2*shgamma(p2+1)*sqrt(sum(imag(bsnew%zb(p1+1:p1+3,bf))**2))
     
  end do ! Loop over basis functions.
  
  ! Constrain Centre of mass momentum?
  if (constraincom) then
     do bf = 1,bsnew%nbasis
        ! ----- Now we want to reset the centre of mass momentum. ----
        
        do cdof = 1,ncartdim
           comp = 0.0d0 ! Centre of mass momentum
           dcomr = (0.0d0,0.0d0)
           dcomp = (0.0d0,0.0d0)
           !        coma = 0.0d0 ! Centre of mass acceleration
           ! Determine the motion of the centre of mass.
           do p1 = 0,nparticles-1
              comp = comp+dimag(bsnew%zb(p1*ncartdim+cdof,bf))
              !!coma = coma+dimag(dbsdt%zb(p1*ncartdim+cdof,bf))

              dcomp =dcomp+dimag(dbsdt%zb(p1*ncartdim+cdof,bf))
              dcomr =dcomr+real(dbsdt%zb(p1*ncartdim+cdof,bf)*m(p1*ncartdim+cdof))
           end do
           ! Now correct the velocities of the particles
           do p1 = 0,nparticles-1
              bsnew%zb(p1*ncartdim+cdof,bf)=bsnew%zb(p1*ncartdim+cdof,bf)&
                   -(0.0d0,1.0d0)*m(p1*ncartdim+cdof)*comp/totalmass
              !!           coma = coma+dimag(dbsdt%zb(p1*ncartdim+cdof,bf))
              dbsdt%zb(p1*ncartdim+cdof,bf) = dbsdt%zb(p1*ncartdim+cdof,bf) &
                   - (dcomr - (0.0d0,1.0d0)*dcomp*m(p1*ncartdim+cdof) )/totalmass
           end do
        end do

     end do ! Loop over basis functions.
  end if
end subroutine cart_constraint

! ==================================================
!          Read params for cartesian system
! ==================================================
subroutine readhamcart
  use options, ONLY: ncartdim,nparticles,verbosity,eu,writeexpect,energyunits
  use files, ONLY: fug,fham
  use errors
  use scalapackstuff, ONLY: isroot
  use arrays, ONLY: dhexpect,hework
  use coherentstates, ONLY: gamma
  implicit none

  integer :: i,j ! For loops
  character(len=50) :: line
  integer :: ierr ! errors from reading stuff from file.
  real(kind=8) :: m1,m2,redmass ! Masses for calculating the reduced mass
  
  ! ---- Allocate more general arrays ----
  ! Just mass?
  call allocatehamarrays 

  ! ---- Read in masses -----
  read(fham,*,iostat=ierr) line
  if  (line .ne. "mass" .or. ierr .ne. 0  ) then
     if(isroot) write(fug,*) "Failed to read keyword mass in hamiltonian.inpt"
     stop
  end if
  do i =1,nparticles
     ! each degree of freedom has a separate mass. 
     read(fham,*,iostat=ierr) m(ncartdim*i)
     if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading masses")
     ! Set the masses for the other degrees of freedom of this particle.
     do j = 1,ncartdim-1
        m(ncartdim*i-j) = m(ncartdim*i)
     end do
  end do

  ! Calculate the total mass of the system.
  ! Because of the generality of the CCS code, each cartesian degree of freedom has a mass
  ! equal to that of the particle
  totalmass = sum(m)/ncartdim

  ! ---- now read in potential ----
  read(fham,*,iostat=ierr) line 
  if (line == "interactions") then
     backspace(fham) 
     read(fham,*,iostat=ierr) line, cart_nints

  else
     call filereaderror(ierr,"hamiltonian.inpt","Can't find any interactions")
  end if

  ! now look for max number of terms in sum of gaussians
   read(fham,*,iostat=ierr) line 
   if (line == "maxterms") then
      backspace(fham) 
      read(fham,*,iostat=ierr) line, cart_maxsgterms
      if (ierr .ne. 0) then
         call filereaderror(ierr,"hamiltonian.inpt","The max number of terms in the sum of gaussians has not been specified.")
      else
         if(isroot)  write(fug,*) "Max number of terms in gaussian sum is",cart_maxsgterms
      end if
   end if
     
  ! now allocate arrays
  ! number of terms in the sums for each mode
  allocate(cart_ntint(cart_nints))
  cart_ntint = 0
  allocate(cart_gsamp(cart_maxsgterms,cart_nints))
  cart_gsamp(:,:) = 0.0d0
  allocate(cart_gssigma(cart_maxsgterms,cart_nints))
  allocate(cart_gstwosigmasq(cart_maxsgterms,cart_nints))
  allocate(cart_part1(cart_nints))
  allocate(cart_part2(cart_nints))
  allocate(cart_gammamodes(cart_nints))
  allocate(modekinzpe(cart_nints))
  cart_gssigma(:,:) = 0.0d0
  cart_gstwosigmasq(:,:) = 0.0d0
  kinzpe = 0.0d0 ! Kinetic zero point energy
  modekinzpe(:) = 0.0d0

  ! For dropping basis functions
  if (writeexpect > 0 ) then
     allocate(dhexpect(1))
     allocate(hework(1))
     dhexpect = 0.0d0
  end if

  ! read in the interactions
  do i = 1,cart_nints
     ! read in type and which particles it is between.
     ! line should be gsum but don't check for now.
     ! gsum, particle 1, particle 2, num terms in sum.
     read(fham,*,iostat=ierr) line,cart_part1(i),cart_part2(i),cart_ntint(i)
     if (cart_ntint(i) > cart_maxsgterms) then
        if(isroot)  then
           write(fug,*) "Interaction ",i," has more terms than the specified maximum"
           write(fug,*) "Please increase the maximum in hamiltonian.inpt and try again"
        end if
        stop
     end if
     if (verbosity .ge. 3 .and. isroot) then
        write(fug,*) line,cart_part1(i),cart_part2(i),cart_ntint(i)
     end if
     do j = 1,cart_ntint(i)
        read(fham,*,iostat=ierr) cart_gsamp(j,i),cart_gssigma(j,i)
        cart_gstwosigmasq(j,i) = 2.0d0*cart_gssigma(j,i)**2
        ! convert energy amplitude to internal energy units
        cart_gsamp(j,i) =  cart_gsamp(j,i)/eu
     end do

     ! Calculate gamma and zero point energy for this mode.
     ! hbar^2*gamma/4m
     cart_gammamodes(i) = 1.0d0/(1.0d0/gamma((cart_part1(i)-1)*ncartdim+1)+1.0d0/gamma((cart_part2(i)-1)*ncartdim+1))
     m1 = m((cart_part1(i)-1)*ncartdim+1)
     m2 = m((cart_part2(i)-1)*ncartdim+1)
     redmass = m1*m2/(m1+m2)
     !debug write(fug,*) "Reduced mass",redmass
     modekinzpe(i) = 0.25*cart_gammamodes(i)/redmass
  end do
  
  kinzpe = sum( modekinzpe(:))
 
  ! Now look to see if there are any constraints. 
  read(fham,*,iostat=ierr) line
  if ( line == "constraints") then
     backspace(fham)
     read(fham,*,iostat=ierr) line,ncart_constraint
     if (ierr .ne. 0) then
        write (fug,*) "Error reading the number of constraints in the Hamiltonian file"
        stop
     end if

     if (ncart_constraint > 1)then
        write(fug,*)   "Sorry, but I don't know how to handle more than one constraint"
        flush(fug)
       stop
     end if
     !check to see if this number is greater than 0
     if (ncart_constraint > 0) then
        allocate(constraint_p1(ncart_constraint))
        allocate(constraint_p2(ncart_constraint))
        allocate(constraint_sep(ncart_constraint))
        
        do j = 1,ncart_constraint
           read(fham,*,iostat=ierr) line, constraint_p1(j), constraint_p2(j), constraint_sep(j)
        end do
     end if
  end if

  ! ---- Now we want to summarise ----
  if (isroot) then
     if (hamtype .eq. clascartham) then
        write(fug,*) "Classical Cartesian Hamiltonian Parameters"
        write(fug,*) "=========================================="
     else
        write(fug,*) "Cartesian Hamiltonian Parameters"        
        write(fug,*) "================================"
     end if
     
     write(fug,*) " Each interaction between pairs of particles is expressed as a sum of gaussians centred at the origin"
     
     write(fug,*) "  Masses:               i     m_i"
     do i = 1,ncartdim*nparticles
        write(fug,"(a21,i5,f8.4)") "                     ",i,m(i)
     end do
     write(fug,*)

     do i =  1,cart_nints
        !     write(fug,"(3a,i4,a3,i4)") "   ",cart_part1(i)," - ",cart_part2(i)
        write(fug,*) "   ",cart_part1(i)," - ",cart_part2(i)
        do j = 1,cart_ntint(i)
           write(fug,*) "                ", cart_gsamp(j,i)*eu, cart_gssigma(j,i)
        end do
        write(fug,*) "                Kinetic zero point energy ", modekinzpe(i)*eu,energyunits
        write(fug,*) "                Mode gamma",cart_gammamodes(i)," 1/Angstrom^2"
     end do
     write(fug,*)
     write(fug,*)  "                Total Kinetic zero point energy per basis fct ",kinzpe*eu,energyunits
  end if
  
  if (hamtype .eq. clascartham) then
     write(fug,*)  "   N.B. The zero point energy is not used for the classical cartesian hamiltonian"
  end if
        

  if (ncart_constraint > 0) then
     write(fug,*) 
     write(fug,*) "    Applying constraints"
     write(fug,*) "         constraint     atom1       atom2      separation"
     do j = 1,ncart_constraint        
        write(fug,*) "     ",j,constraint_p1(j), constraint_p2(j), constraint_sep(j)
     end do
  end if
end subroutine readhamcart

end module cartesianham
