! **************************************************
!                   Create Basis 
!
!                  Stewart K. Reed
!                   7th June 2007
!
! Description:
!
!    Some subroutines to create the initial basis
!  and also to project the initial state on to 
!  these basis funcitons
!  The following methods of creating the initial
!  basis functions are supported
!    1) Use a regular grid - currently doesn't 
!       quite work as number of basis functions 
!       should be grid size^ndf
!    2) Pick functions randomly about initial states
!    3) Use the initial states as the basis 
!       functions
! **************************************************
module makebasis

private ::  creategridbasis,  creategaussbasis, createisbasis, createharmoscbasis  
contains

  subroutine createbasis(basismethod) 
    use coherentstatestype
    use files, ONLY:fug,fib,fCDcheck
    use arrays, ONLY: bs,D,zinit,incbf
    use coherentstates, ONLY: ndf,nbasis,pos,mom,calcC
    use options, ONLY:classical
    use scalapackstuff, ONLY: isroot

    ! functions
    use hamiltonian, ONLY: calcD
    implicit none
    integer, intent(in) :: basismethod
    integer :: i,j
    real(kind=8) :: vals(4*ndf+2)
    complex(kind=8),dimension(ndf) :: q,p
    complex(kind=8),dimension(:),allocatable :: Ccheck
    real(kind=8) :: MS ! mean square
    select case(basismethod)
    case(1)
       call creategridbasis
       call projectbasis(zinit(:,1),bs%zb,bs%C)
    case(2)
       call creategaussbasis
       call projectbasis(zinit(:,1),bs%zb,bs%C)
    case(3)
       call createisbasis
       bs%C(:) = (1.0d0,0.0d0)
       !     call projectbasis(zinit(:,1),zb,C)
    case(4)
       call createharmoscbasis
       ! Calculates C for us. 
    case(5) ! Read basis functions in from a file
       call createfilebasis 
       ! c is either calculated or read in.
    case default
       if (isroot) write(fug,*) "Can't create basis as no method has been chosen"
       stop
    end select

    ! Calculate the Ds which go with these Cs. 
    ! S = 0 probably. Is set when arrays are allocated
    ! don't reset here incase modify later to allow restarts!
    !call calcD(C,zb,S,D,incbf) 


    ! Calculate the Ds which go with these Cs. 
    ! If we are doing a quantum calculation
    ! or semiclassical calculation
    if ( classical .le. 1) then
       call calcD(bs,D,incbf) 

       ! Now check by calculating Cs from these Ds
       allocate(Ccheck(nbasis))

       ! call calcC(D,bs%zb,bs%S,Ccheck,incbf)
       call calcC(D,bs,Ccheck,incbf)

       ! now compare the original C and this newly calculated C.
       open(unit=fCDcheck,file="CDcheck.out")
       write(fCDcheck,*) "# i, |Cinit|,  |Cinit-CfromD|,  Re(CfromD),  Im(CfromD)"
       MS = 0.0d0 ! mean square
       do i = 1,nbasis
          write(fCDcheck,*) i,abs(bs%C(i)),abs(Ccheck(i)-bs%C(i)),real(Ccheck(i)),imag(Ccheck(i))
          MS = MS +(Ccheck(i)-bs%C(i))*dimag(Ccheck(i)-bs%C(i))
       end do
       close(fCDcheck)
       deallocate(Ccheck)
       if (isroot) then
          write(fug,*) "Mean square difference of initial and recalculated C", MS/nbasis
       end if
    else
       bs%C(:)=1.0d0/sqrt(dble(nbasis))
       D(:) = bs%C(:)
    end if

    ! write out initial basis
    if (isroot) then
       open(fib,file="initialbasis.out")
       write(fib,*) "# Initial basis"
       write(fib,*) "# Re(z_1),Im(z_1),...Re(z_df),Im(z_df),Re(C),Im(C)"
       do i = 1,nbasis
          call pos(bs%zb(:,i),dconjg(bs%zb(:,i)),q)
          call mom(bs%zb(:,i),dconjg(bs%zb(:,i)), p)

          do j = 1,ndf
             vals(2*j-1) = real(bs%zb(j,i))
             vals(2*j) = dimag(bs%zb(j,i))
             vals(2*j-1+2*ndf) = real(q(j))
             vals(2*j+2*ndf) = real(p(j))
          end do
          vals(4*ndf+1) = real(bs%C(i))
          vals(4*ndf+2) = dimag(bs%C(i))

          write(fib,"(100es18.8)") (vals(j),j=1,4*ndf+2)
       end do
       close(fib)
    end if
  end subroutine createbasis

! ===============================================
!    projectbasis
!
! A subroutine to project all basis functions
! onto an initial state and thus calculate the
! the overlap funciton.
! ===============================================
subroutine projectbasis( z,basis,C)
  use coherentstates, ONLY:ndf,nbasis
  ! functions
  use coherentstates, ONLY:zoverlap
  implicit none
  complex(kind=8), dimension(ndf), intent(in) :: z
  complex(kind=8), dimension(ndf,nbasis), intent(in) :: basis
  complex(kind=8), dimension(nbasis), intent(out) :: C

  integer :: i

  do i = 1,nbasis
     C(i) = zoverlap(basis(:,i),z)
  end do

end subroutine projectbasis


! ===============================================
!    creategridbasis
!
! Create a basis using a regular grid in q and p
! ===============================================
subroutine creategridbasis
  use coherentstatestype
  use coherentstates, ONLY: ndf,nbasis,s2gammarec,shgamma
  use arrays, ONLY: bs,grid,ngrid
  use files, ONLY: fug
  use scalapackstuff, ONLY: isroot
  implicit none
  real(kind=8) :: start(2,ndf),stride(2,ndf)
  real(kind=8) :: p,q
  integer :: i,pi,qi,bf
  real(kind=8) :: x
  ! Check
  if (nbasis .ne. ngrid(1,1)*ngrid(2,1))then
     if(isroot) then
        write(fug,*) " Have the wrong number of basis functions is creategridbasis!"
        write(fug,*) nbasis,ngrid(1,1)*ngrid(2,1)
     end if
     stop
  end if

  ! Work out grid spacing
  ! position
  stride(1,:) = (grid(2,:) - grid(1,:))/dble(ngrid(1,1))
  start(1,:) = grid(1,:) + stride(1,:)/2.0d0
     
  ! momentum
  stride(2,:) = (grid(4,:) - grid(3,:))/dble(ngrid(2,1))
  start(2,:) = grid(3,:) + stride(2,:)/2.0d0
     
  
  bf = 0 ! Count the basis functions
  do qi = 0,ngrid(1,1)-1
     do pi = 0,ngrid(2,1)-1
 	bf = bf + 1
        do i = 1,ndf
           ! Work out position and momentum of this grid point
           q = start(1,i) + qi *stride(1,i)
           p = start(2,i) + pi *stride(2,i)

           ! Now work out z
           bs%zb(i,bf) =  q*shgamma(i)+(0,1)*p*s2gammarec(i)
 
        end do
     end do
  end do

end subroutine creategridbasis

! ===============================================
!    creategaussbasis
!
! Create a random basis with q and p drawn from
! Gaussian distributions around the initial 
! state
! ===============================================
subroutine creategaussbasis
  use coherentstatestype
  use coherentstates, ONlY: ndf,nbasis,s2gammarec,shgamma
  use arrays, ONLY: zinit,ninitstates,bs,gsigma,enmax
  use hamiltonian, ONLY: h_ord
  use files, ONLY: fug,fibe,fibefail
  use scalapackstuff, ONLY: isroot
  implicit none
  real(kind=8) :: p,q,energy
  integer :: bf,j
  integer :: state ! Which of the states are we sampling around?
  integer :: ntries,maxtries ! so we don't sit for every failing to pick basis vectors.
  
  real(kind=8),external :: ZBQLNOR

  if(isroot) then
     open(fibe,file="ibe.out")   
     open(fibefail,file="ibefail.out")   
     write(fibe,*) "# Initial basis energies"
     write(fug,*) "Creating basis states"
     write(fug,*) "---------------------"
     write(fug,*) "  Randomly distributed from a Gaussian distribution about" 
     write(fug,*) "  the initial state"
     write(fug,*) "  Maximum allowable energy ",enmax
     write(fug,*) "  ",nbasis," functions"
     write(fug,*) 
  end if
  bf = 1
  ntries = 1 ! How many numbers have we picked.
  maxtries = 50 * nbasis
  do while (bf .le. nbasis .and. ntries .le. maxtries)
     !which state are we sampling around?
     state = floor(dble(ninitstates)*dble(bf-1)/dble(nbasis))+1
     do j = 1,ndf
        ! position
!        q = gaussrnd(real(zinit(j,state)),gsigma(1,j))
!        q = gaussrnd(0.0d0,gsigma(1,j))
         q = ZBQLNOR(0.0d0,gsigma(1,j))
        ! momentum
!        p = gaussrnd(dimag(zinit(j,state)),gsigma(2,j))
!        p = gaussrnd(0.0d0,gsigma(2,j))
        p = ZBQLNOR(0.0d0,gsigma(2,j))

        ! Now work out z
!        zb(j,bf) =  q*shgamma(j)+(0,1)*p*s2gammarec(j)
        bs%zb(j,bf) =  q*shgamma(j)+(0,1)*p*s2gammarec(j)+zinit(j,state)
     end do

     !  Include a cutoff?
     ! Calculate energy of this basis function
     energy = real(h_ord (dconjg(bs%zb(:,bf)),bs%zb(:,bf)))
     if ( energy < enmax) then
        write(fibe,*) energy,bs%zb(:,bf)
        bf = bf + 1
     else     
        write(fibefail,*) "not",energy,bs%zb(:,bf)
             
     end if
     ntries = ntries + 1
  end do
  
  ! Check to see if we have reached the max allowable number of basis vectors. 
  ! This is to catch cases when can't choose enough basis vectors within the given constraints.

  if (ntries .ge. maxtries) then
     write(fug,*) "  Chosen ",ntries," basis functions but only accepted ",bf
     write(fug,*) "  Increase distribution width and/or energy cutoff then try again."
     stop
  end if

  close(fibe)
  close(fibefail)

end subroutine creategaussbasis


! ===============================================
!    gaussrnd
!
! A function to calculate a random number from
! a Gaussian distribution.
! ===============================================
function gaussrnd(xbar,sigma)
  implicit none
  real(kind=8),intent(in) :: xbar,sigma
  real(kind=8) :: gaussrnd
  real(kind=8) :: x
  real(kind=8),parameter :: pi=3.1415926535897932385

  real(kind=8),external :: zbqlu01

  ! Generate a random number in the range (0:1)
!  call random_number(x)
  x=ZBQLU01(0.0)
  !  expand so that it is in the range -1:1
  !  if x < 0 then give a random number < xbar
  x = x*2.0-1.0
  
  if ( x > 0) then
  !G(x) = exp(-(x-xbar)^2/(2*sigma2))/(sqrt(2*pi)*sigma)
     gaussrnd = xbar+dsqrt(-2*sigma**2*dlog(x*sigma*sqrt(2*pi)))
  else
     gaussrnd = xbar-dsqrt(-2*sigma**2*dlog(-x*sigma*sqrt(2*pi)))
  end if

  write(65,*) gaussrnd
end function gaussrnd

  
! ===============================================
!    createisbasis
!
! Use the states given in states.inpt as the
! basis.
! ===============================================
subroutine createisbasis
!  use arrays, ONLY: 
  use coherentstatestype
  use coherentstates, ONlY: ndf,nbasis
  use arrays, ONLY: ninitstates,zinit,bs
  use files, ONLY: fug
  use scalapackstuff, ONLY: isroot
  implicit none

  ! There are several things which need redefining.
  if (nbasis .ne. ninitstates) then
     write(fug,*) "Requested initial states as basis functions however &
     the number of basis functions in run.inpt does not equal the number of initial states"
     stop
  end if

  ! Check the number of basis functions
  if (nbasis .ne. ninitstates) then
     write(*,*) "Fatal error: nbasis is note equal to ninitstates in createisbasis"
     stop
!     if (allocated(zb)) deallocate(zb)
!     nbasis = ninitstates
!     allocate(zb(ndf,nbasis))
  end if
  
  if(isroot) then
     write(fug,*) "Creating basis states"
     write(fug,*) "---------------------"
     write(fug,*) "  Using ",ninitstates," initial states as basis functions"
     write(fug,*) 
  end if
  ! Copy the initial states to the basis functions
  bs%zb(:,:) = zinit(:,:)

end subroutine createisbasis


! ===============================================
!    createharmoscbasis
!
! Create a random basis with q and p drawn from
! the Harmonic oscillator first excited state
! or set equal to zero.
! ===============================================
subroutine createharmoscbasis
  use coherentstatestype
  use coherentstates, ONlY: ndf,nbasis,gamma
  use arrays, ONLY: bs,vstate,probscale,zmax,maxtries,harmmethod,harmComp,zoffset,ninitstates
  use files, ONLY: fug,fib

  implicit none

  integer :: ntries
  integer :: i,bf
  integer :: superid
  logical :: accept
  real(kind=8) :: v,prob

  complex(kind=8) :: Cmode ! Contribution to the amplitude of a particular point

  ! Check to make sure that there are enough intial states
  checkinitstates: do i = 1,ndf
     if (harmmethod(i) .eq. 3 .or. harmmethod(i) .eq. 4) then
        if (nbasis .ne. ninitstates) then
           write(fug,*) "Requested initial states as basis functions however &
                the number of basis functions in run.inpt does not equal the number of initial states"
           stop
        else ! nbasis .eq. ninitstates so exit this loop.
           exit checkinitstates 
        end if
     end if
  end do checkinitstates
     
  write(fug,*) 
  write(fug,*) "Initial Basis functions"
  write(fug,*) "-----------------------"
  write(fug,*) "   From harmonic oscillator wavefunction"
  write(fug,*) "     z-max: ",zmax
!  write(fug,*) "     probscale: ",probscale
  write(fug,*) "     maxtries per mode: ",maxtries
  write(fug,*)
  if (allocated(zoffset)) then
     write(fug,*) "     Vibrational states of each mode"
     write(fug,*) "      mode, state,compression,offset"
     do i = 1,ndf
        if (vstate(i) .eq. 0) then
           write(fug,*) "     ",i,vstate(i), harmComp(i),zoffset(i)
        else
           write(fug,*) "     ",i,vstate(i), 1.0d0/harmComp(i),zoffset(i)
        end if
     end do
  else
     write(fug,*) "     Vibrational states of each mode"
     write(fug,*) "      mode, state,compression"
     do i = 1,ndf
        if (vstate(i) .eq. 0) then
           write(fug,*) "     ",i,vstate(i), harmComp(i)
        else
           write(fug,*) "     ",i,vstate(i), 1.0d0/harmComp(i)
        end if
     end do
  end if

  open(fib,file="initial-zbasis.out")
  write(fib,*) "# Initial basis in phase space"
  write(fib,*) "# basisf, degfredom, Re(z), Im(z)"
  do bf = 1,nbasis

     ! Set the amplitude of this basis function
     bs%C(bf) = (1.0d0,0.0d0)

     superid = 0
     ! Loop over the degrees of freedom
     do i=1,ndf
 

!        if (vstate(i) .ge. 0 ) then
        if (harmmethod(i) .eq. 1) then
           ! Pick a value of z
           call  pickharmrnd(zmax,bs%zb(i,bf),vstate(i),harmComp(i),Cmode)

        !else if (vstate(i) .lt. 0 .and. vstate(i) .gt. -100) then
        else if (harmmethod(i) .eq. 2) then
           ! Pick a value of z
!           call  pickharmrnd(zmax,zb(i,bf),-vstate(i)-1,harmComp(i),Cmode)
           call  pickharmrnd(zmax,bs%zb(i,bf),vstate(i),harmComp(i),Cmode)

           ! Displace basis function
           bs%zb(i,bf) = bs%zb(i,bf) + zoffset(i)

        else if (harmmethod(i) .eq. 3 ) then

           call harmuseinitial(bs%zb(i,bf),i,bf,vstate(i),Cmode, .false.,(0.0d0,0.0d0))

        else if (harmmethod(i) .eq.4) then
           call harmuseinitial(bs%zb(i,bf),i,bf,vstate(i),Cmode, .true., zoffset(i))

           ! Choose z at random from a wavefn which is a superposition of others.
        else if (harmmethod(i) .eq. 5) then
           superid = superid + 1
           call pickharmrndsuper(bs%zb(i,bf),i,superid,Cmode)

        else if (harmmethod(i) .eq.7) then
           superid = superid + 1
           call harmuseinitialsuper(bs%zb(i,bf),i,bf,superid,Cmode,.false.,(0.0d0,0.0d0))

        else if (harmmethod(i) .eq.8) then
           superid = superid + 1
           call harmuseinitialsuper(bs%zb(i,bf),i,bf,superid,Cmode, .true., zoffset(i))

!        else if (vstate(i) .eq. -100) then                     
        else if (harmmethod(i) .eq. 0) then
           bs%zb(i,bf) = (0.0d0,0.0d0)
           Cmode =  (1.0d0,0.0d0)
        else
           write(fug,*) "Ooops, you seem to have a basis set method which I don't know how to deal with: ",harmmethod(i)
        end if

        ! Calculate the new coefficient for this whole basis fn.
        bs%C(bf) = Cmode*bs%C(bf)

        ! write out some info to a file.
        write(fib,*) bf,i,real(bs%zb(i,bf)),dimag(bs%zb(i,bf))
     end do

  end do

  close(fib)
end subroutine createharmoscbasis


! ===============================================
!                pickharmrnd
!  subroutine to give a random point and its 
!  probability from the harmonic oscillator 
!  wavefunctions. 
!  The point should be drawn from the correct
!  probability distributions.
!
!  Would use the inverse fct so that the point 
!  was actually drawn from the correct prob. 
!  distribution but the inverse function is not
!  simple therefore need to do it this way and
!  use Monte Carlo sampling.
! 
! A compression factor may be applied to z so
! that z = z/comp. This however only really makes
! sense the groundstate wavefct which has its
! max at z=0. 
! ===============================================
subroutine pickharmrnd(zmax,z,n,comp,ampC)
  use arrays, ONLY: maxtries
  use files, ONLY: fug
  implicit none
  real(kind=8),intent(in) :: zmax
  integer, intent(in) :: n
  ! Compression factor
  real(kind=8), intent(in) :: comp 
  complex(kind=8),intent(out) :: z
  ! amplitude i.e. <z|Phi_n>
  complex(kind=8),intent(out) :: ampC 
  ! vibrational state 0,1,....

  ! Maximum probability
  real(kind = 8) :: pmax,rmax,r0
  integer:: ntries
  logical :: accept
  real(kind=8) :: prob,pr

  integer :: ifact ! for factorial loop
  real(kind=8) :: factn ! n factorial

  real(kind=8),external :: zbqlu01

  ! Make sure n is positive
  if( n < 0) then
      write(fug,*) "Sorry but I don't know the wavefunction for energy level v=",n
      stop
   end if
   
  ! Set maximum probability depending on which mode we are looking at.
  select case(n)
  case(0)
     pmax = 0.1365173622
  case(1)
!     pmax = 0.1304804041
!     rmax =1.22474487139159 ! dsqrt(1.5)
!     r0   = dsqrt(6.0d0)*(1-comp**2)/2.0d0
     rmax = 1.0d0
     r0 = (1-comp**2)
          
     pmax = rmax**3*exp(-(rmax-r0)**2/comp**2)/3.14159265358979
  case default

!     rmax = dsqrt(dble(n)+0.5)
     rmax = dsqrt(dble(n))
     r0 = rmax*(1-comp**2)
     factn=1.0
     do ifact = 2,n
        factn = factn*ifact
     end do
     pmax = rmax**(2*n+1)*exp(-(rmax-r0)**2/comp**2)/3.14159265358979d0/factn


  end select

  ! Initially we haven't tried to choose a fn nor have
  ! we succeeded
  ntries = 0
  accept = .false.
  
  !Keep choosing z's at random until the probability of it
  ! is greater than  pr
  do while ( .not.accept .and. ntries .le. maxtries)
     ntries = ntries + 1

     ! Choose z from appropriate wavefunction
     select case(n)
     case(0)
        call harm0rnd(zmax,z,prob)    
        ! compress z
        z = z/comp

     case(1)
        call harm1rnd(zmax,z,comp,prob) 

     case default
         call harmnrnd(zmax,z,n,comp,prob)

     end select

     ! now see whether or not to accept it.
     !call random_number(pr)
     pr = ZBQLU01(0.0)

     if (pr .le. prob/pmax) then
        accept = .true.
     end if
  end do

  ! We have not chosen an acceptable z within the maximum number of tries, 
  ! stop!
  if (.not.accept) then
     write(fug,*) " Failed to choose a satisfactory value for z within ", &
          maxtries, " tries."
     
!          maxtries, " tries. Try increasing probscale or decreasing rmax"
     stop
  end if


  ! Now we should calculate the amplitude for this mode/fn
   select case(n)
   case(0)

      ampC = exp(-0.5*(abs(z)**2))
   case(1)
      ampC = dconjg(z)*exp(-0.5*(abs(z)**2))
      
   case default
      ampC = dconjg(z)**(n)*exp(-0.5*(abs(z)**2))/sqrt(factn)
!       write(fug,*) "Sorry but I don't know the wavefunction for energy level v=",n
!     stop
  end select

end subroutine pickharmrnd
 
! ===============================================
!                harm0rnd
!  subroutine to give a random point and its 
!  probability from |0> of the harmonic 
!  oscillator. 
! ===============================================
subroutine harm0rnd(rmax,z,prob)

  implicit none
  real(kind=8),intent(in) :: rmax
  complex(kind=8),intent(out) :: z
  real(kind=8),intent(out) :: prob
  
  real(kind=8) :: r,theta

  real(kind=8),external :: zbqlu01

  ! Choose random numbers.
  !call random_number(r)
  r= ZBQLU01(0.0)
  r = r*rmax
  
  !call random_number(theta)
  theta= ZBQLU01(0.0)
  theta = theta*3.14159265358979*2.0
  
  ! Calculate z
  z = r*cos(theta) + r*(0,1)*sin(theta)
  
  ! Calculate probability
  !   prob = z*dconjg(z)*exp(-z*dconjg(z))/3.14159265358979
  ! Probability in polar coords is 
  ! r*exp(-r^2)
  !prob = cdabs(z)*exp(-cdabs(z)**2)/3.14159265358979
  prob = r*exp(-r**2)/3.14159265358979

end subroutine harm0rnd

! ===============================================
!                harm1rnd
!  subroutine to give a random point and its 
!  probability from |1> of the harmonic 
!  oscillator. 
! ===============================================
subroutine harm1rnd(rmax,z,sigma,prob)
  implicit none
   real(kind=8),intent(in) :: rmax
   complex(kind=8),intent(out) :: z
   real(kind=8),intent(in) :: sigma
   real(kind=8),intent(out) :: prob
  
  real(kind=8) :: r,theta,r0

  real(kind=8),external :: zbqlu01
  
  ! Choose random numbers.
  !  call random_number(r)
  r= ZBQLU01(0.0)
  r = r*rmax
  
  ! call random_number(theta)
  theta= ZBQLU01(0.0)
  theta = theta*3.14159265358979d0*2.0d0
  
  ! Calculate z
  z = r*cos(theta) + r*(0,1)*sin(theta)
  
  ! Calculate probability
  !   prob = z*dconjg(z)*exp(-z*dconjg(z))/3.14159265358979
  ! Probability in polar coords is 
  ! r^3*exp(-r^2)
  ! prob = (cdabs(z)**3)*exp(-cdabs(z)**2)/3.14159265358979
  
  ! probability, normalised only if sigma = 1
  r0   = (1-sigma**2)
  prob = r**3*exp(-(r-r0)**2/sigma**2)/3.14159265358979

end subroutine harm1rnd


! ===============================================
!                harmnrnd
!  subroutine to give a random point and its 
!  probability from |n> of the harmonic 
!  oscillator. 
! ===============================================
 subroutine harmnrnd(rmax,z,n,sigma,prob)
!For testing   use coherentstates, ONLY: nbasis

   implicit none
   real(kind=8),intent(in) :: rmax
   complex(kind=8),intent(out) :: z
   integer, intent(in) :: n
   real(kind=8),intent(in) :: sigma
   real(kind=8),intent(out) :: prob
  
  real(kind=8) :: r,theta,r0
  real(kind=8) :: factn
  integer :: i

  real(kind=8),external :: zbqlu01

  !For testing  integer,save ::nbf=0

  ! Choose random numbers.
!   call random_number(r)
   r = ZBQLU01(0.0)
   r = r*rmax

   ! For testing choose r = r_maxprob
!   r =  dsqrt(dble(n)+0.5)

!   call random_number(theta)
   theta= ZBQLU01(0.0)
   theta = theta*3.14159265358979*2.0

   ! For testing evenly distribution in theta
   !   theta = 3.14159265358979d0*2.0d0/nbasis
   !   nbf = nbf + 1

   ! Calculate z
   z = r*cos(theta) + r*(0.0d0,1.0d0)*sin(theta)

   ! Calculate probability
   !   prob = z*dconjg(z)*exp(-z*dconjg(z))/3.14159265358979
   ! Probability in polar coords is 
   ! r^3*exp(-r^2)
   ! prob = (cdabs(z)**3)*exp(-cdabs(z)**2)/3.14159265358979

   ! probability, normalised only if sigma = 1
   !!r0   = dsqrt(dble(n)+0.5)*(1-sigma**2)
   r0 = (1.0d0-sigma**2)*dsqrt(dble(n))

   ! Calculate n factorial
   factn=1.0d0
   do i = 2,n
      factn = factn*i
   end do
   
   prob = r**(2*n+1)*exp(-(r-r0)**2/sigma**2)/3.14159265358979/factn
 
 end subroutine harmnrnd


 
! ===============================================
!                harmuseinitial
!  subroutine to use basis functions from the 
!  intial states and project it on to state 
!  |n> of the harmonic oscillator. 
!
!  if the harmonic mode is offset then 
!   "unoffset" the initial basis function for 
!  projection. We assume that basis functions 
!  in the intial states are already offset.
! ===============================================
 subroutine harmuseinitial(z,i,bf,n,ampC,offset,zo)
   
   use arrays, ONLY: zinit

   implicit none

   complex(kind=8),intent(out) :: z
   integer, intent(in) :: i
   integer, intent(in) :: bf
   integer, intent(in) :: n
  ! amplitude i.e. <z|Phi_n>
   complex(kind=8),intent(out) :: ampC 

   logical, intent(in) :: offset
   complex(kind=8),intent(in) :: zo
   
   integer :: ifact ! for factorial loop
   real(kind=8) :: factn ! n factorial


   ! Get basis function from intial states
   z = zinit(i,bf)

   ! Is the the wavefn offset?   
   if (offset) then
      z = z-zo
   end if

   ! Now we should calculate the amplitude for this mode/fn
   select case(n)
   case(0)
      
      ampC = dexp(-0.5*(cdabs(z)**2))
!!   case(1)
!!      ampC = dconjg(z)*dexp(-0.5*(cdabs(z)**2))
!!      ampC = dconjg(z)*exp(-0.5d0*(abs(z)**2))       
   case default
      ! Calculate n factorial
      factn=1.0d0
      do ifact = 2,n
         factn = factn*ifact
      end do

      ! Now the projection
      ampC = dconjg(z)**(n)*dexp(-0.5*(cdabs(z)**2))/dsqrt(factn)
  end select

  ! Is the the wavefn offset? If so reassign it.   
   if (offset) then
      z = zinit(i,bf) 
   end if

end subroutine harmuseinitial


! ===============================================
!                pickharmrndsuper
!  Pick z randomly from a wave fn which is a 
!  superposition of harmonic oscillator modes.
! ===============================================
subroutine pickharmrndsuper(z,i,sid,ampC)
  use arrays, ONLY: superphi,nsuperbasis,superqmin,superqmax,superpmin,superpmax, &
       supermaxprob,maxtries

  use coherentstates, ONLY: shgamma,s2gammarec
  use files, ONLY:fug
  
  implicit none
  complex(kind=8),intent(out) :: z
  integer, intent(in) :: i ! which mode
  integer, intent(in) :: sid ! which superposition

  ! amplitude i.e. <z|Phi_n>
  complex(kind=8),intent(out) :: ampC
  real(kind=8) :: factn ! n factorial
  integer :: ntries
  logical :: accept

  real(kind=8) :: q,p

  real(kind=8) ::prob,pr
  integer :: j

  real(kind=8),external :: zbqlu01

  complex(kind=8) :: a
   ! Initially we haven't tried to choose a fn nor have
   ! we succeeded
   ntries = 0
   accept = .false.
   

   do while ( .not.accept .and. ntries .le. maxtries)
      ntries = ntries + 1

      ! choose z initially interms of q and p
      q = ZBQLU01(0.0)*(superqmax(sid)-superqmin(sid))+superqmin(sid)
      p = ZBQLU01(0.0)*(superpmax(sid)-superpmin(sid))+superpmin(sid)
      z = q*shgamma(i)+(0,1.0d0)*p*s2gammarec(i)

      ! now work out C
      ! Now we should calculate the amplitude for this mode/fn
      ! multiply by dexp at the end.
   
      ampC = (1.0d0,0)*superphi(1,sid) !*dexp(-0.5*(cdabs(z)**2))

      if (nsuperbasis(sid) .ge. 2) then
         a = (1.0d0,0)*superphi(2,sid)
         ampC =  ampC + a * dconjg(z) !*dexp(-0.5*(cdabs(z)**2))
         
         ! Calculate the factorial
         factn=1.0d0
         
         do j = 2,nsuperbasis(sid)-1
            factn = factn*dble(j)
            
            ! Now the projection
            a = (1.0d0,0) * superphi(j+1,sid)
            ampC =  ampC + a*dconjg(z)**(dble(j))/dsqrt(factn)
         !/3.14159265358979d0
         ! * dexp(-0.5*(cdabs(z)**2))
         end do
      end if

      ! Now multiply by 
      ampC = ampC * dexp(-0.5*(cdabs(z)**2))

      ! Now calculate the probability.
      prob = (cdabs(ampC))**2

        ! now see whether or not to accept it. 

      !call random_number(pr)

      pr = ZBQLU01(0.0)
      
      if (pr .le. prob/supermaxprob(sid)) then
         accept = .true.
      end if
   end do

  ! We have not chosen an acceptable z within the maximum number of tries,
  ! stop!
  if (.not.accept) then
     write(fug,*) " Failed to choose a satisfactory value for z within ", &
          maxtries, " tries."
!          maxtries, " tries. Try increasing probscale or decreasing rmax" 
     stop
  end if

end subroutine pickharmrndsuper
! ===============================================
!                harmuseinitialsuper
!  subroutine to use basis functions from the 
!  intial states and project it on to state 
!  |n> of the harmonic oscillator. 
!
!  if the harmonic mode is offset then 
!   "unoffset" the initial basis function for 
!  projection. We assume that basis functions 
!  in the intial states are already offset.
! ===============================================
 subroutine harmuseinitialsuper(z,i,bf,sid,ampC,offset,zo)
   
   use arrays, ONLY: zinit,superphi,nsuperbasis

   implicit none

   complex(kind=8),intent(out) :: z
   integer, intent(in) :: i
   integer, intent(in) :: bf
   integer, intent(in) :: sid ! which superposition
  ! amplitude i.e. <z|Phi_n>
   complex(kind=8),intent(out) :: ampC 

   logical, intent(in) :: offset
   complex(kind=8),intent(in) :: zo
   
   real(kind=8) :: factn ! n factorial

   integer :: j ! for loop

   ! Get basis function from intial states
   z = zinit(i,bf)

   ! Is the the wavefn offset?   
   if (offset) then
      z = z-zo
   end if
   
   ! Now we should calculate the amplitude for this mode/fn
   ! multiply by dexp at the end.
   ampC = superphi(1,sid) !*dexp(-0.5*(cdabs(z)**2))

   if (nsuperbasis(sid) .ge. 2) then
      ampC =  ampC + superphi(2,sid) * dconjg(z) !*dexp(-0.5*(cdabs(z)**2))
      

      ! The factorial
      factn = 1.0d0

      do j = 2,nsuperbasis(sid)-1
         factn = factn*dble(j)
            
         ! Now the projection
         ampC =  ampC + superphi(j+1,sid) *dconjg(z)**(j)/dsqrt(factn)
         !/3.14159265358979d0
         ! * dexp(-0.5*(cdabs(z)**2))
         end do
      end if

      ! Now multiply by 
      ampC = ampC * dexp(-0.5d0*(cdabs(z)**2))

     
      ! Is the the wavefn offset? If so reassign it.   
      if (offset) then
         z = zinit(i,bf) 
      end if

    end subroutine harmuseinitialsuper

    ! ===============================================
    !                createfilebasis
    !  subroutine to read basis functions in from
    !  a file.
    ! ===============================================
    subroutine createfilebasis
      use coherentstatestype
      use arrays, ONLY: bs
      use coherentstates, ONLY: ndf,nbasis,s2gammarec,shgamma
      use files, ONLY: fug,fbasisin
      use errors
      implicit none
      
      integer ::ierr ! for file errors
      character(len=50) :: filename
      character(len=50) :: line ! for reading in stuff from files
      integer :: nbf ! number of basis functions in the file.
      integer :: ampsinc ! Are the amplitudes included?
                         ! 0, no, 1, yes, 2 set to 1.
      logical :: findoption ! used for identifying options at top of file.
      integer :: i,j       ! for loops

      real(kind=8)q,p,ra,ia  ! position, momentum, real and imag parts of amp.


      ! ---- Name of input file ----
      filename="basis.inpt"
      ! ---- Open file to read basis funcitons ----
      open(fbasisin,file=filename,iostat=ierr,status='old')
      if (ierr .ne. 0) then
         call fileerror(ierr,filename)
      end if

      ! ---- now read in the first line ----
      read(fbasisin,*,iostat=ierr) line
      if  (line == "nbasis" ) then
         backspace(fbasisin) 
         read(fbasisin,*,iostat=ierr) line,nbf
         if(ierr.ne.0) call filereaderror(ierr,filename,"Error reading number of states")
      else
         write(fug,*) "Fatal error: The first line of "//filename//" should be nbasis but it is not."
         stop
      end if

      ! ---- Check number of basis functions ----
      if (nbf .ne. nbasis) then
         write(fug,*) "Number of basis functions in basis.inpt file is not the same as the number specified in run.inpt"
         stop
      end if
      
      ! ---- now check for options ----
      ampsinc = 2 ! By default, amplitudes are not included. Set so that they are all equal. I.e. using unitamp
      findoption = .true.
      do while (findoption)
         read(fbasisin,*,iostat=ierr) line
         findoption = .false.
         if (line == "amplitudes" ) then
            ampsinc = 1
            findoption = .true.
         else if (line == "unitamp") then
            ampsinc = 2
            findoption = .true.
         end if
      end do
      backspace(fbasisin)

      ! ---- Now read in basis functions ----
      do i = 1,nbasis
         read(fbasisin,*,iostat=ierr) line
         if (line == "basis" ) then
            ! check to see if the amplitude has been given and read it if it has.
            ! ignore it if not required.
            read(fbasisin,*,iostat=ierr) line 
            if (line == "amp" ) then
               backspace(fbasisin)
               read(fbasisin,*,iostat=ierr) line,ra,ia
               if (ierr.ne.0) then
                  write(fug,*) "basis function ",i
                  call filereaderror(ierr,filename,"Problem reading amplitude in basis.inpt")
               end if
               if (ampsinc .eq. 1) then
                  bs%C(i) = ra+(0,1.0d0)*ia
               end if
            else if (ampsinc .eq. 1) then
               write(fug,*) "basis function ",i
               call filereaderror(ierr,filename,"Amplitude is missing in basis.inpt")
            else if (ierr .ne. 0 ) then
               call filereaderror(ierr,filename,"Error reading basis functions")
            else
               backspace(fbasisin)
            end if
            
            ! unit amplitude            
            if (ampsinc .eq. 2) then
               bs%C(i) = (1.0d0,0.0d0)
            end if

            ! now read in the positions and momenta
            do j = 1,ndf
               read(fbasisin,*,iostat=ierr) q,p
               if (ierr .ne. 0 ) then
                  write(fug,*) "basis function ",i," degree of freedom ",j
                  call filereaderror(ierr,filename,"Error reading basis functions")
               end if
               ! convert q,p to z
               bs%zb(j,i) = q*shgamma(j)+(0,1.0d0)*p*s2gammarec(j)
            end do
         end if
      end do

    end subroutine createfilebasis
end module makebasis
  
