! **************************************************
!            Coherent States Module
!
!                  Stewart K. Reed
!                   4th June 2007
!
!  Description:
!      This module contains coherent states data
!      and functions for their manipulation
!
! Functions/subroutines include
!  overlap function
!  pos and momentum
!  allocatecsarrays & deallocatecsarrays
! **************************************************
module coherentstates

  implicit none

  ! Initial condition
  integer :: ndf     ! number of degrees of freedom 
  integer :: nbasis  ! number of basis functions
  integer :: nehernfest ! number of ernfest 
  ! Gamma for coherent states, dimension ndf
  real(kind=8),dimension(:),allocatable :: gamma 
  real(kind=8),dimension(:),allocatable :: s2gammarec ! =1/sqrt(2*gamma)
  real(kind=8),dimension(:),allocatable :: shgamma ! = sqrt(gamma/2)
  real(kind=8),dimension(:),allocatable :: sgamma ! = sqrt(gamma)


  real(kind=8) :: overlapTime

contains

  ! ==================================================
  !              Overlap
  ! ================================================== 
  ! Calculate the overlap of two coherent states zi and zj
  ! The overlap matrix is given be
  ! <zi|zj>=int(<zi|x><x|zj> dx = exp(zi^* zj - |zi|^2/2 - |zj|^2/2)
  ! where |zi> is given by the product of the states in each degree
  ! of freedom.
  ! For ehrenfest, the overlap is <z_i|z_j> sum(conjg(a_i)*a_j)
  ! efest is in an integer and it would be useful to set it to the number of ehrenfest degrees of freedom.
  function overlap(zi,zj,ai,aj,Si,Sj,efest)
    use scalapackstuff, ONLY: MPI_Wtime
    implicit none

    complex(kind=8),dimension(ndf), intent(in) :: zi,zj   
    complex(kind=8),dimension(ndf), intent(in) :: ai,aj
    real(kind=8),dimension(:),intent(in) :: Si,Sj
    integer, intent(in) :: efest
    complex(kind=8) :: overlap
    
    integer :: df  ! degree of freedom for loop
    real(kind=8) :: intime
    complex(kind=8) :: exponent
    
    complex(kind=8) :: efestoverlap
    intime = MPI_Wtime()
    !overlap = (1.0d0,0.0d0) 
    exponent = (0.0d0,0.0d0)
    do df = 1,ndf
!       overlap = overlap * &
!            cdexp(dconjg(zi(df))*zj(df) &
!            - cdabs(zi(df))**2/2 -cdabs(zj(df))**2/2)
       exponent = exponent + dconjg(zi(df))*zj(df) - cdabs(zi(df))**2/2 -cdabs(zj(df))**2/2
    end do
    overlap = cdexp(exponent)


    if(efest > 1) then
       overlap = overlap * sum(conjg(ai(:))*aj(:)*exp((0.0d0,1.0d0)*(Sj(:)-Si(:))))
    end if

    overlapTime = overlapTime + MPI_Wtime() - intime
  end function overlap

    function zoverlap(zi,zj)
    use scalapackstuff, ONLY: MPI_Wtime
    implicit none

    complex(kind=8),dimension(ndf), intent(in) :: zi,zj   
    complex(kind=8) :: zoverlap
    
    integer :: df  ! degree of freedom for loop
    real(kind=8) :: intime
    complex(kind=8) :: exponent
    
    intime = MPI_Wtime()
    !overlap = (1.0d0,0.0d0) 
    exponent = (0.0d0,0.0d0)
    do df = 1,ndf
!       overlap = overlap * &
!            cdexp(dconjg(zi(df))*zj(df) &
!            - cdabs(zi(df))**2/2 -cdabs(zj(df))**2/2)
       exponent = exponent + dconjg(zi(df))*zj(df) - cdabs(zi(df))**2/2 -cdabs(zj(df))**2/2
    end do
    zoverlap = cdexp(exponent)

    overlapTime = overlapTime + MPI_Wtime() - intime
  end function zoverlap

  ! ==================================================
  !              Overlap_bs
  ! ================================================== 
  ! Overlap of two basis functions within a basis set
  function overlap_bs(bs,i,j)
    use scalapackstuff, ONLY: MPI_Wtime
    use coherentstatestype
    implicit none
    type(coherentstatebasisset),intent(in) :: bs
    integer :: i,j
    complex(kind=8) :: overlap_bs
    
    integer :: df  ! degree of freedom for loop
    real(kind=8) :: intime
    complex(kind=8) :: exponent
    
    complex(kind=8) :: efestoverlap
    intime = MPI_Wtime()
    !overlap = (1.0d0,0.0d0) 
    exponent = (0.0d0,0.0d0)
    do df = 1,ndf
!       exponent = exponent + dconjg(bs%zi(df,i))*bs%zj(df,j) - cdabs(bs%zi(df,i))**2/2 -cdabs(bs%zj(df,j))**2/2
       exponent = exponent + dconjg(bs%zb(df,i))*bs%zb(df,j) &
            - 0.5d0*((bs%zb(df,i))*conjg(bs%zb(df,i)) + (bs%zb(df,j))*conjg(bs%zb(df,j)))
    end do
    overlap_bs = cdexp(exponent)

    if(bs%nehrenfest > 1) then
       overlap_bs = overlap_bs * sum(conjg(bs%ErnC(:,i))*bs%ErnC(:,j))
    end if

    overlapTime = overlapTime + MPI_Wtime() - intime
  end function overlap_bs


  ! ==================================================
  !              position, q
  ! ================================================== 
  ! Calculate the position of a coherent state
  subroutine pos(zi,zjc,q)
    implicit none
    complex(kind=8),dimension(ndf), intent(in) :: zi,zjc
    complex(kind=8),dimension(ndf), intent(out) :: q

    integer :: i ! for loops

    do i = 1,ndf
       q(i) = (zi(i)+zjc(i))*s2gammarec(i)
    end do

  end subroutine pos

  ! ==================================================
  !              momentum, p
  ! ================================================== 
  ! Calculate the momentum of a coherent state
  subroutine mom(zi,zjc,p)
    implicit none
    complex(kind=8),dimension(ndf), intent(in) :: zi,zjc
    complex(kind=8),dimension(ndf), intent(out) :: p

    integer :: i ! for loops

    do i = 1,ndf
       p(i) = (0d0,-1d0)*(zi(i)-zjc(i))*shgamma(i)
    end do

  end subroutine mom

  ! ==================================================
  !            norm
  ! ================================================== 
  ! Calculate the norm of a set of basis functions
  function norm(C,D)
    implicit none
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis 
    complex(kind=8),dimension(nbasis),intent(in) :: C,D

    real(kind=8) :: norm

    norm =0.0d0

    norm = sum((dconjg(C(:))*D(:)))

  end function norm

  ! ==================================================
  !            calcC
  !  A subroutine to calculate C from D
  !  It is mainly used for testing so there is an 
  !  option to give an array for C.
  ! 
  ! ================================================== 
  !subroutine calcC(D,zb,S,C,incbf)
  subroutine calcC(D,bs,C,incbf)
    use coherentstatestype
    implicit none
    ! ---- interface ----
    complex(kind=8),dimension(nbasis),intent(in) :: D
    type(coherentstatebasisset),intent(inout) :: bs
!    real(kind=8),dimension(nbasis),intent(in) :: S ! action
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: zb ! basis states
    complex(kind=8),dimension(nbasis),intent(out),optional :: C
    
    logical,dimension(nbasis),intent(in) :: incbf

    complex(kind=8),dimension(:,:),allocatable :: olpm
    ! ---- local ----
    integer:: i,j
    real(kind=8) :: omega0 = 1d-10

    allocate(olpm(nbasis,nbasis))
    if (allocated(olpm)) then 
       ! Calculate overlap matrix*exp(i(Sj-Si))
       olpm(:,:) = (0.0,0.0)
       do i = 1,nbasis
          olpm(i,i) = (1.0d0,0.0d0)
       end do
       
       do i = 1,nbasis-1
          if (incbf(i)) then
             do j = i+1,nbasis
                if(incbf(j)) then
                   olpm(i,j) = overlap_bs(bs,i,j)
                   if (abs(olpm(i,j)) .ge. omega0) then
                      ! What should go here for S??????
                      if(bs%nehrenfest .le. 1) then  
                         ! If we are not doing Ehrenfest, multiply by exp(i dS). 
                         ! If we are doing Ehrenfest, this factor forms part of the overlap
                         olpm(i,j) = olpm(i,j)*cdexp((0.0,1.0)*(bs%S(1,j)-bs%S(1,i)))
                      end if
                      olpm(j,i) = dconjg(olpm(i,j))
                   else
                      olpm(i,j) = 0.0d0
                      olpm(j,i) = 0.0d0
                   end if
                end if
             end do
          end if
       end do

       ! Now calculate C * exp(iS) = Overlap * D * exp(iS)
       if (present(C)) then
          do i = 1,nbasis
             C(i) = sum(olpm(i,:)*D(:))
          end do
       else
          do i = 1,nbasis
             bs%C(i) = sum(olpm(i,:)*D(:))
          end do
       end if
       deallocate(olpm)
    else
       write(*,*) "Problem allocating memory in calcC"
       stop
    end if
  end subroutine calcC

  ! ==================================================
  !            calccorfn
  ! ================================================== 
  ! Calculate the correlation function between a set 
  ! of basis functions and an other coherent state
  !subroutine calccorfn(t,basis,zcor,S,C,D,fu)
  subroutine calccorfn(t,bs,zcor,D,fu)
    use coherentstatestype        
    implicit none

    real(kind=8),intent(in) :: t ! time
    type(coherentstatebasisset),intent(inout) :: bs        
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis  ! Basis set
    complex(kind=8),dimension(ndf),intent(in) :: zcor  ! state to correlate with
    !    complex(kind=8),dimension(nbasis),intent(in) :: C,D ! 2 guesses Now in bs
   complex(kind=8),dimension(nbasis),intent(in) :: D
!    real(kind=8),dimension(nbasis),intent(in) :: S ! and another?
    integer,intent(in) :: fu ! File unit to write this fct to

    integer ::ib ! for loop

    complex(kind=8) :: acf ! Value of correlation function
    complex(kind=8) :: ovlpij
!    real(kind=8),external :: norm ! Norm of the current wave packet. 

    acf = (0.0d0,0.0d0)

    if (bs%nehrenfest > 1) then
       write(*,*) "The correlation function routine has not fully been converted to work with Ehrenfest systems"
       stop
    end if

    do ib = 1,nbasis

       !     acf = acf + overlap(zb(:,ib),zinit(:,1))*cdexp((0.0d0,1.0d0)*S(ib))
       !     acf = acf + overlap(basis(:,ib),zinit(:,1))*cdexp((0.0d0,1.0d0)*S(ib))
       !    if (real(sum(basis(:,ib)-zcor(:)*dconjg(basis(:,ib) - zcor(:)))) < 100) then
!       overlap(zi,zj,ai,aj,efest)

       ! Temporary solution, will need to replace first ErnC for Ehrenfest correlation function
       ovlpij=  overlap(zcor,bs%zb(:,ib),bs%ErnC(:,ib),bs%ErnC(:,ib),bs%S(:,ib),bs%S(:,ib),bs%nehrenfest )
       if (cdabs(ovlpij) .ge. 1e-10 ) then
          ! This needs to be changed for Ehrenfest as S will be in the overlap.
          acf = acf + ovlpij*D(ib)*cdexp((0.0d0,1.0d0)*bs%S(1,ib))
       end if
    end do

    write(fu,"(5f15.8)") t,real(acf),dimag(acf),cdabs(acf),norm(bs%C,D)

  end subroutine calccorfn

   ! ==================================================
  !            harmcorfn
  ! ================================================== 
  ! Calculate the correlation function between a set 
  ! of basis functions and an initial harmonic oscillator 
  ! wavefunction.
  subroutine harmcorfn(t,bs,vmode,D,fu)
    use coherentstatestype        
    implicit none

    real(kind=8),intent(in) :: t ! time
    type(coherentstatebasisset),intent(inout) :: bs        
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis  ! Basis set
    integer,dimension(ndf),intent(in) :: vmode  ! state to correlate with
!    complex(kind=8),dimension(nbasis),intent(in) :: C,D ! 2 guesses 
    complex(kind=8),dimension(nbasis),intent(in) :: D
!    real(kind=8),dimension(nbasis),intent(in) :: S ! and another?
    integer,intent(in) :: fu ! File unit to write this fct to

    integer ::ib,i,if ! for loops
    real(kind=8) :: factn ! n factorial
    complex(kind=8) :: acf ! Value of correlation function
    complex(kind=8) :: ovlpij !"Overlap"

    
    if (bs%nehrenfest > 1) then
       write(*,*) "The correlation function routine has not fully been converted to work with Ehrenfest systems"
       stop
    end if

    ! set the correlation fn to zero
    acf = (0.0d0,0.0d0)

    do ib = 1,nbasis
      
       ovlpij = (1.0,0.0)
       ! Calculate the overlap between current basis fn and the initial wavefn.
       do i = 1,ndf

          select case(vmode(i))
          case(0)
              
             ovlpij = ovlpij*exp(-0.5*abs(bs%zb(i,ib))**2)
          case(1)
             ovlpij = ovlpij*bs%zb(i,ib)*exp(-0.5*abs(bs%zb(i,ib))**2)

          case default 
             ! Calculate n factorial
             factn=1.0
              do if = 2,vmode(i)
                 factn = factn*if
              end do
            ovlpij =  ovlpij * bs%zb(i,ib)**(vmode(i))*exp(-0.5*abs(bs%zb(i,ib))**2)/factn

          end select

!          case(-1) ! A Coherent state
!             ovlpij=  overlap(,basis(:,ib))
       end do


       if (cdabs(ovlpij) .ge. 1e-10 ) then
          acf = acf + 1.0*ovlpij*D(ib)*cdexp((0.0d0,1.0d0)*bs%S(1,ib))
       end if
    end do


    write(99,*) " hcf: ",real(acf),dimag(acf),cdabs(acf),norm(bs%C,D)

    write(fu,"(5f15.8)") t,real(acf),dimag(acf),cdabs(acf),norm(bs%C,D)

  end subroutine harmcorfn

  ! **************************************************
  !              Setup routines
  ! **************************************************

  ! ==================================================
  !            Allocate arrays
  ! ================================================== 
  subroutine allocatecsarrays
    implicit none

    if(.not.allocated(gamma)) then
       allocate(gamma(ndf),s2gammarec(ndf),shgamma(ndf),sgamma(ndf))
       gamma(:) = 1.0
       s2gammarec(:) = 1/dsqrt(2.0d0)      
       shgamma(:) = 1/dsqrt(2.0d0)
       sgamma(:) = dsqrt(gamma(:))
    end if
  end subroutine allocatecsarrays

  subroutine deallocatecsarrays
    implicit none

    if(allocated(gamma)) deallocate(gamma)
    if(allocated(s2gammarec)) deallocate(s2gammarec)
    if(allocated(shgamma)) deallocate(shgamma)
    if(allocated(sgamma)) deallocate(sgamma)


  end subroutine deallocatecsarrays

  ! ==================================================
  !            calcgammas
  ! ================================================== 

  ! To speed things up, calculate (2*gamma)^1/2
  !  and hsgamma at beginning
  subroutine calcgammas
    implicit none

    s2gammarec(:) = 1/dsqrt(2.0*gamma(:)) 
    shgamma(:)    = dsqrt(gamma(:)/2.0)
    sgamma(:) = dsqrt(gamma(:))

  end subroutine calcgammas

end module coherentstates
