! **************************************************
!           Coherent States Basis Set Type
!                  Stewart K. Reed
!                22nd February 2010
!
!  Description:
!        This module contains a type for coherent
!        states.
!
!  Note: it the overloaded operators make no attempt
!        to check that two basis sets are of the 
!        same dimension.
! **************************************************

module coherentstatestype
  implicit none

  type coherentstatebasisset
     integer :: ndf
     integer :: nbasis
     ! Number of Ehrenfest equations
     integer :: nehrenfest
     
!     Ideally we should include gamma but leave it out
!     for speed since the code works without it.
!     real(kind=8),dimension(:),allocatable :: gamma 
     complex(kind=8),dimension(:),allocatable :: C
!     complex(kind=8),dimension(:,:),allocatable :: D
     ! The Ehrenfest amplitudes
     complex(kind=8),dimension(:,:),allocatable :: ErnC
     ! Centres of the basis functions ndf,nbasis
     complex(kind=8),dimension(:,:),allocatable :: zb
     ! The actions dimension(nehernfest,nbasis)
     real(kind=8),dimension(:,:),allocatable :: S
     
     

  end type coherentstatebasisset

  ! Define operators. = is given free.
  interface operator(+)
     module procedure csbs_plus_csbs
  end interface
  
  interface operator(*)
     module procedure rconst_times_csbs
  end interface

  interface operator(/)
     module procedure csbs_divide_rcst
  end interface
!  interface const_times_csbs
!     function zconst_times_csbs
  contains

    ! ===================================================================
    ! Function: csbs_plus_csbs
    !  This funtion adds the components of two basis sets together.
    ! ===================================================================
    function csbs_plus_csbs(csbs1,csbs2)
      implicit none
      type(coherentstatebasisset),intent(in) :: csbs1,csbs2
      type(coherentstatebasisset) :: csbs_plus_csbs

      ! set output to first cs basis set
      call createbasisset(csbs1%ndf,csbs1%nbasis,csbs1%nehrenfest,csbs_plus_csbs)
      csbs_plus_csbs = csbs1
      !csbs_plus_csbs%ndf = csbs1%ndf
      !csbs_plus_csbs%nbasis = csbs1%nbasis
      !csbs_plus_csbs%namps = csbs1%namps

      csbs_plus_csbs%C(:)  = csbs1%C(:)  + csbs2%C(:)
      csbs_plus_csbs%ErnC(:,:)  = csbs1%ErnC(:,:)  + csbs2%ErnC(:,:)
      csbs_plus_csbs%zb(:,:)  = csbs1%zb(:,:)  + csbs2%zb(:,:)
      csbs_plus_csbs%S(:,:)  = csbs1%S(:,:)  + csbs2%S(:,:)
      
      
    end function csbs_plus_csbs
    
    subroutine add_basis_sets(csbs1,csbs2,csbsout)
      implicit none
      type(coherentstatebasisset),intent(in) :: csbs1,csbs2
      type(coherentstatebasisset),intent(inout) :: csbsout

      csbsout%C(:)  = csbs1%C(:)  + csbs2%C(:)
      csbsout%ErnC(:,:)  = csbs1%ErnC(:,:)  + csbs2%ErnC(:,:)
      csbsout%zb(:,:)  = csbs1%zb(:,:)  + csbs2%zb(:,:)
      csbsout%S(:,:)  = csbs1%S(:,:)  + csbs2%S(:,:)
    end subroutine add_basis_sets

    subroutine weightedadd_basis_sets(csbs1,csbs2,w,csbsout)
      implicit none
      type(coherentstatebasisset),intent(in) :: csbs1,csbs2
      real(kind=8),intent(in) :: w
      type(coherentstatebasisset),intent(inout) :: csbsout

      csbsout%C(:)  = csbs1%C(:)  + w*csbs2%C(:)
      csbsout%ErnC(:,:)  = csbs1%ErnC(:,:)  + w*csbs2%ErnC(:,:)
      csbsout%zb(:,:)  = csbs1%zb(:,:)  + w*csbs2%zb(:,:)
      csbsout%S(:,:)  = csbs1%S(:,:)  + w*csbs2%S(:,:)
    end subroutine weightedadd_basis_sets

    subroutine add_4basis_sets(csbs1,csbs2,csbs3,csbs4,csbsout)
      implicit none
      type(coherentstatebasisset),intent(in) :: csbs1,csbs2,csbs3,csbs4
      type(coherentstatebasisset),intent(inout) :: csbsout

      csbsout%C(:)  = csbs1%C(:)  + csbs2%C(:)+csbs3%C(:)  + csbs4%C(:)
      csbsout%ErnC(:,:)  = csbs1%ErnC(:,:)  + csbs2%ErnC(:,:)+csbs3%ErnC(:,:)  + csbs4%ErnC(:,:)
      csbsout%zb(:,:)  = csbs1%zb(:,:)  + csbs2%zb(:,:) + csbs3%zb(:,:)  + csbs4%zb(:,:)
      csbsout%S(:,:)  = csbs1%S(:,:)  + csbs2%S(:,:)+csbs3%S(:,:)  + csbs4%S(:,:)
    end subroutine add_4basis_sets

    ! ===================================================================
    ! Function: zconst_times_csbs
    !  This funtion multiplies the components of a basis set by a 
    !  a constant.
    ! ===================================================================
    function zconst_times_csbs(cst,csbs1)
      complex(kind=8),intent(in) :: cst
      type(coherentstatebasisset),intent(in) :: csbs1
      type(coherentstatebasisset) :: zconst_times_csbs

      call createbasisset(csbs1%ndf,csbs1%nbasis,csbs1%nehrenfest,zconst_times_csbs)
      ! set output to first cs basis set
!      zconst_times_csbs%ndf = csbs1%ndf
      zconst_times_csbs%nbasis = csbs1%nbasis
      zconst_times_csbs%nehrenfest = csbs1%nehrenfest
      
      zconst_times_csbs%C  = cst * csbs1%C
      zconst_times_csbs%ErnC(:,:)  = cst * csbs1%ErnC(:,:) 
      zconst_times_csbs%zb(:,:)  = cst * csbs1%zb(:,:)
      zconst_times_csbs%S(:,:)  = cst * csbs1%S(:,:)
      
      
    end function zconst_times_csbs
  
    function rconst_times_csbs(cst,csbs1)
      real(kind=8),intent(in) :: cst
      type(coherentstatebasisset),intent(in) :: csbs1
      type(coherentstatebasisset) :: rconst_times_csbs

      ! set output to first cs basis set
      if (.not.allocated(csbs1%zb))then
         call createbasisset(csbs1%ndf,csbs1%nbasis,csbs1%nehrenfest,rconst_times_csbs)      
      end if
      rconst_times_csbs%C(:)  = cst * csbs1%C(:) 
      rconst_times_csbs%ErnC(:,:)  = cst * csbs1%ErnC(:,:) 

      rconst_times_csbs%zb(:,:)  = cst * csbs1%zb(:,:)
      rconst_times_csbs%S(:,:)  = cst * csbs1%S(:,:)
      
      
    end function rconst_times_csbs

    subroutine rconst_times_csbs_sub(cst,csbs1,csbsout)
      real(kind=8),intent(in) :: cst
      type(coherentstatebasisset),intent(in) :: csbs1
      type(coherentstatebasisset),intent(inout) :: csbsout

      csbsout%C(:)  = cst * csbs1%C(:) 
      csbsout%ErnC(:,:)  = cst * csbs1%ErnC(:,:) 

      csbsout%zb(:,:)  = cst * csbs1%zb(:,:)
      csbsout%S(:,:)  = cst * csbs1%S(:,:)
      
    end subroutine rconst_times_csbs_sub

    ! ===================================================================
    ! Function: csbs_divide_rcst
    !  This funtion divides the components of a basis set by a 
    !  a constant.
    ! ===================================================================
    function csbs_divide_rcst(csbs1,cst)
      real(kind=8),intent(in) :: cst
      type(coherentstatebasisset),intent(in) :: csbs1
      type(coherentstatebasisset) :: csbs_divide_rcst

      ! set output to first cs basis set
      if (.not.allocated(csbs1%zb))then               
         call createbasisset(csbs1%ndf,csbs1%nbasis,csbs1%nehrenfest,csbs_divide_rcst)      
      end if

      csbs_divide_rcst%C(:)  = csbs1%C(:)/cst
      csbs_divide_rcst%ErnC(:,:)  =  csbs1%ErnC(:,:) /cst

      csbs_divide_rcst%zb(:,:)  =  csbs1%zb(:,:)/cst
      csbs_divide_rcst%S(:,:)  = csbs1%S(:,:)/cst
      
      
    end function csbs_divide_rcst


    ! ===================================================================
    ! Subroutine: createbasisset
    !  This subroutine allocates arrays for a basis set.
    ! ===================================================================
    subroutine createbasisset(ndf,nbasis,namps,csbs)
      implicit none
      integer, intent(in) :: ndf,nbasis,namps
      type(coherentstatebasisset),intent(inout) :: csbs
      
      ! save the number of degrees of freedom, number of basis functions
      ! and the number of Ehrenfestamplitudes.
      csbs%ndf = ndf
      csbs%nbasis = nbasis
      csbs%nehrenfest = namps

      ! Now allocate arrays
      if (.not. allocated (csbs%C)) then
         allocate(csbs%C(nbasis))
         csbs%C(:)=(0.0d0,0.0d0)
      else
         write(*,*) "Error creating basis set: C is already allocated"
      end if     
      if (.not. allocated (csbs%ErnC)) then
         allocate(csbs%ErnC(namps,nbasis))
         csbs%ErnC(:,:)=(0.0d0,0.0d0)
      else
         write(*,*) "Error creating basis set: C is already allocated"
      end if
      if (.not. allocated (csbs%zb)) then
         allocate(csbs%zb(ndf,nbasis))
         csbs%zb(:,:)=(0.0d0,0.0d0)
      else
         write(*,*) "Error creating basis set: zb is already allocated"
      end if

      if (.not. allocated (csbs%S)) then
         allocate(csbs%S(namps,nbasis))
         csbs%S(:,:) = 0.0d0
      else
         write(*,*) "Error creating basis set: zb is already allocated"
      end if
    end subroutine createbasisset
end module coherentstatestype
