! **************************************************
!                    Hamiltonian
!
!                  Stewart K. Reed
!                   4th June 2007
!
!  Description:
!      These functions calculate the Hamiltonian 
!    and its derivatives.
!  Modified 1st October 2007.
!    The hamiltonians are now kept in their own 
!    files in a directory called hamiltonians. 
!    To add another hamiltonian create another file
!    along the lines of the existing ones and 
!    include it below. You'll also need to add it to
!    h_ord, dh_ord_dz and readhamparams below.
!   
! Types of hamiltonians, short names and routines 
!  reading in specific info from hamiltonian.inpt
!   1 tunnel        readhamtunbath
!   2 henonheiles   readhamhh
!   3 polymorse     readhampm
!   4 cartesian     readhamcart
!   5 classical polymose 
!   6 classicalcartesian 
! **************************************************
module hamiltonian
  use hamiltoniandata
  use polymorsehammod
  use classicpolymorsehammod
  use cartesianham
  implicit none

  private :: hamtunbath,dhamtunbathdz,readhamtunbath

contains

  ! **************************************************
  !  Include the actual Hamiltonian eqns and derivatives.
  !  Note: there is a subroutine for reading info from
  !        the hamiltonian input file at the bottom 
  !        of this file.
  ! ************************************************** 
  include "hamiltonians/tunnelham.f90"
  include "hamiltonians/henonheilesham.f90"
  !include "hamiltonians/polymorseham.f90"
!  include "hamiltonians/cartesian.f90"

  ! ===============================================
  !  h_ord
  !  Subroutine to calculate the ordered hamiltonian
  ! ===============================================
  function h_ord (zic,zj)

    implicit none

    ! two states <z_i| and |z_j>
    complex(kind=8),dimension(:),intent(in) :: zic,zj
    ! hamiltonian
    complex(kind=8):: h_ord

    select case(hamtype)
    case (1)
       h_ord = hamtunbath(zic,zj)
    case (2)
       h_ord = hamhenonheiles(zic,zj)
    case (3)
        h_ord = hampolymorse(zic,zj)
     case (5)
        h_ord = champolymorse(zic,zj)
     case (cartham) ! Cartesian
        h_ord = hamcart(zic,zj)   
     case (clascartham) ! Classical Cartesian
        h_ord = chamcart(zic,zj)
     case default

    end select

  end function h_ord


  ! ===============================================
  !  h_ord_dz
  !  Subroutine to calculate the derviative of the 
  !  ordered hamiltonian w.r.t. z
  !  This is a vector quantitiy
  ! ===============================================
  function dh_ord_dz (zic,zj)
    use coherentstates, ONLY:ndf
    implicit none

    ! two states <z_i| and |z_j>
    complex(kind=8),dimension(:),intent(in) :: zic,zj

    ! derivative dh/dz - this function
    complex(kind=8) :: dh_ord_dz(ndf)

    select case(hamtype)
    case (1)
       dh_ord_dz = dhamtunbathdz(zic,zj)
    case (2)
       dh_ord_dz = dhamhenonheilesdz(zic,zj)
    case (3)
       dh_ord_dz = dhampolymorsedz(zic,zj)
    case (5)
       dh_ord_dz = dchampolymorsedz(zic,zj)        
    case (cartham) ! Cartesian
       dh_ord_dz = dhamcartdz(zic,zj)   
    case (clascartham) ! Classical cartesian
       dh_ord_dz = dchamcartdz(zic,zj)
    case default

    end select
  end function dh_ord_dz

  ! ===============================================
  !  h_ord_dzc
  !  Subroutine to calculate the derviative of the 
  !  ordered hamiltonian w.r.t. z
  !  This is a vector quantitiy
  ! ===============================================
  function dh_ord_dzc (zic,zj)
    use coherentstates, ONLY:ndf
    implicit none

    ! two states <z_i| and |z_j>
    complex(kind=8),dimension(:),intent(in) :: zic,zj

    ! derivative dh/dz - this function
    complex(kind=8) :: dh_ord_dzc(ndf)

    select case(hamtype)
    case (1)
       dh_ord_dzc = dhamtunbathdz(zic,zj)
       dh_ord_dzc =  dconjg(dh_ord_dzc)
    case (2)
       dh_ord_dzc = dhamhenonheilesdz(zic,zj)
       dh_ord_dzc =  dconjg(dh_ord_dzc)
    case (3)
       dh_ord_dzc = dhampolymorsedz(zic,zj)
       dh_ord_dzc =  dconjg(dh_ord_dzc)
    case (5)
       dh_ord_dzc = dchampolymorsedz(zic,zj)
       dh_ord_dzc =  dconjg(dh_ord_dzc)
    case (cartham) ! Cartesuab
       dh_ord_dzc = dhamcartdzc(zic,zj)   
    case (6) ! Classical cartesian
       dh_ord_dzc = dchamcartdzc(zic,zj)
    case default

    end select
  end function dh_ord_dzc

  ! ===============================================
  !  delta_h_ord
  !  Subroutine to calculate the derviative of the 
  !  ordered hamiltoni0an w.r.t. z
  ! ===============================================
  function delta_h_ord (zi,zic,zj)
    use coherentstates, ONLY: ndf
    implicit none
    ! two states <z_i| and |z_j>
    complex(kind=8),dimension(ndf),intent(in) :: zi,zic,zj

    complex(kind=8) :: delta_h_ord

    complex(kind=8),dimension(ndf) :: dhdz


    dhdz = dh_ord_dz(zic,zi)
    delta_h_ord = h_ord(zic,zj) - h_ord(zic,zi) &
         -sum(dhdz(:)*(zj(:)-zi(:)))

    ! Debug
!!$      write(72,*)  real(h_ord(zic,zj)),dimag(h_ord(zic,zj)), &
!!$           - real(h_ord(zic,zi)),dimag(h_ord(zic,zi)) &
!!$           -real(sum(dhdz(:)*(zj(:)-zi(:)))),&
!!$           -dimag(sum(dhdz(:)*(zj(:)-zi(:)))),&
!!$           real(delta_h_ord),dimag(delta_h_ord)

  end function delta_h_ord

  ! ==================================================
  !    totalenergy
  !   Subroutine to calculate total energy
  ! 
  ! Sum Dj* Di Hord(zj*,zi) - should be real!
  ! ==================================================
  function totalenergy(bs,D,incbf)
    use coherentstatestype
    use coherentstates, ONLY: ndf,nbasis,gamma
    use options, ONLY: eu,omega0 !energy units
    ! functions
    use coherentstates, ONLY:overlap_bs
    use scalapackstuff, ONLY:MPI_Wtime,isroot
    implicit none
    type(coherentstatebasisset) :: bs
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis
    complex(kind=8),dimension(nbasis),intent(in) :: D
!    real(kind=8),dimension(nbasis),intent(in):: S
    logical,dimension(nbasis),intent(in) :: incbf

    complex(kind=8) :: ovlpji
    real(kind=8) :: totalenergy
    complex(kind=8) :: te
    integer :: i,j
    
    real :: entrytime,exittime
 !!debug    real(kind=8) :: ke
    entrytime=MPI_Wtime()
    totalenergy = 0.0d0
!!debug    ke = 0.0d0
    ! ---- diagonal terms ----
!    write(67,*) "# Diagonal terms"
    do i = 1,nbasis
       if (incbf(i)) then
          ! debug, check overlap = 1 which it does.
          !         write(55,*) i,overlap(basis(:,i),basis(:,i))
          totalenergy = totalenergy &
               + dconjg(D(i))*D(i)*h_ord(dconjg(bs%zb(:,i)),bs%zb(:,i))
!!debug the kinetic energy only works for one basis function?
 !!debug          ke = ke+dconjg(D(i))*D(i)* 0.25*sum( (-(dconjg(basis(:,i))-basis(:,i))**2)*gamma(:)/m(:))+kinzpe
       end if
    end do
!debug write(*,*) "DKE", ke*eu
!    write(67,*) "# Off diagonal terms"
    ! ---- off diagonal terms ----
    do i= 1,nbasis-1
       if (incbf(i)) then
          do j = i+1,nbasis
             if (incbf(j)) then
                ovlpji = overlap_bs(bs,j,i) !overlap(bs%zb(:,j),bs%zb(:,i))
                if (cdabs(ovlpji) .ge. omega0) then
                   te =  dconjg(D(j))*D(i)*h_ord(dconjg(bs%zb(:,j)),bs%zb(:,i)) &
                        * cdexp((0,1.0d0)*(bs%S(1,i)-bs%S(1,j)))*ovlpji
!!debug                   ke = ke+2*real(dconjg(D(j))*D(i) * cdexp((0,1.0d0)*(S(i)-S(j)))*ovlpij* &
!!debug                         0.25*sum( (-(dconjg(basis(:,i))-basis(:,i))**2)*gamma(:)/m(:)))+kinzpe
                   totalenergy = totalenergy + 2*real(te)
                end if
             end if
          end do
       end if
    end do
  !  write(67,*) 
    ! ---- convert to known energy units ----
    totalenergy = totalenergy*eu
 !!debug   ke =ke *eu

  !!debug   if (isroot) write(99,*) "Totalenergy",totalenergy,totalenergy-ke,ke
    
 !   call CPU_TIME(exittime)
    ! Total time
!    totengtime = totengtime + exittime-entrytime
    totengtime = totengtime + MPI_Wtime() - entrytime
  end function totalenergy

  ! ==================================================
  !    conservedenergy
  !   Subroutine to calculate classically conserved 
  !   energy
  ! 
  ! Sum Hord(zi*,zi) 
  ! ==================================================
  function conservedenergy(basis)
    use coherentstates, ONLY: ndf,nbasis
    use options, ONLY: eu !energy units

    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis
    integer :: i

    real(kind=8) :: conservedenergy
    conservedenergy = 0.0
    do i = 1,nbasis
       conservedenergy = conservedenergy &
            + real(h_ord (dconjg(basis(:,i)),basis(:,i)))
    end do

    ! ---- convert to known energy units ----
    conservedenergy = conservedenergy*eu
  end function conservedenergy


  ! ==================================================
  !    expectationval
  !   Subroutine to calculate total energy
  ! 
  ! Sum Ci* Di Hord(zi*,zi) - should be real!
  ! ==================================================
  function expectationvalue(basis,C,D,incbf)
#ifdef USE_SCALAPACK
#ifdef USE_MPI1
    use scalapackstuff, ONLY: MPI_DOUBLE_PRECISION, MPI_SUM, comm
    use arrays, ONLY:mpi_tmp_Z_array, mpi_tmp_D
#else
    use scalapackstuff, ONLY: MPI_IN_PLACE, MPI_DOUBLE_PRECISION, MPI_SUM, comm
#endif
#endif
    use coherentstates, ONLY: ndf,nbasis,gamma
    use options, ONLY: eu,omega0 !energy units
    use arrays, ONLY: t
    ! functions
    use coherentstates, ONLY:overlap
    implicit none
    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis
    complex(kind=8),dimension(nbasis),intent(in) :: C,D
    logical,dimension(nbasis),intent(in) :: incbf

    real(kind=8) :: expectationvalue
    integer :: i

    complex(kind=8) :: KE
#ifdef USE_SCALAPACK
    integer :: mpierr
#endif

    expectationvalue = 0.0d0
!debug    write(114,*) "# Step"
    ! ---- diagonal terms ----
!    write(67,*) "# Diagonal terms"
    do i = ibeg_nbasis,iend_nbasis
!    do i = 1,nbasis
       if (incbf(i)) then
          expectationvalue = expectationvalue  &
              + dconjg(C(i))*D(i)*h_ord(dconjg(basis(:,i)),basis(:,i))
!debug          write(114,*) i,real(h_ord(dconjg(basis(:,i)),basis(:,i))*eu),dimag(h_ord(dconjg(basis(:,i)),basis(:,i))*eu)

!              KE = dconjg(C(i))*D(i)* 0.25* (1-(dconjg(basis(1,i))-basis(1,i))**2)*gamma(1)/m(1)*eu
!              write(113,*) t, dconjg(C(i))*D(i),eu*dconjg(C(i))*D(i)*h_ord(dconjg(basis(:,i)),basis(:,i)),&
!                   eu*dconjg(C(i))*D(i)*h_ord(dconjg(basis(:,i)),basis(:,i))-KE,KE

       end if
    end do
!debug    write(114,*)

    ! If we are running in parallel calculate total expectationvalue
#ifdef USE_SCALAPACK    
#ifdef USE_MPI1
    expectationvalue = mpi_tmp_D
    call MPI_Allreduce( mpi_tmp_D, expectationvalue, 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, mpierr)
#else
    call MPI_Allreduce(MPI_IN_PLACE, expectationvalue, 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, mpierr)
#endif
#endif

   ! ---- convert to known energy units ----
    expectationvalue = expectationvalue*eu
!    KE = 0.0d0
!    do i=1,nbasis
!       KE =    KE+dconjg(C(i))*D(i)* 0.25*sum( (1-(dconjg(basis(:,i))-basis(:,i))**2)*gamma(:)/m(:))*eu
!    end do
!    write(113,*) t,expectationvalue,expectationvalue-KE,KE
  end function expectationvalue

  ! ==================================================
  !    modeexpectationval
  !   Subroutine to calculate total energy of a mode
  ! 
  ! Sum Ci* Di Hord(zi*,zi) - should be real!
  ! ==================================================
  subroutine modeexpectationvalue(basis,C,D,incbf,mode,Umode,Ucoup)
#ifdef USE_SCALAPACK
#ifdef USE_MPI1
    use scalapackstuff, ONLY: MPI_DOUBLE_PRECISION, MPI_SUM, comm
    use arrays, ONLY:mpi_tmp_Z_array, mpi_tmp_D
#else
    use scalapackstuff, ONLY: MPI_IN_PLACE, MPI_DOUBLE_PRECISION, MPI_SUM, comm
#endif
#endif
    use coherentstates, ONLY: nbasis,ndf
    use options, ONLY: eu !energy units
    ! functions
    use coherentstates, ONLY:overlap
    implicit none
    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis
    complex(kind=8),dimension(nbasis),intent(in) :: C,D
    logical,dimension(nbasis),intent(in) :: incbf
    integer,intent(in) :: mode
    ! mode and coupling energies
    real(kind=8),intent(out):: Umode, Ucoup
    complex(kind=8)::Umode1,Ucoup1 ! temp values 
    integer :: i

#ifdef USE_SCALAPACK
    integer :: mpierr
#endif

    Umode = 0.0d0
    Ucoup = 0.0d0
    ! ---- diagonal terms ----
!    write(67,*) "# Diagonal terms"
!    do i = 1,nbasis
    do i = ibeg_nbasis,iend_nbasis
       if (incbf(i)) then
          if (hamtype .eq. 3) then
             call hampolymorse1mode(dconjg(basis(:,i)),basis(:,i),mode,Umode1,Ucoup1)
          else
             call champolymorse1mode(dconjg(basis(:,i)),basis(:,i),mode,Umode1,Ucoup1)
          end if
           Umode = Umode + real(dconjg(C(i))*D(i)*Umode1)
           ucoup = Ucoup + real(dconjg(C(i))*D(i)*Ucoup1)

        end if
    end do
    
    ! If we are running in parallel calculate total expectationvalue
#ifdef USE_SCALAPACK    
#ifdef USE_MPI1
    mpi_tmp_D = Umode
    call MPI_Allreduce(mpi_tmp_D, Umode, 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, mpierr)
    mpi_tmp_D = Ucoup
    call MPI_Allreduce(mpi_tmp_D, Ucoup, 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, mpierr)
#else
    call MPI_Allreduce(MPI_IN_PLACE, Umode, 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, mpierr)
    call MPI_Allreduce(MPI_IN_PLACE, Ucoup, 1, MPI_DOUBLE_PRECISION, MPI_SUM, comm, mpierr)
#endif
#endif

    ! ---- convert to known energy units ----
    Umode = Umode*eu
    Ucoup = Ucoup*eu


  end subroutine modeexpectationvalue
  ! ==================================================
  !    polymodeenergy
  !   Subroutine to calculate classically conserved 
  !   energy
  ! 
  ! Sum Hord(zi*,zi) 
  ! ==================================================
  function polymodeenergy(bs,D,mode,Umode,Ucoup,incbf)
    use coherentstatestype
    use coherentstates, ONLY: ndf,nbasis
    use options, ONLY: eu,omega0 !energy units
    ! functions
    use coherentstates, ONLY:overlap_bs
    implicit none
    type(coherentstatebasisset) :: bs
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: basis
    complex(kind=8),dimension(nbasis),intent(in) :: D
!    real(kind=8),dimension(nbasis),intent(in):: S
    integer,intent(in) :: mode

    real(kind=8),intent(out)::Umode,Ucoup
    logical,dimension(nbasis),intent(in) :: incbf

    complex(kind=8)::Umode1,Ucoup1 ! temp values

    complex(kind=8) :: ovlpij
    real(kind=8) :: polymodeenergy
    integer :: i,j
    complex(kind=8) :: te
    polymodeenergy = 0.0d0
    Umode = 0.0
    Ucoup = 0.0
    ! ---- diagonal terms ----
    do i = 1,nbasis
        if (incbf(i)) then
           if(hamtype .eq. 3) then
              call hampolymorse1mode(dconjg(bs%zb(:,i)),bs%zb(:,i),mode,Umode1,Ucoup1)
           else if (hamtype .eq. 5) then
              call champolymorse1mode(dconjg(bs%zb(:,i)),bs%zb(:,i),mode,Umode1,Ucoup1)
           end if
!       polymodeenergy = polymodeenergy &
!            + dconjg(D(i))*D(i)*hampolymorse1mode(dconjg(basis(:,i)),basis(:,i),mode)

           Umode = Umode + real(dconjg(D(i))*D(i)*Umode1)
           ucoup = Ucoup + real(dconjg(D(i))*D(i)*Ucoup1)
        end if

    end do

!    write(99,*) "Mode Energy, mode",mode
!    write(99,*) "Diagonal contribution",Umode
    ! ---- off diagonal terms ----
    do i= 1,nbasis-1
       if (incbf(i)) then
          do j = i+1,nbasis
             if (incbf(j)) then
                ovlpij = overlap_bs(bs,i,j)  !    basis(:,j),basis(:,i))
             
                if (cdabs(ovlpij) .ge. omega0) then
                   !             te =  dconjg(D(j))*D(i)*hampolymorse1mode(dconjg(basis(:,j)),basis(:,i),mode) &
                   !                 * cdexp((0,1.0d0)*(S(i)-S(j)))*ovlpij
                   !            polymodeenergy = polymodeenergy + 2*real(te
                   if (hamtype .eq. 3 ) then
                      call hampolymorse1mode(dconjg(bs%zb(:,j)),bs%zb(:,i),mode,Umode1,Ucoup1)
                   else if (hamtype .eq. 3 ) then
                      call hampolymorse1mode(dconjg(bs%zb(:,j)),bs%zb(:,i),mode,Umode1,Ucoup1)
                   end if
                   Umode = Umode + 2*real(dconjg(D(j))*D(i)*Umode1 * cdexp((0,1.0d0)*(bs%S(1,i)-bs%S(1,j)))*ovlpij)
                   ucoup = Ucoup + 2*real(dconjg(D(j))*D(i)*Ucoup1 * cdexp((0,1.0d0)*(bs%S(1,i)-bs%S(1,j)))*ovlpij)
                end if
             end if
          end do
       end if
    end do

 !   write(99,*) "Total Mode Energy",Umode

    ! ---- convert to known energy units ----
    polymodeenergy = polymodeenergy*eu
    Umode = Umode*eu
    Ucoup = Ucoup *eu

  end function polymodeenergy

  ! ===============================================
  !    dz_dt
  ! Calculate the time derivative of z
  ! ===============================================
  ! dz*/dt = i/hbar*dH(z*,z)/dz
  ! dz/dz =  -i/hbar*dH(z*,z)/dz* = conjg(dz*/dt)

  subroutine dz_dt(z,dzdt)
    use coherentstates, ONLY: ndf
    implicit none

    complex(kind=8),dimension(ndf),intent(in) :: z    !z
    complex(kind=8),dimension(ndf),intent(out) :: dzdt ! itsderivative

    complex(kind=8),dimension(ndf) :: zc

    ! Calculate the complex conjugate of z
    zc = dconjg(z)
    ! Calculate the time derivate of z*
    !dzdt = dconjg(dh_ord_dz(zc,z))*(0,-1)
    dzdt = (dh_ord_dzc(zc,z))*(0,-1)

    ! dz/dt = conjugate of dz*/dt
    !  dzdt = dconjg(dzdt)

  end subroutine dz_dt


  ! ===============================================
  !    dzc_dt
  ! Calculate the time derivative of z*
  ! ===============================================
  ! dz*/dt = i/hbar*dH(z*,z)/dz*
  subroutine dzc_dt(z,dzcdt)
    use coherentstates, ONLY: ndf
    implicit none

    complex(kind=8),dimension(ndf),intent(in) :: z    !zc
    complex(kind=8),dimension(ndf),intent(out) :: dzcdt ! its derivative

    complex(kind=8),dimension(ndf) :: zc

    ! Calculate the complex conjugate of z*
    zc = dconjg(z)

    ! Calculate the time derivate of z*
    dzcdt = dh_ord_dz(zc,z)*(0,1)

  end subroutine dzc_dt

  ! ===============================================
  !  Calculate the time dervative of the action
  !  for a specific basis function.
  ! ===============================================
  subroutine dS_dt(z,zc,dzdt,dzcdt,dSdt)
    use coherentstates, ONLY: ndf
    implicit none
    complex(kind=8),dimension(ndf),intent(in):: z,zc
    complex(kind=8),dimension(ndf),intent(in) :: dzdt,dzcdt

    real(kind=8),intent(out) :: dSdt

    ! Now calculate the action
    dSdt = (0.0,0.5) * sum(zc(:)*dzdt(:) - dzcdt(:)*z(:)) &
         - h_ord (zc,z)

  end subroutine dS_dt

  ! ===============================================
  !  Calculate D from C
  ! ===============================================
    ! subroutine calcD(C,zb,S,D,incbf) 
  subroutine calcD(bs,D,incbf) 
    use coherentstatestype
    use options, ONLY: solvemethod
    implicit none
    type(coherentstatebasisset),intent(in) :: bs
!    complex(kind=8),dimension(:),intent(in) :: C
!    real(kind=8),dimension(:,:),intent(in) :: S ! action
!    complex(kind=8),dimension(:,:),intent(in) :: zb ! basis states
    complex(kind=8),dimension(:),intent(out) :: D
    logical,dimension(:),intent(in) :: incbf

    
#ifdef USE_SCALAPACK
    select case(solvemethod)
    case(3)
       ! call  calcDscalapack(C,zb,S(1,:),D,incbf)
       call  calcDscalapack(bs,D,incbf)
    case default
!       call calcDserial(C,zb,S(1,:),D,incbf)
       call calcDserial(bs,D,incbf)
    end select
#else
!    call calcDserial(C,zb,S(1,:),D,incbf) 
    call calcDserial(bs,D,incbf) 
#endif
  end subroutine calcD


#ifdef USE_SCALAPACK
  
  complex(kind=8) function calc_olpm(bs,i,j,omega0)
    use coherentstatestype
    use coherentstates, ONLY: ndf,nbasis,overlap
    implicit none

    integer,intent(in) :: i,j
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: zb ! basis states
    type(coherentstatebasisset),intent(in) :: bs
!    real(kind=8), intent(in) :: dS
    real(kind=8), intent(in) :: omega0
    
    ! check for a diagonal element
    if (i.eq.j) then
       calc_olpm = (1.0d0,0.0d0)
    else
       ! off diagonal element
       calc_olpm = overlap(,bs,i,j)  !zb(:,i),zb(:,j))
       if (abs(calc_olpm) .ge. omega0) then
          if(bs%nehrenfest .le. 1) then  
             ! If we are not doing Ehrenfest, multiply by exp(i dS). 
             ! If we are doing Ehrenfest, this factor forms part of the overlap.
             calc_olpm = calc_olpm *cdexp((0.0,1.0)*(bs%S(1,ja)-bs%S(1,ia)))
          end if
       else
          calc_olpm = (0.0d0,0.0d0)
       end if
       
    end if    
  end function calc_olpm

  subroutine calcDscalapack(bs,D,incbf)
    ! variables
    use coherentstatestype    
    use coherentstates, ONLY: ndf,nbasis
    use arrays,ONLY: FERR,BERR,RWORK,WORK,LWORK,LRWORK,ipiv,overlpm,overlpmf,t, &
         SP_MB,SP_NB,SP_nbc,SP_nbr,SP_ceb,SP_reb,localC,localD,SCA_R,SCA_C,&
         SP_Row_OLAP,SP_Col_OLAP,NRHS,SP_DESC_OLAP,SP_DESC_C,SP_DESC_D
    use options, ONLY: omega0,verbosity
    ! functions
    use coherentstates, ONLY: overlap
    use files,ONLY: fug
    use scalapackstuff
#ifdef USE_MPI1
    use arrays, ONLY: mpi_tmp_Z_array
#endif
    implicit none
    type(coherentstatebasisset),intent(in) :: bs
!    complex(kind=8),dimension(nbasis),intent(in) :: C
!    real(kind=8),dimension(nbasis),intent(in) :: S ! action
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: zb ! basis states
    complex(kind=8),dimension(nbasis),intent(out) :: D
    logical,dimension(nbasis),intent(in) :: incbf

    !Variables for LAPACK routines + arrays from arrays module
    integer:: info
    real(kind=8) :: rcond
    character :: EQUED

    integer :: ia,ja,il,jl ! for loops
    integer :: rowb,rb,colb,cb
    
    ! For MPI
    integer :: mpierr
    real(kind=8) ::intime
    
    ! Find out entry time
    intime = MPI_Wtime()

    ! ---- Fill local overlap matrix ----
    ! Calculate elements of overlap matrix on this processor.
    loop_rowb: do rowb = 0,SP_nbr-1 ! loop over row blocks
       loop_rb: do rb = 1,SP_NB  ! loop over row block elements
          ! calculate global coordinate
          ia = rowb*nprow*SP_NB + myrow*SP_NB+rb

          ! check to see if this basis function is still included.
          ! If not move onto next one.
          if(.not.incbf(ia)) cycle loop_rb 

          ! calculate local coordinate
          il = rowb*SP_NB + rb

          loop_colb: do colb = 0,SP_nbc-1
             loop_cb: do cb = 1,SP_MB
                ja = colb*npcol*SP_MB + mycol*SP_MB + cb
                ! check to see if this basis function is still included.
                ! If not move onto next one.
                if(.not.incbf(ja)) cycle loop_cb 

                ! calculate local coordinate
                jl = colb*SP_MB+cb

                ! calculate overlap matrix element
                overlpm(il,jl) = calc_olpm(bs.ia,ja,omega0)

             end do loop_cb
          end do loop_colb
          ! do we have an incomplete columb blocks
          if (SP_ceb > 0) then
             colb = SP_nbc
             loop_icb: do cb = 1,SP_ceb
                ja = colb*npcol*SP_MB + mycol*SP_MB + cb

                ! check to see if this basis function is still included.
                ! If not move onto next one.
                if(.not.incbf(ja)) cycle loop_icb 

                ! calculate local coordinate
                jl = colb*SP_MB+cb
                
                ! calculate overlap matrix element
                overlpm(il,jl) = calc_olpm(bs,ia,ja,omega0)

             end do loop_icb
          end if

          ! Fill B (C) array
          ! We only have one RHS therefore only processors in col 0 are needed
          ! This is also why we are not in the column loops
          if(mycol.eq.0) then
             localC(il,1) = bs%C(ia)
          end if
       end do loop_rb
    end do loop_rowb

    ! do we have any incomplete row blocks
    if(SP_reb > 0) then
       rowb = SP_nbr
       loop_irb: do rb = 1,SP_reb        
          ! calculate global coordinate
          ia = rowb*nprow*SP_NB + myrow*SP_NB+rb

          ! check to see if this basis function is still included.
          ! If not move onto next one.
          if(.not.incbf(ia)) cycle loop_irb 

          ! calculate local coordinate
          il = rowb*SP_NB + rb

          loop_irb_colb: do colb = 0,SP_nbc-1
             loop_irb_cb: do cb = 1,SP_MB
                ! calculate global coordinate                
                ja = colb*npcol*SP_MB + mycol*SP_MB + cb

                ! check to see if this basis function is still included.
                ! If not move onto next one.
                if(.not.incbf(ja)) cycle loop_irb_cb 

                ! calculate local coordinate
                jl = colb*SP_MB+cb

                ! calculate overlap matrix element
                overlpm(il,jl) = calc_olpm(bs,ia,ja,omega0)

             end do loop_irb_cb
          end do loop_irb_colb

          ! are there any incomplete columns?
          if (SP_ceb > 0) then
             colb = SP_nbc
             loop_irb_icb: do cb = 1,SP_ceb
                ! calculate global coordinate
                ja = colb*npcol*SP_MB + mycol*SP_MB + cb

                ! check to see if this basis function is still included.
                ! If not move onto next one.
                if(.not.incbf(ja)) cycle loop_irb_icb 

                ! calculate local coordinate
                jl = colb*SP_MB+cb

                ! calculate overlap matrix element
                overlpm(il,jl) = calc_olpm(bs,ia,ja,zb,omega0)

             end do loop_irb_icb
          end if
          if(mycol.eq.0) then
             localC(il,1) = C(ia)
          end if

       end do loop_irb
    end if



    ! Call solver for Hermitian system
    IA = 1
    JA = 1
    call pzgesvx('N','N',nbasis, NRHS, overlpm, IA,JA,SP_DESC_OLAP, &
         overlpmf, IA,JA,SP_DESC_OLAP, &
         IPIV,EQUED,SCA_R,SCA_C,localC, IA,JA,SP_DESC_C, localD, IA,JA,SP_DESC_D,&
         RCOND, FERR, BERR, WORK, LWORK, &
         RWORK, LRWORK,INFO )




    !!    call zhesvx('N','L',nbasis, 1, overlpm, nbasis, overlpmf, nbasis, &
    !!         IPIV, C, nbasis, D, nbasis, RCOND, FERR, BERR, WORK, LWORK, &
    !!         RWORK, INFO )

    ! overlap matrix is now destroyed!
    if(INFO .ne. 0) then
       if (verbosity .gt. 1) then
          write(fug,*)
          write(201+myrank,*) "Processor",myrow,mycol
          write(fug,*) "overlap matrix written out"
          !          open(201+myrank,file = "overlapmatrix")
          write(201+myrank,*) "Overlap matrix"
          write(201+myrank,*) "# i,j,Re(overlap),Im(overlap)"
          do ia = 1,SP_Row_OLAP
             do ja = 1,SP_Col_OLAP
                write(201+myrank,*) ia,ja,real(overlpm(ja,ia)),dimag(overlpm(ja,ia))
             end do
             write(201+myrank,*) 
          end do
          !          close(201+myrank)

          !         open(201+myrank,file = "Dfailed")
          write(201+myrank,*) "D"

          do ia = 1,SP_Row_OLAP
             write(201+myrank,*) D(ia)
          end do
          !          close(81)
       end if
       if (isroot) then
          if (INFO .lt. 0 ) then
             write(fug,*)  "Error solving omega.exp(i(Sj-Si))D = C."
             write(fug,*) "Element ",-INFO," has an illegal value."
             stop
          else if (INFO .ge. 1 .and. INFO.le.nbasis) then
             write(fug,*)  "Error solving omega.exp(i(Sj-Si))D = C."
             write(fug,*) "D(",INFO,",",INFO,") = 0 so matrix is singular."
             stop
          else if (INFO .eq. nbasis+1) then
             write(fug,*)  "Problem solving omega.exp(i(Sj-Si))D = C."
             write(fug,*) " RCOND is less than machine precision.",RCOND
             write(fug,*) " RCOND = 0 indicates that the matrix is singular"
             stop
          else
             write(fug,*) "Error solving omega.exp(i(Sj-Si))D = C. Error number is",info
             stop
          end if
       end if
    end if

    ! ------ Collect answers ------
    D(:) = (0.0d0,0.0d0)
    ! Copy calculated localD into D 
    if (mycol .eq. 0) then
       do rowb = 0,SP_nbr-1
          do rb = 1,SP_NB

             ia = rowb*nprow*SP_NB + myrow*SP_NB+rb
             il = rowb*SP_NB + rb
             D(ia) = localD(il,1)
          end do
       end do

       if(SP_reb > 0) then
          rowb = SP_nbr
          do rb = 1,SP_reb        

             ia = rowb*nprow*SP_NB + myrow*SP_NB+rb
             il = rowb*SP_NB + rb

             D(ia) = localD(il,1)
          end do
       end if
    end if
    ! add up contributions from all processors.
#ifdef USE_MPI1
    mpi_tmp_Z_array(:) = D(:)
    call MPI_Allreduce(mpi_tmp_Z_array, D, nbasis, MPI_DOUBLE_COMPLEX, MPI_SUM, comm, mpierr)
#else
    call MPI_Allreduce(MPI_IN_PLACE, D, nbasis, MPI_DOUBLE_COMPLEX, MPI_SUM, comm, mpierr)
#endif

    ! calculate time spent in this routine
    DsolveTime = DsolveTime + MPI_Wtime() - intime

  end subroutine calcDscalapack

#endif
  ! ===============================================
  !  serial lapack version of calcD.
  ! ===============================================
  ! subroutine calcDserial(C,zb,S,D,incbf)
  subroutine calcDserial(bs,D,incbf)
    ! variables
    use coherentstatestype    
    use coherentstates, ONLY: ndf,nbasis
    use arrays,ONLY: FERR,BERR,RWORK,WORK,LWORK,ipiv,overlpm,overlpmf,t
    use options, ONLY: omega0,verbosity,solvemethod
    ! functions
    use coherentstates, ONLY: overlap_bs
    use files,ONLY: fug
    use scalapackstuff,ONLY: MPI_Wtime
    implicit none
    type(coherentstatebasisset),intent(in) :: bs
!    complex(kind=8),dimension(nbasis),intent(in) :: C
!    real(kind=8),dimension(nbasis),intent(in) :: S ! action
!    complex(kind=8),dimension(ndf,nbasis),intent(in) :: zb ! basis states
    complex(kind=8),dimension(nbasis),intent(out) :: D
    logical,dimension(nbasis),intent(in) :: incbf

    !Variables for LAPACK routines + arrays from arrays module
    integer:: info
    real(kind=8) :: rcond

    integer :: i,j ! for loops
    real(kind=8) ::intime
    
    ! Find out entry time
    intime = MPI_Wtime()

    ! Calculate overlap matrix*exp(i(Sj-Si))
    overlpm(:,:) = (0.0,0.0)
    do i = 1,nbasis
       overlpm(i,i) = (1.0d0,0.0d0)
    end do

    do i = 1,nbasis-1
       if (incbf(i)) then
          do j = i+1,nbasis
             if(incbf(j)) then
                !          write(61,*) overlap(zb(:,i),zb(:,j)),(S(j)-S(i)),cdexp((0.0,1.0)*(S(j)-S(i))),&
!               overlap(zb(:,i),zb(:,j))*cdexp((0.0,1.0)*(S(j)-S(i)))
                 overlpm(i,j) = overlap_bs(bs,i,j)
                 if (abs(overlpm(i,j)) .ge. omega0) then
                    ! What should go here for S??????
                    if(bs%nehrenfest .le. 1) then  
                       ! If we are not doing Ehrenfest, multiply by exp(i dS). 
                       ! If we are doing Ehrenfest, this factor forms part of the overlap
                       overlpm(i,j) = overlpm(i,j)*cdexp((0.0,1.0)*(bs%S(1,j)-bs%S(1,i)))
                    end if                    
                   overlpm(j,i) = dconjg(overlpm(i,j))
                else
                   overlpm(i,j) = 0.0d0
                   overlpm(j,i) = 0.0d0
                   !             write(61,*) "Not including",i,j
                end if
!          write(61,*)
             end if
          end do
       end if
    end do

    if (solvemethod == 2) then
!       D(:) = C(:)
!       D(:) = (0.0d0,0.0d0)
       call cgsolve(D,overlpm,bs%C,nbasis)
    else
    ! Call solver for Hermitian system
    call zhesvx('N','L',nbasis, 1, overlpm, nbasis, overlpmf, nbasis, &
         IPIV, bs%C, nbasis, D, nbasis, RCOND, FERR, BERR, WORK, LWORK, &
         RWORK, INFO )
    ! Call solver for general complex system
!    call zgesvx('N','N',nbasis, 1, overlpm, nbasis, overlpmf, nbasis, &
!         IPIV,'N',GenR,GenC,C, nbasis, D, nbasis, RCOND, FERR, BERR, WORK, &
!         RWORK, INFO )

    ! overlap matrix is now destroyed!
    if(INFO .ne. 0) then
       if (verbosity .gt. 1) then
          write(fug,*)
          write(fug,*) "overlap matrix written out"
          open(81,file = "overlapmatrix")
          write(81,*) "# i,j,Re(overlap),Im(overlap)"
          do i = 1,nbasis
             do j = 1,nbasis
                write(81,*) i,j,real(overlpm(j,i)),dimag(overlpm(j,i))
             end do
             write(81,*) 
          end do
          close(81)

          open(81,file = "Dfailed")
          write(81,*)
          do i = 1,nbasis
             write(81,*) D(i)
          end do
          close(81)
       end if

       if (INFO .lt. 0) then
          write(fug,*)  "Error solving omega.exp(i(Sj-Si))D = C."
          write(fug,*) "Element ",-INFO," has an illegal value."
          stop
       else if (INFO .ge. 1 .and. INFO.le.nbasis) then
          write(fug,*)  "Error solving omega.exp(i(Sj-Si))D = C."
          write(fug,*) "D(",INFO,",",INFO,") = 0 so matrix is singular."
          stop
       else if (INFO .eq. nbasis+1) then
          write(fug,*)  "Problem solving omega.exp(i(Sj-Si))D = C."
          write(fug,*) " RCOND is less than machine precision.",RCOND
          write(fug,*) " RCOND = 0 indicates that the matrix is singular"
          stop
       else
          write(fug,*) "Error solving omega.exp(i(Sj-Si))D = C. Error number is",info
          stop
       end if

    end if
    end if ! solve method
    ! calculate time spent in this routine
    DsolveTime = DsolveTime + MPI_Wtime() - intime

  end subroutine calcDserial


  ! ============================================================
  !     cgsolve
  ! 
  !  Subroutine to solve a system of linear equations using
  !  the conjugate gradient's method.
  ! ============================================================
  subroutine cgsolve(x,Omega,c,N)
    use files,ONLY: fug
    use options, ONLY: maxcgiterations, cgepsilonsq, verbosity
    implicit none
    complex(kind=8),dimension(:),intent(inout) :: x
    complex(kind=8),dimension(:,:),intent(in) :: Omega
    complex(kind=8),dimension(:),intent(in) :: c
    integer, intent(in) :: N
    complex(kind=8), dimension(:),allocatable,save ::cgr,cgd,cgq ! residual
    ! ---- local variables ----
    integer :: i,j
    logical,save :: firsttime = .true.

    real(kind=8) :: deltanew,deltaold
    real(kind=8) :: alpha,beta
    
    if(.not. allocated(cgr)) allocate(cgr(N),cgd(N),cgq(N))
    i = 0

    if (firsttime) then
       x(:) = C(:)
       firsttime = .false.
    end if
    ! First guess D = C
    do i = 1,N
       cgr(i) = C(i) - sum(Omega(i,:)*x(:))
    end do
    cgd(:) = cgr(:)

    deltanew = sum(conjg(cgr(:))*cgr(:))
    
    deltaold = deltanew

    do while ( i < maxcgiterations .and. deltanew .gt. cgepsilonsq)
       !  .and. deltanew > cgepsilonsq * deltaold

       do j =  1, N
          cgq(j) = sum(Omega(j,:)*cgd(:))
       end do
       
!       write(114,*) i,"d* . q",real(sum(conjg(cgd(:))*cgq(:))), imag(sum(conjg(cgd(:))*cgq(:)))
       alpha = deltanew/ real(sum(conjg(cgd(:))*cgq(:)))
       
       x(:) = x(:) + alpha*cgd(:)
       ! Compute residual
       if (mod(i,50) == 0) then
          do j =  1, N
             cgr(j) = c(j) - sum(Omega(j,:)*x(:))
          end do
       else
          cgr(:) = cgr(:) - alpha*cgq
       end if

       deltaold = deltanew
       deltanew = sum(conjg(cgr(:))*cgr(:))
       beta = deltanew/deltaold
       
       if (verbosity > 1 ) then
          write(fug,*) i,deltanew, deltaold,beta,cgepsilonsq
       end if
       
       cgd(:) = cgr(:) + beta*cgd(:)
       i = i + 1       

    end do

    if (i == maxcgiterations) then
       write(fug,*) " The conjugate gradient loop failed to converge."
       flush(fug)
       stop
    else
       write(fug,*) " Conjugate gradient loop converged in ",i," steps"
    end if

  end subroutine cgsolve


  ! ===============================================
  !    dC_dt
  ! Calculate the time derivative of c
  ! ===============================================
  subroutine dC_dt(bs,D,dCdt,incbf)
    use coherentstatestype
    use coherentstates, ONLY: ndf,nbasis
    use arrays, ONLY: overlpm
    ! use functions too
    use coherentstates, ONLY: overlap_bs
    use options, ONLY: omega0
    use files, ONLY: fug ! General output file unit.
#ifdef USE_SCALAPACK
#ifdef USE_MPI1
    use scalapackstuff, ONLY: MPI_Wtime,myrank,comm,mpi_double_complex,mpi_sum
    use arrays, ONLY:mpi_tmp_Z_array, mpi_tmp_D
#else
    use scalapackstuff, ONLY: MPI_Wtime,myrank,comm,mpi_in_place,mpi_double_complex,mpi_sum
#endif
#else
    use scalapackstuff, ONLY: MPI_Wtime
#endif
    implicit none
    type(coherentstatebasisset) :: bs
    ! real(kind=8),dimension(nbasis),intent(in) :: S ! action
    ! complex(kind=8),dimension(ndf,nbasis),intent(in) :: zb ! basis states
    !complex(kind=8),dimension(nbasis),intent(in) :: C,D
    complex(kind=8),dimension(nbasis),intent(in) :: D
    complex(kind=8),dimension(nbasis),intent(out) :: dCdt
    logical,dimension(nbasis),intent(in) :: incbf

    complex(kind=8) :: overlap_ij

    integer :: i,j ! for loops
    real(kind=8) :: intime
    integer :: mpierr

    intime = MPI_Wtime()

    dCdT(:) = (0.0,0.0) 

!    loop_dCdt_i: do i = 1,nbasis
    loop_dCdt_i: do i = ibeg_nbasis,iend_nbasis
       if(incbf(i)) then
          loop_dCdt_j:do j=1,nbasis
             if (i.eq. j) cycle loop_dCdt_j

             if(incbf(j)) then
                !overlap_ij = overlap(zb(:,i),zb(:,j))*cdexp((0.0,1.0)*(S(j)-S(i)))
                overlap_ij = overlap_bs(bs,i,j)
                   if (bs%nehrenfest .le. 1) then
                      overlap_ij = overlap_ij*cdexp((0.0,1.0)*(bs%S(1,j)-bs%S(1,i)))
                   end if          
                    if (abs(overlap_ij) .gt. omega0) then
                   dCdt(i) =  dCdt(i) + delta_h_ord (bs%zb(:,i),dconjg(bs%zb(:,i)),bs%zb(:,j)) &
                        * overlap_ij * D(j)
               ! else 
               !    write(*,*) "Small overlap ",overlap_ij,"dHord", &
               !         delta_h_ord (zb(:,i),dconjg(zb(:,i)),zb(:,j)) &
               !         * overlap_ij * D(j)
                end if
             end if
          end do loop_dCdt_j
       end if
    end do loop_dCdt_i
    dCdt(ibeg_nbasis:iend_nbasis) = (0,-1)*dCdt(ibeg_nbasis:iend_nbasis)

    ! add up contributions from all processors.
#ifdef USE_SCALAPACK
#ifdef USE_MPI1
    mpi_tmp_Z_array(:) = dCdt(:)
    call MPI_Allreduce(mpi_tmp_Z_array, dCdt, nbasis, MPI_DOUBLE_COMPLEX, MPI_SUM, comm, mpierr)
#else
    call MPI_Allreduce(MPI_IN_PLACE, dCdt, nbasis, MPI_DOUBLE_COMPLEX, MPI_SUM, comm, mpierr)
#endif
#endif
    dC_dtTime = dC_dtTime + MPI_Wtime() - intime
    
    
  end subroutine dC_dt


  ! **************************************************
  !                    Integrate
  !
  !                  Stewart K. Reed
  !                   5th June 2007
  !
  !  Description:
  !      This file contains routines to integrate
  !      the "equations of motion"
  ! **************************************************

  !subroutine calcdSanddZ(zb,dzdt,dSdt,incbf)
  subroutine calcdSanddZ(bs,dbsdt,incbf)
    use options, ONLY:dt,classical
    use coherentstates, ONLY: ndf,nbasis
    use coherentstatestype
    implicit none
    type(coherentstatebasisset), intent(in) :: bs
    type(coherentstatebasisset), intent(inout) :: dbsdt
!    complex(kind=8),dimension(ndf,nbasis), intent(in) :: zb
!    complex(kind=8),dimension(ndf,nbasis), intent(out) :: dzdt
!    real(kind=8),dimension(nbasis),intent(out) :: dSdt
    logical,dimension(nbasis),intent(in) :: incbf

    complex(kind=8),dimension(ndf) :: dzcdt
    integer :: i

    do i = 1,nbasis
       if (incbf(i)) then
          ! work out dz/dt and dz*/dt
          call dzc_dt(bs%zb(:,i),dzcdt)
          dbsdt%zb(:,i) = dconjg(dzcdt)
          
          ! Work out dS/dt
          if (bs%nehrenfest .le. 1) then
             call dS_dt(bs%zb(:,i),dconjg(bs%zb(:,i)),dbsdt%zb(:,i),dzcdt,dbsdt%S(1,i))
          else
             write(99,*) "I don't know how to derivs of S for Ehrenfest yet"
          end if
       end if
    end do
  end subroutine calcdSanddZ


  subroutine calcZnew(zb,dzdt)
    use options, ONLY:dt
    use coherentstates, ONLY: ndf,nbasis
    implicit none
    complex(kind=8),dimension(ndf,nbasis), intent(in) :: zb
    complex(kind=8),dimension(ndf,nbasis), intent(out) :: dzdt
    
    complex(kind=8),dimension(ndf) :: dzcdt
    integer :: i
    
    do i = 1,nbasis
       ! work out dz/dt and dz*/dt
       call dz_dt(zb(:,i),dzdt(:,i))
    end do
  end subroutine calcZnew
  
  ! ------------------------------------------------------------------------
  ! Subroutine: calcclassicalderivs
  ! Description: A simple routine to calculate deriviates of a basis
  !              set when doing classical or semi classical simulations.
  ! ------------------------------------------------------------------------
  subroutine calcclassicalderivs(bs,dbsdt)
    use coherentstatestype
    implicit none
    type(coherentstatebasisset),intent(in) :: bs
    type(coherentstatebasisset),intent(inout) :: dbsdt

    ! Calculate the position derivative
    call calcZnew(bs%zb,dbsdt%zb)
    ! Set other derivatives to zero
    dbsdt%C(:) = 0.0d0
    dbsdt%S(:,:) = 0.0d0
    dbsdt%ErnC(:,:) = 0.0d0
    
  end subroutine calcclassicalderivs
     
  ! ------------------------------------------------------------------------
  ! Subroutine: calcderivs
  ! Description: A simple routine to calculate deriviates of a basis
  !              set when doing quantum dynamics simulations.
  ! ------------------------------------------------------------------------
  subroutine calcderivs(bs,dbsdt,D,incbf)
    use coherentstatestype
    implicit none
    type(coherentstatebasisset),intent(in) :: bs
    type(coherentstatebasisset),intent(inout) :: dbsdt
    complex(kind=8),dimension(:) :: D
    logical,dimension(:),intent(in) :: incbf

    !     call dC_dt(bs%S(1,:),bs%zb,bs%C,D,bs1%C,incbf)
     call dC_dt(bs,D,dbsdt%C,incbf)
     !call calcdSanddZ(bs%zb,bs1%zb,bs1%S(1,:),incbf)
     call calcdSanddZ(bs,dbsdt,incbf)
     
     dbsdt%ErnC(:,:) = 0.0d0
     
!     if(ncart_constraint > 0) then
!        call cart_constraint(bs,dbsdt)
!     end if
     
   end subroutine calcderivs

  ! ------------------------------------------------------------------------
  ! Subroutine: step
  ! Description: This routine does the integration. Currently it uses a
  !              fourth order Runge Kutta Algorithm.
  ! ------------------------------------------------------------------------
  subroutine step
    use coherentstatestype
    use options, ONLY: dt,classical, tu, minprob,maxz,writeexpect,newton
    
    use coherentstates, ONLY: ndf,nbasis
    use arrays, ONLY: bs,bs1,bs2,bs3,bs4,bsnew,t,D,Dold, &
         incbf,oldincbf,toteng,dtoteng,dhexpect,hework

! zb,S,C,D,t, zbnew,zk1,zk2,zk3,zk4, &
!         Snew, Sk1,Sk2,Sk3,Sk4,Cnew,Dold, Ck1, Ck2,Ck3,Ck4,&
!         incbf,oldincbf,toteng,dtoteng,dhexpect,hework
    
    use files,ONLY: fug
    use scalapackstuff, ONLY: MPI_Wtime

    implicit none
    complex(kind=8),dimension(ndf):: z,zc
    
    real(kind=8) :: dE ! change in energy from removing these basis functions
    ! dtoteng is change due all removed fcts.
    logical :: dropped
    integer :: i,j
    
    ! For calculating change in expectation value on dropping basis functions
    real(kind=8) :: dhe,mc

    real(kind=8) :: intime ! For doing timings
    intime = MPI_Wtime()
    ! Integrate using a Runge Kutta 4th order algorithm
    ! Except we also have the option of a simple newton method.
    if (classical > 0 .and. newton) then
       call calcclassicalderivs(bs,bs1)
       call rconst_times_csbs_sub( dt,bs1,bs1)
       call add_basis_sets(bs,bs1,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs1,dt)
       end if
       bs%zb(:,:) = bsnew%zb(:,:)
       
    else if(classical > 0) then
       
       ! Calculate k1 
       call calcclassicalderivs(bs,bs1)
!       bs1 = 0.5d0*dt*bs1
!      bsnew = bs + bs1       
       call rconst_times_csbs_sub( 0.5d0*dt,bs1,bs1)
       call add_basis_sets(bs,bs1,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs1,0.5d0*dt)
       end if

       ! Calculate k2
       call calcclassicalderivs(bsnew,bs2)
       !       bs2 = dt*bs2
       !bsnew = bs + 0.5d0*bs2
       call rconst_times_csbs_sub( dt,bs2,bs2)
       call weightedadd_basis_sets(bs,bs2,0.5d0,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs2,0.5d0*dt)
       end if

       ! Calculate k3
       call calcclassicalderivs(bsnew,bs3)
!       bs3 = dt*bs3
!       bsnew = bs + bs3
       call rconst_times_csbs_sub( dt,bs3,bs3)
       call add_basis_sets(bs,bs3,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs3,dt)
       end if

       ! Calculate k4
        call calcclassicalderivs(bsnew,bs4)
!       bs4 = (dt/2.0d0)*bs4
!       bs = bs + (bs1 + bs2 + bs3 + bs4)/3.0d0

        if(ncart_constraint > 0) then
           !           call add_basis_sets(bs,bs4,bsnew)
           ! don't need bnew if not applyingng constraints
           call rconst_times_csbs_sub( dt,bs4,bs4)                      
           call add_basis_sets(bs,bs4,bsnew)
           call cart_constraint(bs,bsnew,bs4,dt)
           call rconst_times_csbs_sub( 0.5d0,bs4,bs4)
        else
           call rconst_times_csbs_sub( 0.5d0*dt,bs4,bs4)
        end if
       
       call add_4basis_sets(bs1,bs2,bs3,bs4,bsnew)
       if (ncart_constraint > 0) then
          bs3%zb(:,:) = bs%zb(:,:)
       end if
       call weightedadd_basis_sets(bs,bsnew,1.0d0/3.0d0, bs)
       if(ncart_constraint > 0) then
          call rconst_times_csbs_sub( 1.0d0/3.0d0,bsnew,bsnew)                 
          ! The old positions are now stored in bs3 and
          ! we can use bsnew as the gradient. Admittedly this is slightly
          ! confusing but it saves defining another array and we don't need 
          ! to use bsnew after this.
           call cart_constraint(bs3,bs,bsnew,dt)
       end if



!!$       call calcZnew(zb,zk1)
!!$       zk1(:,:) = dt*zk1(:,:)/2.0 ! as above
!!$       zbnew(:,:) = zb(:,:) +zk1(:,:)
!!$       
!!$       ! Calculate k2
!!$       call calcZnew(zbnew,zk2)
!!$       zk2(:,:) = dt*zk2(:,:)
!!$       zbnew(:,:) = zb(:,:) +zk2(:,:)/2.0
!!$       
!!$       ! Calculate k3
!!$       call calcZnew(zbnew,zk3)
!!$       zk3(:,:) = dt*zk3(:,:)
!!$       zbnew(:,:) = zb(:,:) +zk3(:,:) 
!!$       
!!$       ! Calculate k4 but don't need recalculate Cnew etc. 
!!$       call calcZnew(zbnew,zk4)
!!$       zk4(:,:) = dt*zk4(:,:)/2.0
!!$       
!!$       ! Now work out the Real new values
!!$       ! The factors of 1/2 for k1 and k4 are already included.
!!$       zb(:,:) = zb(:,:) + (zk1(:,:) + zk2(:,:) + zk3(:,:) + zk4(:,:))/3.0
       
    else
       ! Calculate k1 
       call calcderivs(bs,bs1,D,incbf)
!       bs1 = (dt/2.0d0)*bs1
!       bsnew = bs + bs1
       call rconst_times_csbs_sub( 0.5d0*dt,bs1,bs1)
       call add_basis_sets(bs,bs1,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs1,0.5d0*dt)
       end if
       ! call calcD(bsnew%C,bsnew%zb,bsnew%S,D,incbf)
       call calcD(bsnew,D,incbf) 

      ! Calculate k2
       call calcderivs(bsnew,bs2,D,incbf)
!       bs2 = dt*bs2
!       bsnew = bs + 0.5d0*bs2
       call rconst_times_csbs_sub( dt,bs2,bs2)
       call weightedadd_basis_sets(bs,bs2,0.5d0,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs2,0.5d0*dt)
       end if
       ! call calcD(bsnew%C,bsnew%zb,bsnew%S,D,incbf)
       call calcD(bsnew,D,incbf) 

       ! Calculate k3
       call calcderivs(bsnew,bs3,D,incbf)
!       bs3 = dt*bs3
!       bsnew = bs + 0.5d0*bs3
       call rconst_times_csbs_sub( dt,bs3,bs3)
       call add_basis_sets(bs,bs3,bsnew)
       if(ncart_constraint > 0) then
          call cart_constraint(bs,bsnew,bs3,dt)
       end if
       ! call calcD(bsnew%C,bsnew%zb,bsnew%S,D,incbf)
       call calcD(bsnew,D,incbf) 

       ! Calculate k4
       call calcderivs(bsnew,bs4,D,incbf)
!       bs4 = (dt/2.0d0)*bs4
!       bs = bs + (bs1 + bs2 + bs3 + bs4)/3.0d0
!!$       call rconst_times_csbs_sub( 0.5d0*dt,bs4,bs4)
!!$       if(ncart_constraint > 0) then
!!$          call add_basis_sets(bs,bs4,bsnew)
!!$          call cart_constraint(bs,bsnew,bs4,dt)
!!$       end if
!!$       call add_4basis_sets(bs1,bs2,bs3,bs4,bsnew)
!!$       call weightedadd_basis_sets(bs,bsnew,1.0d0/3.0d0, bs)
       ! call calcD(bs%C,bs%zb,bs%S,D,incbf)
       if(ncart_constraint > 0) then
           !           call add_basis_sets(bs,bs4,bsnew)
           ! don't need bnew if not applyingng constraints
           call rconst_times_csbs_sub( dt,bs4,bs4)                      
           call add_basis_sets(bs,bs4,bsnew)
           call cart_constraint(bs,bsnew,bs4,dt)
           call rconst_times_csbs_sub( 0.5d0,bs4,bs4)
        else
           call rconst_times_csbs_sub( 0.5d0*dt,bs4,bs4)
        end if
       
       call add_4basis_sets(bs1,bs2,bs3,bs4,bsnew)
       if (ncart_constraint > 0) then
          bs3%zb(:,:) = bs%zb(:,:)
       end if
       call weightedadd_basis_sets(bs,bsnew,1.0d0/3.0d0, bs)
       if(ncart_constraint > 0) then
          call rconst_times_csbs_sub( 1.0d0/3.0d0,bsnew,bsnew)                 
          ! The old positions are now stored in bs3 and
          ! we can use bsnew as the gradient. Admittedly this is slightly
          ! confusing but it saves defining another array and we don't need 
          ! to use bsnew after this.
           call cart_constraint(bs3,bs,bsnew,dt)
       end if


       call calcD(bs,D,incbf) 


    end if

    ! ---- Should any basis function be dropped ----
    ! Check position and probability if minimum values are
    ! greater than 0.
    dropped = .false. ! have any functions been dropped?
    
    ! Keep the old list of which basis functions are included
    if (minprob > 0 .or. maxz > 0 ) then
       oldincbf(:) = incbf(:)
    end if
    
    ! Now check to see if any basis functions should be removed
    ! because its probability is too low
    if (minprob > 0) then
       do i = 1,nbasis
          if (incbf(i)) then
             if(abs(dconjg(bs%C(i))*D(i)) < minprob) then
                incbf(i) = .false.
                write(fug,*) "Dropping basis function ",i," prob: ",abs(dconjg(bs%C(i))*D(i))
                dropped = .true.
             end if
          end if
       end do
    end if
    
    ! or because of its position too. 
    ! check each degree of freedom separately
    if (maxz > 0) then
       do i = 1,nbasis
          if (incbf(i)) then
             checkposdf: do j = 1,ndf
                if (cdabs(bs%zb(j,i)) > maxz ) then
                   incbf(i) = .false.
                   write(fug,*) "Dropping basis function ",i," mode ",j," |z|=",cdabs(bs%zb(j,i))
                   dropped = .true.
                   !                      zb(j,i) = dconjg(zb(j,i))
                   !                      write(fug,*) "Reversing basis function ",i," mode ",j," |z|=",cdabs(zb(j,i))
                   
                   exit checkposdf
                end if
             end do checkposdf
          end if
       end do
    end if
       
    ! if any basis functions have been dropped, calculate the resulting change in energy
    ! so that we can add it on again for energy conservation.
    if (dropped) then
       if(writeexpect >0) then
          hework(1) = expectationvalue(bs%zb,bs%C,D,oldincbf)
          if (hamtype .eq. 3 .or. hamtype .eq. 5) then
             do i = 1,ndf
                ! Calculate energy of each mode but ignore coupling
                call modeexpectationvalue(bs%zb,bs%C,D,oldincbf,i,hework(i+1),mc)
             end do
          end if
       end if


       ! Need to recalculate D as the set of basis functions has changed.
       ! 26/5/2009 If we are doing a classical calculation then D=C. 
       if( classical .eq. 0) then
         ! call calcD(bs%C,bs%zb,bs%S,D,incbf)
          call calcD(bs,D,incbf) 
       end if

       ! Now calculate the change in energy
       dE =  totalenergy(bs,D,incbf) -  toteng ! totalenergy(zb,D,S,oldincbf)
       write(fug,*) "Change in energy ", dE
       dtoteng = dtoteng + dE
       write(fug,*) "Total change in energy ", dtoteng

       if(writeexpect >0) then
          ! total energy
          dhe = expectationvalue(bs%zb,bs%C,D,incbf)-hework(1)
          dhexpect(1) = dhexpect(1)+dhe

          if (hamtype .eq. 3 .or. hamtype .eq. 5) then
             do i = 1,ndf
                ! Calculate energy of each mode but ignore coupling
                call modeexpectationvalue(bs%zb,bs%C,D,incbf,i,dhe,mc)
                dhexpect(i+1) = dhexpect(i+1)+dhe-hework(i+1)
             end do
          end if
       end if
    end if
       

!!    end if ! if classical

    ! What is the current time?
    t = t + dt

    ! time spent in this routine
    StepTime =StepTime + MPI_Wtime() - intime
        
  end subroutine step




  ! **************************************************
  !             General Setup/cleanup routines
  ! **************************************************
  ! ==================================================
  !            Work out loop bounds
  ! 
  ! ================================================== 
  subroutine determineloopbounds
    use scalapackstuff, ONLY: myrank,nprocs
    use coherentstates, ONLY: nbasis
    implicit none
    
    integer :: neach,nextra
    
    ! work out how to divide nbasis iterations amoungst nprocs processors
    ! nprocs*neach + nextra = nbasis
    neach = floor(dble(nbasis)/dble(nprocs))
    nextra = mod(nbasis,nprocs)

    ! If all have the same number then the first one for each processor is
    ibeg_nbasis = myrank*neach+1
    ! but we have nextra extra ones. Starting with processor 0, give 1 to
    ! each processor until we run out.
    if (nextra > myrank) then
       ibeg_nbasis = ibeg_nbasis + myrank
    else
       ibeg_nbasis = ibeg_nbasis + nextra
    end if

    ! Now work out end points basiscally it should be beg+neach -1 as start at 1.
    ! Low rank processors get an extra one.
    if (nextra > myrank) then
       iend_nbasis = ibeg_nbasis+neach - 1 + 1
    else
       iend_nbasis = ibeg_nbasis+neach - 1
    end if

    end subroutine determineloopbounds
  ! ==================================================
  !            Read Hamiltomian parameters
  !  allocates arrays too.
  ! ================================================== 
  subroutine readhamparams
    use coherentstates,ONLY:ndf
    use files, ONLY: fug,fham
    use errors
    implicit none
    character(len=50) :: line
    integer :: ierr ! For errors reading from the file
    integer :: n    ! for checking the number of degrees of freedom
    integer :: i    ! For loops, what else?

    if (ndf .le. 0) then
       write(fug,*) "There are no degrees of freedom!"
       stop
    end if

    ! Open file
    open(fham,file="hamiltonian.inpt",status='old',iostat=ierr)
    if(ierr .ne. 0 ) then
       call fileerror(ierr,"hamiltonian.inpt")
    end if

    ! Read the first line which should say what type of potential 
    ! we have.
    read(fham,*,iostat=ierr) hamname
    if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading system name")

    ! Read number of degrees of freedom
    read(fham,*,iostat=ierr) line
    if  (line == "ndf" ) then
       backspace(fham) 
       read(fham,*,iostat=ierr) line,n
       if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading number of degrees of freedom")
    else
       write(fug,*) "Failed to read the number of degrees of freedom in hamiltonian.inpt"
       stop
    end if
    ! Check that we have the correct number of degrees of freedom.
    if (n .ne. ndf) then
       write(fug,*) "There is a miss match between the number of degrees of freedom in run.inpt and in hamiltonian.inpt"
       stop
    end if

    select case (hamname)
    case ("tunnelbath")
       hamtype = 11

       call readhamtunbath
    case ("henonheiles")
       hamtype = 2
       call readhamhh
    case ("polymorse")
       hamtype = 3
       call readhampm
    case ("cartesian")
       hamtype = cartham
       call readhamcart
    case ("classicalcartesian") 
       hamtype = clascartham
       call readhamcart ! Uses the same routine as the cartesian hamiltonian
    case ("classicalpolymorse")
       hamtype = 5
       call readchampm
    case default
       write(fug,*) "Sorry, I don't know how to use a ",hamname," hamiltonian"
       stop
    end select
    !      if (hamtype .eq. "tunnelbath") then
    !      end if


  end subroutine readhamparams

end module hamiltonian
