 ! ==================================================
  !          Tunnel + Harmonic Bath Hamiltomian 
  ! ==================================================
  function hamtunbath(zic,zj)
    use coherentstates, ONLY: ndf,gamma
    ! include subroutines too.
    use coherentstates, ONLY: pos,mom
 
    implicit none

    ! two states <z_i| and |z_j>
    complex(kind=8),dimension(ndf),intent(in) :: zic,zj
    ! hamiltonian
    complex(kind=8):: hamtunbath

    ! position and momentum vectors
    complex(kind=8),dimension(ndf) :: q,p

    ! Set the Hamiltonian to 0.
    hamtunbath = (0.0d0,0.0d0)

    ! Calculate position and momentum vectors
    call pos(zj,zic,q)
    call mom(zj,zic,p)

    ! Now calculate the Hamiltonian
    ! p_i^2/2m_i+gamma_i/4m_i
    hamtunbath = sum( (p(:)**2 + gamma(:)/2) /m(:))/2.0

    ! (q1^4+3q_1/gamma_i+3/4gamma_i^2)/16eta -q1^2/2-1/(4gamma1)
    hamtunbath = hamtunbath &
         + (q(1)**4+3.0*q(1)**2/gamma(1)+0.75/gamma(1)**2)/(16.0*eta) &
         - q(1)**2/2 -0.25/gamma(1)

    ! sum(m_j*omega_j^2*(q_j^2+gamma^j/2),j=2..ndf)/2
    hamtunbath = hamtunbath &
         + sum(m(2:ndf)*omega2(2:ndf)*(q(2:ndf)**2+1/(gamma(2:ndf)*2.0)))/2.0

    ! Coupling
    ! sum(coupling(j)*q(j)*q(j-1)
    hamtunbath = hamtunbath &
         + sum(coup(2:ndf)*q(2:ndf)*q(1:ndf-1))

  end function hamtunbath

  ! ==================================================
  !                 Derivatives of 
  !        the tunnel + Harmonic Bath Hamiltomian 
  ! ==================================================
  function dhamtunbathdz (zic,zj)
    use coherentstates, ONLY: ndf,gamma,s2gammarec,shgamma
    ! include subroutines too.
    use coherentstates, ONLY: pos,mom
    implicit none

    ! two states <z_i| and |z_j>
    complex(kind=8),dimension(ndf),intent(in) :: zic,zj

    ! derivative dh/dz - this function
    complex(kind=8) :: dhamtunbathdz(ndf)


    ! position and momentum vectors
    complex(kind=8),dimension(ndf) :: q,p

    ! Calculate position and momentum vectors
    call pos(zj,zic,q)
    call mom(zj,zic,p)

    ! dq_i/dz_i = 1/(sqrt( 2gamma(i)) = s2gammarec(i)
    ! dp_i/dz_i = -ihbar*sqrt(gamma(i)/2) = (0,-1)*shgamma(i) as hbar = 1
    ! first term
    dhamtunbathdz(1) = p(1)/m(1)*(0,-1)*shgamma(1) &
         + ((4*q(1)**3+6*q(1)/gamma(1))/(16*eta) &
         - q(1)+coup(2)*q(2))*s2gammarec(1)

    ! 2 <= j <= N
    dhamtunbathdz(2:ndf) = p(2:ndf)/m(2:ndf)*(0,-1)*shgamma(2:ndf) &
         + (m(2:ndf)*omega2(2:ndf)*q(2:ndf)  &
         + coup(2:ndf)*q(1:ndf-1))*s2gammarec(2:ndf)

    ! 2<= j <= N-1
    dhamtunbathdz(2:ndf-1) =  dhamtunbathdz(2:ndf-1) &
         + coup(3:ndf)*q(3:ndf)*s2gammarec(2:ndf-1)


  end function dhamtunbathdz

  ! ==================================================
  !            Read Hamiltomian parameters
  !     specific to the Tunnel + Bath Hamiltonian
  !  allocates arrays too.
  ! ==================================================
  subroutine readhamtunbath
    use  coherentstates, ONLY: ndf
    use files, ONLY: fug,fham
    use errors

    character(len=50) :: hamtype,line
    integer :: ierr ! For errors reading from the file
    integer :: n    ! for checking the number of degrees of freedom
    integer :: i    ! For loops, what else?

    ! Allocate more general arrays
    call allocatehamarrays

    ! Allocate Hamiltonian specific arrays
    ! Bath mode frequencies
    allocate(omega(2:ndf),omega2(2:ndf))
    omega(:) = 1.0
    omega2(:) = 1.0

    ! Coupling between degrees of freedom
    allocate(coup(ndf))
    coup(2:ndf) = 0.0

    ! Read in masses
    read(fham,*,iostat=ierr) line
    if  (line .ne. "mass" .or. ierr .ne. 0  ) then
       write(fug,*) "Failed to read keyword mass in hamiltonian.inpt"
       stop
    end if
    do i =1,ndf
       read(fham,*,iostat=ierr) m(i)
       if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading masses")
    end do

    ! Read in eta 
    read(fham,*,iostat=ierr) line,eta
    if  (line .ne. "eta" .or. ierr .ne. 0  ) then
       write(fug,*) "Failed to read keyword eta in hamiltonian.inpt"
       stop
    end if
    backspace(fham)
    read(fham,*,iostat=ierr) line, eta
    if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading eta")

    ! Read in omegas
    read(fham,*,iostat=ierr) line
    if  (line .ne. "omega" .or. ierr .ne. 0  ) then
       write(fug,*) "Failed to read keyword omega in hamiltonian.inpt"
       stop
    end if
    do i = 2,ndf
       read(fham,*,iostat=ierr) omega(i)
       if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading omegas")
       omega2(i) = omega(i)*omega(i)

    end do

    ! Read in coupling constants 
    read(fham,*,iostat=ierr) line
    if  (line .ne. "coupling" .or. ierr .ne. 0  ) then
       write(fug,*) "Failed to read keyword omega in hamiltonian.inpt"
       stop
    end if
    do i = 2,ndf
       read(fham,*) coup(i)
    end do


    ! Summarise
    write(fug,*) "Hamiltonian parameters"
    write(fug,*) "----------------------"
    write(fug,*) "  Studying a tunnelling system with harmonic bath modes"
    write(fug,*) "  Masses:               i     m_i"
    do i = 1,ndf
       write(fug,"(a21,i5,f8.4)") "                     ",i,m(i)
    end do
    write(fug,*) "  eta =",eta
    write(fug,*) "  Bath frequencies.     i   omega_i"
    do i = 2,ndf
       write(fug,"(a21,i5,f8.4)") "                     ",i,omega(i)
    end do
    write(fug,*) "  Coupling strength     i     c_i"
    do i = 2,ndf
       write(fug,"(a21,i5,f8.4)") "                     ",i,coup(i)
    end do
    write(fug,*) 

  end subroutine readhamtunbath
