! **************************************************
!                    initstates
!
!                  Stewart K. Reed
!                   5th June 2007
!
!  Description:
!      This file contains routines to read in /
!      setup initial states.
! **************************************************

subroutine readinitstates
  use arrays, ONLY: ninitstates,zinit
  use files,ONLY: fug,fis
  use coherentstates, ONLY:ndf,nbasis,s2gammarec,shgamma
  use options, ONLY:eu,energyunits,ccorlog ! energy units
  use scalapackstuff, ONLY: isroot
  ! subroutines
  use hamiltonian, ONLY:h_ord
  use errors

  implicit none
  character(len=50) :: filename,line
  real(kind=8) :: q,p ! Initial position and momentum
  integer :: i,j ! For loops
  integer :: ierr ! File I/O errors

  ! Set the name of the file we are reading from
  filename="states.inpt"

  ! Open the file
  open(fis,file=filename,status='old',iostat=ierr)
  if(ierr .ne. 0 ) then
     call fileerror(ierr,filename)
  end if

  ! How many states are given in the file?
  read(fis,*,iostat=ierr) line
  if  (line == "nstates" ) then
     backspace(fis) 
     read(fis,*,iostat=ierr) line,ninitstates
     if(ierr.ne.0) call filereaderror(ierr,filename,"Error reading number of states")
  else
     write(fug,*) "Failed to read the number of states in ",trim(filename)
     stop
  end if

  ! Check that we have enough states
  if (ccorlog > 0 .and.ninitstates .le. 1)  then
     write(fug,*) "Need at least two states in initstates in order to calculate the cross correlation function"
     stop
  end if

  ! Now allocate the arrays
  if(allocated(zinit)) then
     write(fug,*) "Ahhhh....zinit is already allocated in readinitstates!"
     stop
  end if
  allocate(zinit(ndf,ninitstates))
  zinit(:,:) = (0.0,0.0)

  do i = 1, ninitstates
     read(fis,*,iostat=ierr) line
     if ( line .ne. "state".or.ierr .ne. 0) then
        write(fug,*) "Error reading state keyword ",i," in ",trim(filename)
        write(fug,*) "Actually read:",line
        stop
     end if
     do j = 1,ndf
        read(fis,*,iostat=ierr) q,p
        if ( ierr .ne. 0) then
           write(fug,*) "Error reading position/momentum ",j," state", &
                i," in ",trim(filename)
           stop
        end if
        ! z = sqrt(gamma(j)/2)*q+ i*p/sqrt(2*gamma(j))
        zinit(j,i) = q*shgamma(j)+(0,1)*p*s2gammarec(j)
     end do
  end do

  ! Close the file
  close(fis)
  
  ! Summarise
  if (isroot) then
     write(fug,*)
     write(fug,*) "Initial States"
     write(fug,*) "--------------"
     do i = 1,ninitstates
        write(fug,*) "  State",i
        do j = 1,ndf
           write(fug,*) "         z",j,zinit(j,i)
        end do
        
        ! Calculate energy of this state
        ! overlap matrix = 1.
        write(fug,*) "    Energy", h_ord (dconjg(zinit(:,i)),zinit(:,i))*eu," ",trim(energyunits)
        write(fug,*)
     end do
  end if
end subroutine readinitstates
