! **************************************************
!                  Options Module
!
!                  Stewart K. Reed
!                   4th June 2007
!
!  Description:
!    This module contains runtime options such
!    as system size, time step etc.
! **************************************************

module options

  implicit none
  
  ! Time units hbar/Ehartree 
  real(kind=8),parameter :: tfsu =2.418884d-2 ! distance in Bohr radii
  
  real(kind=8),parameter :: tfskcal =48.87306 ! 0.020461166 ! 1 fs = tfskcal tu when [q= A*sqrt(amu) and energy is in kcal/mol  
   ! time unit for distance in Angstrom and mass in unified mass units
  real(kind=8),parameter :: tfsAu  = 1.574609764d2  ! 1fs = 6.350779875974d-3 tu

  ! Conversion factor for time units, to be one of the above
  real(kind=8) :: tu

  ! Energy is in Hartrees
  ! from energy units to eV multiply
  real(kind=8),parameter :: eueVu  =27.2113834736888  
   ! from energy units to Hartrees multiply by
  real(kind=8),parameter :: euHu  =1  
  
  ! for distance in Angstrom, mass in unified mass units and energy in eV or kJ/mol
  real(kind=8),parameter :: euAueV = 4.18015888d-3
  real(kind=8),parameter :: euAukJmol = 0.40332405d0
    

  ! Depending on which energy units are chosen eu will be one of the above.
  real(kind=8) :: eu 
  ! A string for outputing energy units
  character(len=50) :: energyunits
  character(len=5) :: massunits


  ! When doing cartesian simulations,
  integer :: ncartdim   ! system dimension
  integer :: nparticles ! Number of particles = ndf/ncartdim

  ! Run info
  integer :: nsteps               ! Number of steps to perform
  real(kind=8):: dt               ! time step
  character(len=100) :: sysname   ! name of system
  logical :: restartsim           ! Is this a restart simulation?
  logical :: randseed             ! Automatically seed the random number generator?   
  integer(kind=8) :: randgenseed          ! seed for random number generator

  !  should quantum corrections be ignored. For debugging really.
!  logical:: classical
  integer :: classical ! 0 - quantum; 1 - semiclassical; 2 - classical

  ! Basis initialisation     1: grid; 2: gaussian grid; 
  integer :: basismethod

  ! probability below which basis functions should be ignored.
  real(kind=8) :: minprob
  ! maximumum allowed z beyond which basis functions are removed
  real(kind=8) :: maxz

  ! Method with which to calculate D from C
  integer :: solvemethod

  ! For conjugate gradients method
  integer :: maxcgiterations
  real(kind=8) cgepsilon,cgepsilonsq
  
  ! Integration method
  logical :: newton=.false.

  ! output info
  integer :: printfreq ! How often to write step/energy to output file
  integer :: ccorlog ! Calculate the cross correlation function
  integer :: hcorfreq ! Calculate the correlation function with harmonic oscillator
  integer :: writeposfreq ! Frequency with which to write out positions.
  integer :: writeCDfreq ! Frequency with which to write out C and D
  integer :: writeampfreq ! Frequency with which to write out amp and pos.
  integer :: writeSfreq ! Frequency with which to write out action.
  integer :: writeEfreq ! Frequency with which to write out the energy.
  integer :: writeacffreq ! Frequency with which to write out the acf.
  integer :: writenormfreq ! Frequency with which to write out the norm
  integer :: writebasisnormfreq ! Frequency with which to write out the norm
  integer :: writezposfreq ! Frequency with which to write otu z pos 
  integer :: dumpfreq      ! Frequency with which to write restart file other tha n just at the end.
  integer :: writeexpect   !  ditto but for expectation value
  integer :: writeovlpfreq ! write out overlap matrix
  real(kind=8) :: omega0
  integer :: verbosity
  
  contains 
    
    subroutine readoptions
      use coherentstates, ONLY: ndf, nbasis,gamma,nehernfest
      use files, ONLY: fopt,fug
      use scalapackstuff, ONLY: isroot

#ifdef USE_SCALAPACK
      use arrays, ONLY:SP_NB,SP_MB
      use scalapackstuff, ONLY: NPROW,NPCOL
#endif

      ! subroutines
      use coherentstates, ONLY: allocatecsarrays,calcgammas
      use errors
      implicit none
      CHARACTER(len=50) :: line ,basistype,energytype,masstype
      integer :: ierr
      integer :: i,j,s ! for loops / dummy variable
      integer :: gammaset ! Have the gammas been read in?

      open(fopt,file="run.inpt",status='old',iostat=ierr)
      if(ierr .ne. 0 ) then
         call fileerror(ierr,"run.inpt")
      end if
         
      ! ---- set defaults -----
      dt = 1.0
      nsteps = 1
      ndf = 0
      nbasis = 0
      nehernfest = 1
      gammaset = 0
      basismethod = 0
      classical = 0 ! .false. 
      restartsim = .false.

      randseed = .true.
      randgenseed = 0 ! seed for random number generator.
                      ! 0 => gets it from the clock.

      omega0 = 1d-10 ! What counts as zero for overlap matrix

      minprob = -1.0 ! Min probability for including basis functions
                     ! < 0 implies always include all.
      maxz = -1.0

      eu = euHu
      energyunits ="H"

      massunits = "me"
      tu = tfsu
      
      nparticles = -1  ! Number of paricles for cartesian potentials
      ncartdim = 3    ! Number of cartesian dimensions. ndf = nparticles * ncartdim.
      
      ! Solving for D
      solvemethod = 1          ! Lapack = 1, CH = 2 SCALAPACK = 3, see #ifdef below.
      maxcgiterations = 10000  ! Parameters for CG
      cgepsilon = 1d-4
      cgepsilonsq = cgepsilon**2

      ! outputs
      printfreq = 100
      ccorlog = 0
      hcorfreq = 0
      writeCDfreq = 0
      writeampfreq = 0
      writeSfreq = 0
      writeEfreq = 100
      writeacffreq = 0
      writenormfreq = 100
      writebasisnormfreq = 0
      dumpfreq = 1000
      writezposfreq = 0
      writeexpect = 100
      writeovlpfreq = 0

      verbosity = 0

#ifdef USE_SCALAPACK
      SP_MB = 10
      SP_NB = 10
      NPROW = 0   !if unchanged from 0, the scalapack initialisation routine will determine a grid.
      NPCOL = 0

      ! Set default solve method to scalapack.
      solvemethod = 3 
#endif

      read(fopt,*,iostat=ierr) line
      do while (ierr == 0)   

         if (line == "nsteps" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,nsteps
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading number of time steps")

            ! ---- timestep -----
         else if  (line == "dt" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,dt   
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading time step")
            ! Convert timestep in fs to time units
            dt = dt / tu

            ! ---- number of degrees of freedom ----
         else if  (line == "ndf" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,ndf
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading number of degrees of freedom")
            
         else if (line == "nparticles" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,nparticles
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading number of particles for cartesian system")
            ndf = nparticles * ncartdim
         else if (line == "ncartdim" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,ncartdim
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading number of cartesian dimensions")
            ndf = nparticles * ncartdim
            ! ---- number of basis functions----
         else if  (line == "nbasis" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,nbasis
             if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading number of basis functions to use")
             
             ! ---- restart -----
         else if  (line == "restart" ) then
            restartsim = .true.
            
            ! ---- seed random number generator -----
         else if  (line == "timerandseed" ) then
            randseed = .true.
            randgenseed = 0
            
            ! ---- don't seed random number generator -----
         else if  (line == "norandseed" ) then
            randseed = .false.

         else if (line == "randseed" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,randgenseed
            if(isroot) write(fug,*) "Read seed",randgenseed
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error seed for random number generator")
            randseed = .true. 

             ! ---- basis initialisation ----
          else if  (line == "basis" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,basistype
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading basis type")
            if(ndf .eq. 0 ) then
               if(isroot) write(fug,*) &
                    " Need to specify ndf before specifying basis type"
               stop
            else
               call readbasistype(basistype)
            end if

            ! ----energy units ----
          else if  (line == "energy" ) then
             backspace(fopt) 
             read(fopt,*,iostat=ierr) line,energytype ! borrow basistype
             if (energytype .eq. "au" ) then
                eu = 1.0d0
                energyunits = "a.u."
             else if  (energytype .eq. "eV" .or. energytype .eq. "ev") then
                eu = eueVu
                energyunits = "eV"
             else if  (energytype .eq. "H" .or. energytype .eq. "Hartree") then
                eu = euHu
                energyunits = "H"
             else if  (energytype .eq. "kcalmol") then
                eu = 1
                energyunits = "kcal/mol"
                 tu = tfskcal
              else if  (energytype .eq. "akma") then
                 ! default units
                 eu = 1
                 energyunits = "sqrt(kcal/mol/amu) Angstrom"
                 tu = tfskcal

                 ! Should we convert to kcal/mol?
                 backspace(fopt)
                 read(fopt,*,iostat=ierr) line,line,energytype
                 if (ierr .eq. 0) then
                    if(energytype .eq. "kcalmol") then
                       eu = 0.3105
                       energyunits = "kcal/mol"
                    end if
                 end if
                 
              else if  (energytype .eq. "AueV") then
                 eu = euAueV
                energyunits = "eV"
                 tu = tfsAu
                 massunits = "u"

              else if  (energytype .eq. "AukJmol") then
                 eu = euAukJmol
                energyunits = "kJ/mol"
                 tu = tfsAu
                 massunits = "u"            
             else
                if(isroot) write(fug,*) " Unknow energy units in run.inpt",energytype
                stop
             end if

             ! ----mass units ----
          else if  (line == "mass" ) then
             backspace(fopt) 
             read(fopt,*,iostat=ierr) line,masstype ! borrow basistype
             if (masstype .eq. "me" ) then
                massunits = "me"
             else if (masstype .eq. "u") then
                massunits = "u"
             else
                if(isroot) write(fug,*) " Unknow mass units in run.inpt",masstype
                stop
             end if
                

            ! ---- System name----
         else if  (line == "name" ) then
!            backspace(fopt) 
            read(fopt,"(a100)",iostat=ierr) sysname
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading system name")

            ! ---- gammas ----
         else if (line == "gamma") then
!            if(isroot) write(fug,*) "Reading gamma"
            if(ndf .eq. 0 ) then
               if(isroot) write(fug,*) &
                    " Need to specify ndf before specifying gamma"
               stop
            else
               call allocatecsarrays
               do i = 1,ndf
                  read(fopt,*,iostat=ierr) s,gamma(i)
                  if(ierr.ne.0) then
                     if(isroot) write(fug,*) "state: ",i
                     call filereaderror(ierr,"run.inpt","Error reading gamma")
                  end if
               end do
               call calcgammas
               gammaset = 1
            end if
            
            ! ---- particle gammas ----
            ! same as gammas but for cartesian simulations.
            ! this bit assumes there are only nparticle gammas
            ! and sets gammas for each dimension to be the same
         else if (line == "pgamma") then
            if(isroot) write(fug,*) "Reading particle gammas"
            if(ndf .eq. 0) then
               if(isroot) write(fug,*) &
                    " Need to specify ndf before specifying gamma"
               stop
            else if(nparticles .le. 0 ) then
               if(isroot) write(fug,*) "You need to specify the number of particles before gammas"
               stop
            else
               call allocatecsarrays
               do i = 0,nparticles-1
                  read(fopt,*,iostat=ierr) s,gamma(i*ncartdim+1)
                  if(ierr.ne.0) then
                     if(isroot) write(fug,*) "particle: ",i+1
                     call filereaderror(ierr,"run.inpt","Error reading gamma")
                  end if
                  ! set other gammas for this particle
                  do j = 2,ncartdim
                     gamma(i*ncartdim+j) = gamma(i*ncartdim+1)
                  end do
               end do
               gamma(:) = gamma(:)! /tu/eu                                 
               call calcgammas
               gammaset = 1
            end if

            ! ---- Classical ----
         else if (line == "classical") then
            classical = 2 !.true.
         else if (line == "semiclassical") then
            classical = 1 !.true.
            
            ! ---- frequence with which to write to output file ----
         else if (line == "print") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,printfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing positions")

         else if (line == "writepos") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeposfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing positions")

           else if (line == "writezpos") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writezposfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing z-positions")

            ! ---- Cross correlation function ----
         else if (line == "crosscorr") then
            backspace(fopt)
            read(fopt,*,iostat=ierr) line,ccorlog
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error cross correlation frequency")
            
         else if (line == "harmcorr") then
            backspace(fopt)
            read(fopt,*,iostat=ierr) line,hcorfreq
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading harmonic correlation frequency")
      	 else if (line == "writeCD") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeCDfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing C and D")
        else if (line == "writeamp") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeampfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing amplitudes")

        else if (line == "writeaction") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeSfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing action")

        else if (line == "writeenergy") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeEfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing energy")

        else if (line == "writeacf") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeacffreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for acf energy")

        else if (line == "writenorm") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writenormfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for norm")
        else if (line == "writebasisnorm") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writebasisnormfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for norm for each basis function. ")
           
        else if (line == "writeovlp") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeovlpfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing overlap matrix.")

        else if (line == "dumpfreq") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,dumpfreq
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing restart file")
        else if (line == "writeexpect") then       
           backspace(fopt)
           read(fopt,*,iostat=ierr) line,writeexpect
           if(ierr.ne.0) call filereaderror(ierr,"run.inpt", &
                "Error reading frequency for writing expecation value file")
           
        else if (line == "verbosity") then
           backspace(fopt) 
           read(fopt,*,iostat=ierr) line,verbosity
           if(ierr .ne. 0) then
              write (fug,*) "Error setting verbosity level in run.inpt"
              stop
           end if

        else if (line == "omega0") then
           backspace(fopt) 
           read(fopt,*,iostat=ierr) line,omega0
           if(ierr .ne. 0) then
              write (fug,*) "Error setting omega0 level in run.inpt"
              stop
           end if

        else if  (line == "minprob" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,minprob   
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading minium prob.")
         else if  (line == "maxz" ) then
            backspace(fopt) 
            read(fopt,*,iostat=ierr) line,maxz   
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading maximum z.")


         else if (line == "solve") then
            backspace(fopt)
            read(fopt,*,iostat=ierr) line,line
            if ( line == "CG") then
               solvemethod = 2
            else  if ( line == "LAPACK") then
               solvemethod = 1
            else if (line == "SCALACPACK") then
               solvemethod = 3
            end if

         else if (line == "cgepsilon") then
            backspace(fopt)
            read(fopt,*,iostat=ierr) line,cgepsilon
             if(ierr .ne. 0) then
              write (fug,*) "Error reading cgepsilon from run.inpt"
              stop
           end if
           cgepsilonsq = cgepsilon**2
            
               
#ifdef USE_SCALAPACK
         else if (line == "spblock") then
            backspace(fopt)
            read(fopt,*,iostat=ierr) line,SP_MB,SP_NB
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading blocak size for scalapack")
         else if (line == "spgrid") then
            backspace(fopt)
            read(fopt,*,iostat=ierr) line,NPROW,NPCOL
            if(ierr.ne.0) call filereaderror(ierr,"run.inpt","Error reading processor grid for scalapack")
#endif
     
        end if



         ! Read the next line!
         read(fopt,*,iostat=ierr) line
      end do

      ! Close the file
      close(fopt)

      ! Check the options
      if (ndf .eq. 0 .or. nbasis.eq.0) then
         if(isroot) write(fug,*) "You must specify the number of degrees of freedom and the number of basis functions."
         stop
      end if

      if (basismethod .eq. 0) then
          if(isroot) write(fug,*) "You must specify the number method by which to choose the initial basis functions"
         stop
      end if

      if (energytype .eq. "kcalmol" .and. massunits.eq."me") then
         if(isroot) write(fug,*) "Mass needs to be in unified atomic mass units to have energy in kcal/mol"
         stop
      end if

      ! If gamma not set, use defaults of gamma = 1
      if (gammaset .eq. 0) then
         call allocatecsarrays
      end if

      
    end subroutine readoptions
  
    subroutine writeoptions
      use files,ONLY: fug
      use coherentstates, ONLY: ndf,nbasis,gamma
      use scalapackstuff, ONLY: isroot
#ifdef USE_SCALAPACK
      use arrays, ONLY:SP_NB,SP_MB
#endif
      implicit none
      integer :: i
      if(isroot) then
         write(fug,*)
         write(fug,*) "Simulation specifications"
         write(fug,*) "-------------------------"
         write(fug,*) "Run name ", sysname
         write(fug,*) 
         
         write(fug,*) ndf," degrees of freedom"
         write(fug,*) nbasis,"basis functions"
         if(nparticles > 0) then
            write(fug,*) nparticles," particles in cartesians"
            write(fug,*) ncartdim," cartesian dimensions"
         end if
         
         write(fug,*) 
         if (massunits .eq. "u") then
            write(fug,*) "Mass in unified atomic mass units"
         else
            write(fug,*) "Mass in units of electron mass"
         end if
         write(fug,*) "Time step",dt*tu," fs"
         write(fug,*) "Time step",dt, " simulation units"
         write(fug,*) "Perform ",nsteps," time steps"
         write(fug,*)
         write(fug,*) "  Verbosity level: ",verbosity
         write(fug,*) "  Zero for overlap matrix is ",omega0

         write(fug,*) 
         write(fug,*) "  Gammas                i     g_i  "
         do i =1,ndf
            write(fug,"(a21,i5,f10.3)") "                     ",i,gamma(i) 
         end do
         write(fug,*) 
         if (classical.eq.2) then
            write(fug,*) " Not applying quantum correlations"
         else if (classical .eq. 1) then
            write(fug,*) " Semiclassical run - calculate initial D and C but keep them constant"
         end if
         write(fug,*) 
         write(fug,*) "The unit of energy is ",energyunits

         ! write out conversion factors 
         if (verbosity .ge. 1) then
            write(fug,*) "energy conversion factor",eu
            write(fug,*) "time conversion factor",tu
         end if

         write(fug,*)
         if (.not. randseed) then
            write(fug,*) "Not seeding random number generator => reproducible results."
         else if (randgenseed .gt. 0) then
            write(fug,*) "Using seed from run.inpt file"
         else
            write(fug,*) "Using date/time to seed random number generator"         
         end if

         write(fug,*) 
         if (solvemethod == 1) then
            write(fug,*) " Using LAPACK to calculate D from C"
         else if (solvemethod == 2) then
            write(fug,*) " Using CG to calculate D from C"
         end if
            

#ifdef USE_SCALAPACK
         write(fug,*)
         write(fug,*) "Scalapack block sizes", SP_MB," x ",SP_NB
#endif
      end if !if isroot
    end subroutine writeoptions


    subroutine readbasistype(basistype)
      use coherentstates, ONLY: ndf,nbasis
      use files, ONLY:fug,fopt
      use arrays, ONLY:grid,ngrid,gsigma,enmax,vstate,probscale,zmax,maxtries, &
           harmmethod,harmComp,zoffset,nsupermodes,nmaxsuperbasis,nsuperbasis, & 
           superphi,superqmin,superqmax,superpmin,superpmax,supermaxprob
      use errors
      use scalapackstuff, ONLY: isroot

      implicit none
      character(*) :: basistype
      integer :: ierr
      integer :: nbf ! number of basis functions
      integer :: i,j
      integer :: superid,nhb
      CHARACTER(len=50) :: line

      real(kind=8) :: rzoffset,izoffset ! for reading in origin offset

      if (basistype == "grid") then

         basismethod = 1

         ! Allocate arrays
         allocate(grid(4,ndf),ngrid(2,1))
         ! Read in grid details
         ! Number of cells
         read(fopt,*,iostat=ierr) ngrid(1,1),ngrid(2,1)
         if(ierr.ne.0)  call filereaderror(ierr,"run.inpt",& 
              " Error reading in number of bins for grid. Should be nq np")
         ! Grid dimensions
         do i = 1,ndf
            read(fopt,*,iostat=ierr) grid(1,i),grid(2,i),grid(3,i),grid(4,i)
            if(ierr.ne.0) &
                 call filereaderror(ierr,"run.inpt",&
                 "Error reading basis absolute grid size. Should be qmin,qmax,pmin,pmax")
         end do
        
         ! Calculate number of basis functions
         nbf = ngrid(1,1)*ngrid(2,1)
          
         if (nbasis .ne. 0 .and. nbasis .ne. ndf) then
            write (fug,*) "Changing the number of basis functions from ",nbasis," per degree of freedom to", nbf
         end if
         nbasis = nbf

      else if (basistype == "rgauss") then
         basismethod = 2

         ! Allocate arrays
         allocate(gsigma(2,ndf))

         ! ---- default values -----
         enmax = 5

         read(fopt,*) line
         if (line .eq. "enmax") then
            backspace(fopt)         
            read(fopt,*) line,enmax
          endif

         !Read in sigma values for q and p for each degree of freedom
         do i = 1,ndf
            read(fopt,*,iostat=ierr) gsigma(1,i),gsigma(2,i)
            if(ierr.ne.0) &
                 call filereaderror(ierr,"run.inpt",&
                 "Error reading basis absolute grid size. Should be qmin,qmax,pmin,pmax")
            
         end do

         ! Use initial state(s)
      else if (basistype == "istates") then
         basismethod = 3


      else if (basistype == "harmosc") then
         ! Do some initialisation
         basismethod = 4
         
         allocate(harmmethod(ndf))
         allocate(vstate(ndf))
         allocate(harmComp(ndf))

         harmComp(:) = 1.0d0
         zmax = 3.5d0
         probscale = 1.0d0
         maxtries = 20
         nmaxsuperbasis = 0

         read(fopt,*,iostat=ierr) line
         do while (line .ne. "vibstates" .and. ierr.eq.0)
            backspace(fopt)
            if( line.eq."zmax") then
               read(fopt,*,iostat=ierr) line,zmax
               ! Check
               if (zmax .le. 0) then
                  if(isroot) write(fug,*) " Error: zmax < 0 in run.inpt"
                  stop
               else if (ierr .ne. 0) then
                  if(isroot) write(fug,*) " Error ",ierr," while reading zmax in run.inpt"
                  stop
               end if
            else if ( line .eq. "probscale") then
                read(fopt,*,iostat=ierr) line,probscale
                if (probscale .le. 0) then
                   if(isroot) write(fug,*) " Error: probscale factor < 0 in run.inpt"
                   stop
                else if (ierr .ne. 0) then
                  if(isroot) write(fug,*) " Error ",ierr," while reading probscale in run.inpt"
                  stop
                end if
             else if  ( line .eq. "maxtries") then
                read(fopt,*,iostat=ierr) line,maxtries
                if (maxtries .lt. 1) then
                   if(isroot) write(fug,*) " Error: maximum number of tries per basis function is only ",maxtries," in run.inpt"
                   stop
                else if (ierr .ne. 0) then
                   if(isroot) write(fug,*) " Error ",ierr," while reading maxtries in run.inpt"
                   stop
                end if
                
                ! allow modes to be in a super position of harmonic oscillator states
             else if (line .eq. "super") then
                read(fopt,*) line,nsupermodes,nmaxsuperbasis
                allocate(superphi(nmaxsuperbasis,nsupermodes))
                allocate(nsuperbasis(nsupermodes))
                allocate(superqmin(nsupermodes),superqmax(nsupermodes))
                allocate(superpmin(nsupermodes),superpmax(nsupermodes))
                allocate(supermaxprob(nsupermodes))
                
                superphi(:,:) = 0.0d0
                nsuperbasis(:) = 0
                superqmin(:) = 0.0d0
                superqmax(:) = 0.0d0
                superpmin(:) = 0.0d0
                superpmax(:) = 0.0d0
                supermaxprob(:) = 0.0d0

             else
                if(isroot) write(fug,*) "Error: Unrecognised option in harmonic basis:",line
                stop
             end if
             read(fopt,*,iostat=ierr) line
          end do
            
         ! Now read in the vibrational states of each mode
         ! 0 - put at origin
         ! 1,2 harmonic oscillator at origin and offset
         ! 3,4 as 1,2 but using basis functions from initial states.
         ! 5,6 as 1,2 but where the state is a superposition of harmonic states
         ! 7,8 as 3,4 but where the state is a superposition of harmonic states 

         superid = 0 ! For superpositions.

         do i = 1,ndf

            read(fopt,*,iostat=ierr) harmmethod(i) 

            if (harmmethod(i) .eq. 1.or. harmmethod(i) .eq. 3) then
               backspace(fopt) 
               read(fopt,*,iostat=ierr) harmmethod(i),vstate(i),harmComp(i)
               
               if(ierr .ne. 0) then
                  if(isroot) write(fug,*) "Error reading in state and compression for ", &
                       i," degree of freedom."
                  stop
               end if

               ! The compression works the wrong way around for vstate > 0
               if(vstate(i) > 0 ) then
                  harmComp(i) = 1.0d0/harmComp(i)
               end if

               ! Check to see if vstate is < 0
               !if (vstate(i) < 0  .and. vstate(i) > -100) then
            else if (harmmethod(i) .eq. 2.or. harmmethod(i) .eq. 4 ) then
               ! Only allocate the array if it is needed.
               if(.not.allocated(zoffset)) then
                  allocate(zoffset(ndf))
                  zoffset(:) = (0.0,0.0)
               end if
               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i),vstate(i),harmComp(i),rzoffset,izoffset
               if(ierr .ne. 0) then
                  if(isroot) write(fug,*) "Error reading in state, compression and offset  ", &
                       i," degree of freedom."
                  stop
               end if
               zoffset(i) =  rzoffset+(0,1)*izoffset

               ! set vibrational state to be a superposition of harmonic oscillator modes
               ! The compression works the wrong way around for vstate > 0
               if(vstate(i) > 0 ) then
                  harmComp(i) = 1.0d0/harmComp(i)
               end if

               
            else if (harmmethod(i) .eq. 5) then
               if(isroot) write(fug,*) "Functions chosen from a superposition wave function"
               if (nsupermodes .eq. 0 ) then
                  if(isroot) write(fug,*) "Can't readin a wave function superposition as the option super has not been set."
                  stop 
               end if 

               superid = superid + 1
               if (superid .gt. nsupermodes ) then
                  if(isroot) write(fug,*) "Too many wavefunctions superpositions.",&
                       "Change the super statement at top of basis specification."
                  stop
               end if
               
               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i), &
                    superqmin(superid),superqmax(superid),&
                    superpmin(superid),superpmax(superid),&
                    supermaxprob(superid),&
                    nsuperbasis(superid)
               if(isroot) write(fug,*) "read" ,superqmin(superid),superqmax(superid)
               if(ierr.ne.0) then 
                  if(isroot) write(fug,*) "Error reading in harm basis type 5 for mode",i
                  stop
               end if
               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i), &
                    superqmin(superid),superqmax(superid),&
                    superpmin(superid),superpmax(superid),&
                    supermaxprob(superid),&
                    nsuperbasis(superid),&
                    (superphi(j,superid),j=1, nsuperbasis(superid))
               if(ierr.ne.0) then 
                  if(isroot) write(fug,*) "Error reading in harm basis type 5 for mode",i
                  stop
               end if

            else if (harmmethod(i) .eq. 7 ) then
               if (nsupermodes .eq. 0 ) then
                  if(isroot) write(fug,*) "Can't readin a wave function superposition as the option super has not been set."
                  stop
               end if

               superid = superid + 1
               if (superid .gt. nsupermodes ) then
                  if(isroot) write(fug,*) "Too many wavefunctions superpositions. ",&
                       " Change the super statement at top of basis specification."
                  stop
               end if
               
               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i), nsuperbasis(superid)
               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i),  nsuperbasis(superid),&
                    (superphi(j,superid),j=1, nsuperbasis(superid))

           else if (harmmethod(i) .eq. 8 ) then
               ! Only allocate the array if it is needed.
               if(.not.allocated(zoffset)) then
                  allocate(zoffset(ndf))
                  zoffset(:) = (0.0,0.0)
               end if
               
               if (nsupermodes .eq. 0 ) then
                  if(isroot) write(fug,*) &
                       "Can't readin a wave function superposition as the option super has not been set."
                  stop
               end if

               superid = superid + 1
               if (superid .gt. nsupermodes ) then
                  if(isroot) write(fug,*) "Too many wavefunctions superpositions. ",&
                       "Change the super statement at top of basis specification."
                  stop
               end if

               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i),rzoffset,izoffset, nhb
               backspace(fopt)
               read(fopt,*,iostat=ierr) harmmethod(i),rzoffset,izoffset,&
                    nsuperbasis(superid),(superphi(j,superid),j=1,nhb)

 
            else if (harmmethod(i) .ne. 0) then
               if(isroot) write(fug,*) "Unknown harmonic basis set method ",harmmethod(i)
               if(isroot) write(fug,*) "degree of freedom number ",i
               stop
            end if
            
         end do

         !Read basis functions from file
      else if (basistype == "file") then
         basismethod = 5
      else
         if(isroot) write(fug,*) "Unknonw basis type ", trim(basistype)," in run.inpt"
      end if
    end subroutine readbasistype

end module options
