
module polymorsehammod
  use hamiltoniandata

  private ! Keep most functions private

  ! public interface 
  public :: hampolymorse,hampolymorse1mode,dhampolymorsedz
  public :: readhampm
contains
  ! ==================================================
  !          Polynomial + Morse Hamiltomian 
  ! ==================================================
  
function hampolymorse(zic,zj) 
  use coherentstates, ONLY: ndf,gamma,sgamma
  ! include subroutines too.
  use coherentstates, ONLY: pos,mom
  use arrays, ONLY: sqrtpi
  implicit none

  ! two states <z_i| and |z_j>
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  ! hamiltonian
  complex(kind=8)::hampolymorse
  complex(kind=8)::Ucoup ! coupling energy
  complex(kind=8)   ::  coups! ! switch coupling with another mode 
  ! position and momentum vectors
  complex(kind=8),dimension(ndf) :: q,p
  integer :: i,j,index

  !    real(kind=8),external :: derf ! derf isn't actually an intrinsic function 
  !                                  ! but this is needed for the PG compiler.

  interface 
     include "MCERROR.interface"
  end interface

  ! Calculate position and momentum vectors
  call pos(zj,zic,q)
  call mom(zj,zic,p)

  ! Now calculate the Hamiltonian
  ! p_i^2/2m_i+gamma_i/4m_i
  hampolymorse = 0.5*sum( (p(:)**2 + 0.5*gamma(:))/m(:))

  do i = 1,ndf
     ! Polynomial for ith degree of freedom
     if (polydegree(i) > 0) then
        hampolymorse = hampolymorse + polyeng(i,q(i),polydegree(i))
     end if
     ! Morse potential for the ith degree of freedom
     if (morse(i)) then

        hampolymorse = hampolymorse &
             + morseeng(i,morsealpha(i),q(i),morseq0(i),morseA(i),morseC(i))
     end if

     if(gaussian(i)) then
        hampolymorse = hampolymorse &
             + gaussA(i)*gausseng(i,zic(i),zj(i),gaussalpha(i),gaussq0(i))
     end if
  end do

  ! ---- Coupling ----
  index = 1

  do i = 1,ndf-1
     do j = i+1,ndf
        ! ---- Calculate switch mode if applicable ----
        if (coupswitch(index) .eq. 1) then
           coups = 0.5d0*(1  &
           +coupswitchdir(index)*cderror(sgamma(coupswitchmode(index))*(coupswitchpos(index)-q(coupswitchmode(index)))))

           if (abs(coups) .gt. 1d-9) then
              ! If 'on' calculate optionally truncated coupling
              hampolymorse = hampolymorse + coups*hampolymorsecoup(i,j,index,q,p)
           end if
        else if (coupswitch(index) .eq. 2) then
           coups = dsqrt(gamma(coupswitchmode(index))/(gamma(coupswitchmode(index))+coupswitchgamma(index)))&
                *cdexp(-gamma(coupswitchmode(index))*coupswitchgamma(index)*q(coupswitchmode(index))**2&
                       /(gamma(coupswitchmode(index))+coupswitchgamma(index)))
           if (abs(coups) .gt. 1d-9) then
              ! If 'on' calculate optionally truncated coupling
              hampolymorse = hampolymorse + coups*hampolymorsecoup(i,j,index,q,p)
           end if

        else
           ! No switch, calculate  optionally truncated coupling
           hampolymorse = hampolymorse + hampolymorsecoup(i,j,index,q,p)
        end if

        index = index + 1
     end do
  end do
end function hampolymorse

! ==================================================
!          Coupling Hamiltomian 
! 
!  A subroutine to calculate the coupling between
!  two modes. The modes are given by i,j.
! ==================================================
function hampolymorsecoup(i,j,index,q,p)
    use coherentstates, ONLY: gamma,sgamma
  use arrays, ONLY: sqrtpi
  implicit none
  integer,intent(in) :: i,j,index
  complex(kind=8),dimension(*) :: q,p
  
  complex(kind=8)::hampolymorsecoup ! output coupling energy

  complex(kind=8) :: errfn,errfn2,expfn2,expfn ! errorfunction
  complex(kind=8) :: C1i,C1j,C2i,C2j

  ! For maple code
  real(kind=8) :: t6,t9,t52,t71
  complex(kind=8) :: t1,t2,t4,t11,t12,t13,t14,t16,t17,t18,t19,t22,t23,t33,t34,t43,t44,t54,t58,t67,t69

  interface 
     include "MCERROR.interface"
     include "rerfunc.interface"
  end interface

  if (trunc(index) .eq. 0) then
     ! ---- untruncated function ----
     !              hampolymorse = hampolymorse & 
     hampolymorsecoup =  harmcouplea12(index) * (q(i)*q(j)**2+q(i)/(2.0*gamma(j))) &
          + harmcouplea21(index) * (q(i)**2*q(j) + q(j)/(2.0*gamma(i))) & 
          + harmcouplea22(index) * (q(i)**2*q(j)**2 + q(i)**2/(2*gamma(j)) &
          + q(j)**2/(2*gamma(i)) + 0.25/(gamma(i)*gamma(j)))

  else if (trunc(index) .eq. 1 .or. trunc(index) .eq. -1 ) then
     ! truncate the averaged coupling potential smoothly.
     ! trunc(index) appears in the error function to control the direction
     hampolymorsecoup = ( &
          harmcouplea12(index) * (q(i)*q(j)**2+q(i)/(2.0*gamma(j))) &
          + harmcouplea21(index) * (q(i)**2*q(j) + q(j)/(2.0*gamma(i))) & 
          + harmcouplea22(index) * (q(i)**2*q(j)**2 + q(i)**2/(2*gamma(j)) &
          + q(j)**2/(2*gamma(i)) + 0.25/(gamma(i)*gamma(j)))&
          ) &
          *0.5*( rerfunc((-1.0*trunc(index)*&
          real(q(truncm(index)))-truncpos(index))/truncwidth(index)) +1) ! switching fct

  else if (trunc(index) .eq. 2 .or. trunc(index) .eq. -2 ) then
     if (truncm(index) .eq. i) then
        ! trunc(index) appears in the error function to control the direction
        ! The extrafactors of 0.5 occur in front of trunc(index) as trunc(index) = +/- 2 rather than
        ! +/- 1.
        errfn = 1+0.5*trunc(index)*cderror(sgamma(i)*(truncpos(index)-q(i)))
        expfn = 0.5*trunc(index)*cdexp(-gamma(i)*(truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))

        !                 hampolymorse = hampolymorse & 
        hampolymorsecoup = harmcouplea12(index) * (q(j)**2+0.5/gamma(j))*0.5*(q(i)*errfn - expfn ) &
             + harmcouplea21(index) * q(j) &
             *0.5*((q(i)**2+0.5/gamma(i))*errfn-(truncpos(index)+q(i))*expfn) &
             + harmcouplea22(index) * (q(j)**2+0.5/gamma(j)) &
             *0.5*((q(i)**2+0.5/gamma(i))*errfn-(truncpos(index)+q(i))*expfn) 

     else ! truncating according to q(j)
        errfn = 1+0.5*trunc(index)*cderror(sgamma(j)*(truncpos(index)-q(j)))
        expfn = 0.5*trunc(index)*cdexp(-gamma(j)*(truncpos(index)-q(j))**2)/(sqrtpi*sgamma(j))

        !                 hampolymorse = hampolymorse & 
        hampolymorsecoup = harmcouplea21(index) * (q(i)**2+0.5/gamma(i))*0.5*(q(j)*errfn - expfn ) &
             + harmcouplea12(index) * q(i)&
             *0.5*((q(j)**2+0.5/gamma(j))*errfn-(truncpos(index)+q(j))*expfn) &
             + harmcouplea22(index) * (q(i)**2+0.5/gamma(i))&
             *0.5*((q(j)**2+0.5/gamma(j))*errfn-(truncpos(index)+q(j))*expfn) 
     end if

  else if (trunc(index) .eq. 3) then ! truncate in both directions for one mode
     if (truncm(index) .eq. i) then
        !!                errfn  = cderror(sgamma(i)*(truncpos(index)-q(i)))
        !!                errfn2 = cderror(sgamma(i)*(truncpos(index)+q(i)))
        !!                expfn  = cdexp(-gamma(i)*( truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))
        !!                expfn2 = cdexp(-gamma(i)*(-truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))

        !!                C1i = 0.5*(expfn2-expfn)+0.5*q(i)*(errfn+errfn2)
        !!                C2i = 0.5*(q(i)-truncpos(index))*expfn2-0.5*(q(i)+truncpos(index))*expfn &
        !!                     + (0.5*q(i)**2+0.25/gamma(i))*(errfn+errfn2)

        !                C1i = 0.5/(sqrtpi*sgamma(i))*(expfn2-expfn)+0.5*q(i)*(errfn+errfn2)
        !                C2i = 0.5*(q(i)-truncpos(index))*expfn2-0.5*(q(i)-truncpos(index))*expfn &
        !                     + (0.5*q(i)**2+0.25/gamma(i))*(errfn+errfn2)

        t1 = truncpos(index) - q(i)
        t2 = t1 ** 2
        t4 = cdexp(-gamma(i) * t2)
        t6 = 0.1D1 / sgamma(i)

        t9 = 0.1D1 / sqrtpi
        t11 = sgamma(i)* truncpos(index)
        t12 = sgamma(i)* q(i)
        t13 = t11 - t12
        t14 = cderror(t13)
        t16 = -truncpos(index) - q(i)
        t17 = t16 ** 2
        t19 = cdexp(-gamma(i) * t17)
        t22 = -t11 - t12
        t23 = cderror(t22)
        c1i = -t4 * t6 * t9 / 0.2D1 + q(i) * t14 / 0.2D1 + t19 * t6 * t9 / 0.2D1 - q(i) * t23 / 0.2D1


        t33 = -t13
        t34 = -t14 ! erf(t33)

        t43 = -t22
        t44 = -t23 ! erf(t43)

        t52 = 0.1D1 / gamma(i) * t9
        t54 = q(i) ** 2
        t58 = -0.2D1 * gamma(i) * t54 * sqrtpi - sqrtpi
        c2i = -t43 * t4 * t52 / 0.2D1 - t58 * t14 * t52 / 0.4D1 &
             + t33 * t19 * t52 / 0.2D1 + t58 * t23 * t52 / 0.4D1

        !                 hampolymorse = hampolymorse & 
        hampolymorsecoup = harmcouplea12(index) * (q(j)**2+0.5/gamma(j))*C1i & 
             + harmcouplea21(index) * q(j) * C2i &
             + harmcouplea22(index) * (q(j)**2+0.5/gamma(j))*C2i

     else

        !                errfn = cderror(sgamma(j)*(truncpos(index)-q(j)))
        !                errfn2 = cderror(sgamma(j)*(truncpos(index)+q(j)))
        !                expfn = cdexp(-gamma(j)*(truncpos(index)-q(j))**2)/(sqrtpi*sgamma(j))
        !                expfn2 = cdexp(-gamma(j)*(-truncpos(index)-q(j))**2)/(sqrtpi*sgamma(j))

        !                C1j = 0.5*(expfn2-expfn)+0.5*q(j)*(errfn+errfn2)
        !                C2j = 0.5*(q(j)-truncpos(index))*expfn2-0.5*(q(j)+truncpos(index))*expfn &
        !                     + (0.5*q(j)**2+0.25/gamma(j))*(errfn+errfn2)

        t1 = truncpos(index) - q(j)
        t2 = t1 ** 2
        t4 = cdexp(-gamma(j) * t2)
        t6 = 0.1D1 / sgamma(j)

        t9 = 0.1D1 / sqrtpi
        t11 = sgamma(j)* truncpos(index)
        t12 = sgamma(j)* q(j)
        t13 = t11 - t12
        t14 = cderror(t13)
        t16 = -truncpos(index) - q(j)
        t17 = t16 ** 2
        t19 = cdexp(-gamma(j) * t17)
        t22 = -t11 - t12
        t23 = cderror(t22)
        c1j = -t4 * t6 * t9 / 0.2D1 + q(j) * t14 / 0.2D1 + t19 * t6 * t9 / 0.2D1 - q(j) * t23 / 0.2D1


        t33 = -t13
        t34 = -t14 ! erf(t33)

        t43 = -t22
        t44 = -t23 ! erf(t43)

        t52 = 0.1D1 / gamma(j) * t9
        t54 = q(j) ** 2
        t58 = -0.2D1 * gamma(j) * t54 * sqrtpi - sqrtpi
        c2j = -t43 * t4 * t52 / 0.2D1 - t58 * t14 * t52 / 0.4D1 &
             + t33 * t19 * t52 / 0.2D1 + t58 * t23 * t52 / 0.4D1

        !                 hampolymorse = hampolymorse & 
        hampolymorsecoup = harmcouplea12(index) * q(i) * C2j & 
             + harmcouplea21(index) * (q(i)**2+0.5/gamma(i)) * C1j &
             + harmcouplea22(index) * (q(i)**2+0.5/gamma(i)) * C2j
     end if

  else if (trunc(index) .eq. 4) then 
     ! ---- truncate in all four directions for one mode ----
     errfn = cderror(sgamma(i)*(truncpos(index)-q(i)))
     errfn2 = cderror(sgamma(i)*(truncpos(index)+q(i)))
     expfn = cdexp(-gamma(i)*(truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))
     expfn2 = cdexp(-gamma(i)*(-truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))

     C1i = 0.5*(expfn2-expfn)+0.5*q(i)*(errfn+errfn2)
     C2i = 0.5*(q(i)-truncpos(index))*expfn2-0.5*(q(i)-truncpos(index))*expfn &
          + (0.5*q(i)**2+0.25/gamma(i))*(errfn+errfn2)

     errfn = cderror(sgamma(j)*(truncpos2(index)-q(j)))
     errfn2 = cderror(sgamma(j)*(truncpos2(index)+q(j)))
     expfn = cdexp(-gamma(j)*(truncpos2(index)-q(j))**2)/(sqrtpi*sgamma(j))
     expfn2 = cdexp(-gamma(j)*(-truncpos2(index)-q(j))**2)/(sqrtpi*sgamma(j))

     C1j = 0.5*(expfn2-expfn)+0.5*q(j)*(errfn+errfn2)
     C2j = 0.5*(q(j)-truncpos2(index))*expfn2-0.5*(q(j)-truncpos2(index))*expfn &
          + (0.5*q(j)**2+0.25/gamma(j))*(errfn+errfn2)

     !              hampolymorse = hampolymorse & 
     hampolymorsecoup = harmcouplea12(index) * C2j * C1i & 
          + harmcouplea21(index) * C1j * C2i &
          + harmcouplea22(index) * C2j * C2i    

  else
     write(*,*) "I don't know how to calculate the energy with coupling truncation method",trunc(index)
     stop
  end if

end function hampolymorsecoup
   ! ==================================================
   !          Polynomial + Morse Hamiltomian
   !  Calculate the energy of only one mode.
   ! ==================================================
  
subroutine hampolymorse1mode(zic,zj,mode,Umode,Ucoup) 
  use coherentstates, ONLY: ndf,gamma,sgamma
  ! include subroutines too.
  use coherentstates, ONLY: pos,mom
  use arrays, ONLY: pi,sqrtpi,sqrt2

  implicit none

  ! two states <z_i| and |z_j>
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  integer,intent(in) :: mode
  complex(kind=8),intent(out)::Umode,Ucoup

  ! hamiltonian
  !    complex(kind=8)::hampolymorse1mode

  ! position and momentum vectors
  complex(kind=8),dimension(ndf) :: q,p
  integer :: i,j,index,index2

  complex(kind=8) :: errfn,errfn2,expfn2,expfn ! errorfunction
  complex(kind=8) :: C1i,C1j,C2i,C2j

  complex(kind=8) :: Ucouppart  ! a contribution to the coupling energy
  complex(kind=8) :: coups ! coupling switching factor

  !    real(kind=8),external :: derf ! derf isn't actually an intrinsic function
  !                                  ! but this is needed for the PG compiler.

  interface 
     include "MCERROR.interface"
     include "rerfunc.interface"
  end interface

  ! Calculate position and momentum vectors
  call pos(zj,zic,q)
  call mom(zj,zic,p)

  ! Now calculate the Hamiltonian
  ! p_i^2/2m_i+gamma_i/4m_i
  Umode = 0.5* p(mode)**2 + 0.25*gamma(mode)


  ! Polynomial for ith degree of freedom
  if (polydegree(mode) > 0) then
     Umode = Umode &
          + polyeng(mode,q(mode),polydegree(mode))
  end if
  ! Morse potential for the ith degree of freedom
  if (morse(mode)) then

     Umode = Umode &
          + morseeng(mode,morsealpha(mode),q(mode),morseq0(mode),morseA(mode),morseC(mode))
  end if

  if(gaussian(mode)) then
     Umode = Umode &
          + gaussA(mode)*gausseng(mode,zic(mode),zj(mode),gaussalpha(mode),gaussq0(mode))
  end if

  ! ---- Coupling ----
  Ucoup = (0.0,0.0)
 ! write(113,*) "# Next step"
  index = 1
  index2 = 0                  
  do i = 1,ndf-1
     do j = i+1,ndf
        if (i .eq. mode .or. j .eq. mode) then
           
           ! ---- Calculate switch mode if applicable ----
           if (coupswitch(index) .eq. 1) then
              coups = 0.5d0*(1  &
                   +coupswitchdir(index)*cderror(sgamma(coupswitchmode(index))*(coupswitchpos(index)-q(coupswitchmode(index)))))

!              write(113,*)index,real(coups),aimag(coups)
              if (abs(coups) .gt. 1d-9) then
                 ! If 'on' calculate optionally truncated coupling
                            ! Add the (switched) coupling energy
              Ucoup = Ucoup + coups*hampolymorsecoup(i,j,index,q,p)
              end if

           else if (coupswitch(index) .eq. 2) then
              coups = dsqrt(gamma(coupswitchmode(index))/(gamma(coupswitchmode(index))+coupswitchgamma(index)))&
                   *cdexp(-gamma(coupswitchmode(index))*coupswitchgamma(index)*q(coupswitchmode(index))**2&
                   /(gamma(coupswitchmode(index))+coupswitchgamma(index)))
              if (abs(coups) .gt. 1d-9) then
                 ! If 'on' calculate optionally truncated coupling
                 Ucoup = Ucoup + coups*hampolymorsecoup(i,j,index,q,p)
              end if
              
           else
              ! No switch, calculate  optionally truncated coupling
               Ucoup = Ucoup + hampolymorsecoup(i,j,index,q,p)
            end if

         end if
        index = index + 1
     end do
  end do

end subroutine hampolymorse1mode

  ! ==================================================
  !                 Derivatives of 
  !        the Polynomial + Morse Hamiltomian 
  ! ==================================================
function dhampolymorsedz (zic,zj)
  use coherentstates, ONLY: ndf,gamma,s2gammarec,shgamma,sgamma
  ! include subroutines too.
  use coherentstates, ONLY: pos,mom
  use arrays, ONLY: pi,sqrtpi,sqrt2

  implicit none

  ! two states <z_i| and |z_j>
  complex(kind=8),dimension(ndf),intent(in) :: zic,zj
  ! hamiltonian
  complex(kind=8)::dhampolymorsedz(ndf)

  integer :: i,j,index
  ! position and momentum vectors
  complex(kind=8),dimension(ndf) :: q,p

  complex(kind=8) :: errfn,errfn2,expfn2,expfn ! errorfunction
  complex(kind=8) :: C1i,C1j,C2i,C2j,dC1di,dC2di,dC1dj,dC2dj

  ! For maple code
  real(kind=8) :: t6,t9,t52,t71
  complex(kind=8) :: t1,t2,t4,t11,t12,t13,t14,t16,t17,t18,t19,t22,t23,t33,t34,t43,t44,t54,t58,t67,t69


  complex(kind=8) :: coups  ! Coupling switching function.
  complex(kind=8):: dhampolymorsedzi,dhampolymorsedzj,dhampolymorsedzt ! contributions to the derivative

  !    real(kind=8),external :: derf ! derf isn't actually an intrinsic function
  !                                  ! but this is needed for the PG compiler.

  ! include interface for the complex error function
  interface 
     include "MCERROR.interface"
     include "rerfunc.interface"
  end interface

  ! Calculate position and momentum vectors
  call pos(zj,zic,q)
  call mom(zj,zic,p)

  ! Derivatives w.r.t. z_i
  ! dq_i/dz_i = 1/(sqrt( 2gamma(i))) = s2gammarec(i)
  ! dp_i/dz_i = -ihbar*sqrt(gamma(i)/2) = (0,-1)*shgamma(i) as hbar = 1

  ! Kinetic part
  dhampolymorsedz(:) = p(:)*(0,-1)*shgamma(:)/m(:)

  do i = 1,ndf
     ! Polynomial for ith degree of freedom
     if (polydegree(i) > 0) then
        dhampolymorsedz(i) = dhampolymorsedz(i) &
             + dpolyengdq(i,q(i),polydegree(i))*s2gammarec(i)
     end if
     ! Morse potential for the ith degree of freedom
     if (morse(i)) then
        dhampolymorsedz(i) = dhampolymorsedz(i) &
             + dmorseengdq(i,morsealpha(i),q(i),morseq0(i),morseA(i))*s2gammarec(i)
     end if
     if (gaussian(i)) then
        ! don't need the s2gammarec as dgaussengdz does derivative wrt z rather than q.
        dhampolymorsedz(i) = dhampolymorsedz(i) &
             + gaussA(i)*dgaussengdz(i,zic(i),zj(i),gaussalpha(i),gaussq0(i))
     end if
  end do

  ! ---- Coupling ----
  index = 1
  do i = 1,ndf-1
     do j = i+1,ndf

        ! ---- optionally apply a truncation function ----
        ! Apply a switching function too.
        if (coupswitch(index) .eq. 1) then
           coups = 0.5d0*(1  &
                +coupswitchdir(index)*cderror(sgamma(coupswitchmode(index))*(coupswitchpos(index)-q(coupswitchmode(index)))))
        else if (coupswitch(index) .eq. 2) then
           coups = dsqrt(gamma(coupswitchmode(index))/(gamma(coupswitchmode(index))+coupswitchgamma(index)))&
                *cdexp(-gamma(coupswitchmode(index))*coupswitchgamma(index)*q(coupswitchmode(index))**2&
                /(gamma(coupswitchmode(index))+coupswitchgamma(index)))
        else
           coups = 1.0d0
        end if

        if (abs(coups) .gt. 1d-9) then
           dhampolymorsedzi = (0.0d0,0.0d0)
           dhampolymorsedzj = (0.0d0,0.0d0)
           if (trunc(index) .eq.0) then
              ! ---- untruncated function - default ----
              !             dhampolymorsedz(i) = dhampolymorsedz(i) &
              dhampolymorsedzi = (harmcouplea12(index) * (q(j)**2+0.5/gamma(j)) &
                   + 2*harmcouplea21(index) * q(i)*q(j) &
                   + harmcouplea22(index) *(2*q(i)*q(j)**2 + q(i)/gamma(j))) & 
                   * s2gammarec(i)

              !             dhampolymorsedz(j) = dhampolymorsedz(j) &
              dhampolymorsedzj = (harmcouplea21(index) * (q(i)**2+0.5/gamma(i)) &
                   + 2*harmcouplea12(index) * q(i)*q(j) &
                   + harmcouplea22(index) *(2*q(j)*q(i)**2 + q(j)/gamma(i))) &
                   * s2gammarec(j)

           else if (trunc(index) .eq. -1 .or. trunc(index) .eq. 1 ) then
              !dhampolymorsedz(i) = dhampolymorsedz(i) &
              dhampolymorsedzi = (harmcouplea12(index) * (q(j)**2+0.5/gamma(j)) &
                   + 2*harmcouplea21(index) * q(i)*q(j) &
                   + harmcouplea22(index) *(2*q(i)*q(j)**2 + q(i)/gamma(j))) & 
                   * s2gammarec(i) & 
                   *0.5*( rerfunc((-1.0*trunc(index)*real(q(truncm(index)))-truncpos(index))/truncwidth(index)) +1) ! switching fct

              ! dhampolymorsedz(j) = dhampolymorsedz(j) &
              dhampolymorsedzj =  (harmcouplea21(index) * (q(i)**2+0.5/gamma(i)) &
                   + 2*harmcouplea12(index) * q(i)*q(j) &
                   + harmcouplea22(index) *(2*q(j)*q(i)**2 + q(j)/gamma(i))) &
                   * s2gammarec(j)&
                   *0.5*( rerfunc((-1.0*trunc(index)*real(q(truncm(index)))-truncpos(index))/truncwidth(index)) +1) ! switching fct

              ! extra term for the truncated mode
              !dhampolymorsedz(truncm(index)) = dhampolymorsedz(truncm(index)) &
              dhampolymorsedzt =  exp(-((real(q(truncm(index)))-truncpos(index))/truncwidth(index))**2) &
                   * ( harmcouplea12(index) * (q(i)*q(j)**2+q(i)/(2.0*gamma(j))) &
                   + harmcouplea21(index) * (q(i)**2*q(j) + q(j)/(2.0*gamma(i))) & 
                   + harmcouplea22(index) * (q(i)**2*q(j)**2 + q(i)**2/(2*gamma(j)) &
                   + q(j)**2/(2*gamma(i)) + 0.25/(gamma(i)*gamma(j))))

              if (truncm(index) .eq. i) then
                 dhampolymorsedzi = dhampolymorsedzi + dhampolymorsedzt
              else
                 dhampolymorsedzj = dhampolymorsedzj + dhampolymorsedzt
              end if

           else if (trunc(index) .eq. 2 .or. trunc(index) .eq. -2 ) then
              if (truncm(index) .eq. i) then
                 ! The extrafactors of 0.5 occur in front of trunc(index) as trunc(index) = +/- 2 rather than
                 ! +/- 1.
                 errfn = 1+0.5*trunc(index)*cderror(sgamma(i)*(truncpos(index)-q(i)))
                 expfn = 0.5*trunc(index)*cdexp(-gamma(i)&
                      *(truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))

                 !dhampolymorsedz(i) = dhampolymorsedz(i) & 
                 dhampolymorsedzi = s2gammarec(i) &
                      *(harmcouplea12(index) * (q(j)**2+0.5/gamma(j)) &
                      *(0.5*errfn - truncpos(index)*gamma(i)*expfn ) &
                      + harmcouplea21(index) * q(j)&
                      *(q(i)*errfn-(1+gamma(i)*truncpos(index)**2)*expfn)  &
                      + harmcouplea22(index) * (q(j)**2+0.5/gamma(j)) &
                      *(q(i)*errfn-(1+gamma(i)*truncpos(index)**2)*expfn))


                 !dhampolymorsedz(j) = dhampolymorsedz(j) &
                 dhampolymorsedzj = s2gammarec(j) & 
                      *(harmcouplea12(index) * q(j)*(q(i)*errfn - expfn ) &
                      + harmcouplea21(index) &
                      *0.5*((q(i)**2+0.5/gamma(i))*errfn-(truncpos(index)+q(i))*expfn) &
                      + harmcouplea22(index) * q(j)&
                      *((q(i)**2+0.5/gamma(i))*errfn-(truncpos(index)+q(i))*expfn))

              else ! truncating according to q(j)
                 errfn = 1+0.5*trunc(index)*cderror(sgamma(j)*(truncpos(index)-q(j)))
                 expfn = 0.5*trunc(index)*cdexp(-gamma(j)*(truncpos(index)-q(j))**2)/(sqrtpi*sgamma(j))


                 !dhampolymorsedz(j) = dhampolymorsedz(j) & 
                 dhampolymorsedzj = s2gammarec(j) &
                      *(harmcouplea21(index) * (q(i)**2+0.5/gamma(i))&
                      *(0.5*errfn - truncpos(index)*gamma(j)*expfn ) &
                      + harmcouplea12(index) * q(i)&
                      *(q(j)*errfn-(1+gamma(j)*truncpos(index)**2)*expfn) &
                      + harmcouplea22(index) * (q(i)**2+0.5/gamma(i))&
                      *(q(j)*errfn-(1+gamma(j)*truncpos(index)**2)*expfn)) 


                 !                dhampolymorsedz(i) = dhampolymorsedz(i) &
                 dhampolymorsedzi = s2gammarec(i) & 
                      *(harmcouplea21(index) * q(i)*(q(j)*errfn - expfn ) &
                      + harmcouplea12(index) &
                      *0.5*((q(j)**2+0.5/gamma(j))*errfn-(truncpos(index)+q(j))*expfn) &
                      + harmcouplea22(index) * q(i)&
                      *((q(j)**2+0.5/gamma(j))*errfn-(truncpos(index)+q(j))*expfn))
              end if

           else if (trunc(index) .eq. 3) then ! truncate in both directions for one mode
              if (truncm(index) .eq. i) then
                 !!                errfn = cderror(sgamma(i)*(truncpos(index)-q(i)))
                 !!                errfn2 = cderror(sgamma(i)*(truncpos(index)+q(i)))
                 !!                expfn = cdexp(-gamma(i)*(truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))
                 !!                expfn2 = cdexp(-gamma(i)*(-truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))

                 ! Calculate dC1dq1, dC2dq1
                 !                C1i = 0.5*(errfn+errfn2) -sgamma(i)/sqrtpi*truncpos(index)*(expfn+expfn2)
                 !                C2i = (expfn2-expfn)/(sgamma(i)*sqrtpi) + q(i)* (errfn+errfn2)
                 !!                dC1di = 0.5*(errfn2-errfn) -gamma(i)*truncpos(index)*(expfn+expfn2)
                 !!                dC2di = (gamma(i)*truncpos(index)**2+1)*(expfn2-expfn) + q(i)* (errfn2+errfn)


                 t1 = truncpos(index) - q(i)
                 t2 = t1 ** 2
                 t4 = cdexp(-gamma(i) * t2)
                 t6 = 0.1D1 / sgamma(i)

                 t9 = 0.1D1 / sqrtpi
                 t11 = sgamma(i)* truncpos(index)
                 t12 = sgamma(i)* q(i)
                 t13 = t11 - t12
                 t14 = cderror(t13)
                 t16 = -truncpos(index) - q(i)
                 t17 = t16 ** 2
                 t19 = cdexp(-gamma(i) * t17)
                 t22 = -t11 - t12
                 t23 = cderror(t22)
                 c1i = -t4 * t6 * t9 / 0.2D1 + q(i) * t14 / 0.2D1 + t19 * t6 * t9 / 0.2D1 - q(i) * t23 / 0.2D1

                 t33 = -t13
                 t34 = -t14 ! erf(t33)

                 t43 = -t22
                 t44 = -t23 ! erf(t43)
                 dC1di = -(0.2D1 * sgamma(i)* t4 * truncpos(index) + t34 * sqrtpi &
                      + 0.2D1 * sgamma(i)* t19 * truncpos(index) - t44 * sqrtpi) * t9 / 0.2D1

                 t52 = 0.1D1 / gamma(i) * t9
                 t54 = q(i) ** 2
                 t58 = -0.2D1 * gamma(i) * t54 * sqrtpi - sqrtpi
                 c2i = -t43 * t4 * t52 / 0.2D1 - t58 * t14 * t52 / 0.4D1 &
                      + t33 * t19 * t52 / 0.2D1 + t58 * t23 * t52 / 0.4D1

                 t67 = truncpos(index) ** 2
                 t69 = 0.1D1 + gamma(i) * t67
                 t71 = t6 * t9
                 dc2di = -t69 * t4 * t71 + t69 * t19 * t71 - q(i) * t34 + q(i) * t44

                 !             dhampolymorsedz(i) = dhampolymorsedz(i) &
                 dhampolymorsedzi = s2gammarec(i) &
                      * ( harmcouplea12(index) * (q(j)**2+0.5/gamma(j))*dC1di & 
                      + harmcouplea21(index) * q(j) * dC2di &
                      + harmcouplea22(index) * (q(j)**2+0.5/gamma(j))*dC2di)

                 !Calculate the real C1 and C2 (rather than derivatives)
                 !!                C1i = 0.5*(expfn2-expfn)+0.5*q(i)*(errfn+errfn2)
                 !!                C2i = 0.5*(q(i)-truncpos(index))*expfn2-0.5*(q(i)+truncpos(index))*expfn &
                 !!                     + (0.5*q(i)**2+0.25/gamma(i))*(errfn+errfn2)

                 !               dhampolymorsedz(j) = dhampolymorsedz(j) & 
                 dhampolymorsedzj = s2gammarec(j) &
                      * (harmcouplea12(index) * 2*q(j)*C1i & 
                      + harmcouplea21(index) * C2i &
                      + harmcouplea22(index) * 2*q(j)*C2i) 

              else
                 !!                errfn = cderror(sgamma(j)*(truncpos(index)-q(j)))
                 !!                errfn2 = cderror(sgamma(j)*(truncpos(index)+q(j)))
                 !!                expfn = cdexp(-gamma(j)*(truncpos(index)-q(j))**2)/(sqrtpi*sgamma(j))
                 !!                expfn2 = cdexp(-gamma(j)*(-truncpos(index)-q(j))**2)/(sqrtpi*sgamma(j))

                 !Calculate the real C1 and C2 (rather than derivatives)
                 !!                C1j = 0.5*(expfn2-expfn)+0.5*q(j)*(errfn+errfn2)
                 !!                C2j = 0.5*(q(j)-truncpos(index))*expfn2-0.5*(q(j)+truncpos(index))*expfn &
                 !!                     + (0.5*q(j)**2+0.25/gamma(j))*(errfn+errfn2)


                 t1 = truncpos(index) - q(j)
                 t2 = t1 ** 2
                 t4 = cdexp(-gamma(j) * t2)
                 t6 = 0.1D1 / sgamma(j)

                 t9 = 0.1D1 / sqrtpi
                 t11 = sgamma(j)* truncpos(index)
                 t12 = sgamma(j)* q(j)
                 t13 = t11 - t12
                 t14 = cderror(t13)
                 t16 = -truncpos(index) - q(j)
                 t17 = t16 ** 2
                 t19 = cdexp(-gamma(j) * t17)
                 t22 = -t11 - t12
                 t23 = cderror(t22)
                 c1j = -t4 * t6 * t9 / 0.2D1 + q(j) * t14 / 0.2D1 + t19 * t6 * t9 / 0.2D1 - q(j) * t23 / 0.2D1

                 t33 = -t13
                 t34 = -t14 ! erf(t33)

                 t43 = -t22
                 t44 = -t23 ! erf(t43)
                 dC1dj = -(0.2D1 * sgamma(j)* t4 * truncpos(index) + t34 * sqrtpi &
                      + 0.2D1 * sgamma(j)* t19 * truncpos(index) - t44 * sqrtpi) * t9 / 0.2D1

                 t52 = 0.1D1 / gamma(j) * t9
                 t54 = q(j) ** 2
                 t58 = -0.2D1 * gamma(j) * t54 * sqrtpi - sqrtpi
                 c2j = -t43 * t4 * t52 / 0.2D1 - t58 * t14 * t52 / 0.4D1 &
                      + t33 * t19 * t52 / 0.2D1 + t58 * t23 * t52 / 0.4D1

                 t67 = truncpos(index) ** 2
                 t69 = 0.1D1 + gamma(j) * t67
                 t71 = t6 * t9
                 dc2dj = -t69 * t4 * t71 + t69 * t19 * t71 - q(j) * t34 + q(j) * t44


                 !                dhampolymorsedz(i) = dhampolymorsedz(i) & 
                 dhampolymorsedzi = s2gammarec(i) &
                      * (harmcouplea12(index) * 2*q(i)*C1j & 
                      + harmcouplea21(index) * C2j &
                      + harmcouplea22(index) * 2*q(i)*C2j) 

                 ! Calculate dC1dq1, dC2dq1
                 !                C1j = 0.5*(errfn+errfn2) -sgamma(j)/sqrtpi*truncpos(j)*(expfn+expfn2)
                 !                C2j = (expfn2-expfn)/(sgamma(j)*sqrtpi) + q(j)* (errfn+errfn2)
                 !!                dC1dj = 0.5*(errfn+errfn2) -gamma(j)*truncpos(index)*(expfn+expfn2)
                 !!                dC2dj = (gamma(j)*truncpos(index)**2+1)*(expfn2-expfn) + q(j)* (errfn2-errfn)

                 !                dhampolymorsedz(i) = dhampolymorsedz(i) &
                 dhampolymorsedzj = s2gammarec(j) &
                      * ( harmcouplea12(index) * (q(i)**2+0.5/gamma(i))*dC1dj & 
                      + harmcouplea21(index) * q(i) * dC2dj &
                      + harmcouplea22(index) * (q(i)**2+0.5/gamma(i))*dC2dj)
              end if


           else if (trunc(index) .eq. 4) then 
              ! ---- truncate in all four directions for one mode ----
              errfn = cderror(sgamma(i)*(truncpos(index)-q(i)))
              errfn2 = cderror(sgamma(i)*(truncpos(index)+q(i)))
              expfn = cdexp(-gamma(i)*(truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))
              expfn2 = cdexp(-gamma(i)*(-truncpos(index)-q(i))**2)/(sqrtpi*sgamma(i))

              ! Calculate C1i, C2i
              C1i = 0.5/(sqrtpi*sgamma(i))*(expfn2-expfn)+0.5*q(i)*(errfn+errfn2)
              C2i = 0.5*(q(i)-truncpos(index))*expfn2-0.5*(q(i)-truncpos(index))*expfn &
                   + (0.5*q(i)**2+0.25/gamma(i))*(errfn+errfn2)

              ! Calculate dC1dq1, dC2dq1
              dC1di = 0.5*(errfn+errfn2) -sgamma(i)/sqrtpi*truncpos(index)*(expfn+expfn2)
              dC2di = (expfn2-expfn)/(sgamma(i)*sqrtpi) + q(i)* (errfn+errfn2)

              errfn = cderror(sgamma(j)*(truncpos2(index)-q(j)))
              errfn2 = cderror(sgamma(j)*(truncpos2(index)+q(j)))
              expfn = cdexp(-gamma(j)*(truncpos2(index)-q(j))**2)/(sqrtpi*sgamma(j))
              expfn2 = cdexp(-gamma(j)*(-truncpos2(index)-q(j))**2)/(sqrtpi*sgamma(j))

              ! Calculate C1j, C2j
              C1j = 0.5/(sqrtpi*sgamma(j))*(expfn2-expfn)+0.5*q(j)*(errfn+errfn2)
              C2j = 0.5*(q(j)-truncpos2(index))*expfn2-0.5*(q(j)-truncpos2(index))*expfn &
                   + (0.5*q(j)**2+0.25/gamma(j))*(errfn+errfn2)

              ! Calculate dC1dq1, dC2dq1
              dC1dj = 0.5*(errfn+errfn2) -sgamma(j)/sqrtpi*truncpos2(j)*(expfn+expfn2)
              dC2dj = (expfn2-expfn)/(sgamma(j)*sqrtpi) + q(j)* (errfn+errfn2)


              !             dhampolymorsedz(i) =  dhampolymorsedz(i) & 
              dhampolymorsedzi =  s2gammarec(i) &
                   * ( harmcouplea12(index) * C2j * dC1di & 
                   + harmcouplea21(index) * C1j * dC2di &
                   + harmcouplea22(index) * C2j * dC2di)

              !             dhampolymorsedz(j) =  dhampolymorsedz(j) & 
              dhampolymorsedzj =  s2gammarec(j) &
                   * ( harmcouplea12(index) * dC2dj * C1i & 
                   + harmcouplea21(index) * dC1dj * C2i &
                   + harmcouplea22(index) * dC2dj * C2i)

           else
              write(*,*) "Don't know how to calculate the deriviate with coupling truncation method", trunc(index)
              stop
           end if
           
           dhampolymorsedz(i) = dhampolymorsedz(i) + coups*dhampolymorsedzi
           dhampolymorsedz(j) = dhampolymorsedz(j) + coups*dhampolymorsedzj
            
            if (coupswitch(index) .eq. 1) then
               dhampolymorsedz(coupswitchmode(index)) = dhampolymorsedz(coupswitchmode(index)) &
                    - 1/sqrtpi/sqrt2*coupswitchdir(index) &
                    *cdexp(-gamma(coupswitchmode(index))*(coupswitchpos(index)-q(coupswitchmode(index)))**2) &
                    * hampolymorsecoup(i,j,index,q,p)
               !s2gammarec(coupswitchmode(index))*sgamma(coupswitchmode(index)) =  1/sqrt2
            else if (coupswitch(index) .eq. 2) then
               dhampolymorsedz(coupswitchmode(index)) =  dhampolymorsedz(coupswitchmode(index)) &
                    - 2*s2gammarec(coupswitchmode(index)) * coupswitchgamma(index)*q(coupswitchmode(index))&
                     * (dsqrt(gamma(coupswitchmode(index))/(gamma(coupswitchmode(index))+coupswitchgamma(index))))**3 &
                     * cdexp(-gamma(coupswitchmode(index))*coupswitchgamma(index)*q(coupswitchmode(index))**2&
                     /(gamma(coupswitchmode(index))+coupswitchgamma(index))) &
                     * hampolymorsecoup(i,j,index,q,p)
            end if
        end if ! coups > 1e-9 ! ***** This may not be in the correct place for future switches. ****
           index = index + 1
        end do
     end do
   end function dhampolymorsedz

  
  ! ==================================================
  !                Reordered Polynomial
  ! ==================================================
   function polyeng(i,qi,pd) 
    use coherentstates, ONLY: gamma
    implicit none

    integer, intent(in) :: i
    ! degree of poynomial
    integer, intent(in) :: pd 
    !Position of ith degree of freedom
    complex(kind=8), intent(in) :: qi 
    ! energy
    complex(kind=8) :: polyeng

    ! 
    polyeng = pcoeff(i,0)+ pcoeff(i,1)*qi
    if (pd .eq. 1) return

    ! quadratic terms
    polyeng = polyeng &
         +  pcoeff(i,2) * (qi**2 + 0.5/gamma(i))
     if (pd .eq. 2) return

     ! Cubic terms
      polyeng = polyeng + pcoeff(i,3)*( &
           qi*(qi**2 + 1.5/gamma(i)))
      if (pd .eq. 3) return

      ! Quartic terms
      polyeng = polyeng + &
           pcoeff(i,4)*( qi**4 + 3.0*qi**2/gamma(i) + 0.75/(gamma(i))**2)
       if (pd .eq. 4) return

       ! Quintic terms
       polyeng = polyeng + &
             pcoeff(i,5) * (qi*(qi**4+5.0*qi**2/gamma(i)+3.75/(gamma(i)**2)))
       if (pd .eq. 5) return

       ! hexic terms
       polyeng = polyeng + &
             pcoeff(i,6)*(qi**6 + 7.5*qi**4/gamma(i)+11.25*qi**2/(gamma(i)**2) &
            + 1.875/(gamma(i)**3))
       if (pd .eq. 6) return

       ! hepta terms
       polyeng = polyeng + &
            pcoeff(i,7)*qi*(qi**6+10.5*qi**4/gamma(i)+26.25*qi**2/(gamma(i)**2) &
            + 13.125/(gamma(i)**3))
       if (pd .eq. 7) return
        
       !Octa terms
        polyeng = polyeng + &
             pcoeff(i,8)*(qi**8 + 14*qi**6/gamma(i) + 52.5*qi**4/(gamma(i)**2) &
             +52.5*qi**2/(gamma(i)**3) + 6.5625/(gamma(i)**4))
       if (pd .eq. 8) return

       !nonaterms
        polyeng = polyeng + &
             pcoeff(i,9)*qi*(qi**8+18.0*qi**6 + 94.5*qi**4/(gamma(i)**2) &
             + 157.5*qi**2/(gamma(i)**3) + 59.0625/(gamma(i)**4))
        if (pd .eq. 9) return

        ! deca terms
        polyeng = polyeng + &
             pcoeff(i,10)*(qi*10 + 22.5*qi**8/gamma(i) &
             + 157.5*qi**6/(gamma(i)**2) + 393.75*qi**4/(gamma(i)**3) &
             + 295.3125 * qi**2/(gamma(i)**4) + 29.53125/(gamma(i)**5))
        !          if ( pd .eq. 10) return

          
  end function polyeng


  ! ==================================================
  !    Derivative w.r.t. q of Reordered Polynomial
  ! ==================================================  
   function dpolyengdq(i,qi,pd) 
    use coherentstates, ONLY: gamma
    implicit none
    integer, intent(in) :: i
    ! degree of poynomial
    integer, intent(in) :: pd 
    !Position of ith degree of freedom
    complex(kind=8), intent(in) :: qi 
    ! energy
    complex(kind=8) :: dpolyengdq

!    dpolyengdq = 0.0

    ! linear term
    dpolyengdq = pcoeff(i,1)*(1.0,0.0)
    if (pd .eq. 1) return

    ! quadratic terms
    dpolyengdq = dpolyengdq +  pcoeff(i,2)*2*qi
     if (pd .eq. 2) return

     ! Cubic terms
      dpolyengdq = dpolyengdq + pcoeff(i,3)*( &
           (3*qi**2+1.5/gamma(i)))
      if (pd .eq. 3) return

      ! Quartic terms
      dpolyengdq = dpolyengdq + &
           pcoeff(i,4)*( 4.0*qi**3 + 6.0*qi/gamma(i) )
       if (pd .eq. 4) return

       ! Quintic terms
       dpolyengdq = dpolyengdq + &
             pcoeff(i,5) * (5.0*qi**4+15.0*qi**2/gamma(i)+3.75/(gamma(i)**2))
       if (pd .eq. 5) return

       ! hextic terms
       dpolyengdq = dpolyengdq + &
             pcoeff(i,6)*(6.0*qi**5 + 30.0*qi**3/gamma(i)+22.5*qi/(gamma(i)**2))
       if (pd .eq. 6) return

       ! heptic terms
       dpolyengdq = dpolyengdq + &
            pcoeff(i,7)*(7.0*qi**6+52.5*qi**4/gamma(i) &
            +78.75*qi**2/(gamma(i)**2) + 13.125/(gamma(i)**3))
       if (pd .eq. 7) return
        
        dpolyengdq = dpolyengdq + &
             pcoeff(i,8)*(8.0*qi**7 + 84.0*qi**5/gamma(i) &
             + 210.0*qi**3/(gamma(i)**2) +105.0*qi**2/(gamma(i)**3) )
       if (pd .eq. 8) return
       
       !nonaterms
        dpolyengdq = dpolyengdq + &
             pcoeff(i,9)*(9*qi**8+126.0*qi**6 + 472.5*qi**4/(gamma(i)**2) &
             + 472.5*qi**2/(gamma(i)**3) + 59.0625/(gamma(i)**4))
        if (pd .eq. 9) return

        ! deca terms
        dpolyengdq = dpolyengdq + &
             pcoeff(i,10)*(qi*10 + 180.0*qi**7/gamma(i) &
             + 945.0*qi**5/(gamma(i)**2) + 1575.0*qi**3/(gamma(i)**3) &
             + 590.625 * qi/(gamma(i)**4) )
        !          if ( pd .eq. 10) return

  end function dpolyengdq


  ! ==================================================
  !                Reordered Morse fct
  ! ==================================================
  function morseeng(i,alpha,qi,q0,A,mc)
    use coherentstates, ONLY: ndf,gamma
    implicit none
    integer,intent(in) :: i
    real(kind=8),intent(in) :: alpha
    complex(kind=8), intent(in) :: qi
    real(kind=8), intent(in) :: q0
    real(kind=8), intent(in) :: A
    real(kind=8), intent(in) :: mc



    complex(kind=8) :: morseeng
    !          A[m]*(1-2*exp(-alpha*(q-q0))*e^(-alpha*(8*gamma)^(-.5))+exp(-2*alpha*(q-q0))*e^(-alpha*(2*gamma)^(-.5)))
!    gamma(i)
   
  !  morseeng = A*((1.0,0.) - 2* exp(-alpha*((qi-q0)+ dsqrt(0.125/gamma(i)))) &
  !       + exp(-alpha*(2*(qi-q0)+ dsqrt(0.5/gamma(i)))))+mc

    morseeng = A*((1.0,0.) - 2* exp(-alpha*((qi-q0)-0.25*alpha/gamma(i))) &
         + exp(-alpha*(2*(qi-q0)- alpha/gamma(i))))+mc

  end function morseeng

  ! ==================================================
  !   Derivative wrt q of th reordered Morse fct
  ! ==================================================
  function dmorseengdq(i,alpha,qi,q0,A)
    use coherentstates, ONLY: gamma
    implicit none
    integer,intent(in) :: i
    real(kind=8),intent(in) :: alpha
    complex(kind=8), intent(in) :: qi
    real(kind=8), intent(in) :: q0
    real(kind=8), intent(in) :: A

    complex(kind=8) :: dmorseengdq
    ! A[m]*(2*alpha*exp(-alpha*(q-q0))*e^(-.3535533906*alpha/gamma^.5)-2*alpha*exp(-2*alpha*(q-q0))*e^(-.7071067812*alpha/gamma^.5))

!!    dmorseengdq = A*(2.0,0)*alpha* (exp(-alpha*((qi-q0)+ dsqrt(0.125/gamma(i)))) &
!         -exp(-alpha*(2*(qi-q0)+ dsqrt(0.5/gamma(i)))))

    dmorseengdq = A*(2.0,0)*alpha* (exp(-alpha*((qi-q0)-0.25*alpha/gamma(i))) &
         -exp(-alpha*(2*(qi-q0)- alpha/gamma(i))))

!    dmorseengdq = (2.0,0)*alpha* (exp(-alpha*((qi-q0))) &
!         -exp(-alpha*(2*(qi-q0))))

  end function dmorseengdq

  ! ==================================================
  !                Reordered Gaussian fct
  ! ==================================================
  function gausseng(i,zic,zj,alpha,q0)
    use coherentstates, ONLY: gamma
    implicit none
    integer,intent(in) :: i
    complex(kind=8),intent(in) :: zic,zj
        
    real(kind=8),intent(in) :: alpha
    real(kind=8), intent(in) :: q0

    complex(kind=8) :: gausseng

    if(zic .eq. zj) then
       gausseng = dsqrt(gamma(i)/(gamma(i)+alpha)) *exp(-0.5*(zic+zj)*(zic+zj)*(alpha/(gamma(i)+alpha)))
    else
!       gausseng = dsqrt(gamma(i)/(gamma(i)+alpha)) &
!            *exp(-0.5*(zic+zj)*(zic+zj)*(alpha/(gamma(i)+alpha)) + 0.5*(zic*dconjg(zic)-zj*dconjg(zj)))
       gausseng = dsqrt(gamma(i)/(gamma(i)+alpha)) *exp(-0.5*(zic+zj)*(zic+zj)*(alpha/(gamma(i)+alpha)))

    end if
  end function gausseng

  ! ==================================================
  !   Derivative wrt q of th reordered Gaussian fct
  ! ==================================================
  function dgaussengdz(i,zic,zj,alpha,q0)
    use coherentstates, ONLY: ndf,gamma
    implicit none
    integer,intent(in) :: i
    complex(kind=8),intent(in) :: zic,zj
        
    real(kind=8),intent(in) :: alpha
    real(kind=8), intent(in) :: q0

    complex(kind=8) :: dgaussengdz,gausseng

    if(dconjg(zic) .eq. zj) then
       gausseng = dsqrt(gamma(i)/(gamma(i)+alpha)) *exp(-0.5*(zic+zj)*(zic+zj)*(alpha/(gamma(i)*alpha)))
       dgaussengdz = -(zic+zj)*(alpha/(gamma(i)+alpha))*gausseng    
    else
       gausseng = dsqrt(gamma(i)/(gamma(i)+alpha)) &
            *exp(-0.5*(zic+zj)*(zic+zj)*(alpha/(gamma(i)*alpha)) + 0.5*(zic*dconjg(zic)-zj*dconjg(zj)))

       dgaussengdz = -(0.5*dconjg(zj)+(zic+zj)*alpha/(gamma(i)+alpha))*gausseng
       write(*,*) "zic not equal to dconjg(zj)"
       stop
    end if

  end function dgaussengdz



  ! ==================================================
  !            Read Hamiltomian parameters
  !     specific to this Hamiltonian
  !  allocates arrays too.
  ! ==================================================
  subroutine readhampm 
    use options, ONLY: verbosity,writeexpect,eu
    use  coherentstates, ONLY: ndf
    use files, ONLY: fug,fham
    use arrays, ONLY: dhexpect,hework
    use errors
    use scalapackstuff, ONLY: isroot

    character(len=250) ::line
    character(len=50) :: line2 ! for formatted output
    character(len=5) :: line3 ! for formatted output
    integer :: ierr ! For errors reading from the file
    integer :: i,j    ! For loops, what else?
    integer :: index,coupling
    integer :: numcouplings ! How many couplings are read in?
    real(kind=8) a12,a21,a22 ! For reading in coupling.
    integer :: numtruncs,tt,tm ! number of truncations in file and truncation type and mode
    real(kind=8) :: tp,tp2,tw ! truncation position and width


    integer :: numswitches,csw,st,sm,sd
    real(kind=8) :: sp,sg

    ! Allocate more general arrays
    call allocatehamarrays

    ! Allocate Hamiltonian specific arrays
    allocate(polydegree(ndf))
    allocate(pcoeff(ndf,0:10))
    allocate(morse(ndf))
    allocate(morseA(ndf))
    allocate(morsealpha(ndf))
    allocate(morseq0(ndf))
    allocate(morseC(ndf))
    allocate(gaussian(ndf))
    allocate(gaussA(ndf))
    allocate(gaussalpha(ndf))
    allocate(gaussq0(ndf))

    allocate(harmcouplea12(ndf*(ndf-1)/2))
    allocate(harmcouplea21(ndf*(ndf-1)/2))
    allocate(harmcouplea22(ndf*(ndf-1)/2))
    
    ! For coupling truncation
    allocate(trunc(ndf*(ndf-1)/2))
    allocate(truncm(ndf*(ndf-1)/2))
    allocate(truncpos(ndf*(ndf-1)/2))
    allocate(truncpos2(ndf*(ndf-1)/2))
    allocate(truncwidth(ndf*(ndf-1)/2))

    ! For coupling switching
    allocate(coupswitch(ndf*(ndf-1)/2))
    allocate(coupswitchmode(ndf*(ndf-1)/2))
    allocate(coupswitchdir(ndf*(ndf-1)/2))
    allocate(coupswitchpos(ndf*(ndf-1)/2))
    allocate(coupswitchgamma(ndf*(ndf-1)/2))


    ! For dropping basis functions
    if (writeexpect > 0 ) then
       allocate(dhexpect(ndf+1))
       allocate(hework(ndf+1))
       dhexpect = 0.0d0
    end if

    ! Empty arrays
    polydegree(:) = 0
    pcoeff(:,:) = 0.0
    morse(:) = .false.
    morseA(:) = 0.0
    morsealpha(:) = 0.0
    morseq0(:) = 0.0
    gaussian(:) = .false.
    gaussA(:) = 0.0
    gaussalpha(:) = 0.0
    gaussq0(:) = 0.0

    harmcouplea12(:)= 0.0
    harmcouplea21(:)= 0.0
    harmcouplea22(:)= 0.0

    trunc(:) = 0
    truncm(:) = 0
    truncpos(:) = 0.0
    truncpos2(:) = 0.0
    truncwidth(:) = 0.0

    coupswitch(:) = 0
    coupswitchmode(:) = 0
    coupswitchdir(:) = 0
    coupswitchpos(:) = 0.0d0
    coupswitchgamma(:) = 0.0d0

    ! ---- Read in masses -----
    read(fham,*,iostat=ierr) line
    if  (line .ne. "mass" .or. ierr .ne. 0  ) then
       if (isroot) write(fug,*) "Failed to read keyword mass in hamiltonian.inpt"
       stop
    end if
    do i =1,ndf
       read(fham,*,iostat=ierr) m(i)
       if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading masses")
    end do


    ! ---- Now read in mode energy fct ----
    ! Read first key word mode
   read(fham,*,iostat=ierr) line

    do i = 1,ndf
       ! Check that the keyword is mode
       if  (line .ne. "mode" ) then
          if (isroot) write(fug,*) "Failed to read keyword mode in hamiltonian.inpt"
          stop
       else if ( ierr .ne. 0 ) then 
          call filereaderror(ierr,"hamiltonian.inpt","Error reading mode")
       end if

       ! Now read interaction type
       read(fham,*,iostat=ierr) line

       do while (line .ne. "mode" .and. ierr .eq.0 .and. line .ne. "end")

          ! ---- polynomial ----
          if (line .eq. "poly") then

             backspace(fham)
             read(fham,*,iostat=ierr) line,polydegree(i)
             
             if  ( ierr .ne. 0 .or. polydegree(i) .le. 0 ) then
                if (isroot) write(fug,*) "Failed to read suitable degree of polynomial in hamiltonian.inpt for mode ",i
                stop
             end if
             do j = 0,polydegree(i)
                read(fham,*,iostat=ierr) line,pcoeff(i,j)
                if  ( ierr .ne. 0 .or. polydegree(i) .le. 0 ) then
                   if(isroot) write(fug,*) "Failed to read coefficient for  degree",j," for mode ",i
                   stop
                end if
                ! Convert to internal energy units
                pcoeff(i,j) = pcoeff(i,j)/eu
             end do

             ! ---- morse ----
          else if (line .eq. "morse" ) then
             morse(i) = .true.
             read(fham,*,iostat=ierr) line,morseA(i)
             if  ( line .ne. "A" .or.  ierr .ne. 0  ) then
                if (isroot) write(fug,*) "Failed to read coefficient A for morse potential of mode ",i
                stop
             end if
              ! Convert to internal energy units
             morseA(i) = morseA(i)/eu

             read(fham,*,iostat=ierr) line,morsealpha(i)
             if  ( line .ne. "alpha" .or. ierr .ne. 0  ) then
                if (isroot) write(fug,*) "Failed to read coefficient alpha for morse potential of mode ",i
                stop
             end if
             read(fham,*,iostat=ierr) line,morseq0(i)
             if  ( line .ne. "q0" .or. ierr .ne. 0  ) then
                if (isroot) write(fug,*) "Failed to read q0 for morse potential of mode ",i
                stop
             end if
             read(fham,*,iostat=ierr) line,morseC(i)
             if  ( line .ne. "C" .or. ierr .ne. 0  ) then
                if (isroot) write(fug,*) "Failed to read C for morse potential of mode ",i
                stop
             end if
              ! Convert to internal energy units
             morseC(i) = morseC(i) /eu

             ! ---- Gaussian ----
          else if (line .eq. "gaussian" ) then
             gaussian(i) = .true.
             read(fham,*,iostat=ierr) line,gaussA(i)
             if  ( line .ne. "A" .or.  ierr .ne. 0  ) then
                if(isroot) write(fug,*) "Failed to read coefficient A for Gaussian potential of mode ",i
                stop
             end if
              ! Convert to internal energy units
             gaussA(i) = gaussA(i) /eu

             read(fham,*,iostat=ierr) line,gaussalpha(i)
             if  ( line .ne. "alpha" .or. ierr .ne. 0  ) then
                if(isroot) write(fug,*) "Failed to read coefficient alpha for morse potential of mode ",i
                stop
             end if
             read(fham,*,iostat=ierr) line,gaussq0(i)
             if  ( line .ne. "q0" .or. ierr .ne. 0  ) then
                if (isroot) write(fug,*) "Failed to read q0 for morse potential of mode ",i
                stop
             end if
          else 
             if (isroot) write(fug,*)  "Found unknown potential type in hamiltonian.inpt", line
          end if
          ! Read in next line
          read(fham,*,iostat=ierr) line
       end do ! do while
    end do

    ! ---- Mode Coupling ----
    read(fham,*,iostat=ierr) line

    numcouplings = 0

    if (line .eq. "coupling") then

       backspace(fham)
       read(fham,*,iostat=ierr) line,numcouplings
       if ( ierr.ne. 0 )then
          if (isroot) write(fug,*) "Error reading number of couplings from hamiltonian.inpt"
          if (isroot) write(fug,*) line,numcouplings
          stop
       end if

       if (verbosity .ge. 1 .and. isroot) then
          write(fug,*) "Reading ",numcouplings," couplings"
       end if

       do coupling = 1,numcouplings

          read(fham,*,iostat = ierr) i,j, a12,a21,a22
          if ( ierr.ne. 0 )then
             if (isroot) write(fug,*) "Error reading coupling ",coupling," from hamiltonian.inpt"
             stop
          end if
          if ( i .eq. j) then
             if (isroot) write(fug,*) "Can't couple a mode to itself in hamiltonian.inpt"
             stop
          else if ( i .gt.j) then
             if (isroot) write(fug,*) " In coupling potential first index has to be less than the second index."
             stop
          end if
          
          index = j-i + ndf*(ndf-1)/2 - (ndf-i+1)*(ndf-i)/2
          
          harmcouplea12(index)=a12/eu
          harmcouplea21(index)=a21/eu
          harmcouplea22(index)=a22/eu
       end do
       
       ! Check to see if the is a coupling for each pair of degrees of freedom
       if (numcouplings .lt. ndf*(ndf-1)/2) then
          if (isroot) write(fug,*) "  Warning: Only read ",numcouplings, " sets of couplings."
          if (isroot) write(fug,*) "  Setting other couplings to zero" 	
          if (isroot) write(fug,*) 
       end if


       ! ---- Mode Coupling truncation ----
       read(fham,*,iostat=ierr) line

       numtruncs = 0
       
       if (line .eq. "couptrunc") then

          backspace(fham)
          read(fham,*,iostat=ierr) line,numtruncs
          if ( ierr.ne. 0 )then
             if (isroot) then
                write(fug,*) "Error reading number of couplings from hamiltonian.inpt"
                write(fug,*) line,numtruncs
             end if
             stop
          end if
          
          if (verbosity .ge. 1 .and. isroot) then
             write(fug,*) "Reading ",numtruncs," truncations"
          end if
          
          do coupling = 1,numtruncs
             
             read(fham,*,iostat = ierr) i,j, tt
             !,tm,tp,tw

             if ( ierr.ne. 0 )then
                if (isroot) write(fug,*) "Error reading truncation ",coupling," from hamiltonian.inpt"
                stop
             end if
             if ( i .eq. j) then
                if (isroot) write(fug,*) "Can't couple a mode to itself in hamiltonian.inpt"
                stop
             else if ( i .gt.j) then
                if (isroot) write(fug,*) " In coupling potential truncation first index has to be less than the second index."
                stop
             end if

             tw = 0.0
             tp2 = 0.0
             backspace(fham)
             if (tt .eq. 0) then ! No truncation
                read(fham,*,iostat = ierr) i,j, tt
             else if ( tt.eq. -1 .or. tt.eq.1) then
                read(fham,*,iostat = ierr) i,j, tt,tm,tp,tw
             else if  ( tt.eq. -2 .or. tt.eq.2) then
                read(fham,*,iostat = ierr) i,j, tt,tm,tp
             else if  ( tt.eq.3) then
                read(fham,*,iostat = ierr) i,j, tt,tm,tp
             else if  ( tt.eq.4) then
                read(fham,*,iostat = ierr) i,j, tt,tp,tp2
             else
                if (isroot) then
                   write(fug,*) " Error reading coupling trunctation:"
                   write(fug,*) "     Unknown truncation method",tt
                end if
                stop
             end if

             index = j-i + ndf*(ndf-1)/2 - (ndf-i+1)*(ndf-i)/2
             
             ! save details of this truncation
             trunc(index)=tt
             truncm(index)=tm
             truncpos(index)=tp
             truncpos2(index)=tp2
             truncwidth(index)=tw
          end do
       end if
    end if
    
    ! ---- Mode Coupling switching ----
    read(fham,*,iostat=ierr) line
    
    numswitches = 0
    
    if (line .eq. "coupswitch") then
       
       backspace(fham)
       read(fham,*,iostat=ierr) line,numswitches
       if ( ierr.ne. 0 )then
          if (isroot) then
             write(fug,*) "Error reading number of coupling switches from hamiltonian.inpt"
             write(fug,*) line,numswitches
          end if
          stop
       end if
       
       if (verbosity .ge. 1 .and. isroot) then
          write(fug,*) "Reading ",numswitches," switches"
       end if
       
       do csw = 1,numswitches
          ! find out what type of switch is required
          read(fham,*,iostat = ierr) i,j,st
          backspace(fham)
          
          ! Set temporary variables to 0.
          sd = 0
          sp = 0.0d0
          sg = 0.0d0

          ! Read in details of switch 
          if( st .eq. 1) then
             read(fham,*,iostat = ierr) i,j,st,sm,sd,sp

          else if( st .eq. 2) then
             read(fham,*,iostat = ierr) i,j,st,sm,sg
          else
             if (isroot) then
                write(fug,*) "I don't know about switch type ",st," in hamiltonian.inpt"   
                write(fug,*) "i = ",i," j = ",j
             end if
             stop
          end if
          if (ierr.ne. 0 ) then
             if (isroot) write(fug,*) "Problem reading coupling switch ", csw," from hamiltonian file"
             stop
          end if
          ! Calculate the index and check it is valid.
          index = j-i + ndf*(ndf-1)/2 - (ndf-i+1)*(ndf-i)/2
          if (index <1 .or. index > (ndf*(ndf-1)/2) ) then
             if (isroot) then
                write(fug,*) "No such mode combination in  coupling switch ", csw," from hamiltonian file"
                write(fug,*) "i=",i,", j=",j,", index=",index
             end if
             stop
          end if

          ! save details of this switch
          coupswitch(index) = st
          coupswitchmode(index) = sm
          coupswitchdir(index) = sd
          coupswitchpos(index) = sp
          coupswitchgamma(index) = sg
       end do
    end if
    
    if (isroot) then
       write(fug,*) "Finished reading modes"
       write(fug,*) 
    end if

    ! Summarise
    if (isroot) then
       write(fug,*) "Hamiltonian parameters"
       write(fug,*) "----------------------"
       write(fug,*) "  Studying a system with polynomial and morse potentials"
       write(fug,*) "  Masses:               i     m_i"
       do i = 1,ndf
          write(fug,"(a21,i5,f8.4)") "                     ",i,m(i)
       end do
       
       do i = 1,ndf
          write(fug,"(a7,i4)") "  Mode",i
          if(polydegree(i) > 0) then
             
             write(line,"(f12.6)") pcoeff(i,0)
             
             do j = 1,polydegree(i)
                ! Write power to string variable line3
                if (j .lt. 10) then
                   write(line3,"(i1)") j
                else if ( j .lt. 100) then
                   write(line3,"(i2)") j
                else
                   write(line3,"(i3)") j
                end if
                
                ! write coefficient etc.
                if( j .eq. 1) then
                   write(line2,"(a3,f8.4,a2,a4)")  " + ",pcoeff(i,j)*eu,"q"
                else
                   write(line2,"(a3,f8.4,a2,a4)")  " + ",pcoeff(i,j)*eu,"q^",line3
                end if
                line = trim(line)//trim(line2)
             end do
             write(fug,*) "     V =",trim(line)
          end if
          
          if(morse(i)) then
             if (polydegree(i) > 0) then
                write(fug,*) "        + ",morseA(i)*eu," exp(",&
                     -1*morsealpha(i),".(q-",morseq0(i),")) + " ,morseC(i)
             else
                write(fug,*) morseA(i)*eu," exp(",&
                     -1 * morsealpha(i),".(q-",morseq0(i),")) + " ,morseC(i)
             end if
          end if
          
          if(gaussian(i)) then
             if (polydegree(i) > 0 .or. morse(i)) then
                write(fug,*) "        + ",gaussA(i)*eu," exp(",&
                     -1*gaussalpha(i),".(q-",gaussq0(i),")^2)"
             else
                write(fug,*) gaussA(i)*eu," exp(",&
                     -1*gaussalpha(i),".(q-",gaussq0(i),"))"
             end if
          end if
          
       end do

       ! ---- summarise coupling and truncation ----
       if(numcouplings .gt. 0) then
          write(fug,*) 
          write(fug,*) "  Coupling parameters"
          write(fug,*) "       mode 1  mode 2   a12           a21           a22"
          index = 0
          do i = 1,ndf -1
             do j = i+1,ndf
                index = index + 1
                write (fug,"('    ',2I8,3es14.4)") i,j,harmcouplea12(index)*eu,harmcouplea21(index)*eu,harmcouplea22(index)*eu
             end do
          end do
          ! --- truncation ---
          if(numtruncs .gt. 0) then
             write(fug,*) 
             write(fug,*) "  Coupling truncation"
             write(fug,*) "       mode 1  mode 2   type   tcmode     position      width"
             index = 0
             do i = 1,ndf -1
                do j = i+1,ndf
                   index = index + 1
                   if(trunc(index) .ne. 0) then
                      write (fug,"('    ',4I8,'  ',2es14.4)") i,j,trunc(index), truncm(index),  truncpos(index), truncwidth(index)
                   end if
                end do
             end do
          end if
       end if ! if num couplings > 0
       ! --- switching function ---
       if (numswitches .gt. 0) then
          write(fug,*) 
          write(fug,*) "  Coupling switching "
          write(fug,*) "        Coupled modes  switch   type     dir    pos           width"
          index = 0
          do i = 1,ndf -1
             do j = i+1,ndf
                index = index + 1
                if(coupswitch(index) .ne. 0) then
                   write (fug,"('    ',5i8,2es14.4)") i,j,coupswitch(index), &
                        coupswitchmode(index),  coupswitchdir(index), &
                        coupswitchpos(index), coupswitchgamma(index)
                end if
             end do
          end do
       end if ! if numswitches > 0
    end if
  end subroutine readhampm
  
end module polymorsehammod
