! **************************************************
!                 Restart Procedures 
!
!                  Stewart K. Reed
!                 10th Decemner 2007
!
! Description:
!
! This file contains two routines. One to write out 
! the state of the simulation at the end of a run 
! and the second to read in this file so that a run
! may be restarted.
! **************************************************

! ===============================================
! dump
!  Subroutine to write out all the data need to 
!  be able to continue a calculation from where
!  it stopped.
! ===============================================
subroutine dump(step,filename)
  use coherentstatestype
  use files, ONLY: fdump,fbas,fact,facf,feng,fceng,fpos,famp,fccf,fnorm,fC,fD,fdCdT,&
       fdSdT,fdDdT,fovlp,fhcf,fzpos,fexpect,fnormbasis
  use options, ONLY: ccorlog,writeCDfreq,writeampfreq,hcorfreq,writeposfreq,&
       writeSfreq,writeEfreq,writenormfreq,writebasisnormfreq,writeacffreq,&
       writezposfreq,writeexpect,energyunits,writeovlpfreq

  use arrays, ONLY:t,psteps,bs,D,ninitstates,zinit,initnorm,incbf,dtoteng, &
       dhexpect
  use options, ONLY: classical,energyunits,massunits,dt,sysname
  use coherentstates, ONLY: ndf, nbasis

  use scalapackstuff, ONLY: isroot
  implicit none

  integer,intent(in) :: step
  character(* ),intent(in) :: filename
  integer :: i,j

  if (isroot) then
     open(fdump,file=filename)

     write(fdump,*) "System state"
     write(fdump,*) sysname
     if(classical .eq. 2) then
        write(fdump,*) "Simulation type: classical"
     else if (classical .eq. 1) then
        write(fdump,*) "Simulation type: semiclassical"
     else
        write(fdump,*) "Simulation type: quantum"
     end if
     write(fdump,*) "System time: ",t
     write(fdump,*) "Step num: ",step
     write(fdump,*) "Total number of steps: ",step+psteps


     write(fdump,*) "ndf: ",ndf
     write(fdump,*) "nbasis: ",nbasis
     write(fdump,*) "ninitstates: ",ninitstates
     write(fdump,*) "initnorm: ",initnorm
     write(fdump,*) "dtoteng:" ,dtoteng
     if(allocated(dhexpect)) then
        write(fdump,*) "dhexpect:"
        do i =1,size(dhexpect)
           write(fdump,*) dhexpect(i)
        end do
     end if

     if (ninitstates > 0) then
        write(fdump,*) "Initial States"
        do i = 1,ninitstates
           do j = 1,ndf
              write(fdump,*) real(zinit(j,i)),dimag(zinit(j,i))
           end do
        end do
     end if

     ! write out the basis functions
     write(fdump,*) "Basis Functions"
     do i = 1,nbasis
        do j = 1,ndf
           write(fdump,*)  real(bs%zb(j,i)),dimag(bs%zb(j,i))
        end do
     end do

     if (classical .eq. 0) then
        ! write out the action
        write(fdump,*) "Action"
        do i = 1,nbasis
           write(fdump,*)  bs%S(1,i)
        end do
     end if
     ! write out the amplitudes
     write(fdump,*) "C"
     do i = 1,nbasis
        write(fdump,*)  real(bs%c(i)),dimag(bs%c(i))
     end do

     ! write out the amplitudes for semiclassical runs
     if (classical.eq.1) then
        write(fdump,*) "D"
        do i = 1,nbasis
           write(fdump,*)  real(D(i)),dimag(D(i))
        end do
     end if
     
     ! write out the amplitudes
     write(fdump,*) "Include bf"
     do i = 1,nbasis
        write(fdump,*) incbf(i)
     end do

     ! close file
     close(fdump)
     
     ! ====== Flush other output files ======
     if (writeacffreq .gt. 0 ) then
        flush(facf) 
     end if
     
     if (writeEfreq .gt. 0 ) then
        flush(feng)
        flush(fceng)
     end if
     
     if (writeposfreq > 0 ) then
        flush(fpos)
     end if
     
     if (writezposfreq > 0 ) then
        flush(fzpos)
     end if
     
     if (writeampfreq .gt. 0) then
        flush(famp)
     end	if
     
     ! wavefunction amplitudes
     if(writeCDfreq .gt. 0) then
        flush(fC)
        flush(fD)
        flush(fdCdT)
        flush(fdDdT)
     end if
     
     ! Action
     if(writeSfreq .gt. 0) then
        flush(fact)
     end if
     
     ! Cross correlation function
     if ( ccorlog  > 0 ) then
        flush(fccf)
     end if
     
     ! Cross correlation function
     if ( hcorfreq  > 0 ) then
        flush(fhcf)
     end if
     
     ! Energy expectation value
     if (writeexpect > 0 ) then
        flush(fexpect)
     end if
     
     ! Wavefunction norm
     if(writenormfreq > 0) then
        flush(fnorm)
     end if
     ! contribution of each basis function
     if (writebasisnormfreq > 0 ) then
        flush(fnormbasis)
     end if
     
     if (writeovlpfreq > 0 ) then
        flush(fovlp)
     end if

    end if

end subroutine dump

! ===============================================
! restart
!  Subroutine to read in the file written by 
!  dump. To do this the file should be renamed
!  restart.inpt
! ===============================================

subroutine restart()
  use files, ONLY: fdump,fug
  use errors
  use arrays, ONLY:pt,t,psteps,bs,D,ninitstates,zinit,initnorm,incbf,dtoteng,&
       dhexpect
  use options, ONLY: classical,energyunits,massunits,dt,sysname
  use coherentstates, ONLY: ndf, nbasis
  use hamiltonian, ONLY: calcD
  use scalapackstuff, ONLY: isroot
  implicit none

  integer:: istuff
  character(100) :: stuff
  integer :: i,j
  real(kind=8) :: rev,imv ! real and imaginary parts of a complex number
  integer :: ierr ! for errors

  open(fdump,file="restart.inpt",status="old",iostat=ierr)
  if (ierr.ne.0) then
     if (isroot) then
        write(fug,*) "Error opening restart file"
        call fileerror(ierr,"restart.inpt")
     end if
     stop
  end if
  
  read(fdump,*,iostat=ierr) stuff
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading System keyword")
     end if
  end if
  ! Check stuff
  if (stuff .ne. "System") then
     if (isroot) then
        write(fug,*) "File doesn't seem to be a System State file"
        write(fug,*) stuff
     end if
     stop
  end if

  ! read in simulation title
  read(fdump,'(a95)') stuff
  if (isroot) write(fug,*) "Old title: ",stuff

  ! Read system type
  read(fdump,*,iostat=ierr) stuff,stuff,stuff
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading system type")
     end if
  end if
  if(trim(stuff) .eq. "classical") then
     if (isroot) write(fug,*) "Classical restart"
  else if(trim(stuff) .eq. "semiclassical") then
     if (isroot) write(fug,*) "Semiclassical restart"
  else if(trim(stuff) .eq. "quantum") then
     if (isroot) write(fug,*) "Quantum restart"
  else
     if (isroot) write(fug,*) "Unknown restart",trim(stuff),"."
  end if

  read(fdump,*,iostat=ierr) stuff,stuff,t
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading system time")
     end if
  end if
  if (isroot) write(fug,*) "System time: ",t
  pt = t
  ! ---- Read step number ----
  read(fdump,*) stuff 
  read(fdump,*,iostat=ierr) stuff,stuff,stuff,stuff,psteps
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading system type")
     end if
  else if (stuff .ne. "steps:") then
     if(isroot) then       
         call filereaderror(ierr,"restart.inpt","Error reading number of previous steps")
      end if
  end if
  if (isroot) write(fug,*) "Previously done ", psteps, "steps"
  
  ! ---- Read in the number of degrees of freedom ----
  read(fdump,*,iostat=ierr) stuff,istuff
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading number of degrees of freedom")
     end if
  else if ( istuff .ne. ndf) then
     if (isroot) write( fug,*) "Fatal error: ndf in restart file is not equal to ndf in run.inpt"
     stop
  end if

  ! ---- Read in the number of basis functions ----
  read(fdump,*,iostat=ierr) stuff,istuff
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading number of basis functions")
     end if
  else if ( istuff .ne. nbasis) then
     if (isroot) write( fug,*) "Fatal error: nbasis in restart file is not equal to nbasis in run.inpt"
     stop
  end if

  ! ---- Read in the number of initial states -----
  read(fdump,*,iostat=ierr) stuff,istuff
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading number initial states")
     end if
  else  if ( istuff .ne. ninitstates) then
     if (isroot) write( fug,*) "Fatal error: ninitstates in restart file is not equal to ninitstates in states.inpt"
     stop
  end if
  
  ! ---- Read in the initial norm ----
  read(fdump,*,iostat=ierr) stuff,initnorm ! initnorm
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading number the initial norm")
     end if
  else if ( stuff .ne. "initnorm:") then
     if (isroot) write( fug,*) "Fatal error: have not found initnorm in restart file"
     stop
  end if

  ! ---- read in dtoteng value ----
  read(fdump,*,iostat=ierr) stuff,dtoteng
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading dtoteng")
     end if
  else if ( stuff .ne. "dtoteng:") then
     if (isroot) write( fug,*) "Fatal error: have not found dtoteng in restart file"
     stop
  end if

  ! ---- read in dhexpect values if present ----
  read(fdump,*,iostat=ierr) stuff 
  if (ierr .ne. 0) then
     if(isroot) then
        call filereaderror(ierr,"restart.inpt","Error reading dhexpect")
     end if
  else if(stuff .eq. "dhexpect:") then
     if(allocated(dhexpect)) then
        do i =1,size(dhexpect) !ndf+1
           read(fdump,*) dhexpect(i)
        end do
     else
        if (isroot) then
           write(*,*) "Not reading hexpect"
           write(fug,*) "**** Restart file contains dhexpect values but they   ****"
           write(fug,*) "**** don't seem to be being used in this continuation ****"
           write(fug,*) "**** run so they will be ignored                      ****"
           write(fug,*)
        end if
        do i =1,ndf+1
           read(fdump,*) stuff
        end do
     end if
  else ! No dhexpect values
     backspace(fdump)
     if(allocated(dhexpect)) then
        if(isroot) then
           write(fug,*) 
           write(fug,*) "**** Calculating expectation values but restart file   ****"
           write(fug,*) "**** does not contain dhexpect values. Setting to zero.****"
           write(fug,*)
        end if
     end if
  end if
  
  ! ---- read in initial states ----
  if (ninitstates > 0) then
     read(fdump,*) stuff ! "Initial States"
     if ( stuff .ne. "Initial") then
        if (isroot) write( fug,*) "Fatal error: have not found Initial States in restart file"
        stop
     end if

     do i = 1,ninitstates
        do j = 1,ndf
           read(fdump,*) rev,imv
           zinit(j,i) = rev+(0,1)*imv
!           real(zinit(j,i)),dimag(zinit(j,i))
        end do
     end do
  end if
 
  ! ---- Read in the Basis Functions ----
  read (fdump,*)  stuff 
  if ( stuff .ne. "Basis") then
     if (isroot) write( fug,*) "Fatal error: have not found Basis States in restart file"
     stop
  end if

  do i = 1,nbasis
     do j = 1,ndf
         read(fdump,*,iostat=ierr) rev,imv
           if (ierr .ne. 0) then
              if(isroot) then
                 write(fug,*) "Basis function ",i," degree of freedom ",j
                 call filereaderror(ierr,"restart.inpt","Error reading position/momentum pair.")
              end if
           end if
          bs%zb(j,i) = rev+(0,1)*imv
       end do
  end do

  ! ---- Read in the classical action ----
  if ( classical .eq. 0) then
     read(fdump,*)  stuff ! "Action"
     if (trim(stuff) .ne. "Action") then
        if (isroot) write(fug,*) "No action in restart file. Instead found ",stuff
        stop
     end if
  
     do i = 1,nbasis
        read(fdump,*,iostat=ierr)  bs%S(1,i)
        if (ierr .ne. 0) then
           if(isroot) then
              write(fug,*) "Basis function ",i
              backspace(fdump)
              read(fdump,'(a100)') stuff
              write(fug,*) "read ",stuff
              call filereaderror(ierr,"restart.inpt","Error reading the classical action")
           end if
        end if
     end do
  else ! Classical or semi classical
     bs%S(:,:) = (0.0d0,0.0d0)
  end if

  ! ---- Read in the amplitudes C ----
  read(fdump,*)  stuff ! "C"
  if (trim(stuff) .ne. "C") then
     if (isroot) write(fug,*) "No amplitudes (C)  in restart file. Instead found ",stuff
     stop
  end if
  do i = 1,nbasis
     read(fdump,*,iostat=ierr) rev,imv
     if (ierr .ne. 0) then
        if(isroot) then
           write(fug,*) "Basis function ",i
           call filereaderror(ierr,"restart.inpt","Error reading the amplitude C")
        end if
     end if
     bs%C(i) = rev+(0,1)*imv
  end do

  ! ---- Read in the amplitudes C ----
  read(fdump,*)  stuff ! "D" if it is present
  if (trim(stuff) .ne. "D") then
     if (isroot) then
        !write(fug,*) "No amplitudes (D)  in restart file. Instead found ",stuff
        backspace(fdump)
     end if
  else   
     do i = 1,nbasis
        read(fdump,*,iostat=ierr) rev,imv
        if (ierr .ne. 0) then
           if(isroot) then
              write(fug,*) "Basis function ",i
              call filereaderror(ierr,"restart.inpt","Error reading the amplitude D")
           end if
        end if
        D(i) = rev+(0,1)*imv
     end do
  end if

  ! ---- Read in the Inclusion state ----
  read(fdump,*)  stuff ! "Include"
  if (trim(stuff) .ne. "Include") then
     if (isroot) write(fug,*) "No includes  in restart file. Instead found ",stuff
     stop
  end if
  do i = 1,nbasis
     read(fdump,*,iostat=ierr) incbf(i)
     if (ierr .ne. 0) then
        if(isroot) then
           write(fug,*) "Basis function ",i
           call filereaderror(ierr,"restart.inpt","Error reading the inclusion state.")
        end if
     end if
  end do

  ! Calculate the amplitude D if we are doing a quantum calculation or set it to 
  ! C if doing a classical one. If we are doing a semi classical one it should
  ! have been read in above.
  if ( classical .eq. 0) then
     call calcD(bs,D,incbf)
  else if (classical .eq. 2) then
     D(:) = bs%C(:)
  end if
  close(fdump)

end subroutine restart
