! ========================================================================
! Program: autocorrelation
! Date: 7th March 2010
! Author: Stewart Reed
! Email: S.K.Reed@physics.org
! 
! Description: This program calculates autocorrelation functions
! ========================================================================
program autocorrelation
  use basismodule
  use errors

  implicit none

  character(len=100) :: ampfile
  character(len=100),dimension(:),allocatable :: ampfiles
  integer :: nfiles
  integer :: ierr
  integer :: nbasis
  integer :: ndf
  integer :: ncartdim
  integer :: nparticles
  integer :: readconfigerr
  real(kind=8),dimension(:),allocatable :: gamma ! gammas

  integer :: fauto ! file unit for autocorrelation file

  integer :: famp = 10

  complex(kind=8) :: prob

  ! Variables to control partial trajectory
  real(kind=8) :: t0  ! Initial time offset
  logical      :: pad ! Pad out the dissociation rates for t < t0.
  real(kind=8) :: dt     ! time step to use
  integer      :: nsteps ! number of padding steps to write out.

  ! Weighting for simulation
  real(kind=8) ::simweight

  integer :: bf1,bf2,i,j,file ! for loops
  integer :: fug=99
  integer :: fopt=11
  integer :: fdconfig = 12
  integer :: nconfigs
  logical :: firstdissevent ! does this config have the first dissociation event for this bf?
  logical :: writedissconfig ! write out configs when a neon dissociates?

  type(basisset) :: wavefn0,wavefnt
  ! for reading in/output
  character(len=100) :: line
  character(len=50)  :: keyword


  ! ---- open output file ----
  open(unit=fug,file="auto.out")

!!!! Need to set number of basis functions etc.
!!!! allocate/create wavefn.

  nfiles = 0
  !  ampfile="amplitudes.out"
  ncartdim = 3 ! number of cartesian dimensions
  nparticles = 0   ! number of atoms
  ndf = 0      ! number of degrees of freedom

  nbasis = 50

  ! ---- defaults for padding ----
  pad = .false.
  t0 = 0
  dt = 1
  nsteps = 0

  ! weighting for this simulation
  simweight = 1.0d0

  ! ---- Try to open options file ----
  open(unit=fopt,file="auto.inpt",status="old",iostat = ierr)
  if ( ierr .ne. 0 ) then
     write(fug,*) "Failed to open option file auto.inpt" 
     stop
  else
     ! ---- read options from file ----
     ! Read in options
     read (fopt,'(a100)',iostat=ierr) line
     do while (ierr .eq.0)
        read(line,*,iostat=ierr) keyword
        if (ierr.eq. 0) then
           if (keyword=="nbasis") then
              read(line,*) keyword,nbasis
           else if (keyword=="ndf") then
              read(line,*) keyword,ndf
           else if (keyword=="ncartdim") then
              read(line,*) keyword,ncartdim
              ndf = ncartdim * nparticles ! calculate the total number of degrees of freedom
           else  if (keyword=="nparticles") then
              read(line,*) keyword,nparticles
              ndf = ncartdim * nparticles ! calculate the total number of degrees of freedom
              ! --- read in gammas ----
           else if (keyword=="gamma") then
              if (ndf .eq. 0) then 
                 write(fug,*) &
                      "Please specify the number of degrees of freedom or the number of atoms and cartesian dimensions before gamma"
                 stop
              end if
              allocate (gamma(ndf))
              do i=1,ndf 
                 read(fopt,*,iostat=ierr) keyword,gamma(i)
                 if (ierr .ne. 0) then
                    write(fug,*) "D.o.f. ",i
                    call filereaderror(ierr,"array.inpt","Error reading gamma")
                    stop
                 end if
              end do

              ! --- read in gamma for each atom types ----
           else if (keyword=="pgamma") then
              if (ndf .eq. 0) then 
                 write(fug,*) &
                      "Please specify the number of degrees of freedom or the number of atoms and cartesian dimensions before gamma"
                 stop
              else if (nparticles .eq. 0 .or. ncartdim .eq. 0) then 
                 write(fug,*) "You need to specify the number of cartesian dimensions and the number of atoms before pgamma"
                 stop
              end if
              
              allocate (gamma(ndf))
              do i=0,nparticles-1
                 read(fopt,*,iostat=ierr) keyword,gamma(i*ncartdim+1)
                 if (ierr .ne. 0) then
                    write(fug,*) "D.o.f. ",i+1
                    call filereaderror(ierr,"array.inpt","Error reading gamma")
                    stop
                 end if
                 ! store gammas for other cartesian degrees of freedom
                 do j = 2,ncartdim
                    gamma(i*ncartdim+j) = gamma(i*ncartdim+1)
                 end do
              end do
              
           else if (keyword=="t0") then
              read(line,*) keyword,t0
           else if (keyword=="pad") then
              read(line,*) keyword
              pad = .true.
           else if (keyword=="dt") then
              read(line,*) keyword,dt
           else if (keyword=="simweight") then
              read(line,*) keyword,simweight
           else if (keyword=="multifile") then
              if (nfiles .ne. 0) then
                 write(*,*) "Trying to set number of files using multifile option, but it has already been set."
                 stop
              end if
              read(line,*,iostat = ierr) keyword,nfiles
              if (ierr .ne. 0 ) then
                 write(*,*) "Problem reading the number of files from decay.inpt"
                 stop
              end if

              ! Allocate array of file names
              allocate(ampfiles(nfiles))
              ! Read in file names
              do i = 1,nfiles
                 read(fopt,'(a100)') ampfiles(i)
              end do
           else if (keyword=="ampfile") then
              if (nfiles .ne. 0) then
                 write(*,*) "Trying to set number of files using multifile option, but it has already been set."
                 stop
              end if

              nfiles = 1
              allocate(ampfiles(nfiles))
              read(line,'(a100)') ampfiles(1)
           end if
        end if
        read (fopt,'(a100)',iostat=ierr) line
     end do
  end if

  ! Has one or more filenames been given? If not use amplitudes.out
  if (nfiles .eq. 0 ) then
     nfiles = 1
     allocate(ampfiles(nfiles))
     ampfiles(1) = "amplitudes.out"
  end if

  ! Calculate the number of steps for padding
  if(pad) then
     nsteps = t0/dt
  end if

  write(fug,*) "Summary"
  write(fug,*) 
  
  if (nparticles > 0) then
     write(fug,*) "Number of nparticles: ",nparticles
     write(fug,*) "Number of cartesian dimensions: ",ncartdim
  end if
  write(fug,*) "Number of degrees of freedom:",ndf
  write(fug,*) "Number of basis functions:",nbasis
  write(fug,*) "Weight",simweight
  write(fug,*) 
  write(fug,*) "Using ",nfiles," amplitude files:"
  do i = 1,nfiles
     write(fug,'("   ",i4,"  ",a100)') i,ampfiles(i)
  end do


  ! Create the wavefunctions, one for t=0 and one for t>0
  CALL createbasisset(wavefn0,ndf,nbasis,zr=.true.)
  CALL createbasisset(wavefnt,ndf,nbasis,zr=.true.)

  ! Set gammas and masses
  wavefn0%gamma(:) = gamma(:)
  wavefnt%gamma(:) = gamma(:)

  
  fauto = 40
  open(fauto,file="autocf.out")
  write(fauto,*) "# Autocorrelation function"
  write(fauto,*) "# Average over ",simweight," steps"
  if (pad) then
     write(fauto,*) "# Intitial time ",t0
  end if


  ! ---- pad dissociation files if required ----
  if(pad) then
     do j = 0,nsteps-1
        write(fauto,*)  j*dt, 0.0
     end do     
  end if


  do file = 1,nfiles
     ! ---- open amplitudes file ----
     open(unit=famp,file=ampfiles(file),status="old",iostat=ierr) 
     if (ierr .ne. 0) then 
        write(*,*) "Failed to open amplitude file ",ampfiles(file)
        stop
     end if
     write(fug,*) "Reading configurations from file ",file,ampfiles(file)

     ! ---- Read in first configuration from first file ----
     if (file .eq. 1) then
        ! Read in first configuration
        call readconfiguration(famp,wavefn0,readconfigerr,fug)
        if (readconfigerr .ne. 0) then
           write(fug,*) "read config err",readconfigerr
        end if
        wavefnt = wavefn0
     else
        call readconfiguration(famp,wavefnt,readconfigerr,fug)
        if (readconfigerr .ne. 0) then
           write(fug,*) "read config err",readconfigerr
        end if
        ! Skip the first configuration and read in the next.
        call readconfiguration(famp,wavefnt,readconfigerr,fug)
        if (readconfigerr .ne. 0) then
           write(fug,*) "read config err",readconfigerr
        end if

     end if
     
     nconfigs = 0

     ! Now read in the second configuration     
     do while (readconfigerr .eq. 0) 
        nconfigs = nconfigs+1
        write(fug,*) "Config ",nconfigs

        ! Calculate C
        if ( .not.wavefn%Cset) then
           CALL calcC(wavefn)
        end if

        prob = (0.0d0,0.0d0)
        do bf1 = 1,wavefn0%nbasis
           do bf2 = 1,wavefn0%nbasis
              prob = prob + dconjg(wavefn0%D(bf1))*wavefnt%D(bf2)*overlap2wvfn (wavefn0,bf1,wavefnt,bf2)
           end do
        end do
        
        write(fauto,*) wavefnt%time+t0,real(conjg(prob)*prob)
        
        ! Now read in the next configuration
        call readconfiguration(famp,wavefnt,readconfigerr,fug)
     end do

  end do

  close(fauto)
  
  write(fug,*) "Finished. Good bye"

  close(fug)

end program autocorrelation
