!!% ========================================================================
!!%  Subroutine: choosebasisfn
!!%
!!%  This routine chooses basis functions for a tri atomic system
!!%  It was written for Br2Ne.
!!%  ========================================================================
subroutine choosebasisfndiatomic()
  use arrays
  use harmonicoscillator, ONLY: pickharmrnd,overlapzho
  use morsesampling
  
  implicit none

  complex(kind=8) :: amps(ndf)
  real(kind=8):: q(ndf),p(ndf)
  real(kind=8) :: xcom,ycom,zcom
  real(kind=8) :: A,B,zmax,magL2
  integer :: dof,i
  real(kind=8),dimension(3) :: L ! Angular momentum
  ! For choosing which term in the expansion to sample
  integer :: term
  real(kind=8) ::pr,pl
  real(kind=8),external :: zbqlu01 ! random number generator
  integer :: bf
  integer:: ntries
  real(kind=8) :: qpdp
  interface
     include 'readbasisfns.inc'
  end interface

  ! check that the system specifications are appropriate
  if ( nparticles.ne.2 ) then
     write(fug,*) "Error, system specification is wrong"
     write(fug,*) " Needs ndf=3, npartilces=2, ncartdim=2 and ndf=3"
     stop
  end if

  ! ---- Get basis fns from files if appropriate ----
  if (nfiles > 0) then
     call readbasisfns(ndf,nbasis,filenames,qf,pf,cswavefnfile,CDfile,nterms)
  end if


  open(unit=222,file="basisfunctions-morsemodes")
  zmax = 50d0

  do bf = 1,nbasis
     write(fug,*) "Choosing basis fn ",bf
    
     dof = 1
     ! Choose which term of the expansion we should use
     if(method(dof) == 1) then
        term = mode(dof)+1 
        
        ! pick position and momentum from the appropriate harmonic oscillator state
        call pickharmrnd(zmax,zb(dof,bf),term-1,comp(dof),amps(dof))
        
        q(dof) =  sqrt(2/xi(dof))*real(zb(dof,bf))+q0(dof)
        p(dof) =  sqrt(2*xi(dof))*dimag(zb(dof,bf))+p0(dof)
     else if (method(dof) == 2) then
        if (nterms(dof) > 1) then
           pr = ZBQLU01(0.0)
           pl = ec(dof,1)**2
           term = 0
           pl = 0.0d0
           do while ((pr  > pl) .and. (term .lt.nterms(dof)))
              term = term+1
              pr = pr + ec(dof,term)**2
           end do
           ! The first term should be the ground state
           write(fug,*) "Choosing from term ",term-1
        else 
           write(fug,*) "Only one term in expansion"
           term = 1
        end if
        
        ! pick position and momentum from the appropriate harmonic oscillator state
        call pickharmrnd(zmax,zb(dof,bf),term-1,comp(dof),amps(dof))
        
        q(dof) =  sqrt(2/xi(dof))*real(zb(dof,bf))+q0(dof)
        p(dof) =  sqrt(2*xi(dof))*dimag(zb(dof,bf))+p0(dof)
        if (nterms(dof) > 1) then
           !             amps(dof) = ec(dof,term)**2 * amps(dof)
           amps(dof) = (0.0d0,0.0d0)
           do i = 1,nterms(dof)
              amps(dof) =  amps(dof)+ec(dof,i)*overlapzho(zb(dof,bf),i)
           end do
        end if
     else if (method(dof)==3) then 
        ! ---- q,p pairs from a file ----
        amps(dof) = (1.0d0,0.0d0)
        q(dof) = 0.0d0
        p(dof) = 0.0d0
        ! calculate the distance between the two particles in position and momentum
        do i = 1,ncartdim
           q(dof) = q(dof) +  (qf((filep1-1)*ncartdim+i,bf) - qf((filep2-1)*ncartdim+i,bf))**2
           p(dof) = p(dof) +  (pf((filep1-1)*ncartdim+i,bf) - pf((filep2-1)*ncartdim+i,bf))**2
        end do
        q(dof) = sqrt(q(dof)) ! This will be halved later
        p(dof) = sqrt(p(dof))/2.0d0 ! The momentum of each particle will be set to +/- p(dof) so divide by 2 here
        !Find sign of momentum
        qpdp = 0.0d0 ! Calculate the dot production between position and momentum vectors.
        do i = 1,ncartdim
           qpdp =  qpdp+ &
                (qf((filep1-1)*ncartdim+i,bf) - qf((filep2-1)*ncartdim+i,bf))&
                *(pf((filep1-1)*ncartdim+i,bf) - pf((filep2-1)*ncartdim+i,bf))
        end do
        if (qpdp < 0) then
           p(dof) = -1.0d0 * p(dof)
        end if

        ! Now randomise the displacement slightly if required
        if (distpts(dof) .eq. 1) then
           call randomisez(q(dof),p(dof),xi(dof),comp(dof),amps(dof))
        end if
        
     else if (method(dof)==4) then 
        ! ---- Choose q,p from the morse wave function ----
        call pickmorse(mode(dof),q(dof),p(dof),amps(dof))
     else
        write(fug,*) "Sorry, method ",method(dof)," is currently unsupported"
        stop
     end if
     
  
     write(fug,*) "q",q(1:nmodes)
     write(fug,*) "p",p(1:nmodes)

     write(222,*) bf, q(1),p(1),q(2),p(2),q(3),p(3)
     ! set all z components to 0.
     z(:,bf) = 0.0d0
     
     x(1,bf) = -0.5d0*q(1)
     y(1,bf) = 0.0d0
     
     x(2,bf) = 0.5d0*q(1)
     y(2,bf) = 0.0d0
     
     write(fug,*) "qs",(q(i),i=1,2)
     
     ! Calculate centre of mass
     xcom= sum(x(:,bf)*mass(:))/totalmass
     ycom= sum(y(:,bf)*mass(:))/totalmass
     zcom = 0.0d0
     
     ! move centre of mass to the origin
     x(:,bf) = x(:,bf) - xcom
     y(:,bf) = y(:,bf) - ycom
     z(:,bf) = z(:,bf) - zcom
     
     write(fug,*) "Positions"
     do i = 1,2
        write(fug,*) i,x(i,bf),y(i,bf),z(i,bf)
     end do
     
     ! Store amplitude
     C(bf) = (1.0d0,0.0d0)
     !do dof = 1,nmodes
     !      write(fug,*) amps(dof)
     C(bf) = C(bf)* amps(dof)
     !end do
     write(fug,*) "Amplitude",C(bf)
     
     ! ********************************
     ! Now we need to calculate the momentum
     px(:,bf) = 0.0d0
     py(:,bf) = 0.0d0
     pz(:,bf) = 0.0d0
     
     ! Due to first mode
     px(1,bf) = -p(1)  ! other components are zero
     px(2,bf) =  p(1)  ! other components are zero
     
     write(fug,*) "Momenta"
     do i = 1,2
        write(fug,*) i,px(i,bf),py(i,bf),pz(i,bf)
     end do
     
     write(fug,*) "Total Momentum"
     write(fug,*) sum(px(:,bf)),sum(py(:,bf)),sum(pz(:,bf))
     
     write(fug,*) "Angular momentum"
     L(:)=0.0d0
     do i = 1,nparticles
        L(1) =  L(1) + y(i,bf)*pz(i,bf) - z(i,bf)*py(i,bf)
        L(2) =  L(2) + z(i,bf)*px(i,bf) - x(i,bf)*pz(i,bf)
        L(3) =  L(3) + x(i,bf)*py(i,bf) - y(i,bf)*px(i,bf)
     end do
     write(fug,*) L(1),L(2),L(3)

     write(fug,*) 
  end do

  close(222)
end subroutine choosebasisfndiatomic


