! ************************************************************************
!  harmonic oscillator module
!
!  This module contains subroutines to choose coherent states from 
!  harmonic oscillator wavefunctions.
!  The main routine to use is pickharmrnd which allows a coherent state
!  to be chosen from any harmonic oscillator wave function using a given
!  compression factor to reduce the spread. The routine also returns
!  the amplitude as <z|n>.
!
!  Also included is the function overlapzho which calultates the 
!  projection (amplitude) of a harmonic oscillator wave function on to 
!  a coherent state.
! ************************************************************************
module harmonicoscillator
contains
  ! ===============================================
  !                pickharmrnd
  !  subroutine to give a random point and its 
  !  probability from the harmonic oscillator 
  !  wavefunctions. 
  !  The point should be drawn from the correct
  !  probability distributions.
  !
  !  Would use the inverse fct so that the point 
  !  was actually drawn from the correct prob. 
  !  distribution but the inverse function is not
  !  simple therefore need to do it this way and
  !  use Monte Carlo sampling.
  ! 
  ! A compression factor may be applied to z so
  ! that z = z/comp. This however only really makes
  ! sense the groundstate wavefct which has its
  ! max at z=0. 
  ! ===============================================
  subroutine pickharmrnd(zmax,z,n,comp,ampC)
    use arrays, ONLY: maxtries,fug
    !  use files, ONLY: fug
    implicit none
    real(kind=8),intent(in) :: zmax
    integer, intent(in) :: n
    ! Compression factor
    real(kind=8), intent(in) :: comp 
    complex(kind=8),intent(out) :: z
    ! amplitude i.e. <z|Phi_n>
    complex(kind=8),intent(out) :: ampC 
    ! vibrational state 0,1,....

    ! Maximum probability
    real(kind = 8) :: pmax,rmax,r0
    integer:: ntries
    logical :: accept
    real(kind=8) :: prob,pr

    integer :: ifact ! for factorial loop
    real(kind=8) :: factn ! n factorial

    real(kind=8),external :: zbqlu01

    write(*,*) "Choosing from harmonic oscillator ",n,"compression",comp
    ! Make sure n is positive
    if( n < 0) then
       write(fug,*) "Sorry but I don't know the wavefunction for energy level v=",n
       stop
    end if

    ! Set maximum probability depending on which mode we are looking at.
    select case(n)
    case(0)
       pmax = 0.1365173622
    case(1)
       !     pmax = 0.1304804041
       !     rmax =1.22474487139159 ! dsqrt(1.5)
       !     r0   = dsqrt(6.0d0)*(1-comp**2)/2.0d0
       rmax = 1.0d0
       r0 = (1-comp**2)

       pmax = rmax**3*exp(-(rmax-r0)**2/comp**2)/3.14159265358979d0
    case default

       !     rmax = dsqrt(dble(n)+0.5)
       rmax = dsqrt(dble(n))
       r0 = rmax*(1-comp**2)
       factn=1.0
       do ifact = 2,n
          factn = factn*ifact
       end do
       pmax = rmax**(2*n+1)*exp(-(rmax-r0)**2/comp**2)/3.14159265358979d0/factn


    end select

    ! Initially we haven't tried to choose a fn nor have
    ! we succeeded
    ntries = 0
    accept = .false.

    !Keep choosing z's at random until the probability of it
    ! is greater than  pr
    do while ( .not.accept .and. ntries .le. maxtries)
       ntries = ntries + 1

       ! Choose z from appropriate wavefunction
       select case(n)
       case(0)
          call harm0rnd(zmax,z,prob)    
          ! compress z
          z = z*comp

       case(1)
          call harm1rnd(zmax,z,comp,prob) 

       case default
          call harmnrnd(zmax,z,n,comp,prob)

       end select

       ! now see whether or not to accept it.
       !call random_number(pr)
       pr = ZBQLU01(0.0)

       if (pr .le. prob/pmax) then
          accept = .true.
       end if
    end do

    ! We have not chosen an acceptable z within the maximum number of tries, 
    ! stop!
    if (.not.accept) then
       write(fug,*) " Failed to choose a satisfactory value for z within ", &
            maxtries, " tries."

       !          maxtries, " tries. Try increasing probscale or decreasing rmax"
       stop
    end if


    ! Now we should calculate the amplitude for this mode/fn
    select case(n)
    case(0)

       ampC = exp(-0.5*(abs(z)**2))
    case(1)
       ampC = dconjg(z)*exp(-0.5*(abs(z)**2))

    case default
       ampC = dconjg(z)**(n)*exp(-0.5*(abs(z)**2))/sqrt(factn)
       !       write(fug,*) "Sorry but I don't know the wavefunction for energy level v=",n
       !     stop
    end select

  end subroutine pickharmrnd

  ! ===============================================
  !                harm0rnd
  !  subroutine to give a random point and its 
  !  probability from |0> of the harmonic 
  !  oscillator. 
  ! ===============================================
  subroutine harm0rnd(rmax,z,prob)

    implicit none
    real(kind=8),intent(in) :: rmax
    complex(kind=8),intent(out) :: z
    real(kind=8),intent(out) :: prob

    real(kind=8) :: r,theta

    real(kind=8),external :: zbqlu01

    ! Choose random numbers.
    !call random_number(r)
    r= ZBQLU01(0.0)
    r = r*rmax

    !call random_number(theta)
    theta= ZBQLU01(0.0)
    theta = theta*3.14159265358979*2.0

    ! Calculate z
    z = r*cos(theta) + r*(0,1)*sin(theta)

    ! Calculate probability
    !   prob = z*dconjg(z)*exp(-z*dconjg(z))/3.14159265358979
    ! Probability in polar coords is 
    ! r*exp(-r^2)
    !prob = cdabs(z)*exp(-cdabs(z)**2)/3.14159265358979
    prob = r*exp(-r**2)/3.14159265358979

  end subroutine harm0rnd

  ! ===============================================
  !                harm1rnd
  !  subroutine to give a random point and its 
  !  probability from |1> of the harmonic 
  !  oscillator. 
  ! ===============================================
  subroutine harm1rnd(rmax,z,sigma,prob)
    implicit none
    real(kind=8),intent(in) :: rmax
    complex(kind=8),intent(out) :: z
    real(kind=8),intent(in) :: sigma
    real(kind=8),intent(out) :: prob

    real(kind=8) :: r,theta,r0

    real(kind=8),external :: zbqlu01

    ! Choose random numbers.
    !  call random_number(r)
    r= ZBQLU01(0.0)
    r = r*rmax

    ! call random_number(theta)
    theta= ZBQLU01(0.0)
    theta = theta*3.14159265358979d0*2.0d0

    ! Calculate z
    z = r*cos(theta) + r*(0,1)*sin(theta)

    ! Calculate probability
    !   prob = z*dconjg(z)*exp(-z*dconjg(z))/3.14159265358979
    ! Probability in polar coords is 
    ! r^3*exp(-r^2)
    ! prob = (cdabs(z)**3)*exp(-cdabs(z)**2)/3.14159265358979

    ! probability, normalised only if sigma = 1
    r0   = (1-sigma**2)
    prob = r**3*exp(-(r-r0)**2/sigma**2)/3.14159265358979

  end subroutine harm1rnd


  ! ===============================================
  !                harmnrnd
  !  subroutine to give a random point and its 
  !  probability from |n> of the harmonic 
  !  oscillator. 
  ! ===============================================
  subroutine harmnrnd(rmax,z,n,sigma,prob)
    !For testing   use coherentstates, ONLY: nbasis

    implicit none
    real(kind=8),intent(in) :: rmax
    complex(kind=8),intent(out) :: z
    integer, intent(in) :: n
    real(kind=8),intent(in) :: sigma
    real(kind=8),intent(out) :: prob

    real(kind=8) :: r,theta,r0
    real(kind=8) :: factn
    integer :: i

    real(kind=8),external :: zbqlu01

    !For testing  integer,save ::nbf=0

    ! Choose random numbers.
    !   call random_number(r)
    r = ZBQLU01(0.0)
    r = r*rmax

    ! For testing choose r = r_maxprob
    !   r =  dsqrt(dble(n)+0.5)

    !   call random_number(theta)
    theta= ZBQLU01(0.0)
    theta = theta*3.14159265358979*2.0

    ! For testing evenly distribution in theta
    !   theta = 3.14159265358979d0*2.0d0/nbasis
    !   nbf = nbf + 1

    ! Calculate z
    z = r*cos(theta) + r*(0.0d0,1.0d0)*sin(theta)

    ! Calculate probability
    !   prob = z*dconjg(z)*exp(-z*dconjg(z))/3.14159265358979
    ! Probability in polar coords is 
    ! r^3*exp(-r^2)
    ! prob = (cdabs(z)**3)*exp(-cdabs(z)**2)/3.14159265358979

    ! probability, normalised only if sigma = 1
    !!r0   = dsqrt(dble(n)+0.5)*(1-sigma**2)
    r0 = (1.0d0-sigma**2)*dsqrt(dble(n))

    ! Calculate n factorial
    factn=1.0d0
    do i = 2,n
       factn = factn*i
    end do

    prob = r**(2*n+1)*exp(-(r-r0)**2/sigma**2)/3.14159265358979/factn

  end subroutine harmnrnd

  ! ************************************************************************
  !  overlapzho
  ! 
  !  function to calculate overlap between a coherent state and a harmonic
  ! oscillator wave function
  !  arguments:   z complex(kind=8) the coherent state
  !               n integer         harmonic oscillator mode
  !  returns: complex amplitude.
  ! ************************************************************************
  function overlapzho(z,n)
    implicit none
    ! ---- interface ----
    complex(kind=8),intent(in) :: z
    integer,intent(in) :: n
    complex(kind=8) :: overlapzho
    ! ---- local variables ----
    real(kind=8) :: factn
    integer:: ifact

    factn=1.0
    do ifact = 2,n
       factn = factn*ifact
    end do

    overlapzho = dconjg(z)**(n)*exp(-0.5*(abs(z)**2))/sqrt(factn)
  end function overlapzho

end module harmonicoscillator
