! ************************************************************************
!                          readbasisfns
!                        12 December 2008
!                          Stewart Reed
!
!  A subroutine to read in basis fns from a file.
!
!  File format
!       line 1: comment
!       subsequent lines q,p pairs
!
!  Arguments
!     inputs:
!       ndf            - number of degrees of freedom
!       nbf            - number of basis fns per degree of freedom
!       filenames(ndf) - filenames for each degree of freedom.
!                        ignore if equal to ''
!     outputs:
!        q(ndf,nbf)    - positions
!        p(ndf,nbf)    - momenta
! ************************************************************************
subroutine readbasisfns(ndf,nbf,filenames,q,p,cswavefnfile,CDfile,nbasisfns)
  use errors
  use arrays, ONLY: fug,filep1,filep2,ncartdim
  implicit none
  ! ---- interface ----
  integer,intent(in) :: ndf ! number of degrees of freedom
  integer,intent(in) :: nbf ! number of pairs to read
  character(len=*),dimension(ndf) :: filenames ! names of files
  real(kind=8),dimension(ndf,nbf),intent(out) :: q,p
  integer,dimension(ndf) :: cswavefnfile
  complex(kind=8),dimension(ndf,ndf) :: CDfile
  integer,dimension(ndf)::nbasisfns
  ! ---- local variables ----
  integer :: fu = 30 ! file unit to read data from
  integer ::df,bf ! for loops
  integer :: ierr
  real(kind=8) :: rec ! real part of C in file
  character(len=100):: line 
  real(kind=8),dimension(:),allocatable :: values
  integer :: i
  ! ---- set the arrays to zero ----
  q(:,:) = 0.0d0
  p(:,:) = 0.0d0
  allocate(values(ncartdim*4))
  ! ---- main loops to read in basis fns ----
  do df = 1,ndf
     if(.not. (filenames(df)) == "" ) then ! .ne. ' '
        open(fu,file=filenames(df),status="old",iostat=ierr)
        ! Check to see if we have opened the file
        ! if not give a message and stop.
        if (ierr .ne.0) then
           call fileerror(ierr,filenames(df)) 
           stop
        end if
        
        ! read in a comment about the file
        read(fu,*) line
        write(fug,*) "Comment: ",line
       
        if (cswavefnfile(df) .eq. 1) then
           do bf = 1,nbasisfns(df)
              read(fu,*,iostat=ierr) q(df,bf),p(df,bf),rec
              CDfile(df,bf) = rec
              ! check to see if we have succeeded, if not stop.
              if (ierr .ne. 0) then
                 write(line,*) "Failed reading basis fn",bf," with amplitude from "//trim(filenames(df))
                 call filereaderror(ierr,"readbasisfns",line)              
              end if
           end do
        else if (cswavefnfile(df) .eq. 2) then
           do bf = 1,nbf
              read(fu,*,iostat=ierr) (values(i),i=1,ncartdim*4)
              ! store these values in the appropriate parts of q and p
              do i = 1,ncartdim
                 q((filep1-1)*ncartdim+i,bf) = values(2*i-1)
                 p((filep1-1)*ncartdim+i,bf) = values(2*i)
                 q((filep2-1)*ncartdim+i,bf) = values(2*i-1+2*ncartdim)
                 p((filep2-1)*ncartdim+i,bf) = values(2*i+2*ncartdim)
              end do
              write(15,*) q(:,bf)
              write(16,*) p(:,bf)
           end do
        else
           do bf = 1,nbf
              read(fu,*,iostat=ierr) q(df,bf),p(df,bf)
              CDfile(df,bf) = 1.0d0
              ! check to see if we have succeeded, if not stop.
              if (ierr .ne. 0) then
                 write(line,*) "Failed reading basis fn",bf," from "//trim(filenames(df))
                 call filereaderror(ierr,"readbasisfns",line)              
              end if
           end do
        end if
        close(fu)
     end if
  end do
end subroutine readbasisfns
