! This subroutine reads the inputs file

subroutine readinputs
  use arrays
  use errors
  use morsesampling
  implicit none
  integer :: ierr ! for file errors
  CHARACTER(len=100) :: line,meth
  integer :: i,j,s
  logical :: havemorse

  open(unit=finputs,file="createbasis.inpt",status='old',iostat=ierr)
  if (ierr .ne. 0 ) then
     write(fug,*) " Problem opening inputs file createbasis.inpt"
     stop
  end if

  ! File format content
  ! -------------------
  ! ncartdim n
  ! nparticles n
  ! nbasis n
  ! gamma
  ! particle g         ! repeated nparticles times
  ! mass
  ! particle m 
  ! ndf
  ! xi
  ! mode xi        ! repeated ndf times
  ! maxterms n
  ! mode method [comp [nterms ]]   method = single,sum or file
  ! coeffs (nterms)  | repeated ndf times
  ! or filename

  !set some defaults
  systemtype = 1 ! Default to normal mode type.
  ncartdim = 2
  nparticles = 3
  nbasis = 1
  ndf = 3
  maxtries = 10000
  nmodes = -1 ! Number of modes/interactions to read in. Really for non-normal mode code
  reproject = .false.
  havemorse = .false. ! Do we choose any modes from a morse wave function?

  minbfprob = -1.0d0 ! 0 should suffice but this indicates also indicates that it has not been set.
  bfsepmode = -1
  bfsep2 = 0.0d0

  bfengrange = 0 ! By default, don't enforce an energy range for the basis functions

  read(finputs,*,iostat=ierr) line
  do while (ierr == 0)   
     if  (line == "dim" ) then
        backspace(finputs) 
        read(finputs,*,iostat=ierr) line,ncartdim
        if(ierr.ne.0) call filereaderror(ierr,"createbasis.inpt","Error reading number of cartesian dimensions")
        write(fug,*) ncartdim," cartesian dimensions"
        
     else if  (line == "nparticles" ) then
        backspace(finputs) 
        read(finputs,*,iostat=ierr) line,nparticles
        if(ierr.ne.0) call filereaderror(ierr,"createbasis.inpt","Error reading number of particles")
        write(fug,*) nparticles," particles"
     else if  (line == "nbasis" ) then
        backspace(finputs) 
        read(finputs,*,iostat=ierr) line,nbasis
        if(ierr.ne.0) call filereaderror(ierr,"createbasis.inpt","Error reading number of basis functions to choose")
        write(fug,*) "choosing ",nbasis," basis functions"
     else if  (line == "ndf" ) then
        backspace(finputs) 
        read(finputs,*,iostat=ierr) line,ndf
        if(ierr.ne.0) call filereaderror(ierr,"createbasis.inpt","Error reading number of degrees of freedom.")
        write(fug,*) ndf,"vibrational modes (degrees of freedom)"
        if (nmodes < 0) then ! set nmodes to ndf if it hasn't already been set
           nmodes = ndf
        end if

     else if (line == "pgamma") then
        if (nparticles .eq. 0) then
           write(fug,*) "You need to set number of particles first"
           stop
        end if
        allocate(gamma(nparticles))
        write(fug,*) "Gammas"
        do i = 1,nparticles
           read(finputs,*,iostat=ierr) s,gamma(i)
           if(ierr.ne.0) then
              write(fug,*) "particle ",i
              call filereaderror(ierr,"createbasis.inpt","Error reading gamma")
           else
              write(fug,*) i,gamma(i)
           end if
        end do
     else if (line == "gamma") then
        if (ndf .eq. 0) then
           write(fug,*) "You need to set the number of degrees of freedom"
           stop
        end if
        allocate(gamma(ndf))
        write(fug,*) "Gammas"
        do i = 1,ndf
           read(finputs,*,iostat=ierr) s,gamma(i)
           if(ierr.ne.0) then
              write(fug,*) "degree of freedom ",i
              call filereaderror(ierr,"createbasis.inpt","Error reading gamma")
           else
              write(fug,*) i,gamma(i)
           end if
        end do

     else if (line == "mass") then
        if (nparticles .eq. 0) then
           write(fug,*) "You need to set number of particles first"
           stop
        end if
        allocate(mass(nparticles))
        write(fug,*) "Mass"
        totalmass = 0.0d0
        do i = 1,nparticles
           read(finputs,*,iostat=ierr) s,mass(i)
           if(ierr.ne.0) then
              write(fug,*) "particle ",i
              call filereaderror(ierr,"createbasis.inpt","Error reading mass")
           else
              write(fug,*) i,mass(i)
              totalmass = totalmass + mass(i)
           end if
        end do
     else if (line == "xi") then
        if (ndf .eq. 0) then
           write(fug,*) "You need to set ndf first"
           stop
        end if
        allocate(xi(ndf))
        write(fug,*) "xi for wave fn expansion"
        do i = 1,ndf
           read(finputs,*,iostat=ierr) s,xi(i)
           if(ierr.ne.0) then
              write(fug,*) "dof ",i
              call filereaderror(ierr,"createbasis.inpt","Error reading xi")
           else
              write(fug,*) i,xi(i)
           end if
        end do
     else if (line == "mxi") then
        if (ndf .eq. 0) then
           write(fug,*) "You need to set ndf first"
           stop
        end if
        allocate(xi(ndf))
        write(fug,*) "xi for mode wave fn expansion"
        do i = 1,nmodes
           read(finputs,*,iostat=ierr) s,xi(i)
           if(ierr.ne.0) then
              write(fug,*) "Mode ",i," of ",nmodes
              backspace(finputs)
              read(finputs,'(a100)') line
              write(*,*) "Read: ",line
              call filereaderror(ierr,"createbasis.inpt","Error reading mxi")
           else
              write(fug,*) i,xi(i)
           end if
        end do
     else if (line == "displace" ) then
        if (ndf .eq. 0) then
           write(fug,*) "You need to set ndf first"
           stop
        end if
        displace=.true.
        write(fug,*) "Diplacing origin of each mode to "
        allocate(qb0(ndf),pb0(ndf))
        do i = 1,ndf
           read(finputs,*,iostat=ierr) s,qb0(i),pb0(i)
           if(ierr.ne.0) then
              write(fug,*) "dof ",i
              call filereaderror(ierr,"createbasis.inpt","Error reading xi")
           else
              write(fug,*) i,qb0(i),pb0(i)
           end if
        end do
        ! ------ Set system types ----        
     else if (line == "normalmodes" .or. line == "normal") then
        systemtype = 1
     else if (line == "planartriatomic") then
        systemtype =2 
        nmodes = 3         
        write(fug,*) "Creating a basis for a planar triatomic system"
        write(fug,*) nmodes," vibrational modes"
     else if (line == "tetraatomic") then
        systemtype =3 
        nmodes = 6 
        write(fug,*) "Creating a basis for a tetra-atomic system"
        write(fug,*) nmodes," vibrational modes"
     else if (line == "pentaatomic") then
        systemtype =5
        nmodes = 10 
        write(fug,*) "Creating a basis for a penta-atomic system"
        write(fug,*) nmodes," vibrational modes"
     else if (line == "diatomic") then
        systemtype =4
        nmodes = 1 
        write(fug,*) "Creating a basis for a diatomic system"
        write(fug,*) nmodes," vibrational modes"
        
        
        ! ------ read in the stuff about each mode ----
     else if  (line == "maxnterms" ) then
        backspace(finputs) 
        read(finputs,*,iostat=ierr) line,maxnterms
        if(ierr.ne.0) call filereaderror(ierr,"createbasis.inpt","Error reading max number of terms for wave function expansion.")

        allocate(ec(ndf,maxnterms))
        allocate(nterms(ndf))
        allocate(comp(ndf))
        allocate(q0(ndf))
        q0(:)=0.0d0
        allocate(p0(ndf))
        p0(:)=0.0d0
        allocate(method(ndf))
        allocate(mode(ndf))
        ! For reading modes from files.
        allocate(filenames(ndf))
        filenames(:)=""
        allocate(cswavefnfile(ndf))
        cswavefnfile(:) = 0
        allocate(distpts(ndf))
        distpts(:) = 0

        nfiles = 0 ! number of modes from files
        do i = 1,nmodes
           read(finputs,*) s,meth
           write(*,*) "Method ",meth
           if (meth == "single") then 
              method(i) = 1
              nterms(i) = 0
              ec(i,:) = 0.0d0
              backspace(finputs)
              ! Try to read in q0 and p0 if it fails, read just q0.
              read(finputs,'(a100)') line
              read(line,*,iostat=ierr) s,meth,mode(i),comp(i),q0(i),p0(i) !re-read line including 
              if(ierr .ne. 0) then
                 read(line,*,iostat=ierr) s,meth,mode(i),comp(i),q0(i)
                 p0(i) = 0.0d0
              end if
              comp(i) = 1.0d0/comp(i) 
              write(fug,*) "Mode ",mode(i),"comp",comp(i),"q0",q0(i)              
           else if (meth =="sum")then
              method(i) = 2
              backspace(finputs)
              read(finputs,*) s,comp(i),q0(i),nterms(i)
              comp(i) = 1.0d0/comp(i) 
              read(finputs,*) (ec(i,j),j=1,nterms(i))

           else if (meth=="file") then
              method(i) = 3
              nfiles = nfiles + 1
              ! Check for additional options
              backspace(finputs)
              read(finputs,"(a80)",iostat=ierr) line
              if (ierr .eq. 0) then
                 read(line,*,iostat=ierr) s,meth,meth
                 if (meth=="cswavefn") then
                    write(fug,*) "cswavefn"
                    cswavefnfile(i) = 1
                    read(line,*,iostat=ierr) s,meth,meth,meth
                    if (ierr .eq. 0) then
                       if (meth=="randomise") then
                          write(fug,*) "Randomising positions too"
                          distpts(i) = 1
                          ! Check to see how many basis functions are to be given in file
                          read(line,*,iostat=ierr) s,meth,meth,meth,nterms(i)
                          if (ierr.ne.0) then
                             nterms(i) = nbasis
                             if (mod(nbasis,nterms(i)) .ne. 0) then
                                write(fug,*) &
                                     "Number of basis functions is not an integer multiple of number of points in file ",&
                                     filenames(i)
                             end if
                             write(*,*) "Reading ",nterms(i)," from file"
                          end if
                       end if
                    end if
                 else if (meth=="pair") then
                    if (nfiles > 1) then
                       write(*,*) "Sorry but currently you can have only one file for pair interactions"
                       stop
                    end if
                    read(line,*,iostat=ierr) s,meth,meth,filep1,filep2
                    cswavefnfile(i) = 2                    

                    ! allow extra options on the next line
                    read(finputs,'(a80)',iostat=ierr) line
                    read(line,*) meth
                    if (meth == "displace") then
                       distpts(i) = 1 
                       ! find out compression
                       read(line,*) meth, comp(i)
                       comp(i) = 1.0d0/comp(i)
                    else
                       backspace(finputs)
                    end if
                 end if ! if meth = cswavefn or pair
              end if
              read(finputs,"(a80)"),filenames(i)
              write(fug,*) "Reading basis functions for mode",i," from ",filenames(i)

           else if (meth== "morse") then
              method(i) = 4
              havemorse = .true.
              ! ---- choose basis functions from the Morse wave function ----
              backspace(finputs)
              read(finputs,*,iostat=ierr) s,meth,mode(i)
              
           else
              write(fug,*) "Sorry, but I don't know how to produce basis functions using method ",meth
           end if
        end do
     else if (line =="randseed" ) then
        backspace(finputs)
        read(finputs,*,iostat=ierr) line,randseed
     else if (line == "reproject" ) then
        reproject=.true.
        write(fug,*) "Reprojecting wavefunctions on basis with different gamma"
        
     else if (line == "minbfprob") then ! Enforce a minimum probability for basis functions
        backspace(finputs)
        read(finputs,*,iostat=ierr) line,minbfprob
        if (ierr.ne. 0) then
           write(*,*)  "Error reading total probability restriction. Error ",ierr
           stop
        end if
        
     else if (line == "minsep") then  ! Enforce a minimum separation
         backspace(finputs)
         read(finputs,*,iostat=ierr) line,bfsepmode, bfsep2         
         if (ierr.ne. 0) then
            write(*,*)  "Error reading restriction on basis separation. Error ",ierr
            stop
         end if
         ! Square separation
         bfsep2 = bfsep2**2
      else if (line == "engrange") then ! Enforce an energy range for basis functions
         backspace(finputs) 
         read(finputs,*,iostat=ierr) line,minbfeng,maxbfeng
        if (ierr.ne. 0) then
           write(*,*)  "Error reading energy range restriction. Error ",ierr
           stop
        end if
        bfengrange = 1
     else
        write(fug,*) "Unrecognised options",line
     end if


     ! Read the next line!
     read(finputs,*,iostat=ierr) line
  end do

  ! Close the file
  close(finputs)

  ! allocate arrays to store basis fns.
  allocate(zb(ndf,nbasis))
  if(nparticles > 0 ) then
     allocate(x(nparticles,nbasis))
     allocate(y(nparticles,nbasis))
     allocate(z(nparticles,nbasis))
     allocate(px(nparticles,nbasis))
     allocate(py(nparticles,nbasis))
     allocate(pz(nparticles,nbasis))
  else
     allocate(qb(ndf,nbasis))
     allocate(pb(ndf,nbasis))
  end if
  allocate(C(nbasis))

  ! Arrays for reading data in from files.
  if (nfiles > 0) then
     allocate(qf(ndf,nbasis))
     allocate(pf(ndf,nbasis))
     allocate(order(ndf,nbasis))
     allocate(CDfile(ndf,nbasis))
  end if

  ! If we have to do morse mode sampling, read in the morse hamiltonian and other inputs
  if (havemorse) then
     call readhamiltonian("morseham.inpt")
  end if
  write(fug,*)  "Finished reading options"
end subroutine readinputs

