!!% ========================================================================
!!% Subroutine: calcbrengdistnoneons
!!% Author: Stewart Reed
!!% Date: 10/11/2009
!!% Description: This subroutine now contains the code used to calculate 
!!%              the Br2 Energy Distributions for systems that have no Neons
!!% ========================================================================
subroutine calcbrengdistnoneons
  use basismodule 
  use system
  use br2nehamiltonian
  use histogram
  use morsesampling

  implicit none

  ! ---- local variables ----
  integer :: file,i,j,k,bf,dvl ! Loop indices
  
  integer :: ierr ! file error
  integer :: readconfigerr

  ! Histograms 
  type(histogram1D) :: modeenergyprofile
  type(histogram1D) :: br2modeenergyprofile,brNemodeenergyprofile
  type(histogram1D) :: br2quantisedenergy
  type(histogram1D) :: projection
  ! Internal coordinate basis set
  type(basisset) Br2Wfn
  real(kind=8),dimension(:),allocatable :: rBrBr,pBrBr
  real(kind=8) :: dp ! dotproduct rBrBr,pBrBr
  real(kind=8) :: prob ! Probability of a basis function
  real(kind=8) :: x
  ! Energies
  real(kind=8) :: modeenergy

  integer :: nconfigs
  real(kind=8) :: dvlprob
  ! For plotting wave function
  integer :: dof 
  integer :: ndx
  character(len=20) :: charconfig ! For appending configuration number to file name
  complex(kind=8) :: px

  ! ---- Create histogram ----
  call createhistogram(br2modeenergyprofile,histmin,histmax,nbins)

  if (quantise) then
     call createhistogram(br2quantisedenergy,dble(vmin)-0.5d0,dble(vmax)+0.5d0,vmax-vmin+1)
     call createbasisset(Br2Wfn,1,nbasis,zr=.true.)
     Br2Wfn%gamma(1) = gammaBr/2.0d0
     Br2Wfn%mass(1) = massBr/2.0d0
     
     xmin = 1.5
     xmax = 4.0
     ndx = 500
     dx = (xmax-xmin)/dble(ndx)
     
     call createhistogram(projection,xmin-0.5d0*dx,xmin+(ndx+0.5d0)*dx,ndx+1)
  end if

  allocate(rBrBr(ncartdim),pBrBr(ncartdim))

    ! 
  nconfigs = 0

  write(fug,*) "Reading configurations for Br2 Only"

  allfiles: do file = 1,nfiles
     ! ---- open amplitudes file ----
     open(unit=famp,file=ampfiles(file),iostat=ierr) 
     if (ierr .ne. 0) then 
        write(fug,*) "Failed to open amplitude file ",ampfiles(file)
        stop
     end if
     write(fug,*) "File ",file,trim(ampfiles(file))
     ! Read in first configuration in this file
     call readconfiguration(famp,wavefn,readconfigerr)
     if (readconfigerr .ne. 0) then
        write(fug,*) "read config err",readconfigerr
     end if

     ! Keep count of the total number of basis functions to have dissociated.
     configloop: do while (readconfigerr .eq. 0  .and. wavefn%time < tmax ) 
        nconfigs = nconfigs+1
        write(fug,*) "Config ",nconfigs

        
        ! If we haven't yet got to the minimum time, move onto next basis fn.
        if ( wavefn%time < tmin) then
           ! Read in the next configuration
           call readconfiguration(famp,wavefn,readconfigerr)     
           cycle configloop
        end if

        ! Calculate C
        if(.not.wavefn%cset) then
           CALL calcC(wavefn)
        end if
        
        bfloop: do bf = 1,wavefn%nbasis
           ! Calculate the energy of this basis function.           
           modeenergy = brmodehexpect(wavefn,bf)*eu
           ! Calculate probability
           prob = real(dconjg(wavefn%C(bf))*wavefn%D(bf))
           ! Add to histogram
           call addtohistogram(br2modeenergyprofile,modeenergy,prob)

           if (quantise) then
              ! Calculate separation vector
              rBrBr(1:ncartdim) = wavefn%cs(bf)%q(1:ncartdim) - wavefn%cs(bf)%q(ncartdim+1:2*ncartdim)
              ! Calculate momentum - factor of 2 takes account of reduced mass for Br2 mode 
              pBrBr(1:ncartdim) = (wavefn%cs(bf)%p(1:ncartdim) - wavefn%cs(bf)%p(ncartdim+1:2*ncartdim))/2.0d0
              
              ! Determine sign of momentum
              dp = sum(rBrBr(1:ncartdim)*pBrBr(1:ncartdim))           
              if (dp > 0) then
                 Br2Wfn%cs(bf)%p(1) = sqrt(sum(pBrBr(1:ncartdim)**2))
              else
                 Br2Wfn%cs(bf)%p(1) = -sqrt(sum(pBrBr(1:ncartdim)**2))
              end if
              ! Store the separation
              Br2Wfn%cs(bf)%q(1) = sqrt(sum(rBrBr(1:ncartdim)**2))
              
           end if
        end do bfloop
        
        if (quantise) then
           ! ---- Calculate z components ----
           call calczfromqp(Br2Wfn)
               
           
           ! Loop over requested vibrational levels.
           do dvl = vmin,vmax
              
              dvlprob = 0.0d0
              do i = 1,Br2Wfn%nbasis
                 ! Project each basis function onto
                 ! Actually, it is the conjg of C that we need,
                 Br2Wfn%C(i) = projectcsmorsedvib(1,Br2Wfn%cs(i)%q(1),Br2Wfn%cs(i)%p(1),dvl)
                 
                 ! Now add the diagonal contribution.
                 dvlprob = dvlprob + real(conjg(wavefn%D(i))*wavefn%D(i) * Br2Wfn%C(i)*conjg(Br2Wfn%C(i)))
                 
              end do
              do i = 1,Br2Wfn%nbasis
                 do j = i+1,Br2Wfn%nbasis
                    dvlprob = dvlprob + 2*real(conjg(wavefn%D(i))*wavefn%D(j) * Br2Wfn%C(i)*conjg(Br2Wfn%C(j)))
                 end do
              end do
              
              ! Add to histogram                        
              call addtohistogram(br2quantisedenergy,dble(dvl),dvlprob)  
           end do

           ! ---- Plot wave fn ----
           xmin = 1.5
           xmax = 4.0
           ndx = 500
           dx = (xmax-xmin)/dble(ndx)
           
           dof = 1
           
           write(fug,*) "Projecting wave function"
           
           write(charconfig,"(i8)") nconfigs
           do i = 0,ndx
              x = dble(i) * dx + xmin
              
              px = 0.0d0
              do j = 1,nbasis
                 px = px + (abs(wavefn%D(j)* project_xz(x,Br2wfn,j,dof))) **2
                 do k = j+1,nbasis
                    
                    px =  px  &
                         + 2*real(conjg(wavefn%D(j)*project_xz(x,Br2wfn,j,dof))*wavefn%D(k)*project_xz(x,Br2wfn,k,dof) )
                 end do
              end do
              call addtohistogram(projection,x,real(px)) 
           end do

           if (timedep) then
              outfile="Br2wfnprojection4"//trim(adjustl(charconfig))//trim(filesuffix)
              call normalisehistogram(projection)
              call writehistogram(projection,outfile,"Wave function averaged over all configurations")
              call emptyhistogram(projection)
           end if
        end if

        ! ---- write out time dependent distributions ----
        if (timedep) then
            call normalisehistogram(br2modeenergyprofile)
            write(charconfig,"(i8)") nconfigs
            outfile="Br2ModeEnergyMatrix"//trim(adjustl(charconfig))//trim(filesuffix)
            call writehistogram(br2modeenergyprofile,outfile,"Distribution of Br2 mode energies by matrix inversion")

            call emptyhistogram(br2modeenergyprofile)
            if (quantise) then
               ! Now normalise
               call normalisehistogram(br2quantisedenergy)
               br2quantisedenergy%v(:) = br2quantisedenergy%v(:)
               outfile="Br2ModeEnergyQuantised"//trim(adjustl(charconfig))//trim(filesuffix)
               call writehistogram(br2quantisedenergy,outfile,"Distribution of quantised Br2 mode energies")
               call emptyhistogram(br2quantisedenergy)
            end if
        end if
        
        ! Read in the next configuration
        call readconfiguration(famp,wavefn,readconfigerr)     
     end do configloop
     if ( wavefn%time > tmax) exit allfiles
  end do allfiles

   if (.not.timedep) then
      call normalisehistogram(br2modeenergyprofile)
      write(charconfig,"(i8)") nconfigs
      outfile="Br2ModeEnergyMatrix"//trim(filesuffix)
      call writehistogram(br2modeenergyprofile,outfile,"Distribution of Br2 mode energies by matrix inversion")
      
      if (quantise) then
         ! Now normalise
         call normalisehistogram(br2quantisedenergy)
         br2quantisedenergy%v(:) = br2quantisedenergy%v(:)
         outfile="Br2ModeEnergyQuantised"//trim(filesuffix)
         call writehistogram(br2quantisedenergy,outfile,"Distribution of quantised Br2 mode energies")


         call normalisehistogram(projection)
         outfile="Br2wfnprojection"//trim(filesuffix)
         call writehistogram(projection,outfile,"Wave function averaged over all configurations")
      end if
   end if

end subroutine calcbrengdistnoneons
  
