!!$ ************************************************************************
!!$  Module: hamiltonian
!!$  Author: S.K. Reed
!!$  Date:   17th April 2009
!!$
!!$  A module with routines to calculate energy for basis sets and functions
!!$ ************************************************************************
module br2nehamiltonian
  implicit none


  ! ---- module data ----
  integer :: ncartdim ! number of cartesian dimensions
  integer :: nparticles ! number of particles

  ! zero point energy for each potential
  real(kind=8),dimension(:),allocatable ::modekinzpe 
  ! zero point energy for each molecule = sum(modezpe)
  real(kind=8) moleculekinzpe

  ! ---- Cartesian hamiltonian parameters ----
  integer :: cart_nints ! number of interactions
  integer :: cart_maxsgterms
  ! number of terms for each interaction
  integer,dimension(:),allocatable :: cart_ntint
  ! amplitudes for terms in gaussian sum  i,j - j=interaction; i = sum term
  real(kind=8),dimension(:,:),allocatable :: cart_gsamp 
  ! sigmas for terms in gaussian sum
  real(kind=8),dimension(:,:),allocatable :: cart_gstwosigmasq,cart_gssigma
  ! which particles are interacting particle 1 and particle 2
  integer,dimension(:),allocatable :: cart_part1,cart_part2
  real(kind=8),dimension(:),allocatable :: cart_gammamodes
contains

  !!% ========================================================================
  !!%  Function: brmodehexpect
  !!% 
  !!%  Function to calculate the energy expectation value for a particular
  !!%  basis function (basis) of a basis set (bs)
  !!% 
  !!% ========================================================================
  function brmodehexpect(bs,basis)
    use basismodule
    implicit none  
    ! ---- interface ----
    type(basisset),intent(in) :: bs
    integer,intent(in) :: basis
    real(kind=8) :: brmodehexpect
    ! ---- local ----
    real(kind=8) :: A,alpha
    real(kind=8),dimension(ncartdim) :: vBrSep ! Br-Br separation
    real(kind=8) :: BrSep2 
    real(kind=8),dimension(ncartdim) :: vBrNeSep ! Br-Ne separation
    real(kind=8) :: BrNeSep2 
    real(kind=8),dimension(ncartdim) :: vBrMom ! Br-Br momentum
    real(kind=8) :: BrMom 
    integer :: dof
    real(kind=8) ::redmass ! reduced mas
    real(kind=8) :: modeT  ! mode kinetic energy
!!$    real(kind=8) :: a11,a12,a21,a22 ! For matrix inversion

    ! ---- calculate the Br-Br separation vector ----
    vBrSep(1:ncartdim) = bs%cs(basis)%q(1:ncartdim) - bs%cs(basis)%q(ncartdim+1:2*ncartdim)
    ! ---- calculate the square of the scalar separation ----
    BrSep2 = (sum(vBrSep(:)*vBrSep(:)))

    ! ---- calculate the Br-Br momentum ----
    vBrMom(1:ncartdim) = (bs%mass(ncartdim+1)*bs%cs(basis)%p(1:ncartdim)  &
                             - bs%mass(1)*bs%cs(basis)%p(ncartdim+1:2*ncartdim))&
                           /(bs%mass(1)+bs%mass(ncartdim+1))

    write(20,*) basis, vBrMom(1:ncartdim),sqrt(sum( vBrMom(1:ncartdim)**2))
    ! redmass = 1.0d0 /( 1/bs%mass(1) + 1/bs%mass(ncartdim+1))
    redmass = bs%mass(1)/2.0d0
    !debug    write(22,*) bs%time, sum(vBrMom(1:ncartdim)*vBrSep(1:ncartdim)) /sqrt(BrSep2)

    ! Now calculate what proportion is along along the separation vector
    modeT =   (sum(vBrMom(1:ncartdim) *  vBrSep(1:ncartdim) ))**2/(BrSep2*2.0d0*redmass)

    ! Add kinetic zero point energy and vibrational kinetic energy for the Br-Br vibration
    brmodehexpect = (modeT + modekinzpe(1))

    ! Now add potential part
    brmodehexpect = brmodehexpect + hamcartgausssum(bs,basis,basis,1)
    !debug    
    write(21,*) bs%time,brmodehexpect-( modeT + modekinzpe(1)),  modeT + modekinzpe(1)
  end function brmodehexpect


  !!% ========================================================================
  !!%  Function: brrothexpect
  !!% 
  !!%  Function to calculate the Br2 rotational energy expectation value for
  !!%  a particular basis function (basis) of a basis set (bs)
  !!% 
  !!% ========================================================================
  function brrothexpect(bs,basis)
    use basismodule
    implicit none  
    ! ---- interface ----
    type(basisset),intent(in) :: bs
    integer,intent(in) :: basis
    real(kind=8) :: brrothexpect
    ! ---- local ----
    real(kind=8),dimension(ncartdim) :: vBrSep ! Br-Br separation
    real(kind=8) :: BrSep2 
    real(kind=8),dimension(ncartdim) :: vBrMom ! Br-Br momentum
    real(kind=8) :: BrMom 
    integer :: dof
    real(kind=8) ::redmassrec ! reduced mass
    real(kind=8) :: modeT  ! mode kinetic energy

    ! ---- calculate the Br-Br separation vector ----
    vBrSep(1:ncartdim) = bs%cs(basis)%q(1:ncartdim) - bs%cs(basis)%q(ncartdim+1:2*ncartdim)
    ! ---- calculate the square of the scalar separation ----
    BrSep2 = (sum(vBrSep(:)*vBrSep(:)))

    ! ---- calculate the Br-Br momentum ----0.000
    vBrMom(1:ncartdim) = bs%cs(basis)%p(1:ncartdim) - bs%cs(basis)%p(ncartdim+1:2*ncartdim)

    ! redmass = 1.0d0 /( 1/bs%mass(1) + 1/bs%mass(ncartdim+1))
    redmassrec = 2.0/bs%mass(1) 

    ! Now calculate what proportion is along along the separation vector
    modeT =  0.5*redmassrec * (sum(vBrMom(1:ncartdim) *  vBrSep(1:ncartdim) ))**2/BrSep2

    ! Rotational kinetic energy is difference between full kinetic energy
    ! and mode kinetic energy
    brrothexpect = sum(vBrMom(:)*vBrMom(:))*0.5*redmassrec - modeT

    ! There is no zero point energy and no potential part.

  end function brrothexpect


  !!% ========================================================================
  !!%  Function: modeshexpect2D
  !!%  Description: A function to calculate the `bond momenta' and mode 
  !!%               expectation valuses for a 3 particle system. Assumes the 
  !!%               molecule is in the x-y plane.
  !!% ========================================================================
  subroutine modeshexpect2D(bs,basis,modehexpect,pbonds)
    use basismodule
    use vectorutils, ONLY: solve2D
    implicit none
    type(basisset),intent(in) :: bs
    integer,intent(in) :: basis
    real(kind=8),dimension(3),intent(out) :: modehexpect
    real(kind=8),dimension(3),intent(out) :: pbonds ! momenta of the 3 bonds
    

    ! ---- local ----
    real(kind=8),dimension(2) :: pb ! for calculating the bond momentas
    real(kind=8),dimension(2,2) :: A ! matrix
    
    real(kind=8) :: r12,r13,r23 ! Bond lengths

    integer :: q1, q2, q3 ! positions in the array of the x coordinates of each particle
    integer :: i
    real(kind=8),dimension(3) :: redmass ! reduced masses

    q1 = 1
    q2 = ncartdim + 1
    q3 = 2*ncartdim + 1

    ! Calculate the bond lengths
    r12 = sqrt(sum( (bs%cs(basis)%q(1:ncartdim) - bs%cs(basis)%q(ncartdim+1:2*ncartdim))**2))
    r13 = sqrt(sum( (bs%cs(basis)%q(1:ncartdim) - bs%cs(basis)%q(2*ncartdim+1:3*ncartdim))**2))
    r23 = sqrt(sum( (bs%cs(basis)%q(ncartdim+1:2*ncartdim) - bs%cs(basis)%q(2*ncartdim+1:3*ncartdim))**2))

    q1 = 1
    q2 = ncartdim + 1
    q3 = 2*ncartdim + 1

    ! ---- First set ----
    A(1,1) = (bs%cs(basis)%q(q2) - bs%cs(basis)%q(q1)) / r12
    A(1,2) = (bs%cs(basis)%q(q3) - bs%cs(basis)%q(q1)) / r13

    A(2,1) = (bs%cs(basis)%q(q2+1) - bs%cs(basis)%q(q1+1)) / r12
    A(2,2) = (bs%cs(basis)%q(q3+1) - bs%cs(basis)%q(q1+1)) / r13

    !(A (-pa,pb) = p1
    !now calculate pa and pc
    call solve2D(A,bs%cs(basis)%p(1:2), pb)

    ! Transfer to the pbonds array.
    pbonds(1) = -pb(1)
    pbonds(2) = pb(2)

    ! ---- Now the second set ----
    A(1,1) = (bs%cs(basis)%q(q2) - bs%cs(basis)%q(q1)) / r12
    A(1,2) = (bs%cs(basis)%q(q2) - bs%cs(basis)%q(q3)) / r23

    A(2,1) = (bs%cs(basis)%q(q2+1) - bs%cs(basis)%q(q1+1)) / r12
    A(2,2) = (bs%cs(basis)%q(q2+1) - bs%cs(basis)%q(q3+1)) / r23

    !now calculate pa and pc
    call solve2D(A,bs%cs(basis)%p(q2:q2+1), pb)

    ! Transfer
    pbonds(3) = pb(2)
    ! Check that the two give consistent results
    if(abs(pbonds(1)-pb(1)) > 1e-5) then
       write(*,*) "Calculated Pa's are not equal",pbonds(1),pb(1)
    end if

    ! ---- 3rd and final set ----
    ! (Only useful for checking - leave out unless we get the above condition becoming true.

    A(1,1) = (bs%cs(basis)%q(q3)-bs%cs(basis)%q(q1))/r13
    A(1,2) = (bs%cs(basis)%q(q2)-bs%cs(basis)%q(q3))/r23

    A(2,1) = (bs%cs(basis)%q(q3+1)-bs%cs(basis)%q(q1+1))/r13
    A(2,2) = (bs%cs(basis)%q(q2+1)-bs%cs(basis)%q(q3+1))/r23

    !now calculate pa and pc
    call solve2D(A,bs%cs(basis)%p(q3:q3+1), pb)
    
    if(abs(pbonds(2)+pb(1)) > 1e-5) then
       write(*,*) "Calculated Pb's are not equal",pbonds(2),-pb(1)
    end if

    if(abs(pbonds(3)+pb(2)) > 1e-5) then
       write(*,*) "Calculated Pc's are not equal",pbonds(3),-pb(2)
    end if

    ! But we want the bond energyies not the momenta's
    
    ! Calculate the reduced masses for the 3 modes
    redmass(1) =  (bs%mass(1) * bs%mass(ncartdim+1))/  (bs%mass(1) + bs%mass(ncartdim+1))
    redmass(2) =  (bs%mass(ncartdim+1) * bs%mass(2*ncartdim+1))/  (bs%mass(ncartdim+1) + bs%mass(2*ncartdim+1))
    redmass(3) =  (bs%mass(1) * bs%mass(2*ncartdim+1))/  (bs%mass(1) + bs%mass(2*ncartdim+1))

    do i = 1,3
       ! Calculate the kinetic and K. zero point  contributions
       modehexpect(i)  = pbonds(i)**2/(2.0d0*redmass(i)) + modekinzpe(i) + hamcartgausssum(bs,basis,basis,i)
       

       write(22+i,*) bs%time,modehexpect(i)-( pbonds(i)**2/(2.0d0*redmass(i)) + modekinzpe(i)),  &
            pbonds(i)**2/(2.0d0*redmass(i))+ modekinzpe(i)

    end do

    write(30,*) basis,r12,pbonds(1),r13,pbonds(2),r23,pbonds(3)
  end subroutine modeshexpect2D

  !!% ========================================================================
  !!%  Function: hamcartgausssum
  !!% 
  !!%  Function to calculate the energy expectation of a sum of gaussian 
  !!%  potentials. Takes a 
  !!%  From main simulation program
  !!% ========================================================================
  function hamcartgausssum(bs,i,j,pot)
    use basismodule
    implicit none
    ! ---- interface ----
    type(basisset),intent(in) :: bs
    integer, intent(in) :: i,j ! which two basis fns to use.
    integer, intent(in) :: pot ! which interaction are we considering
    real(kind=8) :: hamcartgausssum

    ! ---- local ----
    integer :: g,cdof ! for loops
    integer :: alpha,beta
    real(kind=8) :: Vmorse

    hamcartgausssum = 0.0d0
    do g = 1, cart_ntint(pot)
       ! ---- loop of cartesian dimensions ----
       Vmorse = cart_gsamp(g,pot)
       do cdof = 1,ncartdim
          ! determine global indicies
          alpha = ncartdim*cart_part1(pot)+cdof-ncartdim
          beta  = ncartdim*cart_part2(pot)+cdof-ncartdim

          Vmorse =  Vmorse &
               * hamcartgauss(dconjg(bs%cs(i)%z(alpha)),dconjg(bs%cs(i)%z(beta)), bs%cs(j)%z(alpha),bs%cs(j)%z(beta)&
               ,cart_gstwosigmasq(g,pot),bs%gamma(alpha),bs%gamma(beta))
       end do
       hamcartgausssum=hamcartgausssum + Vmorse !-cart_gsamp(g,i)
    end do

  end function hamcartgausssum

  !!% ========================================================================
  !!%  Function: hamcartgauss
  !!% 
  !!%  Function to calculate the energy expectation of a gaussian potential
  !!%  From main simulation program
  !!% ========================================================================
  function hamcartgauss(z1ac,z1bc,z2a,z2b,twosigmasq,gammap1,gammap2)
    implicit none
    ! ---- interface ----
    complex(kind=8),intent(in) :: z1ac,z1bc,z2a,z2b
    real(kind=8),intent(in) :: twosigmasq,gammap1,gammap2
    complex(kind=8) :: hamcartgauss ! the return type

    ! ---- local ----
    real(kind=8) :: sgammap1,sgammap2

    ! Calculate the sqrt of gammas
    sgammap1 = sqrt(gammap1)
    sgammap2 = sqrt(gammap2)

    !prefactor
    hamcartgauss = dsqrt(twosigmasq /(twosigmasq+1/gammap1+1/gammap2)) &
         *exp(-0.5d0*(sgammap2*(z1ac+z2a)-sgammap1*(z1bc+z2b))**2 &
         /(gammap1*gammap2*twosigmasq+gammap1+gammap2))

  end function hamcartgauss



  ! ==================================================
  !          Read params for cartesian system
  ! ==================================================
  subroutine readhamcart(bs)
    use basismodule
    use system, ONLY : eu,energyunits,fug,fham
    use errors
    !  use options, ONLY: ncartdim,nparticles,verbosity,eu,writeexpect,energyunits
    !  use files, ONLY: fug,fham
    !  use errors
    !  use arrays, ONLY: dhexpect,hework

    implicit none

    ! ---- interface ----
    type(basisset),intent(inout) :: bs

    ! ---- local ----
    integer :: i,j ! For loops
    character(len=50) :: line
    integer :: ierr ! errors from reading stuff from file.
    real(kind=8) :: m1,m2,redmass ! Masses for calculating the reduced mass

    ! ---- Read in masses -----
    read(fham,*,iostat=ierr) line
    if  (line .ne. "mass" .or. ierr .ne. 0  ) then
       write(fug,*) "Failed to read keyword mass in hamiltonian.inpt"
       stop
    end if
    do i =1,nparticles
       ! each degree of freedom has a separate mass. 
       read(fham,*,iostat=ierr) bs%mass(ncartdim*i)
       if(ierr.ne.0) call filereaderror(ierr,"hamiltonian.inpt","Error reading masses")
       ! Set the masses for the other degrees of freedom of this particle.
       do j = 1,ncartdim-1
          bs%mass(ncartdim*i-j) = bs%mass(ncartdim*i)
       end do
    end do

    ! ---- now read in potential ----
    read(fham,*,iostat=ierr) line 
    if (line == "interactions") then
       backspace(fham) 
       read(fham,*,iostat=ierr) line, cart_nints

    else
       call filereaderror(ierr,"hamiltonian.inpt","Can't find any interactions")
    end if

    ! now look for max number of terms in sum of gaussians
    read(fham,*,iostat=ierr) line 
    if (line == "maxterms") then
       backspace(fham) 
       read(fham,*,iostat=ierr) line, cart_maxsgterms
       if (ierr .ne. 0) then
          call filereaderror(ierr,"hamiltonian.inpt","The max number of terms in the sum of gaussians has not been specified.")
       else
          write(fug,*) "Max number of terms in gaussian sum is",cart_maxsgterms
       end if
    end if

    ! now allocate arrays
    ! number of terms in the sums for each mode
    allocate(cart_ntint(cart_nints))
    cart_ntint = 0
    allocate(cart_gsamp(cart_maxsgterms,cart_nints))
    cart_gsamp(:,:) = 0.0d0
    allocate(cart_gssigma(cart_maxsgterms,cart_nints))
    allocate(cart_gstwosigmasq(cart_maxsgterms,cart_nints))
    allocate(cart_part1(cart_nints))
    allocate(cart_part2(cart_nints))
    allocate(cart_gammamodes(cart_nints))
    allocate(modekinzpe(cart_nints))
    cart_gssigma(:,:) = 0.0d0
    cart_gstwosigmasq(:,:) = 0.0d0
    moleculekinzpe = 0.0d0 ! Kinetic zero point energy
    modekinzpe(:) = 0.0d0

    ! read in the interactions
    do i = 1,cart_nints
       ! read in type and which particles it is between.
       ! line should be gsum but don't check for now.
       ! gsum, particle 1, particle 2, num terms in sum.
       read(fham,*,iostat=ierr) line,cart_part1(i),cart_part2(i),cart_ntint(i)
       if (cart_ntint(i) > cart_maxsgterms) then
          write(fug,*) "Interaction ",i," has more terms than the specified maximum"
          write(fug,*) "Please increase the maximum in hamiltonian.inpt and try again"
          stop
       end if
       !     if (verbosity .ge. 3 .and. isroot) then
       !        write(fug,*) line,cart_part1(i),cart_part2(i),cart_ntint(i)
       !     end if
       do j = 1,cart_ntint(i)
          read(fham,*,iostat=ierr) cart_gsamp(j,i),cart_gssigma(j,i)
          cart_gstwosigmasq(j,i) = 2.0d0*cart_gssigma(j,i)**2
          ! convert energy amplitude to internal energy units
          cart_gsamp(j,i) =  cart_gsamp(j,i)/eu
       end do

       ! Calculate gamma and zero point energy for this mode.
       ! hbar^2*gamma/4m

       cart_gammamodes(i) = 1.0d0/(1.0d0/bs%gamma((cart_part1(i)-1)*ncartdim+1)+1.0d0/bs%gamma((cart_part2(i)-1)*ncartdim+1))
       write(fug,*) "Gamma Modes", cart_gammamodes(i) ,bs%gamma((cart_part1(i)-1)*ncartdim+1),bs%gamma((cart_part2(i)-1)*ncartdim+1)
       m1 = bs%mass((cart_part1(i)-1)*ncartdim+1)
       m2 = bs%mass((cart_part2(i)-1)*ncartdim+1)
       redmass = m1*m2/(m1+m2)
       !debug write(fug,*) "Reduced mass",redmass
       !debug

       modekinzpe(i) = 0.25*cart_gammamodes(i)/redmass
    end do

    moleculekinzpe = sum( modekinzpe(:))

    ! ---- Now we want to summarise ----
    write(fug,*) "Cartesian Hamiltonian Parameters"
    write(fug,*) "================================"

    write(fug,*) " Each interaction between pairs of particles is expressed as a sum of gaussians centred at the origin"

    write(fug,*) "  Masses:               i     m_i"
    do i = 1,ncartdim*nparticles
       write(fug,"(a21,i5,f8.4)") "                     ",i,bs%mass(i)
    end do
    write(fug,*)

    do i =  1,cart_nints
       !     write(fug,"(3a,i4,a3,i4)") "   ",cart_part1(i)," - ",cart_part2(i)
       write(fug,*) "   ",cart_part1(i)," - ",cart_part2(i)
       do j = 1,cart_ntint(i)
          write(fug,*) "                ", cart_gsamp(j,i)*eu, cart_gssigma(j,i)
       end do
       write(fug,*) "                Kinetic zero point energy ", modekinzpe(i)*eu,energyunits
       write(fug,*) "                Mode gamma",cart_gammamodes(i)," 1/Angstrom^2"
    end do
    write(fug,*)
    write(fug,*)  "                Total Kinetic zero point energy per basis fct ",moleculekinzpe*eu,energyunits
  end subroutine readhamcart

end module br2nehamiltonian
