!!% ************************************************************************
!!%                          basis set module
!!%                             S. K. Reed
!!%                           6th April 2009
!!%
!!%  A simple module to read in an amplitude file configuration by 
!!%  configuration.
!!% 
!!%  This module defines two types: a coherent state and a basisset. The 
!!%  coherent state is ndimensional and can have a q,p representation, a 
!!%  z-representation or both.
!!%  The basisset type defines an array of coherent states and amplitudes D 
!!%  and C. It also defines the number of dimensions (degrees of freedom)
!!%  and the number of basis functions. Further, it contains the system time,
!!%  gammas and masses. 
!!%  A basis set is created using a type definition statement
!!%  type(basisset) :: mybasisset
!!%  and then initialised with
!!%  CALL CREATEBASISSET(mybasisset, system_dimension, number_of_basisfns)
!!%
!!%  Configurations may be read in from an amplitudes.out file opened at 
!!%  file unit famp using
!!%  CALL READCONFIGURATION(famp,mybasis,readsuccess)
!!%  readsuccess will be 0 if a configuration has been successfully read in.
!!%  If not, it will contain the error number generatedby the read
!!%  statements.
!!% 
!!%  For convenience, a basis set called wavefn is defined but needs to be 
!!%  intialised as above.
!!%
!!%  gammas and masses are initialised to 1.
!!%  There is a gamma and mass associated with each degree of freedom 
!!%  rather than each particle. 
!!%
!!%  To calculate C properly it is necessary to have appropriate gammas. 
!!%  Given a set of Ds, use
!!%  calcC(mybasis) 
!!%  to calculate Cs.
!!% 
!!%  To access the number of basis functions in mybasis use
!!%   mybasis%nbasis
!!%  To access coherent state j use
!!%   mybasis%cs(j)
!!%  To access the position in the kth degree of freedom of the jth coherent
!!%  state use
!!%   mybasis%cs(j)%q(k)
!!%
!!% ************************************************************************
module basismodule
  implicit none
  ! qp representation of a coherent state
  type coherentstate
     logical :: qprep
     logical :: zrep
     real(kind=8),dimension(:),allocatable :: q,p
     complex(kind=8),dimension(:),allocatable :: z
  end type coherentstate

  type basisset
     integer :: ndf 
     integer :: nbasis
     logical :: qprep,zrep
     real(kind=8) :: time
     complex(kind=8),dimension(:),allocatable :: D, C ! amplitudes
     logical :: Cset                                  ! has Cexp(iS/hbar) been calculated?
     type(coherentstate),dimension(:),allocatable :: cs
     real(kind=8),dimension(:),allocatable :: gamma
     real(kind=8),dimension(:),allocatable :: mass
  end type basisset
     
  type(basisset) :: wavefn
  real(kind=8) :: pi = 3.14159265358979d0
  real :: ampfileversion = 1.0 ! Amplitude file type.
contains
  !!% ============================================================
  !!%  Subroutine: createbasisset
  !!%
  !!%  Takes three arguments:
  !!%    bs  - basisset
  !!%    dim - dimension of system = number of degrees of freedom
  !!%    nbf - number of basis functions
  !!% ============================================================
  subroutine createbasisset(bs,dim,nbf,zr)
    implicit none
    type(basisset),intent(inout) :: bs
    integer,intent(in) :: dim,nbf
    logical,intent(in),optional :: zr ! include z representation

    integer :: i
    ! set the system dimensionality and number of basis functions
    bs%ndf = dim
    bs%nbasis = nbf

    ! Allocate array of coherent states
    allocate(bs%cs(nbf))

    ! Choose the representation
    bs%qprep = .true.
    if (present(zr)) then
       if(zr) then
          bs%zrep = zr
       end if
    else
       bs%zrep = .false.
    end if

    ! Allocate each coherent state
    do i = 1,nbf
       call createcoherentstate(bs%cs(i),dim,bs%qprep,bs%zrep)
    end do

    ! Allocate amplitudes
    allocate(bs%C(nbf),bs%D(nbf))
    bs%Cset = .false.

    ! Allocate gammas
    allocate(bs%gamma(dim))
    bs%gamma(:) = 1.0d0

    ! Allocate mass
    allocate(bs%mass(dim))
    bs%mass(:) = 1.0d0

  end subroutine createbasisset
  
  !!% ============================================================
  !!%  Subroutine: createcoherentstate
  !!%
  !!%  Routine to allocate a coherent state
  !!%  Requires
  !!%         cs    - a coherent state
  !!%         dim   - integer - number of degrees of freedom
  !!%         qprep - logical - is the pos-momentum representation
  !!%                           required?
  !!%         zrep  - logical - is the complex-z representation
  !!%                           required?
  !!%         
  !!% ============================================================

  subroutine createcoherentstate(cs,dim,qprep,zrep)
    implicit none
    type(coherentstate), intent(inout) :: cs
    integer, intent(in) :: dim
    logical, intent(in) :: qprep,zrep
    
    cs%qprep = qprep
    cs%zrep = zrep
    
    if (qprep) then
       allocate (cs%q(dim),cs%p(dim))
    end if

    if (zrep) then
       allocate(cs%z(dim))
    end if
    
    ! Catch both false
    if(.not. qprep .and. .not. zrep) then
       write(*,*) "Failed to allocate coherent state. Need to choose at least one representation, either qp or z"
       stop
    end if
  end subroutine createcoherentstate


  !!% ============================================================
  !!%  Subroutine: readconfiguration
  !!%               read next configuration from file
  !!% 
  !!% Requires
  !!%        famp - integer - a file unit as an amplitude
  !!%        bs - a basis set type
  !!%        readsucess -integer - output - indicates the succes 
  !!%                              or otherwise of the operation
  !!%        fu - optional file unit for output.
  !!% ============================================================
  subroutine readconfiguration(famp,bs,readsuccess,fu)
    implicit none
    integer, intent(in) :: famp
    type(basisset),intent(inout) :: bs
    integer, intent(out) :: readsuccess
    integer, intent(in),optional ::fu
    ! ---- local ----
    integer :: ierr,ierr2
    integer :: i,j,df
    character(len=100) :: line,line2
    real(kind=8),dimension(:),allocatable :: qps
    real(kind=8) :: ReAmp,ImAmp,ReAmpC,ImAmpC
    integer:: fuo
    if (present(fu)) then
       fuo = fu;
    else
       fuo=6;
    end if

    ! Look for the next step
    line2 = ""
    ierr = 0
    do while ((.not. (line2=="Step")) .and. ierr .eq. 0)
       read(famp,'(a80)',iostat=ierr) line
       read(line,*,iostat=ierr2) line2,line2

       if (line2 == "Version" ) then
          read(line,*,iostat=ierr2) line2,line2,ampfileversion
          if (ierr2 .ne. 0) then
             write(fuo,*) "Problem reading amplitudes file version"
          end if
          write(fuo,*) "Ampfile version ",ampfileversion
       end if
       !        write(*,*) line2
    end do

    ! have either found a line whose second word is Step
    ! first should be #
    ! or we have reached the end of the file
    if (ierr .eq. 0) then
       ! Allocate array for reading in qps
       if(.not. allocated(qps)) allocate(qps(2*bs%ndf))
       bs%Cset = .false.
       setbasisfn: do j = 1,bs%nbasis
          ! Now read in the positions and momenta
          if (ampfileversion .ge. 2.0) then
             read(famp,*,iostat=ierr) bs%time,ReAmp,ImAmp,ReAmpC,ImAmpC, (qps(df), df = 1,2*bs%ndf)
          else
             read(famp,*,iostat=ierr) bs%time,ReAmp,ImAmp, (qps(df), df = 1,2*bs%ndf)
          end if
          if (ierr .ne. 0 ) exit setbasisfn
          ! Calculate Amplitude
          !          bs%D(j) = complex(ReAmp,ImAmp)
          bs%D(j) = ReAmp+(0.0d0,1.0d0)*ImAmp
          if (ampfileversion .ge. 2.0) then
             bs%C(j) = ReAmpC+(0.0d0,1.0d0)*ImAmpC
             Bs%Cset = .true.
          end if

          ! Store positions and momenta
          if (bs%qprep) then
             do df = 1,bs%ndf
                bs%cs(j)%q(df) = qps(2*df-1)
                bs%cs(j)%p(df) = qps(2*df)
             end do
          end if
          if (bs%zrep) then
!!!!             do df = 1,bs%ndf
!!!!                bs%cs(j)%z(df) = complex((qps(2*df-1)+qps(2*df))/(sqrt(2.0d0*bs%gamma(df))),)
!!!!             end do 
          end if
       end do setbasisfn

       !       if (ierr .ne. 0) then
       readsuccess = ierr
       !       end if
    else
       readsuccess = ierr
       bs%Cset = .false.
    end if

    ! Calculate the z-representation from the q,p representation
    if (readsuccess .eq. 0) then
       if (bs%zrep) then
          call calczfromqp(bs)
       end if
    end if
  end subroutine readconfiguration
  
!!% ============================================================
!!%  Subroutine: calczfromqp
!!% 
!!%  Subroutine to calculate z's
!!% ============================================================
  subroutine calczfromqp(bs)
    implicit none
    type(basisset), intent(inout) :: bs
    
    ! ---- local ----
    integer :: i,df ! for loops
    
    if (.not. bs%zrep) then
       write(*,*) "This basis set doesn't contain a z-representation"
       stop
    end if
    do i = 1,bs%nbasis ! loop over basis funcitons
       do df = 1,bs%ndf ! loop over degrees of freedom
          call zfromqp(bs%cs(i)%q(df),bs%cs(i)%p(df),bs%cs(i)%z(df),bs%gamma(df))
       end do
    end do
  end subroutine calczfromqp

!!% ============================================================
!!%  Subroutine: zfromqp
!!% 
!!%  Subroutine to calculate z given a q,p pair and a gamma
!!% ============================================================
  subroutine zfromqp(q,p,z,gamma)
    implicit none
    real(kind=8),intent(in) ::q,p
    complex(kind=8),intent(out) :: z
    real(kind=8),intent(in) :: gamma

!    z = complex(sqrt(gamma/2.0d0)*q,p/sqrt(2.0d0*gamma))
    z = sqrt(gamma/2.0d0)*q+(0.0d0,1.0d0)*p/sqrt(2.0d0*gamma)

  end subroutine zfromqp
  

!!% ============================================================
!!%  Subroutine: qpfromz
!!% 
!!%  Subroutine to calculate a q,p pair from z and a gamma
!!% ============================================================
  subroutine qpfromz(z,q,p,gamma)
    implicit none
    complex(kind=8),intent(in) :: z
    real(kind=8),intent(out) ::q,p
    real(kind=8),intent(in) :: gamma
    
    q = sqrt(2.0d0/gamma)*real(z)
    p = sqrt(2.0d0*gamma)*dimag(z)

  end subroutine qpfromz
  
  !!% ============================================================
  !!%  Function: overlap
  !!% 
  !!%  function to calculate the overlap between two basis fcts.
  !!%  Arguments 
  !!%          - bs - a basis set
  !!%          - i,j - integer
  !!%          - returns a complex number
  !!% ============================================================
  function overlap (bs,i,j)
    implicit none
    type(basisset), intent(in) :: bs
    integer,intent(in) :: i,j

    complex(kind=8) :: overlap
    ! ---- local ----
    complex(kind=8) :: exponent
    complex(kind=8) :: zi,zj
    integer :: df ! degree of freedom

!    exponent = complex(0.0d0,0.0d0)
    exponent = (0.0d0,0.0d0)
    if (bs%zrep) then 
       do df = 1,bs%ndf
          zi = bs%cs(i)%z(df)
          zj = bs%cs(j)%z(df)
          exponent = exponent + conjg(zi)*zj - 0.5d0*(conjg(zi)*zi + conjg(zj)*zj)
       end do
       overlap = cdexp(exponent)
    else
       do df = 1,bs%ndf
          zi = bs%cs(i)%q(df)*sqrt(0.5d0*bs%gamma(df)) +  (0.0d0,1.0d0) * bs%cs(i)%p(df)/sqrt(2.0d0*bs%gamma(df))
          zj = bs%cs(j)%q(df)*sqrt(0.5d0*bs%gamma(df)) +  (0.0d0,1.0d0) * bs%cs(j)%p(df)/sqrt(2.0d0*bs%gamma(df))

          exponent = exponent + conjg(zi)*zj - 0.5d0*(conjg(zi)*zi  + conjg(zj)*zj)
       end do
       overlap = cdexp(exponent)
    end if
  end function overlap

  !!% ============================================================
  !!%  Function: overlap1dof
  !!% 
  !!%  function to calculate the overlap between two basis fcts.
  !!%  but for one degree of freedom only.
  !!%  Arguments 
  !!%          - bs - a basis set
  !!%          - i,j - integer
  !!%          - df - integer
  !!%          - returns a complex number
  !!% ============================================================
  function overlap1dof (bs,i,j,df)
    implicit none
    type(basisset), intent(in) :: bs
    integer,intent(in) :: i,j
    integer,intent(in) :: df ! degree of freedom

    complex(kind=8) :: overlap1dof
    ! ---- local ----
    complex(kind=8) :: exponent
    complex(kind=8) :: zi,zj


!    exponent = complex(0.0d0,0.0d0)
    exponent = (0.0d0,0.0d0)
    if (bs%zrep) then 
       zi = bs%cs(i)%z(df)
       zj = bs%cs(j)%z(df)
       exponent = exponent + conjg(zi)*zj - 0.5d0*(conjg(zi)*zi + conjg(zj)*zj)
       overlap1dof = cdexp(exponent)
    else
       zi = bs%cs(i)%q(df)*sqrt(0.5d0*bs%gamma(df)) +  (0.0d0,1.0d0) * bs%cs(i)%p(df)/sqrt(2.0d0*bs%gamma(df))
       zj = bs%cs(j)%q(df)*sqrt(0.5d0*bs%gamma(df)) +  (0.0d0,1.0d0) * bs%cs(j)%p(df)/sqrt(2.0d0*bs%gamma(df))
       
       exponent = exponent + conjg(zi)*zj - 0.5d0*(conjg(zi)*zi  + conjg(zj)*zj)
       overlap1dof = cdexp(exponent)
    end if
  end function overlap1dof

  !!% ============================================================
  !!%  Function: overlap2wvfn
  !!% 
  !!%  function to calculate the overlap between two basis fcts.
  !!%  from two different basis sets.
  !!%  Arguments 
  !!%          - bs1 - a basis set
  !!%          - i - integer
  !!%          - bs2 - a basis set
  !!%          - j - integer
  !!%          - returns a complex number
  !!% ============================================================
  function overlap2wvfn (bs1,i,bs2,j)
    implicit none
    type(basisset), intent(in) :: bs1,bs2
    integer,intent(in) :: i,j

    complex(kind=8) :: overlap2wvfn
    ! ---- local ----
    complex(kind=8) :: exponent
    complex(kind=8) :: zi,zj
    integer :: df ! degree of freedom

    ! Check that the two basis sets have the same number of degrees of
    ! freedom
    if(bs1%ndf .ne. bs2%ndf) then
       write(*,*) "Basis sets 1 and 2 have different numbers of degrees of freedom."
       write(*,*) "Panic. Bye Bye!"
       stop
    end if

!    exponent = complex(0.0d0,0.0d0)
    exponent = (0.0d0,0.0d0)
    if (bs1%zrep .and. bs2%zrep) then 
       do df = 1,bs1%ndf
          zi = bs1%cs(i)%z(df)
          zj = bs2%cs(j)%z(df)
          exponent = exponent + conjg(zi)*zj - 0.5d0*conjg(zi)*zi  - 0.5d0*conjg(zj)*zj
       end do
       overlap2wvfn = cdexp(exponent)
    else
       do df = 1,bs1%ndf
          zi = bs1%cs(i)%q(df)*sqrt(0.5d0*bs1%gamma(df)) +  (0.0d0,1.0d0) * bs1%cs(i)%p(df)/sqrt(2.0d0*bs1%gamma(df))
          zj = bs2%cs(j)%q(df)*sqrt(0.5d0*bs2%gamma(df)) +  (0.0d0,1.0d0) * bs2%cs(j)%p(df)/sqrt(2.0d0*bs2%gamma(df))

          exponent = exponent + conjg(zi)*zj - 0.5d0*conjg(zi)*zi  - 0.5d0*conjg(zj)*zj
       end do
       overlap2wvfn = cdexp(exponent)
    end if
  end function overlap2wvfn

  !!% ============================================================
  !!%  Subroutine: project_xz
  !!%
  !!%  subroutine to project the coherent state onto the real axis
  !!%  Requires
  !!%        x   - 
  !!%        bs  - a basis set type
  !!%        bf  - which basis function
  !!%        dof - which degree of freedom
  !!% ============================================================
   function project_xz(x,bs,bf,dof)
   implicit none
   real(kind=8) :: x 
   type(basisset),intent(in) :: bs ! the basis set
   integer, intent(in) :: bf ! which basis funciton
   integer, intent(in) :: dof ! which degree of freedom
   complex(kind=8) :: project_xz

   project_xz = (bs%gamma(dof)/pi)**0.25d0 &
        * exp( -0.5d0*bs%gamma(dof)*(x-bs%cs(bf)%q(dof))**2 &
        + (0.0d0,1.0d0)*bs%cs(bf)%p(dof)*(x-bs%cs(bf)%q(dof)) &
        + 0.5d0*(0.0d0,1.0d0)*bs%cs(bf)%p(dof) * bs%cs(bf)%q(dof) )

 end function project_xz
  !!% ============================================================
  !!%  Subroutine: calcC
  !!%
  !!%  subroutine to calculate Cexp(iS/hbar)
  !!%          C = overlapmatrix D
  !!%  Requires
  !!%        bs - a basis set type
  !!% ============================================================
  subroutine calcC(bs)
    implicit none
    type(basisset),intent(inout) :: bs
    integer :: bf  ! basisfn
    integer :: dof ! degree of freedom
    ! ---- local ----
    integer :: i,j 
    do i =1 ,bs%nbasis
       bs%C(i) = (0.0d0,0.0d0)
       do j = 1,bs%nbasis
          bs%C(i) = bs%C(i) + overlap(bs,i,j)*bs%D(j)          
       end do
    end do

    ! set the flag to say C has been calculated
    bs%Cset = .true.
  end subroutine calcC

  !!% ============================================================
  !!%  Subroutine: calcC
  !!%
  !!%  subroutine to calculate Cexp(iS/hbar)
  !!%          C = overlapmatrix D
  !!%  using Lapack
  !!%
  !!%  Requires
  !!%        bs - a basis set type
  !!%             the z's of which are correct.
  !!% ============================================================
  subroutine calcD(bs)
    implicit none
    type(basisset),intent(inout) :: bs
    
    complex(kind=8),dimension(:,:),allocatable :: overlapmat

    real(kind=8),dimension(:),allocatable :: FERR,BERR,RWORK
    integer,dimension(:),allocatable :: ipiv
    ! nbasis x nbasis
    complex(kind=8),dimension(:,:),allocatable::overlpmf
    
    ! 2*nbasis
    complex(kind=8),dimension(:),allocatable:: WORK
    

    integer :: LWORK ! = 2*nbasis = size(WORK)
    integer :: LRWORK ! size of RWORK
    integer :: INFO
    real(kind=8) :: rcond
    integer :: i,j
    ! The z's need to be set before hand

    ! ---- Calculate the overlap matrix ----
    allocate(overlapmat(bs%nbasis,bs%nbasis))
        
    overlapmat(:,:) = 0.0d0
    do i = 1,bs%nbasis
       overlapmat(i,i) = 1.0d0
    end do

    do i = 1,bs%nbasis-1
       do j = i,bs%nbasis
          overlapmat(i,j) = overlap (bs,i,j)
          if(abs(overlapmat(i,j) ) .lt. 1e-10) then
             overlapmat(i,j) = (0.0d0,0.0d0)
          end if
          overlapmat(j,i) = conjg(overlapmat(i,j))
       end do
    end do

    allocate(FERR(bs%nbasis),BERR(bs%nbasis),RWORK(bs%nbasis))
    allocate(WORK(2*bs%nbasis))
    LWORK = size(WORK)
    LRWORK=  size(RWORK)
    allocate(ipiv(bs%nbasis))
    allocate(overlpmf(bs%nbasis,bs%nbasis))
        
    ! Call solver for Hermitian system
    call zhesvx('N','L',bs%nbasis, 1, overlapmat, bs%nbasis, overlpmf, bs%nbasis, &
         IPIV, bs%C, bs%nbasis, bs%D, bs%nbasis, RCOND, FERR, BERR, WORK, LWORK, &
         RWORK, INFO )
       
#ifdef DEBUG
    write(*,*) "Calculating D from C"
    write(*,*) "RCOND",RCOND
    write(*,*) "INFO ",INFO, " Should be 0"
#endif

    deallocate(overlapmat)
    deallocate(overlpmf)
    deallocate(FERR,BERR)
    deallocate(WORK)
    deallocate(IPIV)
    
  end subroutine calcD

end module basismodule
