!!% ************************************************************************
!!%  Module: histogram
!!%  Author: Stewart Reed
!!%  Date: 24th April 2009
!!%
!!%
!!% ************************************************************************
module histogram
  implicit none

  type histogram1D
     real(kind=8) :: min,max
     integer(kind=8) :: nbins
     real(kind=8) :: binwidth
     real(kind=8),dimension(:),allocatable :: v      
     integer,dimension(:),allocatable :: vcount
     real(kind=8) ::lowest,highest
  end type histogram1D

!  type histogram2D
     
!  end type histogram2D

contains

  !!% ========================================================================
  !!%  Subroutine createhistogram
  !!% 
  !!%  Subroutine to create a 1D histogram.
  !!%  The histogram first needs to be declared using
  !!%  type(histogram1D) :: histogramname
  !!%
  !!%  This subroutine then determines the bin width, stores the range of the
  !!%  histogram and allocates the memory.
  !!% ========================================================================
  subroutine createhistogram(hist,min,max,nbins)
    implicit none
    ! ---- interface ----
    type(histogram1D),intent(inout) :: hist
    real(kind=8),intent(in) :: min,max
    integer,intent(in) :: nbins
    ! ---- do something ----
    ! store parameters in the histogram
    hist%min= min
    hist%max= max
    hist%nbins = nbins
    ! determine binwidth
    hist%binwidth = (max-min)/dble(nbins)
    
    ! Create the histogram with an extra bin on either end
    ! to catch values which are out of range
    allocate(hist%v(0:nbins+1))
    allocate(hist%vcount(0:nbins+1))

    ! Set to zero
    hist%v(:) = 0.0d0
    hist%vcount(:) = 0

    ! Set initial lowest and highest values
    hist%lowest = 1D15
    hist%highest = -1D15
  end subroutine createhistogram
  
  !!% ========================================================================
  !!%  Subroutine: addtohistogram
  !!%  Subroutine to reset a histogram 
  !!% ========================================================================
  subroutine emptyhistogram(hist)
    implicit none
    ! ---- interface ----
    type(histogram1D),intent(inout) :: hist
    ! Set to zero
    hist%v(:) = 0.0d0
    hist%vcount(:) = 0

    ! Set initial lowest and highest values
    hist%lowest = 1D15
    hist%highest = -1D15
  end subroutine emptyhistogram

  !!% ========================================================================
  !!%  Subroutine: addtohistogram
  !!%  
  !!%  Subroutine to add values to a histogram
  !!%  The subroutine takes 3 arguments: hist, pos and value
  !!%    hist  -  1D histogram, 
  !!%    pos   -  position (bin) within histogram to add to
  !!%    value -  optionally how much to add to the bin. If not provide the
  !!%             default is to add 1.
  !!% ========================================================================

  subroutine addtohistogram(hist,pos,value)
    implicit none
    ! ---- interface ----
    type(histogram1D),intent(inout) :: hist
    real(kind=8),intent(in) :: pos
    real(kind=8),intent(in),optional ::  value
    ! ---- local ----
    integer :: bin

    ! ---- Determine bin to add to ----
    if (pos < hist%min) then 
       bin = 0
    else if (pos > hist%max) then
       bin = hist%nbins+1
    else
       bin = floor((pos-hist%min)/hist%binwidth)+1
    end if
    
    ! Keep note of the highest and lowest values
    if (pos < hist%lowest) then
       hist%lowest = pos
    end if
    if (pos > hist%highest) then
       hist%highest = pos
    end if

    ! ---- Add value to bin if provided, other wise add 1 ----
    if(present(value)) then
       hist%v(bin) = hist%v(bin)+value
       hist%vcount(bin) = hist%vcount(bin)+1
    else
       hist%v(bin) = hist%v(bin)+1.0d0
       hist%vcount(bin) = hist%vcount(bin)+1
    end if
  end subroutine addtohistogram

  !!% ========================================================================
  !!%  Subroutine: normalisehistogram
  !!%
  !!%  Subroutine to normalise the histogram so that it's area is 1
  !!%  Values outwith the range of the histogram are ignored.
  !!%  
  !!% ========================================================================
  subroutine normalisehistogram(hist)
    implicit none
    ! ---- interface ----
    type(histogram1D),intent(inout) :: hist
    ! ---- local ----
    real(kind=8) :: total

    ! ---- calculate the total area using simple integration---
    total = sum(hist%v(1:hist%nbins))*hist%binwidth

    ! ---- divide each bin by this total ----
    if (abs(total) > 0) then
       hist%v(1:hist%nbins) = hist%v(1:hist%nbins)/total
    end if

  end subroutine normalisehistogram


  !!% ========================================================================
  !!%  Subroutine: histogrambinmean
  !!%
  !!%  This subroutine calcules the mean value of each bin.
  !!%  Requires a 1D histogram
  !!% ========================================================================
  subroutine histogrambinmean(hist)
    implicit none
    ! ---- interface ----
    type(histogram1D),intent(inout) :: hist
    
    ! ---- local ----
    integer :: i

    ! ---- calculate mean of each bin -----
    do i = 1,hist%nbins
       if( hist%vcount(i) > 0) then
          hist%v(i) = hist%v(i)/dble(hist%vcount(i))
       end if
    end do

  end subroutine histogrambinmean

  !!% ========================================================================
  !!%  Subroutine: writehistogram
  !!%
  !!%  This subroutine writes the histogram out to a file.
  !!%  Arguments: hist, filename, fileunit
  !!%     hist      - a 1D histogram
  !!%     filename  - a character string containing the name of a file to
  !!%                 write to.
  !!%     fileunit  - optional, fileunit to use when writing the file
  !!%                 The default value is 81
  !!% ========================================================================
  subroutine writehistogram(hist,filename,comment,fileunit)
    implicit none
    ! ---- interface ----
    type(histogram1D), intent(in)        :: hist
    character(len=*),intent(in)          :: filename
    character(len=*),intent(in),optional :: comment
    integer,intent(in),optional          :: fileunit

    ! ---- local ----
    integer   :: fu   ! file unit to use
    integer   :: ierr ! for file error
    integer   :: i    ! for the loop

    ! ---- determine file unit to use ----
    if (present(fileunit) ) then
       fu = fileunit
    else
       fu = 81
    end if

    open(unit=fu,file=trim(filename),iostat=ierr)
    if (ierr .ne. 0) then
       write(*,*) "Failed to open file to write histogram to"
    else
       write(fu,*) "# "//trim(comment)
       write(fu,*) "# lowest value",hist%lowest
       write(fu,*) "# highest value",hist%highest
       write(fu,*) "# Values that are too low"
       write(fu,"(22a)") "# Average over 1 steps - for densityav.pl program."
       write(fu,'(" #",2es18.8,i8)') hist%min-0.5*hist%binwidth,hist%v(0),hist%vcount(0)
       write(fu,*) "# pos,   value,     count"
       do i = 1,hist%nbins
          write(fu,'(2es18.8,i8)') (i-0.5d0)*hist%binwidth+hist%min,hist%v(i),hist%vcount(i)
       end do
       write(fu,*) "# Values that are too high"
       write(fu,'(" #",2es18.8,i8)') hist%max+0.5*hist%binwidth,hist%v(hist%nbins+1),hist%vcount(hist%nbins+1)

       close(fu)
             
    end if
  end subroutine writehistogram

end module histogram
