!!% ************************************************************************
!!%  Module: morsesampling
!!%  Author: Stewart Reed
!!%  Email: S.K.Reed@physics.org
!!%  Date: 3rd July 2009
!!%
!!%  Description: This modules provides routines for random sampling of
!!%               points from a morse vibrational mode. The module reads
!!%               in the parameters of the potentials from a file.
!!%  Public routines: readhamiltonian, projectcsmorse, projectcsmorsedvib,
!!%                   pickmorse
!!%                  readhamiltonian - read morse parameters from a file
!!%                                    includes parameters for sampling the
!!%                                    mode
!!%                  projectcsmorse - projects a point q,p in phases space
!!%                                   on to the wave function for 
!!%                                   interaction, mode. Vibrational level
!!%                                   comes from the hamiltonian file
!!%                  projectcsmorsedvib - as projectcsmorse but allows the
!!%                                    the vibrational level to be changed 
!!%                                    by dvib
!!%                  pickmorse - choose a phase space point from the morse 
!!%                              wave function as specified in the 
!!%                              hamiltonian file
!!% 
!!% ************************************************************************

module morsesampling
  implicit none

  private ! keep most stuff private to avoid conflicts

  integer :: nmm ! Number of morse modes

  ! ---- The morse potentials ----
  real(kind=8),dimension(:),allocatable :: morseA
  real(kind=8),dimension(:),allocatable :: morseAlpha
  real(kind=8),dimension(:),allocatable :: morseR0
  real(kind=8),dimension(:),allocatable :: morsegamma 
  real(kind=8),dimension(:),allocatable :: morsemass ! Reduced mass 
  ! Range over which to sample
  real(kind=8),dimension(:),allocatable :: qmin,qmax
  real(kind=8),dimension(:),allocatable :: pmin,pmax
  ! Vibrational level
  integer,dimension(:),allocatable :: viblevel  ! Vibrational energy 
  integer :: maxviblevel
  real(kind=8),dimension(:),allocatable :: vibeng    ! expectation value of the energy (mean)
  integer,dimension(:),allocatable :: srmethod ! method by which to restrict the sampled points
  ! 1 = Use max and min energy.
  ! 2 = use monte carlo selection based on energy
  ! 3 = Use |C|^2 and monte carlo
  ! 4 = Use a grid.
  real(kind=8),dimension(:),allocatable :: engmin,engmax !min and max energies
  reaL(kind=8),dimension(:),allocatable :: twosigmasqrec    ! 2 sigma^2 for p(E) = exp(-(E-En)^2/twosigmasq)
  real(kind=8),dimension(:),allocatable :: WfnNorm 
  real(kind=8),dimension(:),allocatable :: griddq,griddp ! step size for grid
  integer,dimension(:),allocatable :: gridnq,gridnp ! number of grid points
  integer,allocatable,dimension(:) :: qpfileunits
  character(100) :: qpfilename
  ! ---- Energy units ----
  real(kind=8) :: eu = 0.40332405d0 ! euAukJmol
  real(kind=8) :: hbar = 1.0d0
  real(kind=8) :: pi = 3.14159265358979d0

  ! ---- Arrays for the wave function ----
  ! let some routines be publically available
  real(kind=8),dimension(:),allocatable :: laguerre_a, laguerre_bn, laguerre_twobn
  real(kind=8),dimension(:),allocatable :: laguerrex
  real(kind=8) :: zq,zp ! for zxxpsiprojection which is called by integragtec
  
  integer :: morse_fug = 99
  integer :: morse_maxtries = 100000
  integer :: morse_fham = 101
  integer :: morse_fwfn = 102
  integer :: nintpoint = 2000
  public  morse_fug,morse_maxtries,morse_fham,morsegamma,morsemass
    
  public readhamiltonian, projectcsmorse, projectcsmorsedvib, pickmorse,morseenergy
  ! morseprojectopint  

contains
  !!% ========================================================================
  !!%  Function: morseenergy
  !!%  Description: Calcualte the energy of a Morse mode at point q,p
  !!% ========================================================================
  function morseenergy( mode,q,p)
    implicit none
    integer,intent(in) :: mode !
    real(kind=8),intent(in) :: q,p
    real(kind=8) :: morseenergy

    ! ---- local ----
    ! potential part
    morseenergy = MorseA(mode)*(1+exp(-2.0d0*morseAlpha(mode)*(q-morseR0(mode))+morseAlpha(mode)**2/morsegamma(mode))&
         -2.0d0*exp(-MorseAlpha(mode)*(q-morseR0(mode))+0.25*MorseAlpha(mode)**2/morsegamma(mode)))
    ! Add kinetic part
    morseenergy = morseenergy +(0.5*p**2/morsemass(mode) + 0.25*hbar**2*morsegamma(mode)/morsemass(mode))

    ! Convert to kJmol
    morseenergy = morseenergy*eu

  end function morseenergy

  !!% ========================================================================
  !!%  Function: MorseLevelEng
  !!%  Desription: Calculate the energy of the lth vibrational level
  !!%              The calculated energy is in kJ/mol
  !!% ========================================================================
  function morseleveleng(mode,l)
    implicit none
    integer,intent(in) :: mode,l
    real(kind=8) morseleveleng

    morseleveleng=(2*hbar*morsealpha(mode)*sqrt(0.5d0*morseA(mode)/(morsemass(mode)))*(dble(l)+0.5d0) &
         -0.5d0*hbar**2*morsealpha(mode)**2*(dble(l)+0.5d0)**2/morsemass(mode))*eu

  end function morseleveleng

  !!% ========================================================================
  !!%  <z|x> projection
  !!% ========================================================================
  function zxprojection(mode,x,q,p)
    implicit none
    integer :: mode
    real(kind=8), intent(in) :: x,q,p
    complex(kind=8) :: zxprojection

    zxprojection = (morsegamma(mode)/pi)**0.25 &
         * exp(-morsegamma(mode)*(x-q)**2/hbar -(0.0d0,1.0d0)*p*(x-q) - (0.0d0,1.0d0)*0.5d0*q*p)

  end function zxprojection
  
  !!% ========================================================================
  !!%  Wave function projection <x|Psi>
  !!%  Takes a position x and a level 
  !!% ========================================================================
  function xpsiprojection(mode,x)
    implicit none
    integer,intent(in) :: mode
    real(kind=8),intent(in) :: x

    real(kind=8) :: xpsiprojection

    ! ---- local ----
    integer :: i
    real(kind=8) :: z

    interface
       include "laguerre_gen.interface"
    end interface

    z = 2.0d0* laguerre_a(mode) * exp(-morseAlpha(mode)*(x-morseR0(mode)))
    call laguerre_gen ( viblevel(mode), laguerre_twobn(mode), z,Laguerrex(0:viblevel(mode) ))
    
    xpsiprojection = WfnNorm(mode) * exp(-0.5d0*z) * z**laguerre_bn(mode) * Laguerrex(viblevel(mode))
  end function xpsiprojection

  ! ====================================================
  !  Calculate <Psi|x><x|Psi>
  !  For normalisation purposes
  ! ====================================================
  function xpsiprojection2(mode,x)
    implicit none
    real(kind=8) :: x
    integer:: mode
    real(kind=8) :: xpsiprojection2 

    xpsiprojection2 = (xpsiprojection(mode,x))**2
  end function xpsiprojection2

  function zxxpsiprojection(mode,x)
!,zq,zp)
    implicit none
    integer :: mode
    real(kind=8) :: x
    complex(kind=8) :: zxxpsiprojection
    
    zxxpsiprojection = zxprojection(mode,x,zq,zp) * xpsiprojection(mode,x)

  end function zxxpsiprojection

  ! ====================================================
  !  integrate a function which requires an extra 
  !  argument n
  ! ====================================================
  function integrate( xmin,xmax,nsteps,func,n)
    implicit none
    real(kind=8), intent(in) :: xmin,xmax
    integer :: nsteps
    integer :: n ! To pass to the function to be evaluated    
    real(kind=8) :: integrate
    interface
       function func(mode,x)
         integer :: mode 
         real(kind=8) :: x
         real(kind=8) :: func
       end function func
    end interface


    ! ---- local ----
    integer :: i
    real(kind=8) :: h

    if (mod(nsteps,2) .ne. 0) then
       nsteps = nsteps + 1
    end if
    h = (xmax - xmin) /dble(nsteps)

    integrate = func(n,xmin) + func(n,xmax)

    do i = 1,nsteps/2 -1
       integrate = integrate + 2.0d0* func(n,xmin+2*i*h) + 4.0d0*func(n,xmin+(2*i-1)*h)
    end do
    integrate = integrate + 4.0d0*func(n,xmax-h)

    !Now multiply by h/3
    integrate = integrate*h/3.0d0

  end function integrate

  ! ====================================================
  !  integrate a complex function 
  ! ====================================================
  function integratec( xmin,xmax,nsteps,func,n)
    implicit none
    real(kind=8), intent(in) :: xmin,xmax
    integer :: nsteps
    complex(kind=8) :: integratec
    integer :: n ! For passing to the function to be integrated
    interface
       function func(n,x)
         integer :: n
         real(kind=8) :: x
         complex(kind=8) :: func
       end function func
    end interface


    ! ---- local ----
    integer :: i
    real(kind=8) :: h

    if (mod(nsteps,2) .ne. 0) then
       nsteps = nsteps + 1
    end if
    h = (xmax - xmin) /(nsteps)

    integratec = func(n,xmin) + func(n,xmax)

    do i = 1,nsteps/2 -1
       integratec = integratec + 2.0d0* func(n,xmin+2*i*h) + 4.0d0*func(n,xmin+(2*i-1)*h)
    end do
    integratec = integratec + 4.0d0*func(n,xmax-h)

    !Now multiply by h/3
    integratec = integratec*h/3.0d0

  end function integratec

  !!% ========================================================================
  !!%  Function: morseprojectopoint
  !!%  Description: Calculate the overlap of a coherent state with a morse
  !!%               wave function.
  !!% ========================================================================
  function projectcsmorse (mode,q,p)
    implicit none
    integer, intent(in) :: mode
    real(kind=8),intent(in) :: q,p
    complex(kind=8) :: projectcsmorse

    zq = q
    zp = p
    projectcsmorse = integratec(qmin(mode),qmax(mode),nintpoint,zxxpsiprojection,mode)

  end function projectcsmorse

  !!% ========================================================================
  !!%  Function: morseprojectopointdvib
  !!%  Description: Calculate the overlap of a coherent state with a morse
  !!%               wave function viblevel + dvib.
  !!% ========================================================================

  function projectcsmorsedvib (mode,q,p,dvib)
    implicit none
    integer, intent(in) :: mode
    real(kind=8),intent(in) :: q,p
    complex(kind=8) :: projectcsmorsedvib
    integer,intent(in) :: dvib ! vibrational level to project on to.
    
    ! ---- local ----
    integer :: origviblevel ! save the current vibrational level for this mode
    real(kind=8) :: origwfnnorm

   ! if (dvib .ne. 0) then
    ! Store vibrational level that is read in from file
    origviblevel = viblevel(mode)
    
    ! Return 0 if we try to project on to a negative energy level.
    if (origviblevel+dvib < 0 ) then
       projectcsmorsedvib = (0.0d0,0.0d0)
       return 
    end if

    ! change the vibrational level by requested amount
    viblevel(mode) = origviblevel+dvib


    if ( origviblevel+dvib > size(laguerrex)-2) then
       deallocate(laguerrex)
       allocate(laguerrex(0:origviblevel+dvib))
    end if
       
    call  normalisewavefn(mode)
    !       write(*,*) "Wfn Norm",viblevel,WfnNorm(mode)
    !end if

    ! Do the projection
    zq = q
    zp = p
    projectcsmorsedvib = integratec(qmin(mode),qmax(mode),nintpoint,zxxpsiprojection,mode)

   ! if (dvib .ne. 0) then
       
       ! restore original vibrational level
       viblevel(mode) = origviblevel
       ! ---- Calculate parameters for the Laguerre polynomial ----
       laguerre_bn(mode) = laguerre_a(mode) - 0.5d0 - viblevel(mode)
       laguerre_twobn(mode) = 2.0d0*laguerre_bn(mode)
   call  normalisewavefn(mode)
     !  WfnNorm(mode)  = origwfnnorm 
    !  end if

  end function projectcsmorsedvib

  !!% ========================================================================
  !!%  Subroutine: pickmorse
  !!%  Description: Subroutine to pick a coherent state from a morse mode
  !!%               Most of this is controlled by the extra input file
  !!%  Arguments: mode - input - which of the morse modes to choose from
  !!%             z    - output - the chosen coherent state
  !!%             q,p  - output - chosen coherent state
  !!%             amp  - output - the projection of the wave function on to z
  !!% ========================================================================
  subroutine pickmorse(mode,q,p,amp,ntries)
    implicit none
    integer, intent(in) :: mode 
!    complex(kind=8),intent(out) :: z   ! the chosen coherent state
    real(kind=8),intent(out) :: q,p 
    complex(kind=8),intent(out) :: amp ! its amplitude 
    integer,optional :: ntries 
    ! ---- local variables ----
    integer :: i ! for loops
!    real(kind=8) :: q1,p1 
    real(kind=8),external :: ZBQLU01 ! Random number generator
    integer :: seed ! dummmy input for the random number generator
    real(kind=8) :: cseng
    logical :: chosen
    real(kind=8) :: prob,rn  ! For monte carlo acceptance of basis function
    integer,save,allocatable,dimension(:) :: gridq,gridp ! which points on the grid
    logical,save :: gridsetup = .false.

    chosen = .false.
    i = 0

    if (srmethod(mode) .eq. 4 .and. .not. gridsetup) then
       allocate(gridq(nmm),gridp(nmm))
       gridq(:) = 0
       gridp(:) = -1
       gridsetup = .true.
    end if

    do while (  i < morse_maxtries .and. .not. chosen)
       i = i + 1
       ! Choose a q and p
       if (srmethod(mode) .eq. 4) then
          gridp(mode) = gridp(mode) + 1
          if (gridp(mode) .eq. gridnp(mode)) then
             gridp(mode) = 0
             gridq(mode) = gridq(mode) + 1
             if (gridq(mode) .eq. gridnq(mode)) gridq(mode) = 0
          end if
          
          q = qmin(mode) + griddq(mode)*gridq(mode)
          p = pmin(mode) + griddp(mode)*gridp(mode)
          
       else if (srmethod(mode) .eq. 5 ) then ! Read q,p pair from file if requested.
          if (present(ntries)) then
             if (ntries .eq. 1) then
                write(*,*) "Read in q-p pair for mode",mode
                read(qpfileunits(mode),*) q,p
             end if
          else
             write(*,*) "Read in q-p pair for mode",mode
             read(qpfileunits(mode),*) q,p
          end if
       else
          q = (qmax(mode)-qmin(mode))*ZBQLU01(seed) + qmin(mode)
          p = (pmax(mode)-pmin(mode))*ZBQLU01(seed) + pmin(mode)
       end if
       ! Now calculate the energy in kJ/mol
       cseng = morseenergy( mode,q,p)


       ! Is this point acceptable
       if(srmethod(mode) == 2 ) then
          if (cseng >engmin(mode) .and. cseng < engmax(mode)) then
             write(110+2,*) "# Chosen ",cseng,engmin(mode) ,engmax(mode)
             chosen = .true.
          else
             write(111+2*mode,*) mode,q,p,cseng
          end if
       else if(srmethod(mode) == 3) then
          amp =  projectcsmorse (mode,q,p) 

          prob = dconjg(amp)*amp
          rn =  ZBQLU01(seed)
          if (prob > rn) then
             chosen = .true.
          else
              write(111+2*mode,*) mode,real(amp),dconjg(amp),q,p,cseng
           end if
          
        else if(srmethod(mode) == 4) then ! Assign points on a grid.
           ! Only not choose if energy restrictions are requred.
           if (engmin(mode) < engmax(mode)) then
              if (cseng > engmin(mode) .and. cseng < engmax(mode)) then
                 chosen = .true.
              else
                 write(*,*) "Out of energy range: Skipping point ",gridq(mode),gridp(mode)," of mode ",mode, "energy ",cseng
              end if
           else 
              chosen = .true.
           end if
           
        else if (srmethod(mode) == 5 ) then ! always accept points read in from files.
           chosen = .true.
                        
       else
          prob = exp(-(cseng-vibeng(mode))**2*twosigmasqrec(mode))
          rn =  ZBQLU01(seed)
          if (prob > rn) then
             chosen = .true.
          else
              write(111+2*mode,*) mode,q,p,cseng
          end if
       end if
    end do

    if (.not. chosen) then      
       write(morse_fug,*) "Failed to choose a basis fn for morse mode ",mode," in ",morse_maxtries," tries."
       write(morse_fug,*) "Relax restriction or increase morse_maxtries and try again."
       stop
    end if
    
    ! project wave function onto this coherent state
    ! However, if srmethod == 3 then we have already done it so don't do it again.
    if(srmethod(mode) .ne.  3) then
       amp =  projectcsmorse (mode,q,p)    
    end if
    write(510+2*mode,*) mode,real(amp),imag(amp),q,p,cseng
                 
  end subroutine pickmorse

  !!% ========================================================================
  !!%  Subroutine: readhamiltonian
  !!%  Description: Read hamiltonian parameters from a file   
  !!% ========================================================================
  subroutine readhamiltonian(filename)
    implicit none
    character(len=*) :: filename

    ! ---- local ----
    integer :: ierr
    character(len=100):: line,keyword
    integer :: i,j
    character(len=10) :: modec
    real(kind=8) :: x
    real(kind=8) :: sigma
    character(len=50) :: wfnfile
    logical :: getoption ! for reading in options
    logical :: readheader
    character(len=50) :: commentchar
    ! ---- Open file ----
    open(unit=morse_fham,file=trim(filename),status="old",iostat=ierr)
    if (ierr.ne. 0) then
       write(morse_fug,*) "Error opening hamiltonian file",trim(filename),". Error number ",ierr
       stop
    end if

    ! ---- Read stuff in ----
    read(morse_fham,"(a100)") line
    read(line,*) keyword
    if (keyword == "nmodes") then
       read(line,*) keyword,nmm
    else
       write(morse_fug,*) "The hamiltonian must start with the number of modes"
    end if

    ! ---- allocate arrays ----
    allocate(morseA(nmm))      
    allocate(morseAlpha(nmm))
    allocate(morseR0(nmm))
    allocate(morsegamma(nmm))
    allocate(morsemass(nmm))
    allocate(qmin(nmm),qmax(nmm))
    allocate(pmin(nmm),pmax(nmm))
    allocate(viblevel(nmm),vibeng(nmm))
    allocate(srmethod(nmm))
    allocate(engmin(nmm),engmax(nmm))
    allocate(twosigmasqrec(nmm))
    allocate(WfnNorm(nmm))
    allocate(griddq(nmm),griddp(nmm))
    allocate(gridnq(nmm),gridnp(nmm))

    maxviblevel = 0
    ! ---- Now read  parameters for each mode ----
    do i = 1,nmm
       read(morse_fham,*,iostat = ierr) keyword
       do while ( keyword .ne. "mode" .and. ierr .ne. 0)
          read(morse_fham,*,iostat = ierr) keyword
       end do
       if (ierr .ne. 0) then
          write(*,*) "Error reading keyword mode for mode ",i," Error code ",ierr
          stop
       end if
       backspace(morse_fham)

       read(morse_fham,*,iostat=ierr) keyword,line 
       write(morse_fug,*) "Morse mode ",i, trim(line)

       read(morse_fham,*,iostat=ierr) keyword,morseA(i)
       if (keyword .ne. "A" .or. ierr .ne. 0) then
          write(morse_fug,*) "Problem reading Morse A parameter. Read ",trim(keyword)," error code ",ierr
          stop
       end if
       morseA(i) = morseA(i)/eu

       read(morse_fham,*,iostat=ierr) keyword,morseAlpha(i)
       if (keyword .ne. "alpha" .or.ierr .ne. 0 ) then
          write(morse_fug,*) "Problem reading Morse alpha parameter. Read ",trim(keyword)," error code ",ierr
          stop
       end if
       read(morse_fham,*,iostat=ierr) keyword,morseR0(i)
       if (keyword .ne. "r0" .or. ierr .ne. 0 ) then
          write(morse_fug,*) "Problem reading Morse r0 parameter. Read ",trim(keyword)," error code ",ierr
          stop
       end if
       read(morse_fham,*,iostat=ierr) keyword,morsegamma(i)
       if (keyword .ne. "gamma" .or. ierr .ne. 0 ) then
          write(morse_fug,*) "Problem reading Morse gamma parameter. Read ",trim(keyword)," error code ",ierr
          stop
       end if
       read(morse_fham,*,iostat=ierr) keyword,morsemass(i)
       if (keyword .ne. "mass" .or. ierr .ne. 0 ) then
          write(morse_fug,*) "Problem reading Morse mass parameter. Read ",trim(keyword)," error code ",ierr
          stop
       end if
       read(morse_fham,*,iostat=ierr) keyword,viblevel(i)
       if (keyword .ne. "viblevel" .or. ierr .ne. 0 ) then
          write(morse_fug,*) "Problem reading Morse vibrational level. Read ",trim(keyword)," error code ",ierr
          stop
       end if
       ! need to know what the maximum requested vibrational level is so we can allocate arrays for the laguerre polynomials
       if (viblevel(i) > maxviblevel) then
          maxviblevel = viblevel(i)
       end if

       read(morse_fham,*,iostat=ierr) keyword,qmin(i),qmax(i),pmin(i),pmax(i)
       if (keyword .ne. "phasespace" .or. ierr .ne. 0 ) then
          write(morse_fug,*) "Problem reading phase space to sample. Should be phasespace qmin qmax pmin pmax"
          write(morse_fug,*) "Actually read ",trim(keyword)," error code ",ierr
          stop
       end if
       ! Now read in a sampling method
       read(morse_fham,*,iostat=ierr) keyword
       if (keyword == "montecarlo") then          
          read(morse_fham,*) keyword,sigma
          if (keyword .ne. "sigma" .or. ierr .ne. 0) then
             write(morse_fug,*) "Problem reading sigma for mode ",i,". Read ",trim(keyword)," error code ",ierr
             stop
          end if
          srmethod(i) = 1
          twosigmasqrec(i) = 0.5d0/sigma**2
                       
       else if (keyword == "energy") then
          srmethod(i) = 2
          read(morse_fham,*) keyword,engmin(i)
          if (keyword .ne. "mineng" .or. ierr .ne. 0) then
             write(morse_fug,*) "Problem reading mineng for mode ",i,". Read ",trim(keyword)," error code ",ierr
             stop
          end if
          read(morse_fham,*) keyword,engmax(i)
          if (keyword .ne. "maxeng" .or. ierr .ne. 0) then
             write(morse_fug,*) "Problem reading mineng for mode ",i,". Read ",trim(keyword)," error code ",ierr
             stop
          end if
       else if (keyword == "wfnmc") then
          srmethod(i) = 3

       else if (keyword == "grid") then
          srmethod(i) = 4
          getoption =.true.
          ! Set defaults
          gridnq(i) = 21
          gridnp(i) = 21
          ! set min > max to turn off selection
          engmin(i) = 1e10
          engmax(i) = -1e10

          do while (getoption)
             read (morse_fham,*) keyword
             backspace(morse_fham)
             if (keyword .eq. "nq") then
                read(morse_fham,*) keyword,gridnq(i)
                if ( ierr .ne. 0) then
                   write(morse_fug,*) "Problem reading nq for mode ",i,". Read ",trim(keyword)," error code ",ierr
                   stop
                end if
             else  if (keyword .eq. "np") then
                read(morse_fham,*) keyword,gridnp(i)
                if (ierr .ne. 0) then
                   write(morse_fug,*) "Problem reading np for mode ",i,". Read ",trim(keyword)," error code ",ierr
                   stop
                end if
             else  if (keyword .eq. "mineng") then
                read(morse_fham,*) keyword,engmin(i)
                if ( ierr .ne. 0) then
                   write(morse_fug,*) "Problem reading grid engmin for mode ",i,". Read ",trim(keyword)," error code ",ierr
                   stop
                end if
             else  if (keyword .eq. "maxeng") then
                read(morse_fham,*) keyword,engmax(i)
                if (ierr .ne. 0) then
                   write(morse_fug,*) "Problem reading grid engmax for mode ",i,". Read ",trim(keyword)," error code ",ierr
                   stop
                end if
             else
                getoption = .false.
             end if
          end do
          ! Calculate grid point separation
          griddq(i) = (qmax(i) - qmin(i))/dble(gridnq(i)-1)
          griddp(i) = (pmax(i) - pmin(i))/dble(gridnp(i)-1)
          
       else if (keyword == "file") then ! Read in q, p pairs from a file.
          srmethod(i)=5
          if (.not. allocated(qpfileunits)) then
             allocate(qpfileunits(nmm))
             qpfileunits(:) = 0
          end if
          ! Read in file name
          read(morse_fham,*,iostat=ierr) qpfilename
          if (ierr .ne. 0) then
             write(*,*) "Error reading filename for qp pairs in mode ",i
             stop
          end if
          ! Open file
          qpfileunits(i) = 300+i
          open(unit=qpfileunits(i),file=qpfilename,status="old",iostat=ierr)
          if (ierr .ne. 0) then
             write(*,*) "Error opening qp pairs file for mode ",i, " Filename",qpfilename
             write(*,*) "Error code",ierr
             stop
          end if
          ! Read in any comment at top of file.
          readheader = .true.
          do while (readheader)
             read(qpfileunits(i),*) commentchar
             if (commentchar(1:1) .ne. "#") then
                readheader = .false.
                backspace(qpfileunits(i))
             end if
          end do                
          write(*,*) "Opened trajectory file ", trim(qpfilename) , " for mode ",i
       else
          write(*,*) "Don't recognise the sampling method of ",keyword," for mode ",i
          stop
       end if
       write(*,*) "Sampling Method ",trim(keyword),srmethod(i)
    end do

    ! Loop again, this time to tell calculate some things and tell the user what we have got.

    allocate(laguerre_a(nmm))
    allocate(laguerre_bn(nmm))
    allocate(laguerre_twobn(nmm))
    allocate(laguerrex(0:maxviblevel))

    do i = 1,nmm
       write(morse_fug,*) "    A     = ",morseA(i)
       write(morse_fug,*) "    alpha = ",morseAlpha(i)
       write(morse_fug,*) "    r0    = ",morseR0(i)
       write(morse_fug,*) "    gamma = ",morsegamma(i)
       write(morse_fug,*) "    mass  = ",morsemass(i)

       ! ---- Calculate the energy ---- 
       vibeng(i) = morseleveleng(i,viblevel(i))
       write(morse_fug,*) "Vib. level ",viblevel(i)," energy", vibeng(i)
       write(morse_fug,*) "Sampling method ",srmethod(i)
       ! ---- Calculate parameters for the Laguerre polynomial ----
       laguerre_a(i) = sqrt(2*morsemass(i)*morseA(i))/(morseAlpha(i)*hbar)
       laguerre_bn(i) = laguerre_a(i) - 0.5d0 - viblevel(i)
       laguerre_twobn(i) = 2.0d0*laguerre_bn(i)
       write(morse_fug,*) "Laguerre_a",laguerre_a(i)
       write(morse_fug,*) "Laguerre_b",laguerre_bn(i)
       
       call normalisewavefn(i)

!       write (*,*) "New norm = ",  integrate(qmin(i),qmax(i),500,xpsiprojection2,i)
       write(modec,"(i3)") i
       wfnfile = "morsewfn-mode"//trim(adjustl(modec))//"-v"
       write(modec,"(i3)") viblevel(i)
       wfnfile=trim(wfnfile)//trim(adjustl(modec))
       open(morse_fwfn,file=trim(wfnfile))
       do j = 0,nintpoint
          x = j*(qmax(i)-qmin(i))/nintpoint+qmin(i) !j*1.5*(qmax(i)-qmin(i))/500+qmin(i)
          write(morse_fwfn,*) x,xpsiprojection2(i,x),xpsiprojection(i,x)
       end do
       close(morse_fwfn)
    end do

  end subroutine readhamiltonian

  subroutine normalisewavefn(mode)
    implicit none
    integer,intent(in) :: mode
    laguerre_bn(mode) = laguerre_a(mode) - 0.5d0 - viblevel(mode)
    laguerre_twobn(mode) = 2.0d0*laguerre_bn(mode)

    ! ---- Calculate the normalisation constant for the wave function ----
    WfnNorm(mode) = 1.0d0 ! set to one because it is used in xpsiprojection2
    WfnNorm(mode)  = integrate( qmin(mode),qmax(mode),500,xpsiprojection2,mode)
    WfnNorm(mode) = 1.0d0/ sqrt(WfnNorm(mode))
    
  end subroutine normalisewavefn
end module morsesampling
