!!% ************************************************************************
!!%  Module: vectorutils
!!%  Author: Stewart Reed
!!%  Email: S.K.Reed@physics.org
!!%  Date: 15th July 2009
!!%
!!%  Description: A wee for common vector operations. Includes routines for
!!%               rotating vectors, calculating cross and dot products
!!%               Unless otherwise stated, all quantites are real(kind=8) 
!!%               and the vectors have dimension of 3.
!!% ************************************************************************
module vectorutils

contains

  !!% ============================================================
  !!%  Function: rotatex
  !!%  Description: Apply a rotation about x to a 3D vector.
  !!%               The rotation matrix is
  !!%                    (  1        0            0      )
  !!%                    (  0   cos(theta)  -sin(theta)  )
  !!%                    (  0   sin(theta)   cos(theta)  )
  !!% ============================================================
  function rotatex(theta,vect)
    implicit none
    real(kind=8),dimension(3),intent(in)::vect
    real(kind=8),intent(in) :: theta

    real(kind=8),dimension(3) :: rotatex

    rotatex(1) = vect(1)

    rotatex(2) = cos(theta)*vect(2) - sin(theta)*vect(3)
    rotatex(3) = sin(theta)*vect(2) + cos(theta)*vect(3)
    
  end function rotatex

  
  !!% ============================================================
  !!%  Function: rotatey
  !!%  Description: Apply a rotation about y to a 3D vector.
  !!%               The rotation matrix is
  !!%                    (  cos(theta)  0   sin(theta)  )
  !!%                    (      0       1        0      )
  !!%                    ( -sin(theta)  0   cos(theta)  )
  !!% ============================================================
  function rotatey(theta,vect)
    implicit none
    real(kind=8),dimension(3),intent(in)::vect
    real(kind=8),intent(in) :: theta

    real(kind=8),dimension(3) :: rotatey

    rotatey(1) = cos(theta)*vect(1) + sin(theta)*vect(3)
    rotatey(2) = vect(2)
    rotatey(3) = cos(theta)*vect(3) - sin(theta)*vect(1)

  end function rotatey

  !!% ============================================================
  !!%  Function: rotatez
  !!%  Description: Apply a rotation about z to a 3D vector.
  !!%               The rotation matrix is
  !!%                    ( cos(theta)  -sin(theta)  0 )
  !!%                    ( sin(theta)   cos(theta)  0 )
  !!%                    (      0            0      1 )
  
  !!% ============================================================
  function rotatez(theta,vect)
    implicit none
    real(kind=8),dimension(3),intent(in)::vect
    real(kind=8),intent(in) :: theta

    real(kind=8),dimension(3) :: rotatez
    
    rotatez(1) = cos(theta)*vect(1) - sin(theta)*vect(2)
    rotatez(2) = sin(theta)*vect(1) + cos(theta)*vect(2)
    rotatez(3) = vect(3)
  end function rotatez


  !!% ============================================================
  !!% Function: dotprod
  !!% Description: Calculate the dot product of two vectors
  !!%              Works with vectors of any dimension
  !!% ============================================================
  function dotprod(vect1,vect2)
    implicit none
    real(kind=8),dimension(:),intent(in) :: vect1,vect2
    real(kind=8) :: dotprod

    dotprod = sum(vect1(:)*vect2(:))

  end function dotprod


  !!% ============================================================
  !!%  Function: crossprod
  !!%  Description: Calculate the crossproduct of two vectors
  !!% ============================================================
  function crossprod(vect1,vect2)
    implicit none
    real(kind=8),dimension(3),intent(in) :: vect1,vect2
    real(kind=8),dimension(3) :: crossprod

    crossprod(1) = vect1(2)*vect2(3) - vect1(3)*vect2(2)
    crossprod(2) = vect1(3)*vect2(1) - vect1(1)*vect2(3)
    crossprod(3) = vect1(1)*vect2(2) - vect1(2)*vect2(1)
  end function crossprod

  
  !!% ============================================================
  !!%  Subroutine: solve2D
  !!%  Description: A simple subroutine to solve by explicit 
  !!%               inversion a system of linear equations.
  !!%               Solves A*x=b
  !!% ============================================================
  subroutine solve2D(A,b,x)
    implicit none
    real(kind=8),dimension(2,2),intent(in) :: A
    real(kind=8),dimension(2),intent(in) :: b
    real(kind=8),dimension(2),intent(out) :: x
  
    ! ---- local ----
    real(kind=8) :: determinant
    
    determinant = A(1,1)*A(2,2)-A(1,2)*A(2,1)

    x(1) = (A(2,2)*b(1) - A(1,2)*b(2))/determinant
    x(2) = (A(1,1)*b(2) - A(2,1)*b(1))/determinant

  end subroutine solve2D

  !!% ============================================================
  !!%  Subroutine: solve3D
  !!%  Description: A simple subroutine to solve by explicit 
  !!%               inversion a system of linear equations.
  !!%               Solves A*x=b
  !!% ============================================================
  subroutine solve3D(A,b,x)
    implicit none
    real(kind=8),dimension(3,3),intent(in) :: A
    real(kind=8),dimension(3),intent(in) :: b
    real(kind=8),dimension(3),intent(out) :: x
    ! ---- local ----
    real(kind=8) :: determinant

    ! Calculate the determinant of the matrix
    determinant = A(3,1) * A(1,2) * A(2,3) - A(3,1) * A(1,3) * A(2,2) - A(2,1) * A(1,2) * A(3,3) &
         + A(2,1) * A(1,3) * A(3,2) + A(1,1) * A(2,2) * A(3,3) - A(1,1) * A(2,3) * A(3,2)
    
    if(determinant < 1e-7 .and. determinant > -1e-7) then
       write(*,*) "Error in solve3D: Matrix determinant is (almost) zero"
       write(*,*) "Det = ",determinant
    end if

    x(1) = (A(2,2)*A(3,3)-A(2,3)*A(3,2))*b(1) +(A(1,3)*A(3,2)-A(1,2)*A(3,3))*b(2) &
         + (A(1,2)*A(2,3) - A(1,3)*A(2,2))*b(3)

    x(2) = (A(3,1)*A(2,3)-A(2,1)*A(3,3))*b(1) + (A(1,1)*A(3,3)-A(3,1)*A(1,3))*b(2) &
         + (A(2,1)*A(1,3)-A(1,1)*A(2,3))*b(3)

    x(3) = (A(2,1)*A(3,2)-A(3,1)*A(2,2))*b(1) + (A(3,1)*A(1,2)-A(1,1)*A(3,2))*b(2) &
         + (A(1,1)*A(2,2)-A(1,2)*A(2,1))*b(3)

    x(:)= x(:)/determinant

  end subroutine solve3D

end module vectorutils
