! *******************************************
!               plotwavefn
!
!                S.K.Reed
!           Guy Fawkes night 2007
!
!  A wee program to plot a wavefn made up of
!  a sum of coherent states
! *******************************************

program plotwavefn

  implicit none

  character ( len = 80 ) arg ! for command line options
  integer numarg             ! number of command line options
  integer iargc              ! For interpretation loop

  integer i,j,bin            ! for loops


  integer :: ndf    ! number of degrees of freedom
  integer :: nbasis ! number of basis functions
  integer :: dof    ! which degree of freedom to plot
  character(len=80) :: filename 
  character(len=80) :: gfilename  ! filename for gammas
  real(kind=8) :: xmin,xmax ! Range over which to calculate phi
  logical :: minf,pinf  ! are the min/max values -inf/+inf?
  real(kind=8) :: x ! Guess!
  real(kind=8) :: lxmin,lxmax ! min and max values for a particular bin
  integer :: nbins  ! number of bins for phi
  integer :: step   ! for which step should the wavefn be plotted?
  integer :: ierr ! for errors reading inputs
  integer :: verbosity ! how much info to write out.

  ! logical variables to control program behaviour
  logical :: allconfigs   
  logical :: gpsurface
  logical :: haveconfig


  character(len=100) :: line  ! for reading in stuff
  real(kind=8),dimension(:,:),allocatable :: qps ! The set of q and ps
  integer :: q,p  ! for referencing qps
  real(kind=8),dimension(:),allocatable :: gamma ! The set ofgammas
  real(kind=8) :: clgamma ! The set ofgammas
  real(kind=8) :: norm2   ! The norm of the wave function (squared?)
  real(kind=8) :: t ! time
  real(kind=8) :: ReAmp,ImAmp ! real and image parts of the amplitude
  complex(kind=8),dimension(:),allocatable :: Amp ! Complex amplitude
  complex(kind=8) :: cont ! contribution to phi
  complex(kind=8) :: poverlp ! partial overlap matrix element
  complex(kind=8) :: overlp ! overlap matrix element
  complex(kind=8),dimension(:),allocatable :: phi  ! The wave function!
  
  ! file units
  integer, parameter :: famp = 10 ! file unit for amplitude file
  integer, parameter :: fgamma = 11 ! file unit for gammas file

  real,parameter :: pi = 3.14159265358979

  real(kind=8) :: cutoffprob
  real(kind=8) :: rcut2  ! cutoff
  
  ! Include interface for subroutines so that we get compile time type checking etc.
  interface
     subroutine readunwanted( fu, nb,ns,verbosity)
       integer, intent(in) :: fu ! file unit
       integer, intent(in) :: nb ! number of basis functions
       integer, intent(in) :: ns ! number of steps
       integer, intent(in) :: verbosity ! how much info to write out
     end subroutine readunwanted
     
     function overlap_qp(qpi,qpj,ndf,gammas)
       integer,intent(in) :: ndf
       real(kind=8),intent(in) :: qpi(2*ndf),qpj(2*ndf)
       real(kind=8),intent(in) :: gammas(ndf)
     end function overlap_qp

     function partialoverlap_qp(qpi,qpj,ndf,dof,gammas)
       integer,intent(in) :: ndf
       real(kind=8),intent(in) :: qpi(2*ndf),qpj(2*ndf)
       integer,intent(in) :: dof
       real(kind=8),intent(in) :: gammas(ndf)
       complex(kind=8) :: partialoverlap_qp
     end function partialoverlap_qp

     function erfterm(qj,qk,pj,pk,mgamma,xmin,xmax,minf,pinf)
       real(kind=8), intent(in) :: qj,qk,pj,pk,mgamma,xmin,xmax
       logical,intent(in) :: minf,pinf
       complex(kind=8) :: erfterm
     end function erfterm

     function diagerfterm(q,mgamma,xmin,xmax,minf,pinf)
       implicit none
       real(kind=8),intent(in) :: q
       real(kind=8), intent(in):: mgamma,xmin,xmax
       logical,intent(in) :: minf,pinf
       
       real(kind=8) :: diagerfterm
     end function diagerfterm
  end interface


  ! =============== Set some defaults ===============
  xmin = -1      ! Range to plot wave fct over
  xmax = 1
  ndf = 1        ! Number of degrees of freedom
  nbasis = 1     ! Number of basis fcts
  dof = 1        ! Which degree of freedom to plot
  step = 1       ! Which step from the amplitudes file
  nbins = 50     ! how many bins to calculate the wavefct prob. in.

  minf = .false. ! Are either of the limits to be +/-infinity 
  pinf = .false.

  cutoffprob = 1D-7 ! exclude z for which <x|z> < cutoffprob
  
  verbosity = 1   ! How much detail should this program write out
  gfilename = ''  ! name of file containing gammas
  clgamma = 1     ! 'command line' gamma
  allconfigs = .false. ! Plot wave functions for all configs?
  gpsurface = .false.  ! Make a surface suitable for plotting in gnuplot.
  
  ! =============== Find out number command line options ===============
  numarg = iargc ( )
  i = 1
  if (numarg .eq. 0) then
     call help
     stop
  end if

  do while (i .le. numarg)
     call getarg ( i, arg )
     
     ! ----- range ----
     if (arg .eq. "-r") then
        i = i+1
        call getarg ( i, arg )
        if(arg .eq. "-inf" ) then
           minf = .true.
           xmin = -1D6
        else
           READ(arg,FMT=*,IOSTAT=ierr) xmin
           if (ierr .ne. 0) then
              write(*,*) "Error reading xmin",arg
              stop
           end if
        end if
        i = i+1
        call getarg ( i, arg )
        if(arg .eq. "inf" ) then
           pinf = .true.
           xmax = 1D6
        else
           READ(arg,FMT=*,IOSTAT=ierr) xmax        
           if (ierr .ne. 0) then
              write(*,*) "Error reading xmax",arg
              stop
           end if
        end if
        ! ----- which d.o.f. ----
     else if (arg .eq. "-m") then
        i = i+1
        call getarg ( i, arg )
        READ(arg,FMT=*,IOSTAT=ierr) dof
        if (ierr .ne. 0) then
           write(*,*) "Error reading mode to plot", arg
           stop
        end if

         ! ----- number of basis functions ----
     else if (arg .eq. "-n") then
        i = i+1
        call getarg ( i, arg )
        READ(arg,FMT=*,IOSTAT=ierr) nbasis
        if (ierr .ne. 0) then
           write(*,*) "Error reading number of basis functions", arg
           stop
        end if

         ! -----number of degrees of freedom ----
     else if (arg .eq. "-d") then
        i = i+1
        call getarg ( i, arg )
        READ(arg,FMT=*,IOSTAT=ierr) ndf
        if (ierr .ne. 0) then
           write(*,*) "Error reading number of degrees of freedom", arg
           stop
        end if
        
        ! -----step to look at ----
     else if (arg .eq. "-s") then
        i = i+1
        call getarg ( i, arg )
        READ(arg,FMT=*,IOSTAT=ierr) step
        if (ierr .ne. 0) then
           write(*,*) "Error reading step", arg
           stop
        end if
        
        ! -----look at all steps----
     else if (arg .eq. "-all") then
        allconfigs = .true.

         ! -----make a file for gnuplots splot----
     else if (arg .eq. "-gpsurf") then
        gpsurface = .true. 

        ! ----- number of bins ----
     else if (arg .eq. "-nb") then
        i = i+1
        call getarg ( i, arg )
        READ(arg,FMT=*,IOSTAT=ierr) nbins
        if (ierr .ne. 0) then
           write(*,*) "Error reading number of bins", arg
           stop
        end if
        
        ! ----- file containing gammas ----
     else if (arg .eq. "-gf") then
        i = i+1
        call getarg ( i, arg )
        gfilename = trim(arg)

        ! ----- Command line gamma ----
     else if (arg .eq. "-g") then
        i = i+1
        call getarg ( i, arg )
        READ(arg,FMT=*,IOSTAT=ierr) clgamma
        if (ierr .ne. 0) then
           write(*,*) "Error reading gamma from command line", arg
           stop
        end if
        
        ! ----- increased output ----
     else if (arg .eq. "-v") then
        verbosity = 2
        
         ! ----- decreased output ----
     else if (arg .eq. "-q") then
        verbosity = 0
     else if (arg .eq. "-h" .or. arg .eq. "-help" .or. arg .eq. "--help") then
        call help
        stop
     else
        filename = trim(arg)
     end if
     i = i+1
  end do

  ! summarise

  if (verbosity .ge. 1) then
     write(*,*) "# ",ndf," degrees of freedom"
     write(*,*) "# ",nbasis," basis functions"
     write(*,*) "# plotting mode ",dof
     write(*,"(a25,f15.5,a1,f15.5,a2)") " # Plot wavefn over x in [",xmin,",",xmax,"]"
     write(*,*) "# Bin Width ",(xmax-xmin)/nbins
     write(*,*) "# Results set ",step
     write (*,*) "# Reading positions and amplitudes from ",trim(filename)
  end if

  ! =============== Open file ===============
  open(famp,file=filename,status="old",iostat=ierr)
  if (ierr .ne. 0) then
     write(*,*) "Problem opening the file ",filename
     stop
  end if
  
  ! =============== allocate arrays ===============
  allocate(phi(nbins))
  allocate(qps(2*ndf,nbasis))
  allocate(gamma(ndf))

  allocate(Amp(nbasis))

  ! =============== Read Gammas ===============
  if (gfilename .ne. '') then
     open(fgamma,file=gfilename,status="old",iostat=ierr)
     if (ierr.ne. 0) then 
        write(*,*) "Error opening file containing gammas"
        stop
     end if
     if (verbosity .ge. 2) write(*,*) "# Reading in gammas"
     ! Read in a comment line
     read(fgamma,*) line
     do i = 1,ndf
        read(fgamma,*,iostat=ierr) j,gamma(i)
     end do
  else
     gamma(:) = clgamma

  end if
  
  if (verbosity .ge. 1) then
     if (dof > 0) then
        write(*,*) "# Using gamma =",gamma(dof)
     end if
  end if

  ! =============== Calculate cutoff =====================
  if (dof > 0) then
     rcut2 = -2*log(cutoffprob)/gamma(dof)
     if (verbosity .ge. 1) write(*,*) "# Rcut ",rcut2
  else
     rcut2 = 1D5
  end if
  
  ! =============== Read in early amplitudes ===============
  ! read in first comment
  read(famp,*) line
  if (.not. allconfigs) then
     if(verbosity.ge.2)write(*,*) "#Comment: ",trim(line)
     if (step .gt. 1) then
        if (verbosity .ge.2) write(*,*) "# Reading in unwanted sets"
        call readunwanted(famp,nbasis,step-1,verbosity)
        if(verbosity .ge.2) write(*,*) "#" 
     end if
  end if

  ! =============== Read in basis functons ===============
    ! read in comment
  read(famp,"(a80)") line
  if (verbosity .ge.2) write (*,*) trim(line)


  do i = 1,nbasis
     read(famp,*) t,ReAmp,ImAmp,(qps(j,i),j=1,2*ndf)
     ! Complex amplitude
     Amp(i)  = ReAmp+(0,1)*ImAmp
  end do

  haveconfig = .true.

  ! =============== Square of normalisation cst ===============
  norm2 = 1.0d0
  ! normalisation constant is 1!

  !  if (dof > 0) then
  !     norm2 = gamma(dof)/pi
  !     !  norm2 = norm2**(dble(dof)/2.0)
  !     norm2 = dsqrt(norm2)
  !     write(*,*) "# Normalisation factor",norm2, " = gamma/pi"
  !  else
  !     norm2 = 1.0
  !      write(*,*) "# Calculating wavefunction norm"
  !   end if
  
  ! Which part of the array corresponds to q and p
  q = 2*dof-1
  p = 2*dof

  
  do while (haveconfig)

     ! Work out phi
     do bin = 1,nbins
        ! x
        !     x = (bin-0.5)*(xmax-xmin)/nbins + xmin
        ! work out upper and lower bounds for this bin.
        lxmin = (bin-1)*(xmax-xmin)/nbins + xmin
        lxmax = (bin)*(xmax-xmin)/nbins + xmin
        
        ! Set the wave fct prob. to 0.
        phi(bin) = (0.0d0,0.0d0)
        
        ! Diagonal term
        do i = 1,nbasis
           !        if ( (x-qps(q,i))**2 < rcut2 ) then
           !           overlp = overlap_qp(qps(:,i),qps(:,i),ndf,dof,gamma)  =1          
           if (dof > 0 ) then
              phi(bin)  =  phi(bin) + dconjg(Amp(i))*Amp(i)  &
                   *  diagerfterm(qps(q,i),gamma(dof),lxmin,lxmax,minf,pinf)
              !                   *  exp(-gamma(dof)*(x-qps(q,i))**2)
           else
              phi(bin)  =  phi(bin) + dconjg(Amp(i))*Amp(i)
           end if
           !        end if
        end do
        
        ! Off diagonal contributions
        do i = 1,nbasis - 1
           do j = i+1,nbasis
              !           poverlp = partialoverlap_qp(qps(:,i),qps(:,j),ndf,dof,gamma)           
              ! Although calling the partial overlap routine, this actually calculated the full 
              ! overlap. Does the full overlap routine not work?
              overlp =  partialoverlap_qp(qps(:,i),qps(:,j),ndf,0,gamma)
              !           overlp = overlap_qp(qps(:,i),qps(:,j),ndf,gamma)           
              
              if (dof > 0 ) then
                 cont = 2*real(&
                      dconjg(Amp(i))*Amp(j)*overlp & 
                      *erfterm(qps(q,i),qps(q,j),qps(p,i),qps(p,j),gamma(dof),lxmin,lxmax,minf,pinf))
                 
              else ! mode = 0 - i.e. calculate norm
                 cont = 2*real(dconjg(Amp(i))*Amp(j) * overlp)
              end if
              phi(bin) = phi(bin) + cont
              
           end do
        end do
     end do

     ! =============== write out phi ===============
     if (.not. allconfigs .and. verbosity .ge. 1) write(*,*) "# time ",t
     do j = 1,nbins 
        x = (j-0.5)*(xmax-xmin)/nbins + xmin
        if(abs(phi(j)) .lt. 1d-10) phi(j) = (0.0d0,0.0d0)

        !     write(*,*) x,real(phi(j))*norm2,dimag(phi(j))*norm2
        if(.not.allconfigs) then
           write(*,*) x,real(phi(j)),dimag(phi(j))
        else
           write(*,"(g18.8,3es18.8)") t,x,real(phi(j)),dimag(phi(j))
        end if        
     end do
     if(gpsurface) then
        write(*,*)
        write(*,*)
     end if
  
     ! =============== next config ===============
     ! If we only want one config set haveconfig to false, if not try to read in another config
     if (.not.allconfigs) then
        haveconfig = .false.
     else
        if (verbosity .ge.2) write(*,*)  "# reading next config"
        read(famp,"(a80)",iostat=ierr) line
        if (verbosity .ge.2) write (*,*) "# Comment:",trim(line)

        if(trim(line) .eq. "") then
           read(famp,"(a80)",iostat=ierr) line
           if (verbosity .ge.2) write (*,*) "# Comment:",trim(line)
        endif

        if (ierr .eq. 0)then
           if (verbosity .ge.2)  write(*,*)  "# reading next config amplitudes"
           do i = 1,nbasis
              read(famp,*,iostat=ierr) t,ReAmp,ImAmp,(qps(j,i),j=1,2*ndf)
              ! Complex amplitude
              Amp(i)  = ReAmp+(0,1)*ImAmp
           end do
        end if
        if (ierr.ne.0) haveconfig = .false.
     end if
     
  end do
  



  ! =============== Clean up ===============
  deallocate(gamma)
  deallocate(qps)
  deallocate(phi)
  

  
end program plotwavefn


! *******************************************
!          read unwanted data
! *******************************************
subroutine readunwanted( fu, nb,ns,verbosity)
  implicit none 
  integer, intent(in) :: fu ! file unit
  integer, intent(in) :: nb ! number of basis functions
  integer, intent(in) :: ns ! number of steps
  integer, intent(in) :: verbosity ! how much info to write out

  integer :: i
  integer :: nlines ! number of lines to read in.
  integer :: ierr   ! for read error
  character(len=20) :: line

  !  work out number of lines to read
  nlines = (nb+2)*ns

  do i = 1,nlines
     read(fu,'(a10)',iostat=ierr) line
     if (i.eq.1 .or. i .eq.nb+3) then
       if(verbosity .ge. 2) write(*,*) "# line:",line
     end if
     if (ierr .ne. 0) then
        write(*,*) "Error reading unwanted basis functions"
        stop
     end if
  end do

  !write(
end subroutine readunwanted

! *******************************************
!             help subroutine
! *******************************************
subroutine help

  write(*,*) "  plotwavefn"
  write(*,*) "  Stewart Reed"
  write(*,*) "  8th June 2008"
  write(*,*)
  write(*,*) "-r xmin xmax. xmin can be -inf and xmax inf"
  write(*,*) "-m m"
  write(*,*) "-n nbasis"
  write(*,*) "-d ndof"
  write(*,*) "-s step"
  write(*,*) "-all  create histogram for all steps in file"
  write(*,*) "-nb nbins"
  write(*,*) "-g gamma"
  write(*,*) "-gf gammafilename"
  write(*,*) "-v   increased output"
  write(*,*) "-q   only writeout histogram. i.e. minimum output"
  write(*,*) "-gpsurf   include blank lines so that it works as a surface file for gnuplot. For use in conjungtion with -all"

  write(*,*) "ampfilename"

  write(*,*) "Setting m=0 gives the norm of the wavefunction. All bins get the norm so only really need one."
  write(*,*) "Alternatively, can set xmin -inf and xmax inf and choose any mode, including 0."
  
end subroutine help


! ==================================================
!              Partial Overlap
! ================================================== 
! Calculate the partial overlap of two coherent states zi and zj
! The overlap matrix between basis fns zi and zj is given by
! <zi|zj>=prod(int(<zia|x><x|zja> dx,a = 1,ndf) = 
!     =prod(exp(zia^* zja - |zia|^2/2 - |zja|^2/2),a=1,df)
!
! Here we compute the partial overlap defined as
!  prod(exp(zia^* zja - |zia|^2/2 - |zja|^2/2) * (1-delta(a-dof)),a=1,df)
! where dof is the degree of freedom we want to exclude.
!  
function partialoverlap_qp(qpi,qpj,ndf,dof,gammas)
  implicit none

  integer,intent(in) :: ndf
  real(kind=8),intent(in) :: qpi(2*ndf),qpj(2*ndf)
  integer,intent(in) :: dof
  real(kind=8),intent(in) :: gammas(ndf)
  
  complex(kind=8) :: zi,zj
  complex(kind=8) :: partialoverlap_qp
  integer :: df
  
  partialoverlap_qp = (1.0d0,0.0d0) 
  
  do df = 1,ndf
     if (df .ne. dof) then
        ! calculate z
        zi = dsqrt(gammas(df)/2.0) * qpi(2*df -1) &
             + (0,1.0)/(dsqrt(2*gammas(df)))*qpi(2*df) 
        zj = dsqrt(gammas(df)/2.0) * qpj(2*df -1) &
             + (0,1.0)/(dsqrt(2*gammas(df)))*qpj(2*df) 
        
        ! Now calculate contribution to overlap
        partialoverlap_qp = partialoverlap_qp * &
             cdexp(dconjg(zi)*zj &
             - cdabs(zi)**2/2 -cdabs(zj)**2/2)
     end if
  end do
  
end function partialoverlap_qp


! ==================================================
!              Overlap of 2 coherent states
! ================================================== 
! Calculate the  overlap of two coherent states zi and zj
! The overlap matrix between basis fns zi and zj is given by
! <zi|zj>=prod(int(<zia|x><x|zja> dx,a = 1,ndf) = 
!     =prod(exp(zia^* zja - |zia|^2/2 - |zja|^2/2),a=1,df)
function overlap_qp(qpi,qpj,ndf,gammas)
  implicit none

  integer,intent(in) :: ndf
  real(kind=8),intent(in) :: qpi(2*ndf),qpj(2*ndf)
  real(kind=8),intent(in) :: gammas(ndf)
  
  complex(kind=8) :: zi,zj
  complex(kind=8) :: overlap_qp
  integer :: df
  
  overlap_qp = (1.0d0,0.0d0) 
  
  do df = 1,ndf
     ! calculate z
     zi = dsqrt(gammas(df)/2.0) * qpi(2*df -1) &
          + (0,1.0)/(dsqrt(2*gammas(df)))*qpi(2*df) 
     zj = dsqrt(gammas(df)/2.0) * qpj(2*df -1) &
          + (0,1.0)/(dsqrt(2*gammas(df)))*qpj(2*df) 
     
     ! Now calculate contribution to overlap
     overlap_qp = overlap_qp * &
          cdexp(dconjg(zi)*zj &
          - cdabs(zi)**2/2 -cdabs(zj)**2/2)
  end do
  
end function overlap_qp


! ==================================================
!             Calculate erfterms
! ================================================== 
! This function calculates 
function erfterm(qj,qk,pj,pk,mgamma,xmin,xmax,minf,pinf)
  implicit none
  real(kind=8), intent(in) :: qj,qk,pj,pk,mgamma,xmin,xmax
  logical,intent(in) :: minf,pinf

  complex(kind=8) :: erfterm

  complex(kind=8) :: errarg
  interface
     function CDERROR(Z)
       complex(kind=8) :: Z
       complex(kind=8) :: cderror
     end function CDERROR
  end interface

  errarg =0.5D0*((0,1D0)*(pj-pk)/mgamma-qj-qk)
  if (pinf) then
     erfterm = 1.0d0
  else
     erfterm = cderror(dsqrt(mgamma)*(xmax+errarg))
  end if


  if (minf) then
     erfterm = erfterm + 1.0d0
  else
     erfterm = erfterm - cderror(dsqrt(mgamma)*(xmin+errarg))
  end if
  
  erfterm = erfterm*0.5d0
  
end function erfterm

! ==================================================
!             Calculate diagerfterms
! ================================================== 
! This function calculates 
function diagerfterm(q,mgamma,xmin,xmax,minf,pinf)
  implicit none
  real(kind=8),intent(in) :: q
  real(kind=8), intent(in):: mgamma,xmin,xmax
  logical,intent(in) :: minf,pinf

  real(kind=8) :: diagerfterm

  interface
     function rerfunc(x)
       real(kind=8) :: x
       real(kind=8) :: rerfunc
     end function rerfunc
  end interface

  if (pinf) then
     diagerfterm = 1.0d0
  else
     diagerfterm = rerfunc(dsqrt(mgamma)*(xmax-q))
  end if


  if (minf) then
     diagerfterm = diagerfterm + 1.0d0
  else
     diagerfterm = diagerfterm -rerfunc(dsqrt(mgamma)*(xmin-q))
  end if
  
  diagerfterm = diagerfterm*0.5d0
  
end function diagerfterm
