! **************************************************
!                redistribute
!               27th Feb. 2007
!                   SKR
!
!  A wee program to read in a trajectory in phase 
!  space and write out n points in q,p but evenly 
!  distributed in phase space
! **************************************************

module redist

  ! How many basis functions
  integer :: nbasis
  ! The basis functions
  complex(kind=8),dimension(:), allocatable :: zb
  ! new positions and momenta
  real(kind=8),dimension(:), allocatable:: q,p

  ! Probability density at each initial point
  real(kind=8),dimension(:),allocatable :: prob
  ! Probability density at each new point
  real(kind=8),dimension(:),allocatable :: rprob

  real(kind=8) :: gamma

  ! Trajectory length
  real(kind=8) :: totlen

  ! Is the first point also the last point?
  logical :: returnstart

  ! How many segments to split the trajectory into
  integer :: nsegs

  ! Program control
  ! ---------------
  logical :: usetraj   ! use trajectory files rather than zpos and amp files

  integer :: displacefirst 
  real(kind=8) :: dinitlen

  integer :: fug=6     ! file unit to write output to.
  integer(kind=8) :: randseed=0 ! for random number generator

  ! file units
  ! ----------
  integer :: fcont=10    ! input file
  integer :: fzpos=11    ! zpos file
  integer :: ftraj=12    ! Trajectory file in
  integer :: famp=20     ! outputfile
  integer :: ftrajout=21 ! trajectory file out
  
  ! file names
  ! ----------
  character(len=100) :: zposfile
  character(len=100) :: ampfile
  character(len=100) :: controlfile
  character(len=100) :: trajfile       ! input trajectory file
  character(len=100) :: redisttrajfile ! redistributed trajectory file


contains
  ! ******************** readinputs ******************
  !  A subroutine to read the control file
  ! **************************************************
  subroutine readinputs
    implicit none
    
    character(len=100) :: stuff 
    
    integer :: ierr ! for file errors

    ! some defaults
    ampfile = "newamps.out"
    zposfile= "zpos.out"
    redisttrajfile = "redistqpc"

    nbasis = 1
    nsegs   = 1
    gamma  = 1.0d0
    returnstart = .false.
    usetraj = .false.
    displacefirst = 0

    read(fcont,*,iostat=ierr) stuff

    ! Check the options in the control file
    do while (ierr .eq.0)
       ! --- number of steps in input file ----
       if(stuff=="nsteps") then 
          backspace(fcont)
          read(fcont,*,iostat=ierr) stuff,nbasis
          if (ierr .ne. 0) then 
             write(*,*) "Error reading nsteps"
          end if
          ! --- number of redistributed points ----
       else if(stuff=="nsegments") then
          backspace(fcont)
          read(fcont,*) stuff,nsegs
          if (ierr .ne. 0) then 
             write(*,*) "Error reading nsegments"
          end if

          ! --- gamma ---
       else if(stuff=="gamma") then
          backspace(fcont)
          read(fcont,*,iostat=ierr) stuff,gamma
          if (ierr .ne. 0) then 
             write(*,*) "Error reading value for gamma"
          end if

          ! --- use trajfiles rather than zpos ---
       else if (stuff=="trajfiles") then
          usetraj = .true.
          
          ! ---- name of traj file ---
       else if (stuff=="traj") then
          read(fcont,*) trajfile

          ! ---  specify name of output trajectory ---
       else if (stuff=="redisttraj") then
          read(fcont,*) redisttrajfile

          ! --- Is the input trajectory closed? --- 
       else if (stuff=="closed") then
          returnstart= .true.

          ! ---- displace the first new point ----
       else if (stuff=="displacefirst") then
          displacefirst  = 1
          backspace(fcont) 
          read(fcont,*) stuff,dinitlen
          if (dinitlen < 0) then
             write (fug,*) "Displacement of intial point must be greater than 0."
             stop
          end if
          ! --- displace the first point by a random amount ---
       else if (stuff=="randdisplacefirst") then
          displacefirst = 2          

          ! --- print the help message ---
       else if(stuff=="help") then
          call help
          stop
       end if
       read(fcont,*,iostat=ierr) stuff
    end do


    if (returnstart) then
       nbasis = nbasis+1
    end if

    ! ---- now allocate arrays ----
    allocate(zb(nbasis))
    zb(:)=(0.0d0,0.0d0)
    
    allocate(q(nsegs),p(nsegs))

    if (usetraj) then
       allocate(prob(nbasis))
       allocate(rprob(nsegs))
    end if

  end subroutine readinputs

  ! ******************** readzpos ********************
  !  A subroutine to read the zpos.out file
  ! **************************************************
  subroutine readzpos
    implicit none
    
    integer :: ierr ! for file errors
    character(len=100) :: stuff  ! for reading stuff!

    real(kind=8) :: x,y ! for reading in Re(z) and Im(z)

    integer :: bf ! for loops
    real(kind=8)::t ! for reading in the time
    logical :: header  ! For reading in any comments at the top of the file

    ! try to open the zpos.out file
    open(fzpos,file=zposfile,status="old",iostat=ierr)
    if (ierr .eq. 29 ) then
       write(fug,*) "zpos file ",zposfile," does not exist"
       call help
       stop
    else if ( ierr .ne. 0) then
       write(fug,*) "Problem opening zpos file. Error number ",ierr
       call help
       stop
    end if

    if (returnstart) then
       nbasis = nbasis -1
    end if
    ! Read in top line(s)
    header = .true.
    do while (header)
       read(fzpos,*,iostat=ierr) stuff
       if (ierr.ne.0) then
          write(fug,*) "Error reading header of zpos.out. Error number",ierr
          stop
       end if
       if (stuff(1:1) .ne. "#") then
          header = .false.
          backspace(fzpos)
       end if
    end do
    ! Unread the last comment line which should say "# Step"
    backspace(fzpos)
    
    readbfs: do bf = 1,nbasis
       read(fzpos,*) stuff ! read in step line
       read(fzpos,*) t,x,y ! read in time, Re(z), Im(z)
       ! blank line is ignored.
       
       ! Store this z.
       zb(bf) = x+(0.0d0,1.0d0)*y
    end do readbfs

    ! close the file
    close(fzpos)

    if (returnstart) then
       nbasis = nbasis +1
       zb(nbasis) = zb(1)
    end if
    
  end subroutine readzpos
  
  ! ******************** readtraj ********************
  !  A subroutine to read a trajectory file
  !  First lines are comments
  !  Other lines q p prob
  ! **************************************************
  subroutine readtraj
    implicit none

    character(len=100) :: stuff  ! for reading stuff!

    real(kind=8) :: qq,pp ! for reading in q,p
    
    integer :: pt ! for loops
    logical :: header  ! For reading in any comments at the top of the file
    integer :: ierr
    
    ! try to open the trajectory file
    open(ftraj,file=trajfile,status="old",iostat=ierr)
    if (ierr .eq. 29 ) then
       write(fug,*) "Trajectory file ",trajfile," does not exist"
       call help
       stop
    else if ( ierr .ne. 0) then
       write(fug,*) "Problem opening trajectory file. Error number ",ierr
       call help
       stop
    end if
    
    if (returnstart) then
       nbasis = nbasis -1
    end if

    ! Read in top line(s)
    header = .true.
    do while (header)
       read(ftraj,*,iostat=ierr) stuff
       if (ierr.ne.0) then
          write(fug,*) "Error reading header of zpos.out. Error number",ierr
          stop
       end if
       if (stuff(1:1) .ne. "#") then
          header = .false.
          backspace(ftraj)
       end if
    end do
    
    readtrajpoints: do pt = 1,nbasis
       read(ftraj,*,iostat=ierr) qq,pp,prob(pt)
       if (ierr .ne. 0) then
          write(fug,*) "Error reading q,p, prob form file. Basis fn number",pt
          stop
       end if
       ! Now calculate z
       zb(pt) = sqrt(gamma/2.0d0)*qq+(0.0d0,1.0d0)*pp/sqrt(2.0d0*gamma)
    end do readtrajpoints

    ! close the file
    close(ftraj)

    if (returnstart) then
       nbasis = nbasis +1
       zb(nbasis) = zb(1)
       prob(nbasis) = prob(1)
    end if
  end subroutine readtraj


  ! ==================================================
  !              position, q
  ! ================================================== 
  ! Calculate the position of a coherent state
  function pos(zi)
    implicit none
    complex(kind=8), intent(in) :: zi
    real(kind=8) :: pos

    pos = real(zi+dconjg(zi)) /dsqrt(2.0d0*gamma)
    !s2gammarec(i)
    
  end function pos
  ! ==================================================
  !              momentum, p
  ! ================================================== 
  ! Calculate the momentum of a coherent state
  function mom(zi)
    implicit none
    complex(kind=8), intent(in) :: zi
    real(kind=8) :: mom

    mom = real((0d0,-1d0)*(zi-dconjg(zi)))*dsqrt(gamma/2.0d0)
    !shgamma(i)

  end function mom


  ! ******************** calclen *********************
  !  A subroutine to calculate the total length of the
  !  trajectory
  ! **************************************************
  subroutine calclen
    implicit none
    real(kind=8) :: dx,dy
    integer :: i
    ! set length to zero
    totlen = 0.0d0

    do i = 2,nbasis
       dx = real(zb(i)) - real(zb(i-1))
       dy = dimag(zb(i)) - dimag(zb(i-1))
       
       totlen = totlen + dsqrt(dx**2+dy**2)

    end do
 
    ! This is now done by increasing the number of points and
    ! Setting the extra point at the end equal to the first.
    ! I.e. it is already calculated above and if true the
    ! following will be zero!
    ! Is the first point also the last point?
!    if (returnstart) then
!       dx = real(zb(1)) - real(zb(nbasis))
!       dy = dimag(zb(1)) - dimag(zb(nbasis))
!       
!       totlen = totlen + dsqrt(dx**2+dy**2)
!    end if

  end subroutine calclen

  ! ******************** makeqp **********************
  !  A subroutine to work out qps which are evenly
  !  distributed in phase space
  ! **************************************************
  subroutine makeqp
    implicit none

    ! basis function
    integer :: bf

    ! current position around loop
    real(kind=8) :: len
    ! Change in position
    real(kind=8) :: dlen
    real(kind=8) :: dx,dy

    real(kind=8) :: extralen ! how much past the last bf is the required point

    ! segment length
    real(kind=8) :: slen
    integer :: seg

    ! new z and prob
    complex(kind=8) :: znew
    real(kind=8) :: pnew
    
    ! Random number generator
    real(kind=8),external :: ZBQLU01

    slen = totlen/dble(nsegs)

    ! Current position
    len = 0.0d0

    bf = 1
    seg = 1
    if (displacefirst > 0) then

       ! random amount otherwise use a fixed amount
       if (displacefirst == 2) then 
          call ZBQLINI(randseed,fug)
          
          ! choose a random len. randseed doesn't actually do anything
          dinitlen = slen*ZBQLU01(randseed)
       else
          dinitlen = dinitlen * slen
       end if

       dx = real(zb(bf+1)) - real(zb(bf))
       dy = dimag(zb(bf+1)) - dimag(zb(bf))
       dlen = dsqrt(dx**2+dy**2)
       do while (len+dlen < dinitlen)
          bf = bf + 1
          len = len + dlen

          dx = real(zb(bf+1)) - real(zb(bf))
          dy = dimag(zb(bf+1)) - dimag(zb(bf))
          dlen =  dsqrt(dx**2+dy**2)
       end do
       !distance from last point to where we want the new point
       extralen = dinitlen - len
       
       ! New point is 
       znew = extralen / dlen * (zb(bf+1) - zb(bf)) + zb(bf)
       q(1) = pos(znew)
       p(1) = mom(znew)
       
       ! probability if required
       if (usetraj) then
          pnew = extralen / dlen * (prob(bf+1) - prob(bf)) + prob(bf)
          rprob(1) = pnew
       end if

       write (fug,*) "Displacing initial point around the trajectory by ",dinitlen
    else
       ! First z is first new node
       q(1) = pos(zb(1))
       p(1) = mom(zb(1))

       ! make sure dinitlen = 0
       dinitlen = 0.0d0

       if(usetraj) rprob(1) = prob(1)

    end if
    ! 
    segments: do seg = 2,nsegs

       dx = real(zb(bf+1)) -  real(zb(bf))
       dy = dimag(zb(bf+1)) - dimag(zb(bf))
       dlen = dsqrt(dx**2+dy**2)

       do while (len+dlen < dble(seg-1)*slen +dinitlen)
          bf = bf + 1
          if (bf .ge. nbasis) then
             bf = 1
          end if

          len = len + dlen

          dx = real(zb(bf+1)) - real(zb(bf))
          dy = dimag(zb(bf+1)) - dimag(zb(bf))
          dlen =  dsqrt(dx**2+dy**2)

       end do

       !distance from last point to where we want the new point
       extralen = dble(seg-1)*slen+dinitlen - len
       
       ! New point is 
       znew = extralen / dlen * (zb(bf+1) - zb(bf)) + zb(bf)
       q(seg) = pos(znew)
       p(seg) = mom(znew)
       ! probability if required
       if (usetraj) then
          pnew = extralen / dlen * (prob(bf+1) - prob(bf)) + prob(bf)
          rprob(seg) = pnew
       end if

    end do segments

  end subroutine makeqp

  ! ******************** writeamp ********************
  !  A subroutine to write the amplitudes file
  ! **************************************************
  subroutine writeamp
    implicit none
    
    integer :: seg

    ! Open file
    open(famp,file=ampfile)

    ! write header
    write(famp,*) "# wave packet amplitudes"

    do seg = 1,nsegs
       write(famp,*) "#  Step  Re(amp) Img(amp) q1,p1,.....qn,pn"
       write(famp,"(5es18.8)") dble(seg),1.0d0 ,0.0d0,q(seg),p(seg)
       write(famp,*)
    end do
    
    close(famp)
  end subroutine writeamp

  ! ******************** writetraj ********************
  !  A subroutine to write a trajectory file with 
  !  amplitude.
  ! **************************************************
  subroutine writetraj
    implicit none
    
    integer :: seg

    ! Open file
    open(ftrajout,file=redisttrajfile)

    ! write header
    write(ftrajout,*) "# Redistributed trajector q,p,c"
    do seg = 1,nsegs
       write(ftrajout,"(3es18.8)") q(seg),p(seg),sqrt(rprob(seg))
    end do
    
    close(ftrajout)
  end subroutine writetraj

  ! ******************** help ************************
  !  A subroutine to give info about this program
  ! **************************************************
   subroutine help
    
    implicit none
    
    write(*,*) "The control file,",controlfile," must contain at least"
    write(*,*) "  nsteps <ns>       Number of steps in initial trajectory."
    write(*,*) "  nsegments <ns>    Number of steps required in final trajectory"
    write(*,*) "It may also contain"
    write(*,*) "  gamma <gamma>     Gamma for the mode required if converting between q,p and z."
    write(*,*) "                    Redistribution is in terms of z so it is needed if input is in q,p"
    write(*,*) "  trajfiles         Input file is a q,p trajectory file"
    write(*,*) "  traj              Name of trajectory file should be on the next line"
    write(*,*) "  redistrtraj       Name of output trajectory file should be on the next line."
    write(*,*) "                    Default: redistqpc "
    write(*,*) "  closed            Initial trajectory is closed. I.e. the last point is supposed to"
    write(*,*) "                    be the same as the first. If this option is specified, the first"
    write(*,*) "                    point is used instead of the last in order to make it truely closed."       
    write(*,*) "  displacefirst <v> Displace the first redistributed point around the trajectory from"
    write(*,*) "                    the first point in the input trajectory by v times distance beteen"
    write(*,*) "                    each new point i.e. v * total length / nsegments."
    write(*,*) "  randdisplacefirst As for displacefirst except that v is chosen randomly in [0,1]."
    write(*,*) "  help              Gives this message."
  end subroutine help
end module redist

! ******************** main program ****************
!  Basically just calls routines from the module above.
! **************************************************

program redistribute

  use redist

  implicit none

  integer :: ierr ! for file errors

  write(fug,*) "Welcome to redsistribute"

  ! ---- read input file ----
  ! Set name of input file
  controlfile = "redistribute.inpt"

  ! try to open the input file
  open(fcont,file=controlfile,status="old",iostat=ierr)
  if (ierr .eq. 29 ) then
     write(fug,*) "Input file ",controlfile," does not exist"
     call help
     stop
  else if ( ierr .ne. 0) then
     write(fug,*) "Problem opening input file. Error number ",ierr
     call help
     stop
  end if

  write(fug,*) "Reading inputs"
  call readinputs

  close(fcont)

  if (usetraj) then
     ! ---- Read trajectory ----
     call readtraj
     write(fug,*) "Read trajectory"
  else
     ! ---- Read zpositions ----
     call readzpos
     write(fug,*) "Reading basis functions"
  end if

  ! ---- Calculate trajectory length ----
  write(fug,*) "Calculating trajectory length"
  call calclen

  write(fug,*) "Total trajectory length ",totlen

  ! ---- determine q,ps ----
  write(fug,*) "Determining {q,p}"
  call makeqp


  if (usetraj) then
     ! ---- write out new q,p,amplitude as a trajectory ----
     write(fug,*) "Writing trajectory file"
     call writetraj
     write(fug,*) "Written new {q,p,c} to ",redisttrajfile
       
  else
     ! ---- write out q,p as an amplitude file ----
     write(fug,*) "Writing output file"
     call writeamp
     write(fug,*) "Written new {q,p} to ",ampfile
       
  end if

  ! ---- Say good bye ----
  write(fug,*) "Bye bye"

end program redistribute
